#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#include <math.h>
#include "hibutil.h"

#define MAXDEVS 16

static UINT64 *piowbuf;
static UINT64 *piowbuf2[NHIB];


static void eraserom(int argc, char **argv);
static void writerom(int argc, char **argv);
static void readromid(int argc, char **argv);

static void writepllconf(int argc, char **argv);

static void get_cputime(double *laptime, double *sprittime);


static void dmarperf(int devid)
{
    Hib *h;
    UINT64 *wbuf, *rbuf;
    int sendfunc;
    int i, j, ntry, nng, size0, off;
    int size; /* in 32-bit words */
    double lt = 0.0, st = 0.0;
    double nloop;

    h = hib_openMC(devid);
    hib_set_test_modeMC(devid, TESTMODE_REFDESIGN_RAM);

    fprintf(stderr, "\n# hib[%d] DMA read (host -> HIB)\n", devid);
    rbuf = h->dmar_buf;
    wbuf = h->dmaw_buf;

    for (i = 0; i < (h->dmabuf_bytes>>3); i++) {
	rbuf[i] = 0x123456789abc0000ll|i;
	wbuf[i] = 0xfedcba9876540000ll|i;
	//      fprintf(stderr, "rbuf[0x%02x]: 0x%016llx\n", i, rbuf[i]);
    }

    nloop = 1e7;

    fprintf(stderr, "First test call to DMAW interface start\n");

    
    if (1)    {
	int i;
	for (i=0;i<64; i++){
	    hib_mem_writeMC(devid, 0x60, i);
	    hib_mem_writeMC(devid, 0x64, i*3);
	}
	hib_set_dma_extended_regMC(devid, 0, 0, 0x01f);
	hib_register_dma_seqMC(devid, 0, 0, 128, rbuf);
	
	for (i=32;i<48; i++){
	    hib_mem_writeMC(devid, 0x60, i);
	    fprintf(stderr, "extended register  0x%08lx: 0x%08x\n",
		    i, hib_mem_readMC(devid, 0x60));
	}
	int addr;
	hib_mem_writeMC(devid, 0x10, 0x1234);
	hib_mem_writeMC(devid, 0x14, 0x5678);
	hib_mem_writeMC(devid, 0x18, 0xabcd);
	usleep(100);
	hib_start_dma_seqMC(devid, 0, 0);
	usleep(100);
	fprintf(stderr, "Now wait for dma to finish?\n");
	hib_finish_dmarMC(devid);
	fprintf(stderr, "DMA finished\n");
	
	for(addr=0x10;addr < 0x20; addr+= 4){
	    fprintf(stderr, "hib[%d] 0x%08lx: 0x%08x\n",
		    devid, addr, hib_mem_readMC(devid, addr));
	}
    }

    int maxreqsize;
    for(maxreqsize=1; maxreqsize<7; maxreqsize++){
	hib_config_setmaxrequestMC(0, maxreqsize);
	printf("Max req size set to %d\n", maxreqsize);
	for (size = 128; size <= 4096; size *= 2) {
	    for (ntry = 0; ntry < 1; ntry++) {
		get_cputime(&lt, &st);
		
		for (j = 0; j < nloop/size; j++) {
		    hib_start_dmarMC(devid, size, rbuf);
		    hib_finish_dmarMC(devid);
		}
		get_cputime(&lt, &st);
		printf("size: %d DMA read: %f sec  %f MB/s\n",
		       size*8, lt, sizeof(UINT64)*nloop/1e6/lt);
	    }
	}
	printf("# With interal hibreg used\n", devid);

	for (size = 128; size <= 8192; size *= 2) {
	    for (ntry = 0; ntry < 1; ntry++) {
		hib_register_dma_seqMC(devid, 0, 0, size, rbuf);
		get_cputime(&lt, &st);
		
		for (j = 0; j < nloop/size; j++) {
		    hib_start_dma_seqMC(devid, 0, 0);
		    hib_finish_dmarMC(devid);
		}
		get_cputime(&lt, &st);
		printf("size: %d DMA read: %f sec  %f MB/s\n",
		       size*8, lt, sizeof(UINT64)*nloop/1e6/lt);
	    }
	}
    }

    // wait long enough before unset test mode
    // so that PIOW burst is completed.
    usleep(10);
    hib_set_test_modeMC(devid, TESTMODE_NONE);
    hib_closeMC(devid);
}

static void dmarperfMC(int ndevs)
{
    Hib *h[MAXDEVS];
    UINT64 *wbuf[MAXDEVS], *rbuf[MAXDEVS];
    int sendfunc;
    int i, j, k, ntry, nng, size0, off;
    int size; /* in 32-bit words */
    double lt = 0.0, st = 0.0;
    double nloop;

    for(j=0;j<ndevs;j++){
    
	h[j] = hib_openMC(j);
	hib_set_test_modeMC(j, TESTMODE_REFDESIGN_RAM);
	
	fprintf(stderr, "\n# hib[%d] DMA read (host -> HIB)\n", j);
	rbuf[j] = h[j]->dmar_buf;
	wbuf[j] = h[j]->dmaw_buf;
	
	for (i = 0; i < (h[j]->dmabuf_bytes>>3); i++) {
	    rbuf[j][i] = 0x123456789abc0000ll|i;
	    wbuf[j][i] = 0xfedcba9876540000ll|i;
	    //      fprintf(stderr, "rbuf[0x%02x]: 0x%016llx\n", i, rbuf[i]);
	}
    }

    nloop = 1e7;

    fprintf(stderr, "First test call to DMAW interface start\n");

    int maxreqsize;
    for(maxreqsize=1; maxreqsize<7; maxreqsize++){
	for(k=0;k<ndevs;k++) hib_config_setmaxrequestMC(k, maxreqsize);
	printf("Max req size set to %d\n", maxreqsize);
	for (size = 128; size <= 8192; size *= 2) {
	    for (ntry = 0; ntry < 1; ntry++) {
		get_cputime(&lt, &st);
		
		for (j = 0; j < nloop/size; j++) {
		    for(k=0;k<ndevs;k++) hib_start_dmarMC(k, size, rbuf);
		    for(k=0;k<ndevs;k++) hib_finish_dmarMC(k);
		}
		get_cputime(&lt, &st);
		printf("size: %d DMA read: %f sec  %f MB/s\n",
		       size*8, lt, sizeof(UINT64)*nloop*ndevs/1e6/lt);
	    }
	}
	printf("# With interal hibreg used\n" );

	for (size = 128; size <= 8192; size *= 2) {
	    for (ntry = 0; ntry < 1; ntry++) {
		for(k=0;k<ndevs;k++) hib_register_dma_seqMC(k, 0, 0, size, rbuf[k]);
		get_cputime(&lt, &st);
		
		for (j = 0; j < nloop/size; j++) {
		    for(k=0;k<ndevs;k++) hib_start_dma_seqMC(k, 0, 0);
		    //for(k=0;k<ndevs;k++) {
		    //			hib_finish_dmarMC(k);
		    //			hib_start_dma_seqMC(k, 0, 0);
		    //		    }
		    for(k=0;k<ndevs;k++) hib_finish_dmarMC(k);
		}
		get_cputime(&lt, &st);
		printf("size: %d DMA read: %f sec  %f MB/s\n",
		       size*8, lt, sizeof(UINT64)*nloop*ndevs/1e6/lt);
	    }
	}
    }

    // wait long enough before unset test mode
    // so that PIOW burst is completed.
    usleep(10);
    for(k=0;k<ndevs;k++){
	hib_set_test_modeMC(k, TESTMODE_NONE);
	hib_closeMC(k);
    }
}


static void get_cputime(double *splittime, double *laptime)
{
    struct timeval x;

    gettimeofday(&x, NULL);

    *splittime = x.tv_sec + x.tv_usec/1000000.0 - *laptime;
    *laptime = x.tv_sec + x.tv_usec/1000000.0;
}

int main(int argc, char **argv)
{
    int mode, i;
   dmarperf(0);
   //     dmarperfMC(4);
    exit (0);
}

