#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#include <math.h>
#include <getopt.h>

#include "hibutil.h"

#include <xmmintrin.h>
#include <emmintrin.h>

#define MAXDEVS 16

static UINT64 *piowbuf;
static UINT64 *piowbuf2[NHIB];



static void get_cputime(double *laptime, double *sprittime);


static void reset_card(int devid, int reset_code)
{
    Hib *h;
    h = hib_openMC(devid);
    fprintf(stderr,"resetting card %d with code %d\n", devid, reset_code);
    hib_send_resetMC(devid, reset_code);
    hib_closeMC(devid);
}


static void dmawperf(int devid)
{
    Hib *h;
    UINT64 *wbuf, *rbuf;
    int i, j, ntry, nng, size0, off;
    int size; /* in 32-bit words */
    double lt = 0.0, st = 0.0;
    double nloop;

    h = hib_openMC(devid);
    hib_reset_dma(devid);
    hib_set_test_modeMC(devid, TESTMODE_REFDESIGN_RAM);
    rbuf = h->dmar_buf;
    wbuf = h->dmaw_buf;

    for (i = 0; i < (h->dmabuf_bytes>>3); i++) {
	rbuf[i] = 0x123456789abc0000ll|i;
	wbuf[i] = 0xfedcba9876540000ll|i;
	//      fprintf(stderr, "rbuf[0x%02x]: 0x%016llx\n", i, rbuf[i]);
    }

    nloop = 1e7;
    fprintf(stderr, "First test call to DMAW interface start\n");

    
    if (1)    {
	int i;
	for (i=0;i<64; i++){
	    hib_mem_writeMC(devid, 0x60, i);
	    hib_mem_writeMC(devid, 0x64, i*3);
	}
	hib_set_dma_extended_regMC(devid, 0, 0, 0x01f);
	hib_register_dma_seqMC(devid, 0, 1, 128, wbuf);
	
	//	hib_start_dma_seqMC(devid, 0, 1);
	//	hib_mem_writeMC(devid, 0x68, 2);
	for (i=32;i<48; i++){
	    hib_mem_writeMC(devid, 0x60, i);
	    fprintf(stderr, "extended register  0x%08lx: 0x%08x\n",
		    i, hib_mem_readMC(devid, 0x60));
	}
	int addr;
	hib_mem_writeMC(devid, 0x10, 0x1234);
	hib_mem_writeMC(devid, 0x14, 0x5678);
	hib_mem_writeMC(devid, 0x18, 0xabcd);
	usleep(100);
	hib_start_dma_seqMC(devid, 0, 1);
	usleep(100);
	fprintf(stderr, "Now wait for dma to finish?\n");
	hib_finish_dmawMC(devid);
	fprintf(stderr, "DMA finished (not started..)\n");
	
	for(addr=0x10;addr < 0x20; addr+= 4){
	    fprintf(stderr, "hib[%d] 0x%08lx: 0x%08x\n",
		    devid, addr, hib_mem_readMC(devid, addr));
	}
	// wait long enough before unset test mode
	// so that PIOW burst is completed.
	//	usleep(10);
	//	hib_set_test_modeMC(devid, TESTMODE_NONE);
	//	hib_closeMC(devid);
	//	exit(0);
	hib_start_dmawMC(devid, 128, wbuf);
	hib_finish_dmawMC(devid);
	fprintf(stderr, "First test call to DMAW interface end\n");
    }

    fprintf(stderr, "\n# hib[%d] DMA write (host <- HIB)\n", devid);
    //    for (size = 128; size <= 8192; size *= 2) {
    hib_set_dma_extended_regMC(devid, 0, 0, 0x01f);
    for (size = 128; size <= 8192; size *= 2) {
        for (ntry = 0; ntry < 1; ntry++) {
	    get_cputime(&lt, &st);
            for (j = 0; j < nloop/size; j++) {
		hib_start_dmawMC(devid, size, wbuf);
		hib_finish_dmawMC(devid);

	    }
	    get_cputime(&lt, &st);
	    printf("size: %d DMA write: %f sec  %f MB/s\n",
		   size*8, lt, sizeof(UINT64)*nloop/1e6/lt);
	}
    }

    if(0){
	fprintf(stderr, "# With interal hibreg used\n", devid);
	for (size = 128; size <= 8192; size *= 2) {
	    for (ntry = 0; ntry < 1; ntry++) {
		hib_register_dma_seqMC(devid, 0, 1, size, wbuf);
		get_cputime(&lt, &st);
		for (j = 0; j < nloop/size; j++) {
		    hib_start_dma_seqMC(devid, 0, 1);
		    hib_finish_dmawMC(devid);
		    
		}
		get_cputime(&lt, &st);
		printf("size: %d DMA write: %f sec  %f MB/s\n",
		       size*8, lt, sizeof(UINT64)*nloop/1e6/lt);
	    }
	}
    }

    // wait long enough before unset test mode
    // so that PIOW burst is completed.
    usleep(10);
    hib_set_test_modeMC(devid, TESTMODE_NONE);
    hib_closeMC(devid);
}

static void testfunc(int devid)
{
    Hib *h;
    UINT64 *wbuf, *rbuf;
    int i, j, ntry, nng, size0, off;
    int size; /* in 32-bit words */
    double lt = 0.0, st = 0.0;
    double nloop;

    h = hib_openMC(devid);
    hib_reset_dma(devid);
    hib_set_test_modeMC(devid, TESTMODE_REFDESIGN_RAM);
    rbuf = h->dmar_buf;
    wbuf = h->dmaw_buf;

    for (i = 0; i < (h->dmabuf_bytes>>3); i++) {
	rbuf[i] = 0x123456789abc0000ll|i;
	wbuf[i] = 0xfedcba9876540000ll|i;
	//      fprintf(stderr, "rbuf[0x%02x]: 0x%016llx\n", i, rbuf[i]);
    }

    nloop = 1e7;
    
    for (i=0;i<64; i++){
	hib_mem_writeMC(devid, 0x60, i);
	hib_mem_writeMC(devid, 0x64, i*3);
    }
    hib_set_dma_extended_regMC(devid, 0, 0, 0x01f);
    hib_register_dma_seqMC(devid, 0, 1, 128, wbuf);
    
    //	hib_start_dma_seqMC(devid, 0, 1);
    //	hib_mem_writeMC(devid, 0x68, 2);
    for (i=32;i<48; i++){
	hib_mem_writeMC(devid, 0x60, i);
	fprintf(stderr, "extended register  0x%08lx: 0x%08x\n",
		i, hib_mem_readMC(devid, 0x60));
    }
    int addr;
    hib_mem_writeMC(devid, 0x10, 0x1234);
    hib_mem_writeMC(devid, 0x14, 0x5678);
    hib_mem_writeMC(devid, 0x18, 0xabcd);
    fprintf(stderr, "DMA finished (not started..)\n");

    hib_set_dma_extended_regMC(devid, 1, 0xff, 0x01f);

    for(addr=0x10;addr < 0x30; addr+= 4){
	fprintf(stderr, "hib[%d] 0x%08lx: 0x%08x\n",
		devid, addr, hib_mem_readMC(devid, addr));
    }
    hib_set_dma_extended_regMC(devid, 0, 0x0, 0x01f);

    for(addr=0x10;addr < 0x30; addr+= 4){
	fprintf(stderr, "hib[%d] 0x%08lx: 0x%08x\n",
		devid, addr, hib_mem_readMC(devid, addr));
    }

    // wait long enough before unset test mode
    // so that PIOW burst is completed.
    usleep(1000);
    hib_set_test_modeMC(devid, TESTMODE_NONE);
    usleep(1000);
    hib_closeMC(devid);
    usleep(1000);
}

static inline UINT64 rdtsc2()
{
	UINT64 ret;
	asm __volatile__("rdtsc  \n\t" \
			 "shlq $32, %%rdx \n\t" \
			 "orq  %%rdx, %%rax \n\t" \
			 "movq %%rax, %0" : "=g"(ret) : :"rax", "rdx");
	return ret;
}

static void xwait(int wait)
{
    UINT64 st = rdtsc2();
    while((rdtsc2()-st)<wait) {}
}

static void dmawperfMC(int ndevs, int firstdev)
{
    Hib *h[MAXDEVS];
    UINT64 *wbuf[MAXDEVS], *rbuf[MAXDEVS];
    int i, j, k, ntry, nng, size0, off;
    int enddev;
    int size; /* in 32-bit words */
    double lt = 0.0, st = 0.0;
    double nloop;
    enddev = firstdev + ndevs;
    for(j=firstdev;j<enddev;j++){
	h[j] = hib_openMC(j);
	hib_reset_dma(j);
	hib_set_test_modeMC(j, TESTMODE_REFDESIGN_RAM);
	rbuf[j] = h[j]->dmar_buf;
	wbuf[j] = h[j]->dmaw_buf;
	fprintf(stderr,"hib[%d] dmabuf_bytes=%d\n", j, h[j]->dmabuf_bytes);
	for (i = 0; i < (h[j]->dmabuf_bytes>>3); i++) {
	    rbuf[j][i] = 0x123456789abc0000ll|i;
	    wbuf[j][i] = 0xfedcba9876540000ll|i;
	    //      fprintf(stderr, "rbuf[0x%02x]: 0x%016llx\n", i, rbuf[i]);
	}
    }
    nloop = 1e7;

    fprintf(stderr, "\n# hib[%d] DMA write (host <- HIB)\n", ndevs);
    for (size = 128; size <= 8192; size *= 2) {
	get_cputime(&lt, &st);
	for (j = 0; j < nloop/size; j++) {
	    for(k=firstdev;k<enddev;k++)hib_start_dmawMC(k, size, wbuf[k]);
	    xwait(5000);
	    for(k=firstdev;k<enddev;k++)hib_finish_dmawMC(k);
	    
	}
	get_cputime(&lt, &st);
	printf("size: %d DMA write: %f sec  %f MB/s\n",
	       size*8, lt, sizeof(UINT64)*ndevs*nloop/1e6/lt);
    }
    fprintf(stderr, "\n# With interal hibreg used, ndevs=%d\n", ndevs);

    if(1)for (size = 128; size <= 8192; size *= 2) {
	for (ntry = 0; ntry < 1; ntry++) {
	    for(k=firstdev;k<enddev;k++){
		hib_set_dma_extended_regMC(k, 0, 0, 0x01f);  
		usleep (100);
		hib_register_dma_seqMC(k, 0, 1, size, wbuf[k]);
		usleep (100);
	    }
	    get_cputime(&lt, &st);
	    for (j = 0; j < nloop/size; j++) {
		for(k=firstdev;k<enddev;k++) hib_start_dma_seqMC(k, 0, 1);
		xwait(5000);
		for(k=firstdev;k<enddev;k++) hib_finish_dmawMC(k);
	    }
	    get_cputime(&lt, &st);
	    printf("size: %d DMA write: %f sec  %f MB/s\n",
		   size*8, lt, sizeof(UINT64)*ndevs*nloop/1e6/lt);
	}
    }
    if (0){
	//
	// this part seems to cause kernel panic for ndevs=4
	//
	fprintf(stderr, "\n# With interal hibreg used, two DMAW issued\n");
	
	int kstart=0;
	int kend = ndevs;
	for (size = 128; size <= 2048; size *= 2) {
	    for (ntry = 0; ntry < 1; ntry++) {
		for(k=firstdev;k<enddev;k++){
		    hib_register_dma_seqMC(k, 0, 1, size, wbuf[k]);
		    hib_register_dma_seqMC(k, 1, 1, size, wbuf[k]+size);
		}
		get_cputime(&lt, &st);
		for (j = 0; j < nloop/size; j++) {
		    for(k=firstdev;k<enddev;k++) hib_start_dma_seqMC(k, 0, 1);
		    xwait(5000);
		    for(k=firstdev;k<enddev;k++) hib_start_dma_seqMC(k, 1, 1);
		    xwait(5000);
		    for(k=firstdev;k<enddev;k++) hib_finish_dmawMC(k);
		    
		}
		get_cputime(&lt, &st);
		printf("size: %d DMA write: %f sec  %f MB/s\n",
		       size*8, lt, sizeof(UINT64)*ndevs*2*nloop/1e6/lt);
	    }
	}
    }
    // wait long enough before unset test mode
    // so that PIOW burst is completed.
    // wait long enough before unset test mode
    // so that PIOW burst is completed.
    usleep(10);
    for(k=firstdev;k<enddev;k++){
	// test to fix page table structure
	for (i = 0; i < (h[k]->dmabuf_bytes>>3); i++) {
	    rbuf[k][i] = 0x123456789abc0000ll|i;
	    wbuf[k][i] = 0xfedcba9876540000ll|i;
	}
    }
    usleep(100);
    for(k=firstdev;k<enddev;k++){
	fprintf(stderr,"Reset test mode for %d\n", k);
	hib_set_test_modeMC(k, TESTMODE_NONE);
	usleep(1000);
	hib_closeMC(k);
	usleep(1000);
    }
}


static void dmarperf(int devid)
{
    Hib *h;
    UINT64 *wbuf, *rbuf;
    int sendfunc;
    int i, j, ntry, nng, size0, off;
    int size; /* in 32-bit words */
    double lt = 0.0, st = 0.0;
    double nloop;

    h = hib_openMC(devid);
    hib_reset_dma(devid);

    hib_set_test_modeMC(devid, TESTMODE_REFDESIGN_RAM);

    fprintf(stderr, "\n# hib[%d] DMA read (host -> HIB)\n", devid);
    rbuf = h->dmar_buf;
    wbuf = h->dmaw_buf;

    for (i = 0; i < (h->dmabuf_bytes>>3); i++) {
	rbuf[i] = 0x123456789abc0000ll|i;
	wbuf[i] = 0xfedcba9876540000ll|i;
	//      fprintf(stderr, "rbuf[0x%02x]: 0x%016llx\n", i, rbuf[i]);
    }

    nloop = 1e7;

    fprintf(stderr, "First test call to DMAW interface start\n");

    
    if (1)    {
	int i;
	for (i=0;i<64; i++){
	    hib_mem_writeMC(devid, 0x60, i);
	    hib_mem_writeMC(devid, 0x64, i*3);
	}
	hib_set_dma_extended_regMC(devid, 0, 0, 0x01f);
	hib_register_dma_seqMC(devid, 0, 0, 128, rbuf);
	
	for (i=32;i<48; i++){
	    hib_mem_writeMC(devid, 0x60, i);
	    fprintf(stderr, "extended register  0x%08lx: 0x%08x\n",
		    i, hib_mem_readMC(devid, 0x60));
	}
	int addr;
	hib_mem_writeMC(devid, 0x10, 0x1234);
	hib_mem_writeMC(devid, 0x14, 0x5678);
	hib_mem_writeMC(devid, 0x18, 0xabcd);
	usleep(100);
	hib_start_dma_seqMC(devid, 0, 0);
	usleep(100);
	fprintf(stderr, "Now wait for dma to finish?\n");
	hib_finish_dmarMC(devid);
	fprintf(stderr, "DMA finished\n");
	
	for(addr=0x10;addr < 0x20; addr+= 4){
	    fprintf(stderr, "hib[%d] 0x%08lx: 0x%08x\n",
		    devid, addr, hib_mem_readMC(devid, addr));
	}
    }

    int maxreqsize;
    for(maxreqsize=1; maxreqsize<7; maxreqsize++){
	hib_config_setmaxrequestMC(0, maxreqsize);
	printf("Max req size set to %d\n", maxreqsize);
	for (size = 128; size <= 4096; size *= 2) {
	    for (ntry = 0; ntry < 1; ntry++) {
		get_cputime(&lt, &st);
		
		for (j = 0; j < nloop/size; j++) {
		    hib_start_dmarMC(devid, size, rbuf);
		    hib_finish_dmarMC(devid);
		}
		get_cputime(&lt, &st);
		printf("size: %d DMA read: %f sec  %f MB/s\n",
		       size*8, lt, sizeof(UINT64)*nloop/1e6/lt);
	    }
	}
	printf("# With interal hibreg used\n", devid);

	for (size = 128; size <= 8192; size *= 2) {
	    for (ntry = 0; ntry < 1; ntry++) {
		hib_register_dma_seqMC(devid, 0, 0, size, rbuf);
		get_cputime(&lt, &st);
		
		for (j = 0; j < nloop/size; j++) {
		    hib_start_dma_seqMC(devid, 0, 0);
		    hib_finish_dmarMC(devid);
		}
		get_cputime(&lt, &st);
		printf("size: %d DMA read: %f sec  %f MB/s\n",
		       size*8, lt, sizeof(UINT64)*nloop/1e6/lt);
	    }
	}
    }

    // wait long enough before unset test mode
    // so that PIOW burst is completed.
    usleep(10);
    hib_set_test_modeMC(devid, TESTMODE_NONE);
    hib_closeMC(devid);
}

static void dmarperfMC(int ndevs, int firstdev, int maxreqmax)
{
    Hib *h[MAXDEVS];
    UINT64 *wbuf[MAXDEVS], *rbuf[MAXDEVS];
    int sendfunc;
    int i, j, k, ntry, nng, size0, off;
    int size; /* in 32-bit words */
    double lt = 0.0, st = 0.0;
    double nloop;
    int enddev = firstdev+ndevs;
    for(j=firstdev;j<enddev;j++){
    
	h[j] = hib_openMC(j);
	hib_reset_dma(j);

	hib_set_test_modeMC(j, TESTMODE_REFDESIGN_RAM);
	
	fprintf(stderr, "\n# hib[%d] DMA read (host -> HIB)\n", j);
	rbuf[j] = h[j]->dmar_buf;
	wbuf[j] = h[j]->dmaw_buf;
	fprintf(stderr, "buffsize = %d\n",h[j]->dmabuf_bytes);

	for (i = 0; i < ((h[j]->dmabuf_bytes)>>3); i++) {
	    rbuf[j][i] = 0x123456789abc0000ll|i;
	    wbuf[j][i] = 0xfedcba9876540000ll|i;
	    //      fprintf(stderr, "rbuf[0x%02x]: 0x%016llx\n", i, rbuf[i]);
	}
    }

    nloop = 1e7;

    fprintf(stderr, "First test call to DMAW interface start\n");

    int maxreqsize;
    for(maxreqsize=1; maxreqsize<maxreqmax; maxreqsize++){
	for(k=firstdev;k<enddev;k++) hib_config_setmaxrequestMC(k, maxreqsize);
	printf("Max req size set to %d\n", maxreqsize);
	for (size = 128; size <= 8192; size *= 2) {
	    for (ntry = 0; ntry < 1; ntry++) {
		get_cputime(&lt, &st);
		
		for (j = 0; j < nloop/size; j++) {
		    for(k=firstdev;k<enddev;k++) hib_start_dmarMC(k, size, rbuf);
		    for(k=firstdev;k<enddev;k++) hib_finish_dmarMC(k);
		}
		get_cputime(&lt, &st);
		printf("size: %d DMA read: %f sec  %f MB/s\n",
		       size*8, lt, sizeof(UINT64)*nloop*ndevs/1e6/lt);
	    }
	}
	printf("# With interal hibreg used\n" );

	for (size = 128; size <= 16384; size *= 2) {
	    for (ntry = 0; ntry < 1; ntry++) {
		for(k=firstdev;k<enddev;k++) hib_register_dma_seqMC(k, 0, 0, size-1, rbuf[k]);
		get_cputime(&lt, &st);
		
		for (j = 0; j < nloop/size; j++) {
		    for(k=firstdev;k<enddev;k++) hib_start_dma_seqMC(k, 0, 0);
		    //for(k=firstdev;k<enddev;k++) {
		    //			hib_finish_dmarMC(k);
		    //			hib_start_dma_seqMC(k, 0, 0);
		    //		    }
		    for(k=firstdev;k<enddev;k++) hib_finish_dmarMC(k);
		}
		get_cputime(&lt, &st);
		printf("size: %d DMA read: %f sec  %f MB/s\n",
		       size*8, lt, sizeof(UINT64)*nloop*ndevs/1e6/lt);
	    }
	}
    }

    // wait long enough before unset test mode
    // so that PIOW burst is completed.
    usleep(10);
    for(k=firstdev;k<enddev;k++){
	// test to fix page table structure
	for (i = 0; i < (h[k]->dmabuf_bytes>>3); i++) {
	    rbuf[k][i] = 0x123456789abc0000ll|i;
	    wbuf[k][i] = 0xfedcba9876540000ll|i;
	}
    }
    for(k=firstdev;k<enddev;k++){
	usleep(10);
	hib_set_test_modeMC(k, TESTMODE_NONE);
	usleep(10);
	hib_closeMC(k);
    }
}

static void dmaperf(int devid, int stride)
{
    Hib *h;
    UINT64 *wbuf, *rbuf;
    int sendfunc;
    int i, j, ntry, nng, size0, off;
    int size; /* in 32-bit words */
    double lt = 0.0, st = 0.0;
    double nloop;
    h = hib_openMC(devid);
    hib_reset_dma(devid);
    hib_set_test_modeMC(devid, TESTMODE_REFDESIGN_FIFO);

    rbuf = h->dmar_buf;
    wbuf = h->dmaw_buf;


    nloop = 1e7;

    for (size = 32; size <= 4096; size *= 2) {
	for (i = 0; i < size; i++) {
	    rbuf[i] = 0x123456789abc0000ll|(i*stride);
	    wbuf[i] = 0xfedcba9876540000ll|((h->dmabuf_bytes>>3)-i);
	}
	//	printf("\n Write buffer before DMA\n");
	//	for(j=0;j<size;j++){
	//	    printf(" %6d", wbuf[j]&0xffff);
	//	    if ( (j%10)==9) printf("\n");
	//	}
	//	    printf("\n");
	for (ntry = 0; ntry < 1; ntry++) {
	    get_cputime(&lt, &st);
	    hib_set_dma_extended_regMC(devid, 1, 0x0, 0x01f);
	    hib_register_dma_seqMC(devid, 0, 0, size, rbuf);
	    hib_register_dma_seqMC(devid, 0, 1, size, wbuf);
	    for (j = 0; j < nloop/size; j++) {
		//		hib_start_dmarMC(devid, size, rbuf);
		hib_start_dma_seqMC(devid, 0, 0);
		//		hib_start_dmawMC(devid, size, wbuf);
		//		xwait(1000);
		hib_start_dma_seqMC(devid, 0, 1);
		//		hib_finish_dmarMC(devid);
		hib_finish_dmawMC(devid);
	    }
	    get_cputime(&lt, &st);
	    printf("size: % 5d DMA read & DMA write: % 4.1f sec  % 7.2f MB/s\n",
		   size*8, lt, 2*sizeof(UINT64)*nloop/1e6/lt);
	    //	    for(j=0;j<size;j++){
	    //		printf(" %6d", wbuf[j]&0xffff);
	    //		if ( (j%10)==9) printf("\n");
	    //	    }
	    //	    printf("\n");
	}
    }
    // wait long enough before unset test mode
    // so that PIOW burst is completed.
    usleep(10);
    hib_set_test_modeMC(devid, TESTMODE_NONE);
    hib_closeMC(devid);
}
static void dmaperfMC(int firstcard, int ncards, int stride)
{
    Hib *h[MAXDEVS];
    UINT64 *wbuf[MAXDEVS], *rbuf[MAXDEVS];
    int sendfunc;
    int i, j, k, ntry, nng, size0, off;
    int size; /* in 32-bit words */
    double lt = 0.0, st = 0.0;
    double nloop;
    int endcard = firstcard+ncards;
    for(j=firstcard; j<endcard; j++){
	h[j] = hib_openMC(j);
	hib_reset_dma(j);
	hib_set_test_modeMC(j, TESTMODE_REFDESIGN_FIFO);
	rbuf[j] = h[j]->dmar_buf;
	wbuf[j] = h[j]->dmaw_buf;
    }

    fprintf(stderr,"ncards = %d\n", ncards);
    nloop = 2e7;

    for (size = 32; size <= 8192; size *= 2) {
	for(j=firstcard; j<endcard; j++){
	    for (i = 0; i < size; i++) {
		rbuf[j][i] = 0x123456789abc0000ll|(i*stride);
		wbuf[j][i] = 0xfedcba9876540000ll|((h[j]->dmabuf_bytes>>3)-i);
	    }
	}
	for (ntry = 0; ntry < 1; ntry++) {
	    for(j=firstcard; j<endcard; j++){
		hib_set_dma_extended_regMC(j, 1, 0x0, 0x01f);
		hib_register_dma_seqMC(j, 0, 0, size, rbuf[j]);
		hib_register_dma_seqMC(j, 0, 1, size, wbuf[j]);
	    }
	    get_cputime(&lt, &st);
	    for (j = 0; j < nloop/size; j++) {
		//		hib_start_dmarMC(devid, size, rbuf);
		for(k=firstcard; k<endcard; k++) hib_start_dma_seqMC(k, 0, 1);
		for(k=firstcard; k<endcard; k++) hib_start_dma_seqMC(k, 0, 0);
		for(k=firstcard; k<endcard; k++) hib_finish_dmawMC(k);
	    }
	    get_cputime(&lt, &st);
	    printf("size: % 5d DMA read & DMA write: % 4.1f sec  % 7.2f MB/s\n",
		   size*8, lt, ncards*2*sizeof(UINT64)*nloop/1e6/lt);
	}
    }
    // wait long enough before unset test mode
    // so that PIOW burst is completed.
    usleep(10);
    for(k=firstcard; k<endcard; k++) {
	hib_set_test_modeMC(k, TESTMODE_NONE);
	hib_closeMC(k);
    }
}

static void dump_dmarbuf(int devid, int ndata)
{
    int i;
    int x0, x1;
    char * regname[4]={"indata_count","outdata_count,", "size_reg", "dbmask"};
    for(i=0x200;i<0x204;i++){
	hib_mem_writeMC(devid, 0x00, i);
	x0=hib_mem_readMC(devid, 0x04);
	printf("dmarbuf regs: %-16s %8x\n", regname[i-0x200], x0);
    }
    for(i=0;i<ndata;i++){
	hib_mem_writeMC(devid, 0x00, i);
	x1=hib_mem_readMC(devid, 0x4);
	printf("dmarbuf memory: %4x %8x\n", i, x1);
    }
}
	
	
static void dmatest(int devid, int stride, int nloop, int dmasize)
{
    Hib *h;
    UINT64 *wbuf, *rbuf;
    int sendfunc;
    int i, j, ntry, nng, size0, off;
    int size; /* in 32-bit words */
    double lt = 0.0, st = 0.0;
    h = hib_openMC(devid);
    hib_reset_dma(devid);
    hib_set_test_modeMC(devid, TESTMODE_REFDESIGN_FIFO);

    rbuf = h->dmar_buf;
    wbuf = h->dmaw_buf;

    srandom(1);

    size = dmasize;
    hib_set_dma_extended_regMC(devid, 1, 0x3f, 0x01f);
    hib_register_dma_seqMC(devid, 0, 0, size, rbuf);
    hib_register_dma_seqMC(devid, 0, 1, size, wbuf);
    int nerror=0;
    for (ntry = 0; ntry < nloop; ntry++) {
	if ((nloop <= 10)||((ntry %(nloop/10))== 0))printf("test %d\n", ntry);
	if (ntry == 0){
	    for (i = 0; i < size; i++) {
		rbuf[i] = 0x023456789abc0000ll|((i+1)*stride+ntry);
		wbuf[i] = 0xfedcba9876540000ll|((h->dmabuf_bytes>>3)-i);
	    }
	}else{
	    for (i = 0; i < size; i++) {
		UINT64 r0, r1;
		r0 = random();
		r1 = random();
		rbuf[i] = (r0 << 32)|r1;
		wbuf[i] = 0;
	    }
	    
	}
	//	printf("\n Write buffer before DMA\n");
	//	for(j=0;j<size;j++){
	//	    printf(" %16lx", wbuf[j]);
	//	    if ( (j%4)==3) printf("\n");
	//	}
	//	printf("\n");
	//		hib_start_dmarMC(devid, size, rbuf);
	hib_start_dma_seqMC(devid, 0, 0);
	//		hib_start_dmawMC(devid, size, wbuf);
	xwait(1000);
	hib_start_dma_seqMC(devid, 0, 1);
	if (nloop == 1){
	    // register dump here for the case of single test
	    xwait(1000);
	    int addr;
	    for(addr=0x00;addr < 0x40; addr+= 4){
		fprintf(stderr, "hib[%d] 0x%08lx: 0x%08x\n",
			devid, addr, hib_mem_readMC(devid, addr));
	    }
	}
	    
	hib_finish_dmarMC(devid);
	if (nloop == 1){
	    fprintf(stderr,"dma read end\n");
	    for(j=0;j<size;j++){
		printf(" %16lx", wbuf[j]);
		if ( (j%4)==3) printf("\n");
	    }
	    printf("\n");
	    dump_dmarbuf(devid, size);
	}

	hib_finish_dmawMC(devid);
	if (nloop == 1)fprintf(stderr,"dma write end\n");
	//	for(j=0;j<size;j++){
	//	    printf(" %16lx", wbuf[j]);
	//	    if ( (j%4)==3) printf("\n");
	//	}
	//	printf("\n");
	for (i = 0; i < size; i++) {
	    if(rbuf[i] != wbuf[i]){
		nerror++;
		printf("loop %d, location %d, %16lx != %16lx\n",
		       j,i,rbuf[i],wbuf[i]);
	    }
	}
	
    }
    printf("test end with %d error(s)\n", nerror);
    if(0){
	hib_set_dma_extended_regMC(devid, 0, 0x0, 0x01f);
	for (i = 0; i < size; i++) {
	    rbuf[i] = 0x023456789abc0000ll|((i+1));
	}
	hib_start_dma_seqMC(devid, 0, 0);
	xwait(1000);
	hib_start_dma_seqMC(devid, 0, 1);
	hib_finish_dmarMC(devid);
	hib_finish_dmawMC(devid);
	printf("Normal DMAR/DMAW without using dmarbuf\n");
	for(j=0;j<size;j++){
	    printf(" %16lx", wbuf[j]);
	    if ( (j%4)==3) printf("\n");
	}
	printf("\n");
    }
    if(nloop==1){
	int addr;
	for(addr=0x00;addr < 0x40; addr+= 4){
	    fprintf(stderr, "hib[%d] 0x%08lx: 0x%08x\n",
		    devid, addr, hib_mem_readMC(devid, addr));
	}
    }
    hib_set_dma_extended_regMC(devid, 0, 0x0, 0x01f);
    if(nloop==1)dump_dmarbuf(devid, size+16);
    // wait long enough before unset test mode
    // so that PIOW burst is completed.
    usleep(10);
    hib_set_test_modeMC(devid, TESTMODE_NONE);
    hib_closeMC(devid);
}

static void showstatus(int devid)
{
    Hib *h;
    UINT64 *wbuf, *rbuf;
    int i, j, busmode;
    double freq;
    int hib0, nhibs;

    h = hib_openMC(devid);

    fprintf(stderr, "## hib%d:\n", devid);
    UINT32 val, gen;
    
    fprintf(stderr, "protocol : PCIe\n");
    
    // link width
    val = hib_config_readMC(devid, 0x90) >> 4+16;
    val = val & 0x3f;
    fprintf(stderr, "link width negotiated : x%d\n", val);
    val = hib_config_readMC(devid, 0x8c) >> 4;
    val = val & 0x3f;
    fprintf(stderr, "            supported : x%d\n", val);
    
    // link speed
    val = hib_config_readMC(devid, 0x90) >> 0+16;
    val = val & 0xf;
    fprintf(stderr, "link speed negotiated : %3.1f Gb/s\n", 2.5 * val);
    val = hib_config_readMC(devid, 0x8c) & 0xf;
    fprintf(stderr, "           supported  : %3.1f Gb/s\n", 2.5 * val);
    
    // payload size
    val = hib_config_readMC(devid, 0x88) >> 5;
    val = val & 0x7;
    val = 128 << val;
    fprintf(stderr, "max payload size negotiated : %d byte\n", val);
    val = hib_config_readMC(devid, 0x84) & 0x7;
    val = 128 << val;
    fprintf(stderr, "                 supported  : %d byte\n", val);
    
    // read request size
    val = hib_config_readMC(devid, 0x88) >> 12;
    val = val & 0x7;
    val = 128 << val;
    fprintf(stderr, "max read request size : %d byte\n", val);
    fprintf(stderr, "\n");
    fprintf(stderr, "configuration register:\n");
    for (i = 0; i < 16; i++) {
	if (4 <= i && i <=7) {
	    fprintf(stderr, "0x%08x: 0x%08x 0x%08x\n",
		    i*4, hib_config_readMC(devid, i*4),
		    (hib_config_readMC(devid, i*4)>>4)<<4);
	}
	else {
	    fprintf(stderr, "0x%08x: 0x%08x\n",
		    i*4, hib_config_readMC(devid, i*4));
	}
    }
    fprintf(stderr, "PCI Express Capability Register:\n");
    for (i = 0x80; i <= 0x90; i += 4) {
	fprintf(stderr, "0x%08x: 0x%08x\n",
		i, hib_config_readMC(devid, i));
    }
    int addr;
    fprintf(stderr, "\nHIB local registers:\n");
	
    for(addr=0x00;addr < 0x70; addr+= 4){
	fprintf(stderr, " 0x%08lx: 0x%08x\n", addr,
		hib_mem_readMC(devid, addr));
    }
    
    fprintf(stderr, "\nHIB dmareg registers:\n");
	
    for (i=0;i<64; i++){
	hib_mem_writeMC(devid, 0x60, i);
	if ((i % 4) == 0)fprintf(stderr, " 0x%08lx: ", i);
	fprintf(stderr, " 0x%08x", hib_mem_readMC(devid, 0x64));
	if ((i % 4) == 3)fprintf(stderr, "\n");
    }
    fprintf(stderr, "\nCIB extended status:\n");
	
    for (i=0;i<4; i++){
	hib_mem_writeMC(devid, 0x0, i);
	fprintf(stderr, " misc:0x%08lx: 0x%016lx\n", i, hib_mem_read64MC(devid, 0x60));
    }
    for (i=4;i<15; i++){
	hib_mem_writeMC(devid, 0x0, i);
	fprintf(stderr, " misc:0x%08lx: 0x%08x\n", i, hib_mem_readMC(devid, 0x60));
    }
    dump_dmarbuf(devid,0);
}

static void dump_syncstatus(int ndevs)
{
    int i;
    for(i=0;i<ndevs;i++){
	hib_mem_writeMC(i, 0x0, 0x10);
	printf( " [%d]= %02x", i, hib_mem_readMC(i, 0x60));
    }
    printf("\n");
}
    
static void testsync()
{
    int ndevs = 4;
    Hib *h[MAXDEVS];
    int i, j, k;
    int iboard, mode;
    int firstdev=0;
    int enddev=ndevs;
    iboard = 0;
    double nloop;
    
    for(j=0;j<ndevs;j++){
	h[j] = hib_openMC(j);
    }
    while(iboard >= 0){
	fprintf(stderr,"Enter cardid, mode to set (cardid=-1 to finish):");
	scanf("%d%d",&iboard, &mode);
	if (iboard >= 0){
	    if (mode){
		hib_mem_writeMC(iboard, 0x20, 1 << 27);
	    }else{
		hib_mem_writeMC(iboard, 0x20, 0);
	    }
	    printf("card[%d]  %d ", iboard, mode);
	}
	usleep(100);
	dump_syncstatus(ndevs);
	for(i=0;i<ndevs;i++)hib_dump_sreg(i);
	usleep(100);
    }
    
    for(k=firstdev;k<enddev;k++){
	usleep(1000);
	hib_closeMC(k);
	usleep(1000);
    }
}


static void
get_cputime(double *splittime, double *laptime)
{
    struct timeval x;

    gettimeofday(&x, NULL);

    *splittime = x.tv_sec + x.tv_usec/1000000.0 - *laptime;
    *laptime = x.tv_sec + x.tv_usec/1000000.0;
}

typedef struct parmstruct{
    int testtype;
    int cardid;
    int ncards;
    int stride;
    int nloop;
    int dmasize;
    int maxreqmax;
} PARMS, *PPARMS;

void usage()
{
    fprintf(stderr,"dmatest options:\n");
    fprintf(stderr,"  -h: This help\n");
    fprintf(stderr,"  -t: type of test (default=0)\n");
    fprintf(stderr,"      0: dmaw single-card test\n");
    fprintf(stderr,"      1: dmaw multi-card test\n");
    fprintf(stderr,"      2: dmar single card test\n");
    fprintf(stderr,"      3: dmar multi-card test\n");
    fprintf(stderr,"      4: read-write test single-card test\n");
    fprintf(stderr,"      5: read-write test multi-card test\n");
    fprintf(stderr,"      6: call testfunc (for hardware test)\n");
    fprintf(stderr,"      7: call dmatest  (for hardware test)\n");
    fprintf(stderr,"      8: show status\n");
    fprintf(stderr,"      9: test sync operation\n");
    fprintf(stderr,"      101-115: send reset\n");
    fprintf(stderr,"         reset code 1: transiever related logic\n");
    fprintf(stderr,"         reset code 2: transievers\n");
    fprintf(stderr,"         reset code 4: backend pipeline\n");
    fprintf(stderr,"         reset code 8: DMA regs\n");
    fprintf(stderr,"  -i: cardid (default=0) for single-card test\n");
    fprintf(stderr,"  -n: number of cards (default=1) for multi-card test\n");
    fprintf(stderr,"  -s: stride for data in DMAR buffer(default=1)\n");
    fprintf(stderr,"  -l: loop count for dmatest)\n");
    fprintf(stderr,"  -m: dma size for dmatest\n");
    fprintf(stderr,"  -r: max value for max read req size (default=3)\n");
}


void print_parms(FILE* stream, PPARMS parms)
{
    fprintf(stream,"Type=%d cardid=%d ncards=%d stride=%d\n",
	    parms->testtype, parms->cardid, parms->ncards, parms->stride);
    fprintf(stream,"Nlloop=%d dma_size=%d max_req_max=%d\n",
	    parms->nloop, parms->dmasize,parms->maxreqmax);
}

void read_parms(int argc, char * argv[], PPARMS parms)
{
    int ch;
    static struct option longopts[] = {
	{ "help",      no_argument,      0,           'h' },
	{ "test_type",   required_argument,      NULL,           't' },
	{ "card_id",   required_argument,      NULL,           'i' },
	{ "n_cards",   required_argument,      NULL,           'n' },
	{ "stride",   required_argument,      NULL,           's' },
	{ "nloop",   required_argument,      NULL,           'l' },
	{ "dmasize",   required_argument,      NULL,           'm' },
	{ "maxreqmax", required_argument,      NULL,           'r' },
	{ NULL,         0,                      NULL,           0 }
    };
    parms->testtype=0;
    parms->cardid=0;
    parms->ncards=1;
    parms->stride=1;
    parms->nloop=1;
    parms->dmasize=128;
    parms->maxreqmax=3;
    while((ch=getopt_long(argc,argv,"hi:l:m:n:r:s:t:",longopts, NULL))!= -1){
	fprintf(stderr,"optchar = %c optarg=%s\n", ch,optarg);
	switch (ch) {
	    case 't': parms->testtype = atoi(optarg); break;
	    case 'i': parms->cardid = atoi(optarg); break;
	    case 'n': parms->ncards = atoi(optarg); break;
	    case 's': parms->stride = atoi(optarg); break;
	    case 'l': parms->nloop  = atoi(optarg); break;
	    case 'm': parms->dmasize= atoi(optarg); break;
	    case 'r': parms->maxreqmax= atoi(optarg); break;
	    case 'h': usage(); exit(1);
	    case '?':usage(); exit(1);
		break;
	    default:break;
	}
    }
    argc -= optind;
    argv += optind;
    print_parms(stderr, parms);
    print_parms(stdout, parms);
}

int main(int argc, char *argv[])
{
    int i,n;
    PARMS parms;
    read_parms(argc, argv, &parms);
    i=parms.cardid;
    n=parms.ncards;

    if (parms.testtype == 0){
	dmawperf(i);
    }else if (parms.testtype == 1){
	dmawperfMC(n,i);
    }else if (parms.testtype == 2){
	dmarperf(i);
    }else if (parms.testtype == 3){
	dmarperfMC(n,i,parms.maxreqmax+1);
    }else if (parms.testtype == 4){
	dmaperf(i,parms.stride);
    }else if (parms.testtype == 5){
	dmaperfMC(i,n,parms.stride);
    }else if (parms.testtype == 6){
	testfunc(i);
    }else if (parms.testtype == 7){
	dmatest(i,parms.stride,parms.nloop, parms.dmasize);
    }else if (parms.testtype == 8){
	showstatus(i);
    }else if (parms.testtype == 9){
	testsync();
    }else if ((parms.testtype >100) &&(parms.testtype <116) ){
	reset_card(i,parms.testtype-100);
    }
}

