#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#include <math.h>
#include "hibutil.h"

#include <xmmintrin.h>
#include <emmintrin.h>

#define MAXDEVS 16

static UINT64 *piowbuf;
static UINT64 *piowbuf2[NHIB];

static void dmawperf(int devid);
static void dmawperfMC(int ndevs);
static void dmarperf(int argc, char **argv);
static void resetbackend(int argc, char **argv);
static void rawperf(int argc, char **argv);
static void dmaperf2(int argc, char **argv);
static void dmawperf2(int argc, char **argv);
static void dmarperf2(int argc, char **argv);

static void eraserom(int argc, char **argv);
static void writerom(int argc, char **argv);
static void readromid(int argc, char **argv);

static void writepllconf(int argc, char **argv);

static void get_cputime(double *laptime, double *sprittime);


static void dmawperf(int devid)
{
    Hib *h;
    UINT64 *wbuf, *rbuf;
    int i, j, ntry, nng, size0, off;
    int size; /* in 32-bit words */
    double lt = 0.0, st = 0.0;
    double nloop;

    h = hib_openMC(devid);
    hib_set_test_modeMC(devid, TESTMODE_REFDESIGN_RAM);
    rbuf = h->dmar_buf;
    wbuf = h->dmaw_buf;

    for (i = 0; i < (h->dmabuf_bytes>>3); i++) {
	rbuf[i] = 0x123456789abc0000ll|i;
	wbuf[i] = 0xfedcba9876540000ll|i;
	//      fprintf(stderr, "rbuf[0x%02x]: 0x%016llx\n", i, rbuf[i]);
    }

    nloop = 1e7;
    fprintf(stderr, "First test call to DMAW interface start\n");

    
    if (1)    {
	int i;
	for (i=0;i<64; i++){
	    hib_mem_writeMC(devid, 0x60, i);
	    hib_mem_writeMC(devid, 0x64, i*3);
	}
	hib_set_dma_extended_regMC(devid, 0, 0, 0x01f);
	hib_register_dma_seqMC(devid, 0, 1, 128, wbuf);
	
	//	hib_start_dma_seqMC(devid, 0, 1);
	//	hib_mem_writeMC(devid, 0x68, 2);
	for (i=32;i<48; i++){
	    hib_mem_writeMC(devid, 0x60, i);
	    fprintf(stderr, "extended register  0x%08lx: 0x%08x\n",
		    i, hib_mem_readMC(devid, 0x60));
	}
	int addr;
	hib_mem_writeMC(devid, 0x10, 0x1234);
	hib_mem_writeMC(devid, 0x14, 0x5678);
	hib_mem_writeMC(devid, 0x18, 0xabcd);
	usleep(100);
	hib_start_dma_seqMC(devid, 0, 1);
	usleep(100);
	fprintf(stderr, "Now wait for dma to finish?\n");
	hib_finish_dmawMC(devid);
	fprintf(stderr, "DMA finished (not started..)\n");
	
	for(addr=0x10;addr < 0x20; addr+= 4){
	    fprintf(stderr, "hib[%d] 0x%08lx: 0x%08x\n",
		    devid, addr, hib_mem_readMC(devid, addr));
	}
	// wait long enough before unset test mode
	// so that PIOW burst is completed.
	//	usleep(10);
	//	hib_set_test_modeMC(devid, TESTMODE_NONE);
	//	hib_closeMC(devid);
	//	exit(0);
	hib_start_dmawMC(devid, 128, wbuf);
	hib_finish_dmawMC(devid);
	fprintf(stderr, "First test call to DMAW interface end\n");
    }

    fprintf(stderr, "\n# hib[%d] DMA write (host <- HIB)\n", devid);
    //    for (size = 128; size <= 8192; size *= 2) {
    hib_set_dma_extended_regMC(devid, 0, 0, 0x01f);
    for (size = 128; size <= 8192; size *= 2) {
        for (ntry = 0; ntry < 1; ntry++) {
	    get_cputime(&lt, &st);
            for (j = 0; j < nloop/size; j++) {
		hib_start_dmawMC(devid, size, wbuf);
		hib_finish_dmawMC(devid);

	    }
	    get_cputime(&lt, &st);
	    printf("size: %d DMA write: %f sec  %f MB/s\n",
		   size*8, lt, sizeof(UINT64)*nloop/1e6/lt);
	}
    }

    if(0){
	fprintf(stderr, "# With interal hibreg used\n", devid);
	for (size = 128; size <= 8192; size *= 2) {
	    for (ntry = 0; ntry < 1; ntry++) {
		hib_register_dma_seqMC(devid, 0, 1, size, wbuf);
		get_cputime(&lt, &st);
		for (j = 0; j < nloop/size; j++) {
		    hib_start_dma_seqMC(devid, 0, 1);
		    hib_finish_dmawMC(devid);
		    
		}
		get_cputime(&lt, &st);
		printf("size: %d DMA write: %f sec  %f MB/s\n",
		       size*8, lt, sizeof(UINT64)*nloop/1e6/lt);
	    }
	}
    }

    // wait long enough before unset test mode
    // so that PIOW burst is completed.
    usleep(10);
    hib_set_test_modeMC(devid, TESTMODE_NONE);
    hib_closeMC(devid);
}

static inline UINT64 rdtsc2()
{
	UINT64 ret;
	asm __volatile__("rdtsc  \n\t" \
			 "shlq $32, %%rdx \n\t" \
			 "orq  %%rdx, %%rax \n\t" \
			 "movq %%rax, %0" : "=g"(ret) : :"rax", "rdx");
	return ret;
}

static void xwait(int wait)
{
    UINT64 st = rdtsc2();
    while((rdtsc2()-st)<wait) {}
}

static void dmawperfMC(int ndevs)
{
    Hib *h[MAXDEVS];
    UINT64 *wbuf[MAXDEVS], *rbuf[MAXDEVS];
    int i, j, k, ntry, nng, size0, off;
    int size; /* in 32-bit words */
    double lt = 0.0, st = 0.0;
    double nloop;

    for(j=0;j<ndevs;j++){
	h[j] = hib_openMC(j);
	hib_set_test_modeMC(j, TESTMODE_REFDESIGN_RAM);
	rbuf[j] = h[j]->dmar_buf;
	wbuf[j] = h[j]->dmaw_buf;
	
	for (i = 0; i < (h[j]->dmabuf_bytes>>3); i++) {
	    rbuf[j][i] = 0x123456789abc0000ll|i;
	    wbuf[j][i] = 0xfedcba9876540000ll|i;
	    //      fprintf(stderr, "rbuf[0x%02x]: 0x%016llx\n", i, rbuf[i]);
	}
    }
    nloop = 1e7;

    fprintf(stderr, "\n# hib[%d] DMA write (host <- HIB)\n", ndevs);
    for (size = 128; size <= 8192; size *= 2) {
	get_cputime(&lt, &st);
	for (j = 0; j < nloop/size; j++) {
	    for(k=0;k<ndevs;k++)hib_start_dmawMC(k, size, wbuf[k]);
	    xwait(5000);
	    for(k=0;k<ndevs;k++)hib_finish_dmawMC(k);
	    
	}
	get_cputime(&lt, &st);
	printf("size: %d DMA write: %f sec  %f MB/s\n",
	       size*8, lt, sizeof(UINT64)*ndevs*nloop/1e6/lt);
    }
    fprintf(stderr, "\n# With interal hibreg used, ndevs=%d\n", ndevs);

    if(1)for (size = 128; size <= 8192; size *= 2) {
	for (ntry = 0; ntry < 1; ntry++) {
	    for(k=0;k<ndevs;k++){
		hib_set_dma_extended_regMC(k, 0, 0, 0x01f);  
		usleep (100);
		hib_register_dma_seqMC(k, 0, 1, size, wbuf[k]);
		usleep (100);
	    }
	    get_cputime(&lt, &st);
	    for (j = 0; j < nloop/size; j++) {
		for(k=0;k<ndevs;k++) hib_start_dma_seqMC(k, 0, 1);
		xwait(5000);
		for(k=0;k<ndevs;k++) hib_finish_dmawMC(k);
	    }
	    get_cputime(&lt, &st);
	    printf("size: %d DMA write: %f sec  %f MB/s\n",
		   size*8, lt, sizeof(UINT64)*ndevs*nloop/1e6/lt);
	}
    }
    if (0){
	//
	// this part seems to cause kernel panic for ndevs=4
	//
	fprintf(stderr, "\n# With interal hibreg used, two DMAW issued\n");
	
	int kstart=0;
	int kend = ndevs;
	for (size = 128; size <= 2048; size *= 2) {
	    for (ntry = 0; ntry < 1; ntry++) {
		for(k=0;k<ndevs;k++){
		    hib_register_dma_seqMC(k, 0, 1, size, wbuf[k]);
		    hib_register_dma_seqMC(k, 1, 1, size, wbuf[k]+size);
		}
		get_cputime(&lt, &st);
		for (j = 0; j < nloop/size; j++) {
		    for(k=0;k<ndevs;k++) hib_start_dma_seqMC(k, 0, 1);
		    xwait(5000);
		    for(k=0;k<ndevs;k++) hib_start_dma_seqMC(k, 1, 1);
		    xwait(5000);
		    for(k=0;k<ndevs;k++) hib_finish_dmawMC(k);
		    
		}
		get_cputime(&lt, &st);
		printf("size: %d DMA write: %f sec  %f MB/s\n",
		       size*8, lt, sizeof(UINT64)*ndevs*2*nloop/1e6/lt);
	    }
	}
    }
    // wait long enough before unset test mode
    // so that PIOW burst is completed.
    usleep(100);
    for(k=0;k<ndevs;k++){
	fprintf(stderr,"Reset test mode for %d\n", k);
	hib_set_test_modeMC(k, TESTMODE_NONE);
	usleep(1000);
	hib_closeMC(k);
	usleep(1000);
    }
}

static void
get_cputime(double *splittime, double *laptime)
{
    struct timeval x;

    gettimeofday(&x, NULL);

    *splittime = x.tv_sec + x.tv_usec/1000000.0 - *laptime;
    *laptime = x.tv_sec + x.tv_usec/1000000.0;
}

int main(int argc, char **argv)
{
    int mode, i;


    //    dmawperf(3);
    dmawperfMC(4);
    exit (0);
}

