#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <hibutil.h>
#include <stdlib.h>
#include "ctrl.h"

#define CERR(errstr) fprintf(stderr, errstr); \
assert(0);

//#define CTRL_DEBUG
//#define VERBOSE

#define MODULE_HET     0
#define MODULE_HIT     1
#define MODULE_EIT     2
#define MODULE_COC     3
#define MODULE_OHT     4
#define MODULE_CONFIG  5

#define STATUS_EIT_BUSY             (1 << 0)
#define STATUS_COC_BUSY             (1 << 1)
#define STATUS_RESULTS_RETURNED     (1 << 2)
#define STATUS_EIT_DONE             (1 << 3)
#define STATUS_HET_FULL             (1 << 4)
#define STATUS_HET_IDLE             (1 << 5)

#define NCHIPMAX 8
static CTRLinfo *chip_state[NCHIPMAX];
static int active_chip = 0;

void setdqsdelay(CTRLinfo *c)
{
	UINT64 dqs , delayctl[5];
#if 0
	delayctl[0] = 0x14; //0x10; //0x10;    /* 6bit */  
	delayctl[1] = 0x14; //0x10; //0x10;    /* 6bit */
	delayctl[2] = 0x14; //0x10; //0x10;    /* 6bit */
	delayctl[3] = 0x12; //0x5; //0x5   /* 6bit */
	delayctl[4] = 0x08; //0x4; //0x00;    /* 6bit */
#else
	delayctl[0] = 0x20; //0x10; //0x10;    /* 6bit */  
	delayctl[1] = 0x20; //0x10; //0x10;    /* 6bit */
	delayctl[2] = 0x20; //0x10; //0x10;    /* 6bit */
	delayctl[3] = 0x20; //0x5; //0x5   /* 6bit */
	delayctl[4] = 0x16; //0x4; //0x00;    /* 6bit */
#endif

	dqs = (delayctl[0] << 24) |  (delayctl[1] << 18) |(delayctl[2] << 12) | (delayctl[3] << 6) | (delayctl[4] );
	CTRL_config_write(c, 9, &dqs, 1);
}

static inline UINT64 rdtsc2()
{
	UINT64 ret;
	asm __volatile__("rdtsc  \n\t" \
			 "shlq $32, %%rdx \n\t" \
			 "orq  %%rdx, %%rax \n\t" \
			 "movq %%rax, %0" : "=g"(ret) : :"rax", "rdx");
	return ret;
}


static void xwait(int wait)
{
    UINT64 st = rdtsc2();
    while((rdtsc2()-st)<wait) {}
}

CTRLinfo* CTRL_open(int id)
{
  CTRLinfo *c;
  //  fprintf( stderr, "CTRL_open id = %d\n", id);
  c = malloc(sizeof(CTRLinfo));
  c->h = hib_openMC(id);
  hib_set_sendfuncMC(id, SENDFUNC_PIOW);
  c->hib_id = id;
  c->buff_ptr = 0;
  c->buff_mode = 0;
  c->bcastflag = 0;
  if(c->h == NULL){
    free(c);
    return NULL;
  } else {
      CTRL_coc_reg_write(c,2,(UINT64)0);
    chip_state[active_chip] = c;
    active_chip++;
    return c;
  }
}

void CTRL_reset(CTRLinfo *c)
{
  /*
  hib_mem_writeMC(c->hib_id, c->h->dmastat, (1<<c->h->dmastat_reset_backend_bit));
  */
}

int CTRL_close(CTRLinfo *c)
{
  hib_closeMC(c->hib_id);
  free(c);
  return 0;
}

int CTRL_init(CTRLinfo *c)
{
  CTRL_reset(c);
  setdqsdelay(c);
  CTRL_coc_reset(c);
  CTRL_result_clear(c);
  return 0;
}

int CTRL_all_reset(CTRLinfo *c)
{
  return CTRL_init(c);
}

void CTRL_beginbcast(CTRLinfo *c)
{
  c->bcastflag = 1;
}

void CTRL_endbcast(CTRLinfo *c)
{
  c->bcastflag = 0;
}

static int dummy_val;

static int CTRL_write_flushmode = 1;

void CTRL_write_setflushmode(int mode)
{
    CTRL_write_flushmode = mode;
}

int CTRL_write(CTRLinfo *c, int module, UINT64 addr, const void *data, int len) 
{
    UINT64 *p = (UINT64 *)data;
    UINT64 header;
    int devid = c->hib_id;
    int i, j, size;
    Hib *hib = c->h;
    //int s = (hib->piowbuf_bytes>>3)-1;
    //int s=256;
    //int s=128;
    int s=384;
    //int s=512;

    if(c->bcastflag){
	header = ((UINT64)len << 40) | ((UINT64)1 << 39) | ((UINT64)module<<36) | addr;
    } else {
	header = ((UINT64)len << 40) | ((UINT64)module << 36) | addr; //For new hic header KK
    }
#if 0
    hib_sendMC(c->hib_id, 1, &header);
    hib_sendMC(c->hib_id, len, p);
#else
    size = len+1;
    //    fprintf(stderr, "id , len = %d , %d \n", devid, len);
    for (i = 0; i < size; i += s) {
	if (i + s > size) {
	    s = size - i;
	}
	//	fprintf(stderr, "i, s = %d , %d \n", i,s);

	if(i==0){
	    hib->piow_dblbuf[0] = header;
	    for (j = 0; j < s-1; j++) {
		hib->piow_dblbuf[j+1] = p[i + j];
	    }
	} else {
	    for (j = 0; j < s; j++) {
		hib->piow_dblbuf[j] = p[i + j -1];
	    }
	}
	hib_mem_writeMC(devid, hib->dma0misc,
			(1<<(hib->dma0misc_swap_sram_bit)) |
			(s<<(hib->dma0misc_sram_wcnt_bit)));

	//      _mm_mfence();
	//      usleep(100000);
	// write '1' to HIB local register 'swap_sram'
	// write data size to 'sram_wcnt'
	if (CTRL_write_flushmode){
	    hib_flush_piow_channel(devid);
	    hib_flush_piow_channel(devid);
	}
    }
#endif
#ifdef CTRL_DEBUG
    fprintf(stderr, "ctrl write : %d words. \n", len+1);
#endif
    return len;
}

int CTRL_write_shortmessage(CTRLinfo *c, int module, UINT64 addr,
			    const void *data,
			    int mode, int startp, int len) 
{
    UINT64 *p = (UINT64 *)data;
    UINT64 header;
    int devid = c->hib_id;
    int i, j, size;
    Hib *hib = c->h;

    if(c->bcastflag){
	header = ((UINT64)len << 40) | ((UINT64)1 << 39) | ((UINT64)module<<36) | addr;
    } else {
	header = ((UINT64)len << 40) | ((UINT64)module << 36) | addr; //For new hic header KK
    }
    
    hib->piow_dblbuf[startp] = header;
    for (j = 0; j < len; j++) {
	hib->piow_dblbuf[j+startp+1] = p[j];
    }
    if (mode){
	hib_mem_writeMC(devid, hib->dma0misc,
			(1<<(hib->dma0misc_swap_sram_bit)) |
			((startp+len+1)<<(hib->dma0misc_sram_wcnt_bit)));
	return 0;
    }
    return startp+len+1;
}

int CTRL_rawwrite_shortmessage(CTRLinfo *c,   const void *data, int startp, int len) 
{
    UINT64 *p = (UINT64 *)data;
    UINT64 header;
    int devid = c->hib_id;
    int i, j, size;
    Hib *hib = c->h;

    for (j = 0; j < len; j++) {
	hib->piow_dblbuf[j+startp] = p[j];
    }
    return startp+len;
}

int CTRL_write_usingdmar(CTRLinfo *c, int module, UINT64 addr, const void *data, int len) 
{
    UINT64 *p = (UINT64 *)data;
    UINT64 header;
    int devid = c->hib_id;
    int i, j, size;
    Hib *hib = c->h;
    if(c->bcastflag){
	header = ((UINT64)len << 40) | ((UINT64)1 << 39) | ((UINT64)module<<36) | addr;
    } else {
	header = ((UINT64)len << 40) | ((UINT64)module << 36) | addr; //For new hic header KK
    }
    hib_clear_dmarMC(devid);
    if(0){
	hib_register_data_dmarMC(devid, 1, &header);
	hib_register_data_dmarMC(devid, len, p);
	hib_start2_dmarMC(devid);
    }else{
	hib_register_data_dmarMC_multiplane(devid, 1, &header);
	hib_register_data_dmarMC_multiplane(devid, len, p);
	hib_start2_dmarMC_multiplane(devid);
    }	
    return len;
}
int CTRL_write_dummypacket(CTRLinfo *c)
// send a packet which is not used by any module
// a bug workaround for modules which cannot receive two packets without
// idle cycles -- JM
{
    int len=1;
    UINT64 data[len];
    UINT64 *p = data;
    UINT64 header;
    int devid = c->hib_id;
    int i, j, size;
    Hib *hib = c->h;
    for(i=0;i<len;i++)data[i]=0;
    
    header = ((UINT64)len << 40) | (((UINT64)(7)) << 36) ; 
    hib_register_data_dmarMC_multiplane(devid, 1, &header);
    hib_register_data_dmarMC_multiplane(devid, len, p);
}

void CTRL_write_usingdmar_startonly(CTRLinfo *c)
{
    hib_start2_dmarMC_multiplane(c->hib_id);
}

int dmar_buffered_mode = 0;

int CTRL_write_set_dmar_buffered_mode(int mode)
{
    dmar_buffered_mode = mode;
}


int CTRL_write_usingdmar_setbufonly(CTRLinfo *c, int module, UINT64 addr, const void *data, int len) 
{
    UINT64 *p = (UINT64 *)data;
    UINT64 header;
    int devid = c->hib_id;
    int i, j, size;
    Hib *hib = c->h;
    if(c->bcastflag){
	header = ((UINT64)len << 40) | ((UINT64)1 << 39) | ((UINT64)module<<36) | addr;
    } else {
	header = ((UINT64)len << 40) | ((UINT64)module << 36) | addr; //For new hic header KK
    }
    if (dmar_buffered_mode ==0)hib_clear_dmarMC(devid);
    hib_register_data_dmarMC_multiplane(devid, 1, &header);
    hib_register_data_dmarMC_multiplane(devid, len, p);
    CTRL_write_dummypacket(c);
    return len;
}


void CTRL_write_usingdmar_wait(CTRLinfo *c)
{
    int devid = c->hib_id;
    hib_finish_dmarMC(devid);
}


int CTRL_write_multichip_usingdmar(int nchips, CTRLinfo *c[nchips], int module,
				   UINT64 addr, const void *data, int len,
				   int increment) 
{
    int ichip;
    //    fprintf(stderr,"CTRL_write_multichip_usingdmar called\n");
    for(ichip=0;ichip<nchips;ichip++){
	UINT64 *p = ((UINT64 *)data)+ichip*increment;
	CTRL_write_usingdmar(c[ichip], module, addr, (void *) p, len);
    }
    for(ichip=0;ichip<nchips;ichip++){
	CTRL_write_usingdmar_wait(c[ichip]);
    }
    //    fprintf(stderr,"CTRL_write_multichip_usingdmar end\n");
    return len;
}
int CTRL_write_multichip_usingdmar_buffered_control(int nchips,
						    CTRLinfo *c[nchips],
						    int flag)
// flag = 0 : initialize
//        1 : finalize
{
    int ichip;
    fprintf(stderr,"dmar_buffer_control called %d\n", flag);
    for(ichip=0;ichip<nchips;ichip++){
	int devid = c[ichip]->hib_id;
	fprintf(stderr,"first loop ichip= %d status = %d %d\n", ichip,
		hib_status_dmarMC(devid),
		hib_bufferd_words_dmarMC(devid));
	if (flag == 1){
	    if ((hib_status_dmarMC(devid)== 0)&&
		(hib_bufferd_words_dmarMC(devid)>0)){
		hib_start2_dmarMC(devid);
	    }
	}
    }
    for(ichip=0;ichip<nchips;ichip++){
	int devid = c[ichip]->hib_id;
	fprintf(stderr,"second loop ichip= %d\n", ichip);
	hib_clear_dmarMC(devid);
    }
    fprintf(stderr,"return from dmar_control\n");
}    

int CTRL_write_multichip_usingdmar_simple_da(int nchips, CTRLinfo *c[nchips],
					     int module,
					     UINT64 addr,
					     void *data[nchips], int len)
{
    int ichip;
    //    static int icalls = 0;
    //    fprintf(stderr,"CTRL_write_multichip_usingdmar called for length= %d callid=%d\n", len,icalls);
    //    icalls++;

    for(ichip=0;ichip<nchips;ichip++){
	CTRL_write_usingdmar_setbufonly(c[ichip], module, addr, data[ichip], len);

       //	fprintf(stderr,"calling hib_finish_dmarMC %d\n", c[ichip]->hib_id);
    }
    for(ichip=0;ichip<nchips;ichip++){
	if (dmar_buffered_mode==0)CTRL_write_usingdmar_startonly(c[ichip]);
    }
    
    //    fprintf(stderr,"CTRL_write_multichip_usingdmar end\n");
    return len;
}
int CTRL_write_multichip_usingdmar_simple(int nchips, CTRLinfo *c[nchips], int module,
				   UINT64 addr, const void *data, int len,
				   int increment) 
{
    void * dataarray[nchips];
    int i;
    for(i=0;i<nchips;i++){
	dataarray[i] = (void*)(((UINT64 *)data)+i*increment);
    }
    //    fprintf(stderr,"call CTRL_write_multichip_usingdmar_simple_da\n");
    int ret= CTRL_write_multichip_usingdmar_simple_da(nchips, c,module, addr,
						    dataarray, len);
    return ret;

}

int CTRL_write_multichip_usingdmar_dataarray_start(int nchips, CTRLinfo *c[nchips],
						    int module, UINT64 addr,
						    const void *data[nchips], int len)
{
    int ichip;
    //    fprintf(stderr,"CTRL_write_multichip_usingdmar called\n");
    for(ichip=0;ichip<nchips;ichip++){
	CTRL_write_usingdmar(c[ichip], module, addr, data[ichip], len);
    }
    return 0;
}

int CTRL_write_multichip_usingdmar_dataarray_wait(int nchips, CTRLinfo *c[nchips])

{
    int ichip;
    for(ichip=0;ichip<nchips;ichip++){
	CTRL_write_usingdmar_wait(c[ichip]);
    }
    return 0;
}





int CTRL_write_start(CTRLinfo *c) 
{
  return 0;
}


int CTRL_config_write(CTRLinfo *c, int addr, const void *data, int len) 
{
  CTRL_write(c, MODULE_CONFIG, addr, data, len);
  CTRL_write_start(c);
  return 0;
}

int CTRL_config(CTRLinfo *c, const void *data, int len)
{
  CTRL_config_write(c, 0, data, len);
  return 0;
}

int CTRL_config_isp0(CTRLinfo *c, const void *data, int len) 
{
  CTRL_config_write(c, 2, data, len);
  return 0;
}

int CTRL_config_isp1(CTRLinfo *c, const void *data, int len) 
{
  CTRL_config_write(c, 3, data, len);
  return 0;
}

int CTRL_config_idp00(CTRLinfo *c, const void *data, int len) 
{
  CTRL_config_write(c, 4, data, len);
  return 0;
}

int CTRL_config_idp01(CTRLinfo *c, const void *data, int len) 
{
  CTRL_config_write(c, 5, data, len);
  return 0;
}

int CTRL_config_idp10(CTRLinfo *c, const void *data, int len) 
{
  CTRL_config_write(c, 6, data, len);
  return 0;
}

int CTRL_config_idp11(CTRLinfo *c, const void *data, int len) 
{
  CTRL_config_write(c, 7, data, len);
  return 0;
}

int CTRL_config_run(CTRLinfo *c) 
{
  UINT64 data = 0;
  CTRL_config_write(c, 8, &data, 1);
  return 0;
}

int CTRL_results_returned(CTRLinfo *c) 
{
  return (hib_mem_readMC(c->hib_id,  c->h->dma1misc) & 0x7ff) / 8;
}

int CTRL_number_of_results(CTRLinfo *c) 
{
  return (hib_mem_readMC(c->hib_id,  c->h->dma1misc) & 0x7ff)/8;
}

#define N 256

int CTRL_sing_chip_init(CTRLinfo *c) 
{
  UINT64 isp0[N], isp1[N];
  UINT64 idp00[4*N], idp01[4*N], idp10[4*N], idp11[4*N];
  UINT64 config[N];
  int i, n;

  n = 100;
  for (i = 0; i < n; i++) config[i] = 0x100000;
  config[n] = 0x100001;
  CTRL_config(c, config, n+1);
  CTRL_config_run(c);

  usleep(1000);

  isp0[0] = isp1[0] = 0;
  n = 150;
  for(i=0;i<n;i++){
  	isp0[i] = 0x0;
  	isp1[i] = 0x0;
  }
  CTRL_config_isp0(c, isp0, n);
  CTRL_config_isp1(c, isp1, n);

  for (i = 0; i < 100; i++) {
    if (0 == i % 2) {
      idp00[i] = idp10[i] = 0xffffffffffffffffULL;
      idp01[i] = idp11[i] = 0xff80000000000000ULL;
    } else {
      // really need data strobe?
      idp00[i] = idp10[i] = 0;
      idp01[i] = idp11[i] = 0x0080000000000000ULL;
    }
  }
  for (i = 100; i < n; i++) {
    // really need data strobe?
    idp00[i] = idp10[i] = 0;
    idp01[i] = idp11[i] = 0x0080000000000000ULL;
   }
  // need data strobe?
  idp00[n] = idp01[n] = idp10[n] = idp11[n] = 0;
  CTRL_config_idp00(c, idp00, n+1);
  CTRL_config_idp01(c, idp01, n+1);
  CTRL_config_idp10(c, idp10, n+1);
  CTRL_config_idp11(c, idp11, n+1);

  for (i = 0; i < 50; i++) config[i] = 0x100001;
  for (; i < 80; i++) config[i] = 0x100005; // 0x100003;
  /* for (; i < 80; i++) config[i] = 0x100003; */
  for (; i < n; i++) config[i] = 0x100007;
  config[n] = 0x000007;
  CTRL_config(c, config, n+1);
  CTRL_config_run(c);

  usleep(1000);

  return 0;

}

int CTRL_result_clear(CTRLinfo *c)
{
#if 0 //for old hib
  hib_mem_writeMC(c->hib_id,  c->h->dmastat, (1<<(c->h->dmastat_dma_reset_bit)));
#else //for new hib
  hib_mem_writeMC(c->hib_id,  c->h->command, (1<<(c->h->command_dma_reset_bit)));
#endif

  return 0;
}

int CTRL_read_as_many_as(CTRLinfo *c, void *data, int len) 
{
  int i;
  UINT64 *p = (UINT64 *)data;
  double *a;
  
  hib_dmawMC(c->hib_id, len, c->h->dmaw_buf);
  for (i = 0; i < len; i++){
    p[i] = c->h->dmaw_buf[i];
#ifdef CTRL_DEBUG
#ifdef VERBOSE
    a = (double*)(&(p[i]));
    fprintf(stderr, "%llx : %e\n", p[i], *a);
#endif
#endif
  }
  return hib_dmaw_timeout_error();
}

int CTRL_read_as_many_as_async_start(CTRLinfo *c, void *data, int len) 
{
  int i;
  UINT64 *p = (UINT64 *)data;
  double *a;
  
  hib_start_dmawMC(c->hib_id, len, c->h->dmaw_buf);
#if 0
  for (i = 0; i < len; i++){
    p[i] = c->h->dmaw_buf[i];
#ifdef CTRL_DEBUG
#ifdef VERBOSE
    a = (double*)(&(p[i]));
    fprintf(stderr, "%llx : %e\n", p[i], *a);
#endif
#endif
  }
#endif
  return 0;
}

int CTRL_read_as_many_as_async_end(CTRLinfo *c, void *data, int len) 
{
  int i;
  UINT64 *p = (UINT64 *)data;
  double *a;
 
  hib_finish_dmawMC(c->hib_id);
  for (i = 0; i < len; i++){
    p[i] = c->h->dmaw_buf[i];
#ifdef CTRL_DEBUG
#ifdef VERBOSE
    a = (double*)(&(p[i]));
    fprintf(stderr, "%llx : %e\n", p[i], *a);
#endif
#endif
  }
  return hib_dmaw_timeout_error();
}



int CTRL_read2(CTRLinfo *c, void *data) 
{
  UINT64 *p = (UINT64 *)data;
  int n;
  if (!CTRL_results_returned(c)) return -1;
  while (CTRL_results_returned(c)) {
    n = CTRL_number_of_results(c);
#ifdef CTRL_DEBUG
    fprintf(stderr, "CTRL_read2: %d results returned\n", n);
#endif
    CTRL_read_as_many_as(c, p, n);
    p += n;
  };
  return 0;
}

void loopback(CTRLinfo *c)
{
  UINT64 config[256], idp00[256], idp01[256], idp10[256], idp11[256], isp0[256], isp1[256];
  int i,nresults,  ch = 20;
  UINT64 result, *wbuf;
#if 0
  hib_mem_writeMC(c->hib_id,  c->h->dmastat, (1<<c->h->dmastat_dma_reset_bit));
#else
  hib_mem_writeMC(c->hib_id,  c->h->command, (1<<c->h->command_dma_reset_bit));
#endif
  usleep(1000);
  for(i=0;i<32;i++){
  	config[i] = 0x100407 | (ch<<4);
  	idp00[i] = 0xffffffffffffffff;
  	idp01[i] = 0x0080000000000000;
  	idp10[i] = 0xaa00000000000000;
  	idp11[i] = 0x0080000000000000;
  	isp0[i] = 0xf102030405060708;
  	isp1[i] = 0xf90a0b0c0d0e0f10;
  }
  for(;i<256;i++){
    idp00[i] = 0;
    idp01[i] = 0;
    idp10[i] = 0;
    idp11[i] = 0;
    isp0[i] = 0;
    isp1[i] = 0;
    config[i] = 0x100007 | (ch<<4);
  }
  CTRL_config(c, config, 256);
  CTRL_config_isp0(c, isp0, 256);
  CTRL_config_isp1(c, isp1, 256);
  CTRL_config_idp00(c, idp00, 256);
  CTRL_config_idp01(c, idp01, 256);
  CTRL_config_idp00(c, idp10, 256);
  CTRL_config_idp01(c, idp11, 256);
  CTRL_config_run(c);
  sleep(1);
  
  nresults = CTRL_results_returned(c);
  printf("results : %d \n", nresults);
  wbuf = c->h->dmaw_buf;
  hib_dmawMC(c->hib_id, nresults, wbuf);
  for(i=0;i<nresults;i++){
        result = wbuf[i];
  	printf("%llx \n", result);
  }
}

void reset_vector(CTRLinfo *c)
{
  UINT64 config[8];
  config[0] = 0x0;
  config[1] = 0x1;
  config[2] = 0x5;
  config[3] = 0x5;
  config[4] = 0x7;
  config[5] = 0x7;
  config[6] = 0x7;
  config[7] = 0x7;

  CTRL_write(c, MODULE_CONFIG, 0x0, config, 8);
  CTRL_write_start(c);
  CTRL_write(c, MODULE_CONFIG, 0x8, config, 1);
  CTRL_write_start(c);
}


int read_and_run(CTRLinfo *c) 
{
  UINT64 isp0[N], isp1[N];
  UINT64 idp00[4*N], idp01[4*N], idp10[4*N], idp11[4*N];
  UINT64 config[N];
  UINT64 *wbuf;
  static char strln[4096];
  static char cisp[1024];  
  static char cds[4][1024];
  static char cidp[4][1024];
  static char tmp[1000];
  static char odp[2][1024];
  static char odps[2][32];
  int i, ic, ii, nisp, nidp, nresults;

  //Stop hib & reset
#if 0 //for old hib
  hib_mem_writeMC(c->hib_id,  c->h->dmastat, (1<<c->h->dmastat_dma_reset_bit));
#else
  hib_mem_writeMC(c->hib_id,  c->h->command, (1<<c->h->command_dma_reset_bit));
#endif
  usleep(1000);

  nisp = nidp = 0;
  while(fgets(strln, 4090,stdin)!=NULL) {
    if ('#' == strln[0]) continue;

    sscanf(strln,"%s %s %s %s %s %s %s %s %s %s %s %s %s",cisp,cds[0],cidp[0],cds[1],cidp[1],
	   cds[2],cidp[2],cds[3],cidp[3], odps[0], odp[0], odps[1], odp[1]);
    
    ic = strlen(cisp);
    for (i = ic; i < 32; i++) cisp[i] = '0';
    cisp[32] = '\0';
    
    for(ii=0;ii<4;ii++){
      ic = strlen(cidp[ii]);
      for (i = ic; i < 32; i++) cidp[ii][i] = '0';
      cidp[ii][32] = '\0';
    }

    strncpy(tmp,cisp,16);
    sscanf(tmp,"%llx",&isp0[nisp]);

    strncpy(tmp,cisp+16,16);
    sscanf(tmp,"%llx",&isp1[nisp]);
    
    strncpy(tmp,cidp[0],16);
    sscanf(tmp,"%llx",&idp00[nidp]);
    
    strncpy(tmp,cidp[0]+16,16);
    sscanf(tmp,"%llx",&idp01[nidp]);
    if(strcmp(cds[0],"1")==0) idp01[nidp] |= 0x0080000000000000ULL;
    
    strncpy(tmp,cidp[1],16);
    sscanf(tmp,"%llx",&idp10[nidp]);
    
    strncpy(tmp,cidp[1]+16,16);
    sscanf(tmp,"%llx",&idp11[nidp]);
    if(strcmp(cds[1],"1")==0) idp11[nidp] |= 0x0080000000000000ULL;

    nisp++;
    nidp++;
    
    strncpy(tmp,cidp[2],16);
    sscanf(tmp,"%llx",&idp00[nidp]);
    
    strncpy(tmp,cidp[2]+16,16);
    sscanf(tmp,"%llx",&idp01[nidp]);
    if(strcmp(cds[2],"1")==0) idp01[nidp] |= 0x0080000000000000ULL;
    
    strncpy(tmp,cidp[3],16);
    sscanf(tmp,"%llx",&idp10[nidp]);
    
    strncpy(tmp,cidp[3]+16,16);
    sscanf(tmp,"%llx",&idp11[nidp]);
    if(strcmp(cds[3],"1")==0) idp11[nidp] |= 0x0080000000000000ULL;

    nidp++;
  }
  
  for (i = 0; i < nisp; i++)
    printf("isp: %016llx %016llx\n", isp0[i], isp1[i]);

  for (i = 0; i < nidp; i++)
    printf("idp: %016llx %016llx %016llx %016llx\n", idp00[i], idp01[i], idp10[i], idp11[i]);

  //config[0] = 0x110007;
  for(i=0;i<N;i++){
  	config[i] = 0x100007 ;
  }
  printf("nidp %d \t nisp %d \n", nidp, nisp);
 
  CTRL_config(c, config, N);
  CTRL_config_run(c);

  CTRL_config_isp0(c, isp0, nisp);
  CTRL_config_isp1(c, isp1, nisp);
  CTRL_config_idp00(c, idp00, nidp);
  CTRL_config_idp01(c, idp01, nidp);
  CTRL_config_idp10(c, idp10, nidp);
  CTRL_config_idp11(c, idp11, nidp);
  CTRL_config(c, config, N);
  CTRL_config_run(c);
  usleep(1000);
  nresults = CTRL_results_returned(c);
  printf("results : %d \n", nresults);
  wbuf = c->h->dmaw_buf;
  hib_recvMC(c->hib_id, nresults, wbuf);
  for(i=0;i<nresults;i++){
  	printf("%llx \n", wbuf[i]);
  }

  return 0;
}

int CTRL_het_length(CTRLinfo *c, int length)
{
  UINT64 len = length;
  CTRL_write(c, MODULE_HET, 1, &len, 1);
  return 0;
}
int CTRL_het_length_shortmessage(CTRLinfo *c, int length,
				 int mode, int startp)
{
  UINT64 len = length;
  return CTRL_write_shortmessage(c, MODULE_HET, 1, &len, mode, startp, 1);
}

int CTRL_het_convaddr(CTRLinfo *c, int convaddr) 
{
  UINT64 addr = (UINT64)convaddr + 0x400ULL;
  CTRL_write(c, MODULE_HET, 0, &addr, 1);
  //  fprintf(stderr, "CTRL_het_convaddr called\n");
  return 0;
}
int CTRL_het_convaddr_shortmessage(CTRLinfo *c, int convaddr,
				   int mode, int startp) 
{
  UINT64 addr = (UINT64)convaddr + 0x400ULL;
  return CTRL_write_shortmessage(c, MODULE_HET, 0, &addr, mode, startp, 1);
}

int CTRL_het_convtable(CTRLinfo *c, int addr, const void *data, int len) 
{
  CTRL_write(c, MODULE_HET, addr + 0x400ULL, data, len);
  return 0;
}
int CTRL_het_convtable_shortmessage(CTRLinfo *c, int addr, const void *data,
				    int mode, int startp, int len) 
{
    return CTRL_write_shortmessage(c, MODULE_HET, addr + 0x400ULL, data, mode,
			    startp, len);
}

int CTRL_het_write(CTRLinfo *c, int addr, const void *data, int len) 
{
#ifdef CTRL_DEBUG
#ifdef VERBOSE
    const double *p = data;
    int i;
    for (i = 0; i < len; i++) 
	fprintf(stderr, "addr = %x, data[%d] = %016llx\n", addr + i, i, ((UINT64*)p)[i]);
#endif
#endif
    if(c->buff_mode==0){
	CTRL_write(c, MODULE_HET, addr + (0x1ULL << 35), data, len);
    } else {
	UINT64 header, *p=(UINT64*)data;
	int i;
	if(c->buff_ptr+len > BUFFSIZE){
	    printf("flushbuff! : ptr=%d \n", c->buff_ptr);
	    CTRL_flushbuff(c);
	    CTRL_beginbuff(c);
	}
	
	header = ((UINT64)len << 40) | ((UINT64)MODULE_HET << 36) | (addr+(0x1ULL<<35)); //For new hic header KK
	c->buff[c->buff_ptr+0] = header;
	for(i=0;i<len;i++){
	    c->buff[c->buff_ptr+i+1] = p[i];
	}
	c->buff_ptr += len+1;
    }
    return 0;
}


int CTRL_het_write_shortmessage(CTRLinfo *c, int addr, const void *data,
				int mode, int startp, int len) 
{

    return CTRL_write_shortmessage(c, MODULE_HET, addr + (0x1ULL << 35), data,
				   mode, startp, len);
}

int CTRL_het_write_multichip(int nchips,
			     CTRLinfo *c[nchips],
			     int addr, const void *data, int len,int increment)
{
    int i;
    //    CTRL_write_multichip_usingdmar_buffered(nchips, c, MODULE_HET,
    //    CTRL_write_multichip(nchips, c, MODULE_HET,
    CTRL_write_multichip_usingdmar_simple(nchips, c, MODULE_HET,
					    addr + (0x1ULL << 35),
					    (UINT64*)data, len,increment);
    return 0;
}

/*
  addresses are spefified by the long word except for bmsadr 
*/
int CTRL_eit_register(CTRLinfo *c, 
		      int idpid,
		      int emsadr,
		      int emiadr0,
		      int emiadr1,
		      int bmsadr,
		      int nbm,
		      int len,
		      int mode,
		      int start) 
{
  UINT64 data[4];
  data[0] = emsadr;
  data[1] = emiadr0;
  data[2] = emiadr1;
  if (256 < len) {
    fprintf(stderr, "CTRL_eit_register: length %d must be no more than 256\n", len);
    return -1;
  }
  /*
  if(len==256){
    len=0;
  }
  */
  data[3] = ((UINT64)start << 63) + ((UINT64)mode << 32) + ((UINT64)len << 20) + ((UINT64)nbm << 12) + (UINT64)bmsadr;
#if 0
  fprintf(stderr, "idpid = %d, emsadr = %x, emiadr0 = %x, emiadr1 = %x, bmsadr = %x, nbm = %d, len = %d, mode = %d, start = %d\n", 
	  idpid,
	  emsadr,
	  emiadr0,
	  emiadr1,
	  bmsadr,
	  nbm,
	  len,
	  mode,
	  start);
  fprintf(stderr, "data[0] = %llx \n", data[0]);
  fprintf(stderr, "data[1] = %llx \n", data[1]);
  fprintf(stderr, "data[2] = %llx \n", data[2]);
  fprintf(stderr, "data[3] = %llx \n", data[3]);
#endif
  CTRL_write(c, MODULE_EIT, (UINT64)idpid << 2, data, 4);
  hib_flush_piow_channel(c->hib_id);
  if(start==1){
      //      while(CTRL_eit_busy(c)){hib_xwait(1000);   }
      // waiting by CTRL_eit_budy causes random stoppage
      // on P6T6 WS Revolution with two cards
      // -- JM 2010/5/14
      hib_xwait(10000);
  }
  return 0;
}

void CTRL_flush_piow_channel(CTRLinfo *c)
{
    hib_flush_piow_channel(c->hib_id);
}

int CTRL_coc_reset(CTRLinfo *c) 
{
  static UINT64 data[] = {1, 0}; /* reset on and off, respectively */
  CTRL_write(c, MODULE_COC, 0, &(data[0]), 1);
  CTRL_write(c, MODULE_COC, 0, &(data[1]), 1);
  return 0;
}

void CTRL_coc_reg_write(CTRLinfo *c, int addr, UINT64 data) 
{
    UINT64 local_data = data;
    //    fprintf(stderr,"card id=%d, reg[0x0c]=%x\n", c->hib_id,hib_mem_readMC(c->hib_id, 0xc));
    //    fprintf(stderr,"write COC reg %x data %lx\n", addr, local_data);
    CTRL_write(c, MODULE_COC, addr, &local_data, 1);
    //    fprintf(stderr,"card id=%d, reg[0x0c]=%x\n", c->hib_id,hib_mem_readMC(c->hib_id, 0xc));
}

void CTRL_dump_syncstatus(int ndevs, char * s)
{
    int i;
    fprintf(stderr,"%s: ", s);
    for(i=0;i<ndevs;i++){
	hib_mem_writeMC(i, 0x0, 0x10);
	fprintf(stderr, " [%d]= %02x", i, hib_mem_readMC(i, 0x60));
    }
    fprintf(stderr,"\n");
}

static int coc_wait_mode=0;
static int coc_dmar_id=-1;
static int coc_dmaw_id=-1;

void CTRL_coc_setflags(int wait_mode, int dmar_id, int dmaw_id)
{
    // coc_wait_mode has 3 effective bits
    // bit 0: wait coc
    // bit 1: wait emi
    // bit 2: wait sync
    // for normal operation, the value must be 7.
    coc_wait_mode=wait_mode;
    coc_dmar_id=dmar_id;
    coc_dmaw_id=dmaw_id;
}    

int CTRL_coc_start(CTRLinfo *c, int saddr, int ninst, int niter, int nonblock) 
{
  UINT64 cmr = ((UINT64)niter << 26) | ((UINT64)ninst << 13) | (UINT64)saddr;
  if (coc_wait_mode) cmr |= ((UINT64)coc_wait_mode) <<39;
  if (coc_dmar_id >= 0) cmr |= ((UINT64)1) <<(42 + coc_dmar_id);
  if (coc_dmaw_id >= 0) cmr |= ((UINT64)1) <<(46 + coc_dmaw_id);

  CTRL_coc_setflags(0,-1,-1);// reset coc flags to be set in the next call
      
#ifdef CTRL_DEBUG
  fprintf(stderr, "CTRL_coc_start ");
  fprintf(stderr, "saddr = %x, ninst = %d, niter = %d\n", saddr, ninst, niter);
#endif
  //  fprintf(stderr, "CTRL_coc_start ");
  //  fprintf(stderr, "saddr = %x, ninst = %d, niter = %d\n", saddr, ninst, niter);
  //  fprintf(stderr, "buff_mode = %d noblock=%d\n",c->buff_mode, nonblock);
  if(c->buff_mode==0){
      CTRL_write(c, MODULE_COC, 1, &cmr, 1);
      CTRL_write_start(c);
      if (nonblock) return 0;
      xwait(10000);
      int iwait = 0;
      do {
	  xwait(1000);
	  iwait += 10;
	  if (iwait > 100000000){
	      fprintf(stderr,"COC_start, timeout\n");
	      return -1;
	  }
      } while (CTRL_coc_busy(c));
  } else {
      UINT64 header;
      header = ((UINT64)1 << 40) | ((UINT64)MODULE_COC << 36) | 0x1ULL; //For new hic header KK
      c->buff[c->buff_ptr+0] = header;
      c->buff[c->buff_ptr+1] = cmr;
      c->buff_ptr += 2;
  }
  return 0 ;
}

int CTRL_coc_write(CTRLinfo *c, int startaddr, const void *data, int len) 
{
    return CTRL_write(c, MODULE_COC, startaddr + 0x4000, data, len);
}

int CTRL_oht_lat(CTRLinfo *c, int lat) 
{
    UINT64 data = (UINT64)lat;
    CTRL_write(c, MODULE_OHT, 1, &data, 1);
    return 0;
}

int CTRL_eit_busy(CTRLinfo *c) 
{ 
  int f = hib_mem_readMC(c->hib_id, c->h->boardinfo) & STATUS_EIT_DONE;
  if(f!=0){
    return 0;
  } else {
    return 1;
  }
}

int CTRL_coc_busy(CTRLinfo *c) 
{
  int f = hib_mem_readMC(c->hib_id, c->h->boardinfo) & STATUS_COC_BUSY;
  return f;
}

int CTRL_het_full(CTRLinfo *c)
{
  int f = hib_mem_readMC(c->hib_id, c->h->boardinfo) & STATUS_HET_FULL;
  return f;
}

int CTRL_het_idle(CTRLinfo *c)
{
  int f = hib_mem_readMC(c->hib_id, c->h->boardinfo) & STATUS_HET_IDLE;
  return f;
}

UINT64 CTRL_sing_status(CTRLinfo *c)
{
  return 0;
}

void CTRL_hit_reset_count(CTRLinfo *c) {
  UINT64 dummy = 0;
  CTRL_write(c, MODULE_HIT, 0, &dummy, 1);
}

void CTRL_hit_length(CTRLinfo *c, int len) {
  UINT64 data = (UINT64)len;
  CTRL_write(c, MODULE_HIT, 1, &data, 1);
}

void CTRL_hit_header(CTRLinfo *c, int len, int addr, int bbn, int bbnm, int seq) {
  UINT64 data = 
    ((UINT64)len << 56) +
    ((UINT64)addr << 43) + 
    ((UINT64)bbn << 38) +
    ((UINT64)bbnm << 33) +
    ((UINT64)seq << 32);
    //fprintf(stderr, "hit header : %016llx\n", data);
  CTRL_write(c, MODULE_HIT, 2, &data, 1);
}

/* need not specify addr: it always should be zero */
void CTRL_hit_convtable(CTRLinfo *c, int addr, const void *data, int len) {
  CTRL_write(c, MODULE_HIT, 0x100ULL + addr, data, len);
}

void CTRL_hit_data(CTRLinfo *c, const void *data, int len) {
  CTRL_write(c, MODULE_HIT, 0x200ULL, data, len);
}

void CTRL_hit_headerdata(CTRLinfo *c, int len, int addr, int bbn, int bbnm, int seq, UINT64 *data)
{
  int size;
  UINT64 datahead = 
    ((UINT64)len << 56) +
    ((UINT64)addr << 43) + 
    ((UINT64)bbn << 38) +
    ((UINT64)bbnm << 33) +
    ((UINT64)seq << 32);
  int s = 128;
  int devid = c->hib_id;
  Hib *hib = c->h;
  int i, j;

  UINT64 header1 = ((UINT64)1 << 40)   | ((UINT64)MODULE_HIT << 36) | 0x2ULL;    //for hitheader
  UINT64 header2 = ((UINT64)len << 40) | ((UINT64)MODULE_HIT << 36) | 0x200ULL;  //for data
  size = len+3; //header + hithead + header + data(len)
  //printf("id , len = %d , %d \n", devid, len);
  for (i = 0; i < size; i += s) {
    if (i + s > size) {
      s = size - i;
    }
    if(i==0){
      hib->piow_dblbuf[0] = header1;
      hib->piow_dblbuf[1] = datahead;
      hib->piow_dblbuf[2] = header2;
      for (j = 0; j < s-3; j++) {
	hib->piow_dblbuf[j+3] = data[i + j];
      }
    } else {
      for (j = 0; j < s; j++) {
	hib->piow_dblbuf[j] = data[i + j - 3];
      }
    }
    hib_mem_writeMC(devid, hib->dma0misc,
		    (1<<(hib->dma0misc_swap_sram_bit)) |
		    (s<<(hib->dma0misc_sram_wcnt_bit)));

      //      _mm_mfence();
      //      usleep(100000);
      // write '1' to HIB local register 'swap_sram'
      // write data size to 'sram_wcnt'
  }
}


void CTRL_beginbuff(CTRLinfo *c)
{
  c->buff_mode = 1;
  c->buff_ptr = 0;
}


void CTRL_flushbuff(CTRLinfo *c)
{
  //hib_piowMC(c->hib_id, c->buff_ptr, c->buff);
  UINT64 start, end;

  start = rdtsc2();
  if(c->buff_mode==1){
    int i, j;
    int s = (c->h->piowbuf_bytes>>3)-1;
    int size = c->buff_ptr;
    for (i = 0; i < size; i += s) {
      
      if (i + s > size) {
	s = size - i;
      }
      for (j = 0; j < s; j++) {
          c->h->piow_dblbuf[j] = c->buff[i + j];
      }
      hib_mem_writeMC(c->hib_id, c->h->dma0misc,
                      (1<<(c->h->dma0misc_swap_sram_bit)) |
                      (s<<(c->h->dma0misc_sram_wcnt_bit)));

      //      _mm_mfence();
      //      usleep(100000);
      // write '1' to HIB local register 'swap_sram'
      // write data size to 'sram_wcnt'
    }
  }
  end = rdtsc2();
  //printf("size: %d elapse %lld rate %e B/s \n", c->buff_ptr, end-start, (double)(c->buff_ptr*8)/(double)(end-start)*2.6e9);
  c->buff_mode = 0;
  c->buff_ptr = 0;
}


void CTRL_hit_header_data(CTRLinfo *c, int len, int addr, int bbn,
			  int bbnm, int seq,
			  const void *senddata, int sendlen)
{
    UINT64 header, buf[256], *p;
    int i;
    p= (UINT64*)senddata;
    header = ((UINT64)1 << 40) | ((UINT64)MODULE_HIT << 36) | 2; //For new hic header KK
    UINT64 data = 
	((UINT64)len << 56) +
	((UINT64)addr << 43) + 
	((UINT64)bbn << 38) +
	((UINT64)bbnm << 33) +
	((UINT64)seq << 32);
    //fprintf(stderr, "hit header : %016llx\n", data);
    
#if 0
    CTRL_write(c, MODULE_HIT, 2, &data, 1);
    CTRL_write(c, MODULE_HIT, 0x200ULL, senddata, sendlen);
#else 
    if(c->buff_mode == 0){
	buf[0] = header;
	buf[1] = data;
	header = ((UINT64)sendlen << 40) | ((UINT64)MODULE_HIT << 36) | 0x200ULL; //For new hic header KK
	buf[2] = header;
	for(i=0;i<sendlen;i++){
	    buf[3+i] = p[i];
	}
	hib_piowMC(c->hib_id, sendlen+3, buf);
    } else {
	//printf("buff hit header");
	c->buff[c->buff_ptr+0] = header;
	c->buff[c->buff_ptr+1] = data;
	header = ((UINT64)sendlen << 40) | ((UINT64)MODULE_HIT << 36) | 0x200ULL; //For new hic header KK
	c->buff[c->buff_ptr+2] = header;
	for(i=0;i<sendlen;i++){
	    c->buff[c->buff_ptr+3+i] = p[i];
	}
	c->buff_ptr += sendlen+3;
    }
#endif
}

void CTRL_setodpcom(CTRLinfo *c, UINT64 data)
{
    CTRL_write(c, MODULE_OHT, 0x2, &data, 1);
}

int CTRL_remaining_words_dmaw(CTRLinfo * c)
{
    return hib_remaining_words_dmawMC(c->hib_id);
}



//#define CTRL_TEST

#ifdef CTRL_TEST

#if 0
int main()
{
  CTRLinfo *c;

  printf("open hib.\n");
  c = CTRL_open(0);
  if(c==NULL){
    printf("error to open hib!\n");
    return 1;
  }
  printf("ctrl backend reset.\n");
  //CTRL_init(c);  sleep(2);
 
 
  printf("send reset vector for sing \n");
  CTRL_sing_chip_init(c);
  sleep(1);
  CTRL_sing_chip_init(c);
  sleep(1);
  
  printf("test vector for sing \n");  
  loopback(c);
  

  CTRL_close(c);
}
#else
int main(int argc, char **argv)
{
  CTRLinfo *c;
  int chip;

  if(argc==2){
    chip = atoi(argv[1]);
  }
  printf("open hib.\n");
  c = CTRL_open(chip);
#ifndef NOBCAST
  CTRL_beginbcast(c);
  printf("set bcast mode\n");
#endif  
  if(c==NULL){
    printf("error to open hib!\n");
    return 1;
  }
  printf("ctrl backend reset.\n");
  CTRL_init(c);  sleep(2);
 
 
  printf("send reset vector for sing \n");
  CTRL_sing_chip_init(c);
  usleep(100000);
  CTRL_sing_chip_init(c);
  usleep(100000);
  
  printf("test vector for sing \n");  
  read_and_run(c);

  CTRL_endbcast(c);


  CTRL_close(c);
}
#endif

#endif
