#ifndef _CTRL_H
#define _CTRL_H

#include <typedef.h>
#include <hibutil.h>

#define NBB 16
#define NPE 32

#define BUFFSIZE 2048

typedef struct
{
  int hib_id;
  Hib *h;

  UINT64 buff[BUFFSIZE];
  int buff_ptr;
  int buff_mode;
  int bcastflag;
} CTRLinfo;

#if defined(__cplusplus)
extern "C"
{
#endif

    int CTRL_sing_chip_init(CTRLinfo* c);
    int CTRL_all_reset(CTRLinfo* c);
    int CTRL_result_clear(CTRLinfo* c);
    int CTRL_init(CTRLinfo *c);
    CTRLinfo* CTRL_open(int id);
    int CTRL_close(CTRLinfo* c);
    int CTRL_write(CTRLinfo* c, int module, UINT64 addr, const void *data, int len);
    void CTRL_write_setflushmode(int mode);
    int CTRL_write_start(CTRLinfo* c);
    int CTRL_read_as_many_as(CTRLinfo* c, void *data, int len);
    int CTRL_read2(CTRLinfo *c, void *data);
    int CTRL_het_length_shortmessage(CTRLinfo *c, int length,
				     int mode, int startp);
    int CTRL_het_length(CTRLinfo* c, int length);
    int CTRL_het_convaddr(CTRLinfo* c, int convaddr);
    int CTRL_het_convaddr_shortmessage(CTRLinfo *c, int convaddr,
				       int mode, int startp) ;
    
    int CTRL_het_convtable(CTRLinfo* c, int addr, const void *data, int len);
    int CTRL_het_convtable_shortmessage(CTRLinfo *c, int addr, const void *data,
					int mode, int startp, int len) ;
    int CTRL_het_write(CTRLinfo* c, int addr, const void *data, int len);
    int CTRL_het_write_shortmessage(CTRLinfo *c, int addr, const void *data,
				    int mode, int startp, int len) ;
    int CTRL_rawwrite_shortmessage(CTRLinfo *c,   const void *data, int startp, int len) ;

    int CTRL_eit_register(CTRLinfo* c, int idpid, int emsadr, int emiadr0, int emiadr1, int bmsadr, int nbm, int len, int mode, int start);
    int CTRL_coc_reset(CTRLinfo* c);
    void CTRL_coc_reg_write(CTRLinfo *c, int addr, UINT64 data);
    int CTRL_coc_start(CTRLinfo* c, int saddr, int ninst, int niter, int nonblock);
    int CTRL_coc_write(CTRLinfo* c, int startaddr, const void *data, int len);
    int CTRL_oht_lat(CTRLinfo* c, int lat);
    int CTRL_config_write(CTRLinfo* c, int addr, const void *data, int len);
    int CTRL_config(CTRLinfo* c, const void *data, int len);
    int CTRL_config_isp0(CTRLinfo* c, const void *data, int len);
    int CTRL_config_isp1(CTRLinfo* c, const void *data, int len);
    int CTRL_config_idp00(CTRLinfo* c, const void *data, int len);
    int CTRL_config_idp01(CTRLinfo* c, const void *data, int len);
    int CTRL_config_idp10(CTRLinfo* c, const void *data, int len);
    int CTRL_config_idp11(CTRLinfo* c, const void *data, int len);
    int CTRL_config_run(CTRLinfo* c);
    int CTRL_eit_busy(CTRLinfo* c);
    int CTRL_coc_busy(CTRLinfo* c);
    int CTRL_het_full(CTRLinfo* c);
    int CTRL_het_idle(CTRLinfo* c);
    int CTRL_results_returned(CTRLinfo* c);
    UINT64 CTRL_sing_status(CTRLinfo* c);
    int CTRL_number_of_results(CTRLinfo* c);
    void CTRL_hit_reset_count(CTRLinfo *c);
    void CTRL_hit_length(CTRLinfo *c, int len);
    void CTRL_hit_header(CTRLinfo *c, int len, int addr, int bbn, int bbnm, int seq);
    void CTRL_hit_headerdata(CTRLinfo *c, int len, int addr, int bbn, int bbnm, int seq, UINT64 *data);
    void CTRL_hit_data(CTRLinfo *c, const void *data, int len);
    void CTRL_hit_convtable(CTRLinfo *c, int addr, const void *data, int len);
    void CTRL_beginbuff(CTRLinfo *c);
    void CTRL_flushbuff(CTRLinfo *c);
    void CTRL_beginbcast(CTRLinfo *c);
    void CTRL_endbcast(CTRLinfo *c);
    void CTRL_setodpcom(CTRLinfo *c, UINT64 data);
    int CTRL_read_as_many_as_async_end(CTRLinfo *c, void *data, int len);
    int CTRL_het_write_multichip(int nchips,
				 CTRLinfo *c[nchips],
				 int addr, const void *data,
				 int len,int increment);
    int CTRL_write_multichip_usingdmar_buffered_control(int nchips,
							CTRLinfo *c[nchips],
							int flag);
    int CTRL_write_multichip_usingdmar_dataarray_simple(int nchips, CTRLinfo *c[nchips],
							int module, UINT64 addr,
							const void *data[nchips],
							int len);
    int CTRL_write_multichip_usingdmar_dataarray_start(int nchips, CTRLinfo *c[nchips],
						       int module, UINT64 addr,
						       const void *data[nchips],
						       int len);
    int CTRL_write_multichip_usingdmar_dataarray_wait(int nchips, CTRLinfo *c[nchips]);
   int CTRL_write_multichip_usingdmar_simple_da(int nchips, CTRLinfo *c[nchips], int module,
					     UINT64 addr, void *data[nchips], int len);

    void CTRL_coc_setflags(int wait_mode, int dmar_id, int dmaw_id);

    int CTRL_write_set_dmar_buffered_mode(int mode);

    int CTRL_remaining_words_dmaw(CTRLinfo * c);
    void CTRL_flush_piow_channel(CTRLinfo *c);

#if defined(__cplusplus)
}
#endif
#endif /* __CTRL__*/
