//
// function_table.c
//  Time-stamp: <11/01/17 17:48:16 makino>
// JM 

#include <stdio.h>
#include <sys/time.h>
#include <stdlib.h>
//#include "simplegravity-16x32.h"
//#include "g5sim-gdr.h"
#include <math.h>
#include <signal.h>

#ifndef UINT64
 typedef unsigned long long int UINT64;
#endif

typedef union datapack{
    unsigned int i32[2];
    UINT64        i64;
    double       d;
}DATAPACK;


typedef union double_and_uint64{
    double d;
    UINT64 u;
}DOUBLE_AND_UINT64;


#define N_CUTOFF_TABLES 4
#define CUTOFF_TABLE_ADDR_SIZE 7
#define CUTOFF_TABLE_EXP_SIZE 4
#define CUTOFF_TABLE_MANTISSA_SIZE (CUTOFF_TABLE_ADDR_SIZE-CUTOFF_TABLE_EXP_SIZE )
#define CUTOFF_TABLE_ENTRIES (1<<CUTOFF_TABLE_ADDR_SIZE)
#define ORDER_OF_POLYNOMIAL 2
#define NCOEFFS (ORDER_OF_POLYNOMIAL+1)

double g5work_function_table(double r2,
		      double rscale,
		      double c2[],
		      double c1[],
		      double c0[])
{
    DATAPACK pack;
    pack.d = r2;
    UINT64 adr0 = pack.i64 >> 49;
    // adr0 bit length is 15. 1 sign, 11 exp and 3 mantissa
    //    fprintf(stderr, "r2 = %e d = %Lx, adr0 =%Lx\n", r2, pack.i64, adr0);
    int adroff = 0;
    if (adr0 > 0x207fL) {
	adr0 = 0x7f;
    }else if (adr0 <= 0x2000L) {
	adr0 = 0x0;
	adroff = (0x400 - (adr0>>3) )*3;
    }else{
	adr0 &= 0x7fL;
    }
    UINT64 dru = pack.i64 << 3;
    dru &= 0xfff0000000000000LL;
    dru |= 0x3ff0000000000000LL;
    pack.i64 = dru;
    double dx = pack.d - 1.0;
    pack.d = (dx*c2[adr0]+ c1[adr0])*dx * c0[adr0];
    pack.i64 += ((UINT64)adroff) << 52;
    double result = pack.d;
    return result;
}


double cutoff_table[N_CUTOFF_TABLES][CUTOFF_TABLE_ENTRIES][NCOEFFS];




double g5work_internal_dll_cutoff(double x)
{
    x = sqrt(x);
    x *= 8;
    double result;
    if (x < 1.0){
	result=1.0;
    }else if (x < 2.0){
	double x2 = (2-x)*(2-x);
	result= ((6*x2-15)*x2+10)*x2*x2*x2;
    }else{
	result= 0.0;
    }
    return result+1;
}

double g5work_internal_dll_cutoff_inverted(double x)
{
    return 1.0 - g5work_internal_dll_cutoff(x);
}

double g5work_internal_unity(double x)
{
    return 1.0;
}


double g5work_rinv_cubed(double x)
{
    x = sqrt(x);
    return 1.0/(x*x*x);
}

double g5work_rinv_cubed_with_dll_cutoff(double x)
{
    return g5work_rinv_cubed(x)*g5work_internal_dll_cutoff(x);
}

static void get_index_and_dx(double x, int* index,
			     double * dx, int *exp_offset)
{
    int exp, offset;
    double mantissa = frexp(x, &exp);
    exp --;
    exp += (1 << CUTOFF_TABLE_EXP_SIZE) ;
    offset = 0;
    //    fprintf(stderr,"get_index x, exp, mantissa= %g, %d, %g\n",x, exp, mantissa);
    if (exp < 0){
	offset = - exp;
	exp = 0;
	if (offset &1){
	    exp++;
	    offset ++;
	}
	offset  *= -1;
    }
    if (exp >=(1 << CUTOFF_TABLE_EXP_SIZE)){
	offset = exp -(1 << CUTOFF_TABLE_EXP_SIZE)+1;
	exp -= offset;
	if (offset &1){
	    exp--;
	    offset ++;
	}
    }
    
    mantissa *= 2;
    exp &= ((1 << CUTOFF_TABLE_EXP_SIZE) - 1);
    //    fprintf(stderr,"get_index x, exp, mantissa, exp_off = %g, %d, %g %d\n",x, exp,
    //	    mantissa, offset);
    double scaled_x = (mantissa-1) * (1 << CUTOFF_TABLE_MANTISSA_SIZE) ;
    int man_i = floor(scaled_x);
    *dx = scaled_x - man_i;
    *index = man_i + (exp <<CUTOFF_TABLE_MANTISSA_SIZE);
    *exp_offset = offset;
//  fprintf(stderr,"get_index x, index, dx= %g, %d, %g\n",x, *index, *dx);
}



//
// third-order Chebyshev polynomical is given by
//
// y = 4x^3 -3x
//
// Therefore, zero points are
// +- sqrt(3)/2, 0
//

static void set_table(double x[3],
		      double f[3],
		      double coef[3])
{
    // first create coef for midpoint expansion
    double c0, c1, c2;
    c0 = f[1];
    c1 = (f[2]-f[0])/(x[2]-x[0]);
    double dx = x[1]-x[0];
    c2 = ((f[2]+f[0])*0.5-f[1])/(dx*dx);
    // then shift this by 0.5
    // c2(x-0.5)^2 + c1(x-0.5) +c0
    // = c2 x^2 + (c1-c2) x + c2/4 - c1/2+ c0
    coef[0] = c0 -c1*0.5 + c2*0.25;
    coef[1] = c1-c2;
    coef[2] = c2;
}

static void calculate_x_and_dx_from_index(int index,
					  double * x,
					  double * dx)
{
    int exp = index >> CUTOFF_TABLE_MANTISSA_SIZE;
    int subcount = (1<<CUTOFF_TABLE_MANTISSA_SIZE);
    int subid = index &( subcount-1);
//    fprintf(stderr, "index=%d, exp=%d, subcount=%d, subid=%d ",
//	    index, exp, subcount, subid);
    double x0 = ldexp(1.0, exp-(1<<CUTOFF_TABLE_EXP_SIZE));
    *dx = x0 / subcount;
    *x = x0 + subid * (*dx);
//    fprintf(stderr," x, dx = %e, %e\n",  *x, *dx);
}

static void calculate_coef(int index,
		    double (*func)(double),
		    double coef[])
{
    double xp[3]={-sqrt(3.0)*0.25+0.5, 0.5, sqrt(3.0)*0.25+0.5};
    double fval[3];
    double dx, x0;
    calculate_x_and_dx_from_index(index,&x0, &dx);
    int i;
    for(i=0;i<3;i++)fval[i]=(*func)(x0+dx*xp[i]);
//    for(i=0;i<3;i++)fprintf(stderr,"x, f= %e %e ", x0+dx*xp[i], fval[i]);
//    fprintf(stderr,"\n");
    set_table(xp, fval, coef);
    
}

static void create_cutoff_table(int table_id, double (*func)(double))
{
    int i;
    for (i=0; i< CUTOFF_TABLE_ENTRIES; i++){
	double * pd =cutoff_table[table_id][i];
	calculate_coef(i, func, pd);
    }
}


static void dump_table(int table_id)
{
    
    int i;
    for (i=0; i< CUTOFF_TABLE_ENTRIES; i++){
	double * p = cutoff_table[table_id][i];
	printf("%3d %20.10g %20.10g %20.10g\n",i, p[0],p[1],p[2]);
    }
}
    

double polynomial_double(int table_id, double x)
{
    int tabid,  exp_offset;
    double dx;
    get_index_and_dx(x, &tabid, &dx, &exp_offset);
    double *p = cutoff_table[table_id][tabid];
    return ((p[2]*dx+p[1])*dx + p[0])*ldexp(1.0, -(exp_offset*3)/2);
}

void g5work_initialize_cutoff_table(int tableid, double (*func)(double))
{
    create_cutoff_table(tableid, func);
}
void g5work_initialize_cutoff_table_with_dll(int tableid)
{
    create_cutoff_table(tableid, g5work_internal_dll_cutoff);
}

typedef struct {
    double (*func)(double);
} Func;

#ifdef TEST_CUTOFF_TABLE

int main()
{
    double (*func)(double);
    func = g5work_rinv_cubed_with_dll_cutoff;
    create_cutoff_table(0, func);
//    dump_table(0);
    double dx = 1.0/50;
    double x0 = 1e-7;
    double x;
    for(x = x0; x < 10; x*= (1+dx)){
	//	if (x > 0.03) dx = 0.001;
	double f = (*func)(x);
	double ff = polynomial_double(0, x);
	printf("%.7g %16.8g %16.8g %14.7e %14.7e\n",x, f, ff,  ff-f,
	       (ff-f)/g5work_rinv_cubed(x));
    }
    return 0;

}

#endif
