//
// g5sim.c
//  Time-stamp: <11/01/20 00:15:47 makino>
// JM 

#include <stdio.h>
#include <sys/time.h>
#include <stdlib.h>
#include "simplegravity-16x32.h"
#include "g5sim-gdr.h"
#include <math.h>
#include <signal.h>

#define MAXCHIPS 8
#define SHIFT_INDEX 
static inline UINT64 rdtsc2()
{
	UINT64 ret;
	asm __volatile__("rdtsc  \n\t" \
			 "shlq $32, %%rdx \n\t" \
			 "orq  %%rdx, %%rax \n\t" \
			 "movq %%rax, %0" : "=g"(ret) : :"rax", "rdx");
	return ret;
}

static UINT64 jptime = 0;
static UINT64 iptime = 0;
static UINT64 runtime = 0;
static int jpcalls = 0;
static int ipcalls = 0;
static int runcalls = 0;

static int debug_level = 0;
int nja[MAXCHIPS];
int nja_old[MAXCHIPS];

void g5_set_debug_level(int level)
{
    debug_level = level;
}


static int signal_flag_for_exit = 1;
static int signal_flag_for_check = 0;

static void exit_or_set_flag_on_int()
{
    if (signal_flag_for_exit == 1){
	fprintf(stderr,"SIG interrupt - immediate exit \n");
	exit(-1);
    }else{
	signal_flag_for_check = 1;
    }
}

static void exit_on_flag_on_int()
{
    if (signal_flag_for_check == 1){
	fprintf(stderr,"SIG interrupt - delayed exit \n");
	exit(-1);
    }
}

static void set_signal_flag_for_exit()
{
    exit_on_flag_on_int();
    signal_flag_for_exit=1;
}
static void reset_signal_flag_for_exit()
{
    signal_flag_for_exit=0;
}

	
static void remove_lock_and_exit_quit();
static void remove_lock_and_exit_bus();
static void remove_lock_and_exit_segv();
static void remove_lock_and_exit_term();

static void sig_set()
{
    static int called = 0;
    if(called == 0){
	called = 1;
	signal(SIGINT,  exit_or_set_flag_on_int);       /* and ^C also */
	signal(SIGQUIT,  exit_or_set_flag_on_int);       /* and ^C also */
	signal(SIGTERM,  exit_or_set_flag_on_int);       /* and ^C also */
    }
}

void g5_report_timers()
{
    printf("JP= %e %d IP= %e %d RUN = %e %d \n", ((double)jptime)/jpcalls, (int) jpcalls,
	   ((double)iptime)/ipcalls, (int) ipcalls,
	   ((double)runtime)/runcalls, (int) runcalls);
}


void g5_openMC(int devid)
{
    SING_grape_initMC(devid);
    nja_old[devid]=-1;
    sig_set();

}

void g5_closeMC(int devid)
{
    SING_grape_closeMC(devid);
}

void g5_set_rangeMC(int devid, double xmin, double xmax, double mmin)
{}

static struct SING_elt_struct0 *jpbufp[MAXCHIPS] = {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};


void g5_set_jpMC_old(int devid, int adr, int nj, double *m, double (*x)[3])
{
    UINT64 start = rdtsc2();
    if(debug_level)fprintf(stderr, "set_jp devid=%d, adr=%d, nj=%d\n", devid, adr, nj);

    if (jpbufp[devid] == NULL){
	jpbufp[devid]= (struct SING_elt_struct0 *)
	    malloc(SING_elt_size0()* sizeof(struct SING_elt_struct0));
	if (jpbufp[devid] == NULL){
	    fprintf(stderr,"g5_set_jpMC: buffer allocation failed\n");
	    exit(-1);
	}
    }
    int j,k;
    int jpmax = 1024;

    reset_signal_flag_for_exit();

    for(j=0;j<nj;j+=jpmax){
	int jend = j+jpmax;
	if (jend > nj) jend = nj;
	int np = jend - j;
	struct SING_elt_struct0 * jpp = jpbufp[devid];
	int jj;
	for(jj=0;jj<np; jj++){
	    int jindex = j+jj;
	    jpp->xj = x[jindex][0];
	    jpp->yj = x[jindex][1];
	    jpp->zj = x[jindex][2];
	    jpp->mj = m[jindex];
	    jpp->eps2 = 0.0;
	    //	SING_send_j_particleMC(devid,  jpp, adr+j);
	    jpp++;
	}
	SING_send_elt_data0_mMC(devid, jpbufp[devid], adr+j, np);
    }
    jptime += rdtsc2()-start;
    jpcalls ++;

    set_signal_flag_for_exit();

}

void g5s_set_jpMC_old(int devid, int adr, int nj, double *m, double (*x)[3], double *eps2)
{
    UINT64 start = rdtsc2();
    if(debug_level)fprintf(stderr, "set_jp devid=%d, adr=%d, nj=%d\n", devid, adr, nj);

    if (jpbufp[devid] == NULL){
	jpbufp[devid]= (struct SING_elt_struct0 *)
	    malloc(SING_elt_size0()* sizeof(struct SING_elt_struct0));
	if (jpbufp[devid] == NULL){
	    fprintf(stderr,"g5_set_jpMC: buffer allocation failed\n");
	    exit(-1);
	}
    }
    int j,k;
    int jpmax = 1024;
    for(j=0;j<nj;j+=jpmax){
	int jend = j+jpmax;
	if (jend > nj) jend = nj;
	int np = jend - j;
	struct SING_elt_struct0 * jpp = jpbufp[devid];
	int jj;
	for(jj=0;jj<np; jj++){
	    int jindex = j+jj;
	    jpp->xj = x[jindex][0];
	    jpp->yj = x[jindex][1];
	    jpp->zj = x[jindex][2];
	    jpp->mj = m[jindex];
	    jpp->eps2 = eps2[jindex];
	    //	SING_send_j_particleMC(devid,  jpp, adr+j);
	    jpp++;
	}
	SING_send_elt_data0_mMC(devid, jpbufp[devid], adr+j, np);
    }
    jptime += rdtsc2()-start;
    jpcalls ++;
}


void g5sn_set_jpMC(int devid, int adr, int nj, double *m, double (*x)[3],
		   double *eps2, int * index)
{
    UINT64 start = rdtsc2();
    if(debug_level)fprintf(stderr, "g5sn_set_jp devid=%d, adr=%d, nj=%d\n", devid, adr, nj);

    if (jpbufp[devid] == NULL){
	jpbufp[devid]= (struct SING_elt_struct0 *)
	    malloc(SING_elt_size0()* sizeof(struct SING_elt_struct0));
	if (jpbufp[devid] == NULL){
	    fprintf(stderr,"g5_set_jpMC: buffer allocation failed\n");
	    exit(-1);
	}
    }
    int j,k;
    int jpmax = 1024;
    int njhard = ((nj + SING_number_of_bb()-1)/SING_number_of_bb())*SING_number_of_bb();
    for(j=0;j<njhard;j+=jpmax){
	int jend = j+jpmax;
	int jendsoft = j+jpmax;
	if (jend > njhard) jend = njhard;
	if (jendsoft > nj) jendsoft = nj;
	int np = jendsoft - j;
	int nphard = jend - j;
	struct SING_elt_struct0 * jpp = jpbufp[devid];
	int jj;
	for(jj=0;jj<np; jj++){
	    int jindex = j+jj;
	    jpp[jj].xj = x[jindex][0];
	    jpp[jj].yj = x[jindex][1];
	    jpp[jj].zj = x[jindex][2];
	    jpp[jj].mj = m[jindex];
	}
	for(jj=np;jj<nphard; jj++){
	    int jindex = j+jj;
	    jpp[jj].xj =  jpp[jj].yj = jpp[jj].zj = 1e30;
	    jpp[jj].mj = 0;
	}
	if (eps2 != NULL){
	    for(jj=0;jj<np; jj++){
		int jindex = j+jj;
		jpp[jj].eps2 = eps2[jindex];
	    }
	}
	if (index != NULL){
	    for(jj=0;jj<np; jj++){
		int jindex = j+jj;
#ifndef SHIFT_INDEX	
		jpp[jj].idj = index[jindex];
#else
		jpp[jj].idj = (index[jindex]<<8);
		
#endif
		
	    }
	}
	SING_send_elt_data0_mMC(devid, jpbufp[devid], adr+j, nphard);
    }
    jptime += rdtsc2()-start;
    jpcalls ++;
}
void g5n_set_jpMC(int devid, int adr, int nj, double *m, double (*x)[3],
		  int * index)
{
    g5sn_set_jpMC(devid, adr, nj, m, x, NULL, index);
}

void g5s_set_jpMC(int devid, int adr, int nj, double *m, double (*x)[3], double *eps2)
{
    g5sn_set_jpMC(devid, adr, nj, m, x, eps2, NULL);
}

void g5_set_jpMC(int devid, int adr, int nj, double *m, double (*x)[3])
{
    g5sn_set_jpMC(devid, adr, nj, m, x, NULL, NULL);
}


static struct  grape_i_particle_struct *ipbufp[MAXCHIPS]=
    {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
static struct  grape_result_struct *resbufp[MAXCHIPS]=
    {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
static int nia[MAXCHIPS];

static void setup_ipbufp(int devid)
{
    if (ipbufp[devid] == NULL){
	ipbufp[devid]= (struct grape_i_particle_struct *)
	    malloc(SING_number_of_pe()
		   *4* sizeof(struct grape_i_particle_struct));
	if (ipbufp[devid] == NULL){
	    fprintf(stderr,"g5: setup_ipbufp: buffer allocation failed\n");
	    exit(-1);
	}
    }
}

static void setup_resbufp(int devid)
{
    if (resbufp[devid] == NULL){
	resbufp[devid]= (struct grape_result_struct *)
	    malloc(SING_number_of_pe()
		   *4* sizeof(struct grape_result_struct));
	if (resbufp[devid] == NULL){
	    fprintf(stderr,"g5:setup_resbufp: buffer allocation failed\n");
	    exit(-1);
	}
    }
}

void g5_set_xiMC(int devid, int ni, double (*x)[3])
{
    setup_ipbufp(devid);
    int i;
    struct grape_i_particle_struct * ip = ipbufp[devid];
    for(i=0;i<ni;i++){
	ip->xi = x[i][0];
	ip->yi = x[i][1];
	ip->zi = x[i][2];
	ip++;
    }
    nia[devid] = ni;
}

struct grape_i_particle_struct * g5_ibufp(int devid)
{
    return ipbufp[devid];
}

int  clear_unused_i_particles(int devid,
			      struct grape_i_particle_struct * ip, 
			      int nused )
{
    int nend =g5_get_number_of_pipelinesMC(devid);
    int i;
    //    fprintf(stderr, "clear i: %d %d\n", nused, nend);
    for(i=nused; i<nend; i++){
	(ip+i)->xi = (ip+i)->yi = (ip+i)->zi = (ip+i)->eps2i=1e30;
	(ip+i)->idi = -1;
    }
}
int  clear_unused_j_particles(int devid, int nused, int nmin )
{
    int nend = ((nmin + SING_number_of_bb()-1)/SING_number_of_bb())
	*SING_number_of_bb();
    int j;
    //    fprintf(stderr, "clear j: %d %d\n", nused, nend);
    struct grape_j_particle_struct  jp; 
    jp.xj = jp.yj = jp.zj = jp.eps2=1e30; jp.mj=0, jp.idj=-1;
    if (nused != nja_old[devid]){
	for(j=nused;j<nend; j++){
	    SING_send_j_particleMC(devid,  &jp, j);
	}
    }
    //    nja_old[devid]=nused;
}


void g5_runMC(int devid)
{
    UINT64 start = rdtsc2();
    int nireal=g5_get_number_of_pipelinesMC(devid);
    sig_set();
    reset_signal_flag_for_exit();

    clear_unused_i_particles(devid, ipbufp[devid], nia[devid]);
    //    clear_unused_j_particles(devid, nja[devid], nja[devid]);
    SING_send_i_particleMC(devid,ipbufp[devid], nireal);
    iptime += rdtsc2()-start;
    ipcalls++;
    start = rdtsc2();
    int njreal = (nja[devid]+SING_number_of_bb()-1)/SING_number_of_bb();
    SING_grape_runMC(devid,njreal);
    runtime += rdtsc2()-start;
    runcalls++;
    set_signal_flag_for_exit();
}

void g5_set_nMC(int devid, int n){nja[devid]=n;}

void g5_set_eps2MC(int devid, int ni, double *eps2)
{
    setup_ipbufp(devid);
    int i;
    struct grape_i_particle_struct * ip = ipbufp[devid];
    for(i=0;i<ni;i++){
	ip->eps2i = eps2[i];
	ip++;
    }
    nia[devid] = ni;
}    
void g5_set_eps2_to_allMC(int devid, double eps2)
{
    setup_ipbufp(devid);
    int i;
    struct grape_i_particle_struct * ip = ipbufp[devid];
    int nireal=g5_get_number_of_pipelinesMC(devid);
    for(i=0;i<nireal;i++){
	ip->eps2i = eps2;
	ip++;
    }
}

void g5_set_indexMC(int devid, int ni, int *index)
{
    setup_ipbufp(devid);
    int i;
    struct grape_i_particle_struct * ip = ipbufp[devid];
    for(i=0;i<ni;i++){
#ifndef SHIFT_INDEX	
	ip->idi = index[i];
#else
	ip->idi = (index[i]<<8);
#endif	
	ip++;
    }
    nia[devid] = ni;
}    

void g5_get_forceMC_internal_old(int devid, int ni, double (*a)[3], double *pot,
			    int mode)
// mode == 0: clear
// mode != 1: accumurate
{
    setup_resbufp(devid);
    struct grape_result_struct * rp;
    int k;
    SING_get_resultMC(devid,resbufp[devid]);
    if (mode == 0){
        for(k=0;k<ni; k++)  a[k][0]=a[k][1]= a[k][2]= pot[k] = 0.0;
    }
    rp = resbufp[devid];
    for(k=0;k<ni; k++){
	a[k][0] +=rp-> accx;
	a[k][1] +=rp->accy;
	a[k][2] +=rp->accz;
	pot[k]  += -rp->pot;
	rp++;
    }
}

typedef union datapack{
    unsigned int i32[2];
    UINT64        i64;
    double       d;
}DATAPACK;

void g5n_get_forceMC_internal(int devid, int ni, double (*a)[3], double *pot,
			      double * rnnb2, int * innb,  int mode)
// mode == 0: clear
// mode != 1: accumurate
{
    setup_resbufp(devid);
    struct grape_result_struct * rp;
    int k;
    if(debug_level){
	fprintf(stderr,"g5n_get_forceMC_internal called %d %d\n", devid, ni);
    }
    SING_get_resultMC(devid,resbufp[devid]);
    if (mode == 0){
        for(k=0;k<ni; k++)  a[k][0]=a[k][1]= a[k][2]= pot[k] = 0.0;
	if (rnnb2 != NULL)for(k=0;k<ni; k++)  rnnb2[k] = VERY_LARGE_NUMBER;
    }
    rp = resbufp[devid];
    for(k=0;k<ni; k++){
	a[k][0] +=rp-> accx;
	a[k][1] +=rp->accy;
	a[k][2] +=rp->accz;
	pot[k]  += -rp->pot;
	rp++;
    }
    if (rnnb2 != NULL){
	rp = resbufp[devid];
	for(k=0;k<ni; k++){
	    DATAPACK pack;
	    pack.d = rp->nnb;
 //	    fprintf(stderr,"k= %d rnnb2 = %e %Lx\n", k, rp->nnb, pack.i64);
	    if (pack.d < rnnb2[k]){
		innb[k] =(int)( pack.i64 & 0xffffffL);
		rnnb2[k] = pack.d;
	    }
	    rp++;
	}
    }
}
void g5_get_forceMC_internal(int devid, int ni, double (*a)[3], double *pot,
			    int mode)
{
    g5n_get_forceMC_internal(devid, ni, a, pot, NULL, NULL, mode);
}


void g5_get_forceMC(int devid, int ni, double (*a)[3], double *pot)
{
    g5_get_forceMC_internal(devid, ni, a, pot,0);
    return ;
}

void g5n_get_forceMC(int devid, int ni, double (*a)[3], double *pot,
		     double * rnnb2, int * innb)
{
    g5n_get_forceMC_internal(devid, ni, a, pot, rnnb2, innb, 0);
    return ;
}

void g5sn_calculate_force_on_xMC(int devid, double (*x)[3], double * eps2,
				 int * index,
				 double (*a)[3], double *p,
				 double * rnnb2,
				 int * innb, int ni)
{
    int off, nii, np;
    np = g5_get_number_of_pipelinesMC(devid);
    //    fprintf(stderr,"g5sn_calculate_force_on_xMC ni= %d\n", ni);
    for (off = 0; off < ni; off += np) {
        nii = np;
        if (off+nii > ni) {
            nii = ni - off;
        }
	g5_set_xiMC(devid, nii, x+off);
	if (eps2 != NULL) g5_set_eps2MC(devid, nii, eps2+off);
	if (index != NULL) g5_set_indexMC(devid, nii, index+off);
        g5_runMC(devid);
	double * pr = rnnb2+off;
	int * pi    = innb+off;
	if (rnnb2 == NULL){
	    pr = NULL;
	    pi = NULL;
	}
        g5n_get_forceMC(devid, nii, a+off, p+off,pr, pi);
    }
}

void g5n_calculate_force_on_xMC(int devid, double (*x)[3], 
				 int * index,
				 double (*a)[3], double *p,
				 double * rnnb2,
				 int * innb, int ni)
{
    double * eps2 = NULL;
    g5sn_calculate_force_on_xMC(devid, x, eps2, index, a, p, rnnb2, innb, ni);
}

void g5s_calculate_force_on_xMC(int devid, double (*x)[3], 
				double * eps2,
				double (*a)[3], double *p,
				int ni)
{
    int  * index = NULL;
    g5sn_calculate_force_on_xMC(devid, x, eps2, index, a, p, NULL, NULL, ni);
}

void g5_calculate_force_on_xMC(int devid, double (*x)[3], 
				double (*a)[3], double *p,
				int ni)
{
    int  * index = NULL;
    double * eps2 = NULL;
    g5sn_calculate_force_on_xMC(devid, x, eps2, index, a, p, NULL, NULL, ni);
}


struct grape_result_struct * g5_resbufp(int devid)
{
    return resbufp[devid];
}


int g5_get_number_of_pipelinesMC(int devid)
{
    return SING_number_of_pe()*4;
}

int g5_get_jmemsizeMC(int devid)
{
    return SING_elt_size0();
}

void g5_set_etaMC(int devid, double eta){}
void g5_set_h_to_allMC(int devid, double h){}
void g5_set_hMC(int devid, int ni, double *h){}
int g5_read_neighbor_listMC(int devid){return 0;}
int g5_get_neighbor_listMC(int devid, int ip, int *list){return 0;}
int g5_get_nbmemsizeMC(int devid){return 0;}
int g5_set_nbmemsizeMC(int devid, int size){return 0;}


//
//

static int nchips = 4;
void g5_setnchips(int n)
{
    nchips = n;
}

static int chipids[MAXCHIPS];

void g5_open(void)
{
    if(debug_level)fprintf(stderr, "g5_open called\n");
    int i; for(i=0;i<nchips; i++) {
	g5_openMC(i);
	chipids[i] = i;
    }
}

void g5_close(void)
{
    if(debug_level)fprintf(stderr, "g5_close called\n");
    int i; for(i=0;i<nchips; i++) g5_closeMC(i);
}
void g5_set_range(double xmin, double xmax, double mmin){}

static int njparallel;
static int njsaved0 = -1;
static int njsaved1 = -1;

static int nmax_for_singlechip = 1024;

static UINT64 last_time_for_jp;


void g5sn_set_jp(int adr, int nj, double *m, double (*x)[3], double *eps2,
		 int *index)
{
    int i;
    if (debug_level) fprintf(stderr,"g5s_set_jp nj = %d\n", nj);
    njsaved0 = nj;
    if (nj <=nmax_for_singlechip){
	//	fprintf(stderr,"call setxj %d\n", nj);
	g5sn_set_jpMC(0, adr,  nj, m, x,eps2, index);
	return;
    }
    if (adr != 0){
	fprintf(stderr,"g5_set_jp: current version supports only adr=0\n");
	exit(-1);
    }
    njparallel = (nj+nchips-1)/nchips;
    for(i=0;i<nchips; i++) {
	int istart = njparallel* i;
	int njlocal = njparallel;
	if (istart + njlocal  > nj) njlocal = nj-istart;
	if (njlocal < 0) njlocal = 0;
	if (njlocal) g5sn_set_jpMC(i, 0, njlocal, m+istart, x+istart,
				   eps2+istart, index+istart);
	g5_set_nMC(i, njlocal);
    }
    SING_nwait(70000);
    last_time_for_jp = rdtsc2();
}

void g5n_set_jp(int adr, int nj, double *m, double (*x)[3], int * index)
{
    g5sn_set_jp(adr, nj, m, x, NULL, index);
}

void g5_set_jp(int adr, int nj, double *m, double (*x)[3])
{
    int i;
    if (debug_level) fprintf(stderr,"g5_set_jp nj = %d\n", nj);
    njsaved0 = nj;
    if (nj <=nmax_for_singlechip){
	//	fprintf(stderr,"call setxj %d\n", nj);
	g5_set_jpMC(0, adr,  nj, m, x);
	return;
    }
    if (adr != 0){
	fprintf(stderr,"g5_set_jp: current version supports only adr=0\n");
	exit(-1);
    }
    njparallel = (nj+nchips-1)/nchips;
    for(i=0;i<nchips; i++) {
	int istart = njparallel* i;
	int njlocal = njparallel;
	if (istart + njlocal  > nj) njlocal = nj-istart;
	if (njlocal < 0) njlocal = 0;
	if (njlocal) g5_set_jpMC(i, 0, njlocal, m+istart, x+istart);
	g5_set_nMC(i, njlocal);
    }
    SING_nwait(70000);
    last_time_for_jp = rdtsc2();
}


void g5s_set_jp(int adr, int nj, double *m, double (*x)[3], double *eps2)
{
    int i;
    if (debug_level) fprintf(stderr,"g5s_set_jp nj = %d\n", nj);
    njsaved0 = nj;
    if (nj <=nmax_for_singlechip){
	//	fprintf(stderr,"call setxj %d\n", nj);
	g5s_set_jpMC(0, adr,  nj, m, x,eps2);
	return;
    }
    if (adr != 0){
	fprintf(stderr,"g5_set_jp: current version supports only adr=0\n");
	exit(-1);
    }
    njparallel = (nj+nchips-1)/nchips;
    for(i=0;i<nchips; i++) {
	int istart = njparallel* i;
	int njlocal = njparallel;
	if (istart + njlocal  > nj) njlocal = nj-istart;
	if (njlocal < 0) njlocal = 0;
	if (njlocal) g5_set_jpMC(i, 0, njlocal, m+istart, x+istart);
	g5_set_nMC(i, njlocal);
    }
    SING_nwait(70000);
    last_time_for_jp = rdtsc2();
}

void g5_set_n(int nj)
{
    if (debug_level) fprintf(stderr,"g5_set_n nj = %d\n", nj);
    njsaved1 = nj;
    if (nj <=nmax_for_singlechip){
	//	fprintf(stderr,"call setn %d\n", nj);
	g5_set_nMC(0, nj);
	return;
    }
}

void g5_calculate_force_on_x_old(double (*x)[3], double (*a)[3], double *p, int ni)
{
    int off, nii, np;
    np = g5_get_number_of_pipelines();
    for (off = 0; off < ni; off += np) {
        nii = np;
        if (off+nii > ni) {
            nii = ni - off;
        }
	g5_set_xi(nii, x+off);
        g5_run();
        g5_get_force(nii, a+off, p+off);
    }
}

void g5sn_calculate_force_on_x(double (*x)[3], double * eps2, int * index,
			       double (*a)[3], double *p,
			       double * rnnb2,
			       int * innb, int ni)
{
    int off, nii, np;
    np = g5_get_number_of_pipelines();
    //    fprintf(stderr,"g5sn_calculate_force_on_x ni= %d\n", ni);
    for (off = 0; off < ni; off += np) {
        nii = np;
        if (off+nii > ni) {
            nii = ni - off;
        }
	g5_set_xi(nii, x+off);
	if (eps2 != NULL) g5_set_eps2(nii, eps2+off);
	if (index != NULL) g5_set_index(nii, index+off);
        g5_run();
	double * pr = rnnb2+off;
	int * pi    = innb+off;
	if (rnnb2 == NULL){
	    pr = NULL;
	    pi = NULL;
	}
        g5n_get_force(nii, a+off, p+off,pr, pi);
    }
}

void g5n_calculate_force_on_x(double (*x)[3], int * index,
			      double (*a)[3], double *p,
			      double * rnnb2,
			      int * innb, int ni)
{
    g5sn_calculate_force_on_x(x, NULL, index, a, p, rnnb2, innb, ni);
}
    

void g5_calculate_force_on_x(double (*x)[3], double (*a)[3], double *p, int ni)
{
    g5sn_calculate_force_on_x(x, NULL, NULL, a, p, NULL, NULL, ni);
}


void g5_set_xi(int ni, double (*x)[3])
{
    if (debug_level){
	fprintf(stderr,"g5_set_xi, ni= %d, x[0]= %e %e %e\n",
		ni, x[0][0], x[0][1], x[0][2]);
    }
    g5_set_xiMC(0,ni, x);
}
void g5_run(void)
{
    if  (njsaved1 != njsaved0){
	fprintf(stderr,"g5_set_n and g5_set_jp should have used same nj, different = %d %d\n",
		njsaved0, njsaved1);
	exit(-1);
    }
    int devid = 0;
    if (njsaved0 <=nmax_for_singlechip){
	//	fprintf(stderr,"call run %d\n", nja[0]);
	g5_runMC(0);
	return ;
    }
    reset_signal_flag_for_exit();
    UINT64 start = rdtsc2();
    if (start > last_time_for_jp){
	while(rdtsc2() < last_time_for_jp+1000000);
    }
    int nireal=g5_get_number_of_pipelinesMC(devid);
    clear_unused_i_particles(devid, ipbufp[devid], nia[devid]);
    int ic;
    //    for(ic=0;ic<nchips; ic++){
    //	clear_unused_j_particles(ic, nja[ic], njparallel);
    //    }
    //    fprintf(stderr,"call SING_send_i_particleMCP\n");
    SING_send_i_particleMCP(nchips, chipids, ipbufp[devid], nireal);
    //    fprintf(stderr,"return from SING_send_i_particleMCP\n");
    iptime += rdtsc2()-start;
    ipcalls++;
    start = rdtsc2();
    int njreal = (nja[devid]+SING_number_of_bb()-1)/SING_number_of_bb();
    //    fprintf(stderr,"call SING_grape_runMCP\n");
    if (debug_level){
	fprintf(stderr,"g5_run, nj= %d, njreal=%d\n", njsaved0, njreal);
    }
    SING_grape_runMCP(nchips,chipids,njreal);
    //    fprintf(stderr,"return from SING_grape_runMCP\n");
    runtime += rdtsc2()-start;
    runcalls++;
    set_signal_flag_for_exit();
}
void g5_set_eps2(int ni, double *eps2)
{
    if(debug_level)fprintf(stderr, "g5_set_eps2 called ni=%d\n",ni);
    g5_set_eps2MC(0, ni, eps2);

}
void g5_set_eps2_to_all(double eps2)
{
    g5_set_eps2_to_allMC(0, eps2);
}
void g5_set_index(int ni, int *index)
{
    if(debug_level)fprintf(stderr, "g5_set_index called ni=%d\n",ni);
    g5_set_indexMC(0, ni, index);

}
void g5_get_force_old(int ni, double (*a)[3], double *pot)
{
    int ic;
    if (njsaved0 <=nmax_for_singlechip){
	g5_get_forceMC_internal(0, ni, a, pot,0);
	return;
    }
    for(ic=0;ic<nchips; ic++){
	g5_get_forceMC_internal(ic, ni, a, pot,ic);
    }
}

void g5n_get_force(int ni, double (*a)[3], double *pot,
		   double* rnnb2, int* innb)
{
    int ic;
    if (njsaved0 <=nmax_for_singlechip){
	g5n_get_forceMC_internal(0, ni, a, pot,rnnb2, innb, 0);
	return;
    }
    for(ic=0;ic<nchips; ic++){
	g5n_get_forceMC_internal(ic, ni, a, pot,rnnb2, innb, ic);
    }
}

void g5_get_force(int ni, double (*a)[3], double *pot)
{
    g5n_get_force(ni, a, pot, NULL, NULL);
}


int g5_get_number_of_pipelines(void)
{
    return g5_get_number_of_pipelinesMC(0);
}

int g5_get_jmemsize(void)
{
    return g5_get_jmemsizeMC(0);
}
void g5_set_eta(double eta)
{
    g5_set_etaMC(0, eta);
}
void g5_set_h_to_all(double h){}
void g5_set_h(int ni, double *h){}
int g5_read_neighbor_list(void){}
int g5_get_neighbor_list(int ip, int *list){}
int g5_get_nbmemsize(void){return 0;}
int g5_set_nbmemsize(int size){return 0;}
