//
// g5sim-gdr.h
//  Time-stamp: <11/01/19 22:43:54 makino>
// JM
// G5 interface functions according to g5user-j.pdf
//

//
// Low-level multichip functions
//

#ifndef _GRAPE5_H
#define _GRAPE5_H

#if defined(__cplusplus)
#define _cplusplus_
#endif

#if defined(_cplusplus_)
extern "C" {
#endif
    
    void g5_set_debug_level(int level);
    
    
    void g5_openMC(int devid);
    void g5_closeMC(int devid);
    void g5_set_rangeMC(int devid, double xmin, double xmax, double mmin);
    void g5_set_jpMC(int devid, int adr, int nj, double *m, double (*x)[3]);
    void g5s_set_jpMC(int devid, int adr, int nj, double *m, double (*x)[3],
		      double *eps2);
    void g5n_set_jpMC(int devid, int adr, int nj, double *m, double (*x)[3],
		      int *index);
    void g5sn_set_jpMC(int devid, int adr, int nj, double *m, double (*x)[3],
		       double *eps2, int *index);
    void g5_set_xiMC(int devid, int ni, double (*x)[3]);
    void g5_runMC(int devid);
    void g5_set_nMC(int devid, int n);
    void g5_set_eps2MC(int devid, int ni, double *eps2);
    void g5_set_eps2_to_allMC(int devid, double eps2);
    void g5_set_indexMC(int devid, int ni, int *index);
    void g5_get_forceMC(int devid, int ni, double (*a)[3], double *pot);
    void g5n_get_forceMC(int devid, int ni, double (*a)[3], double *pot,
			 double* rnnb2, int* innb);
    int g5_get_number_of_pipelinesMC(int devid);
    int g5_get_jmemsizeMC(int devid);
    void g5_set_etaMC(int devid, double eta);
    void g5_set_h_to_allMC(int devid, double h);
    void g5_set_hMC(int devid, int ni, double *h);
    int g5_read_neighbor_listMC(int devid);
    int g5_get_neighbor_listMC(int devid, int ip, int *list);
    int g5_get_nbmemsizeMC(int devid);
    int g5_set_nbmemsizeMC(int devid, int size);
    void g5sn_calculate_force_on_xMC(int devid, double (*x)[3], double *eps2,
				     int *index,
				     double (*a)[3], double *p, double * rnnb2,
				     int * innb, int ni);
    void g5n_calculate_force_on_xMC(int devid, double (*x)[3], 
				    int * index,
				    double (*a)[3], double *p,
				    double * rnnb2,
				    int * innb, int ni);
    void g5s_calculate_force_on_xMC(int devid, double (*x)[3], 
				    double * eps2,
				    double (*a)[3], double *p,
				    int ni);
    void g5_calculate_force_on_xMC(int devid, double (*x)[3], 
				   double (*a)[3], double *p,
				   int ni);
    
    struct grape_result_struct * g5_resbufp(int devid);
    struct grape_i_particle_struct * g5_ibufp(int devid);
    
    //
    // High-level aggregate functions
    //
    
    //
    // function not in manual
    //
    void g5_setnchips(int n);
    
    
    void g5_open(void);
    void g5_close(void);
    void g5_set_range(double xmin, double xmax, double mmin);
    void g5_set_jp(int adr, int nj, double *m, double (*x)[3]);
    void g5s_set_jp(int adr, int nj, double *m, double (*x)[3],
		    double *eps2);
    void g5n_set_jp(int adr, int nj, double *m, double (*x)[3],
		    int *index);
    void g5sn_set_jp(int adr, int nj, double *m, double (*x)[3],
		     double *eps2, int *index);
    void g5_calculate_force_on_x(double (*x)[3], double (*a)[3],
				 double *p, int ni);
    void g5n_calculate_force_on_x(double (*x)[3], int * index,
				  double (*a)[3],
				  double *p, double * rnnb2,
				  int * innb, int ni);
    void g5sn_calculate_force_on_x(double (*x)[3], double *eps2, int * index,
				   double (*a)[3], double *p, double * rnnb2,
				  int * innb, int ni);
    void g5_set_xi(int ni, double (*x)[3]);
    void g5_run(void);
    void g5_set_n(int nj);
    void g5_set_eps2(int ni, double *eps2);
    void g5_set_eps2_to_all(double eps2);
    void g5_set_index(int ni, int *index);
    void g5_get_force(int ni, double (*a)[3], double *pot);
    void g5n_get_force(int ni, double (*a)[3], double *pot,
		       double* rnnb2, int* innb);
    int g5_get_number_of_pipelines(void);
    int g5_get_jmemsize(void);
    void g5_set_eta(double eta);
    void g5_set_h_to_all(double h);
    void g5_set_h(int ni, double *h);
    int g5_read_neighbor_list(void);
    int g5_get_neighbor_list(int ip, int *list);
    int g5_get_nbmemsize(void);
    int g5_set_nbmemsize(int size);
    void g5_report_timers(void);
   
#if defined(_cplusplus_)
} /* extern "C" */
#endif    
#endif /* _GRAPE5_H*/

// constants and defs for internal use
#define VERY_LARGE_NUMBER (1e300)
