//
// grapesing.c
// grape-like simple N-squre inerface for simplegravity.
//
// J. Makino
//    Time-stamp: <10/08/27 16:06:23 makino>

#include <stdio.h>
#include <math.h>
#ifdef SIMPLEGRAVITY
#include "simplegravity.h"
#define HEADER
#endif

#ifdef SIMPLEGRAVITYFULL
#include "simplegravity-16x32.h"
#define HEADER
#define gravity_grape_init        SING_grape_init
#define gravity_send_j_particleMC SING_send_j_particleMC
#define gravity_send_i_particleMC SING_send_i_particleMC
#define gravity_grape_runMC       SING_grape_runMC
#define gravity_get_resultMC      SING_get_resultMC

#endif


#ifndef HEADER
#include "gravity.h"
#define HEADER
#define AVOID_SELF
#endif

void calculate_accel_by_singsim(int n,
				double x[][3],
				double m[],
				double a[][3],
				double pot[],
				double eps2)
{
    static first=1;
    int i, niparallel;
    double epsinv = 1.0/sqrt(eps2);
    if (first){
#ifdef 	SIMPLEGRAVITYFULL
	SING_grape_initMC(0);
#else	
	gravity_grape_init();
#endif	
	first=0;
    }
    for(i=0;i<n;i++){
	struct grape_j_particle_struct jp;
	jp.xj=x[i][0];
	jp.yj=x[i][1];
	jp.zj=x[i][2];
	jp.mj=m[i];
	jp.eps2=eps2;
#ifdef AVOID_SELF
	jp.idxj=i;
#endif	
	gravity_send_j_particleMC(0,&jp, i);
    }
    niparallel = SING_number_of_pe()*4;
    for (i=0; i< n;i+=niparallel){
	int k;
	int njreal = n/SING_number_of_bb();
	struct grape_i_particle_struct ip[2048];
	struct grape_result_struct result[2048];
	for(k=0;k<niparallel; k++){
	    ip[k].xi=x[i+k][0];
	    ip[k].yi=x[i+k][1];
	    ip[k].zi=x[i+k][2];
#ifdef  SIMPLEGRAVITYFULL
	    ip[k].eps2i=0;
#endif	    
#ifdef AVOID_SELF
	    ip[k].idxi=i+k;
#endif	
	}
	gravity_send_i_particleMC(0,ip, niparallel);
	gravity_grape_runMC(0,njreal);
	gravity_get_resultMC(0,result);
	for(k=0;k<niparallel; k++){
	    a[i+k][0]=result[k].accx;
	    a[i+k][1]=result[k].accy;
	    a[i+k][2]=result[k].accz;
	    pot[i+k] = -result[k].pot;
	}
	
    }
#ifndef AVOID_SELF
    for(i=0;i<n;i++) pot[i] += m[i]*epsinv;
#endif
}
