//
// testg5.c
// test driver for simplegravity.vsm-generated interface library
//
// J. Makino
//    Time-stamp: <11/01/05 03:42:47 makino>

#include <stdio.h>
#include <sys/time.h>
#include <stdlib.h>
#include <math.h>
#include <getopt.h>
#include "simplegravity-16x32.h"
#include "g5sim-gdr.h"

#define NJ 32
#define NI 32
#define NJMAX 65536
#define MAXNCHIPS 8
struct grape_j_particle_struct xj[32768];
struct grape_i_particle_struct xi[32768];
struct grape_result_struct result[MAXNCHIPS][32768];
struct grape_result_struct result_host[32768];

double x[NJMAX][3];
double m[NJMAX];
double eps2[NJMAX];
int jindex[NJMAX];
double a[NJMAX][3];
double p[NJMAX];
double rnnb2[NJMAX];
int nnbindex[NJMAX];
double ahost[NJMAX][3];
double phost[NJMAX];
double rnnb2host[NJMAX];
int nnbindexhost[NJMAX];


void calc_on_host(int n, double x[][3], double *m, int* index, double * eps2,
		  double a[][3], double * p, double * rnnb2, int * nnbindex)
{
    int i;
    for(i=0;i<n; i++){
	a[i][0]=a[i][1]=a[i][2]=p[i]=0;
	rnnb2[i]=1e200;
    }
    //#pragma omp parallel for private(i)
    for(i=0;i<n; i++){
	int j;
	for(j=0;j<n; j++){
	    double dx = x[j][0]-x[i][0];
	    double dy = x[j][1]-x[i][1];
	    double dz = x[j][2]-x[i][2];
	    double r2 = (dx*dx+dy*dy+dz*dz+eps2[j]+eps2[i]);
	    double r2inv = 1.0/r2;
	    double mrinv = m[j]*sqrt(r2inv);
	    double mr3inv = mrinv*r2inv;
 //	    fprintf(stderr, "%d %d %e %e %e %e\n", i, j, dx, dy, dz, mr3inv);
	    p[i] -= mrinv;
	    a[i][0] += mr3inv * dx;
	    a[i][1] += mr3inv * dy;
	    a[i][2] += mr3inv * dz;
	    if (i!= j){
		if (r2 < rnnb2[i]){
		    rnnb2[i]=r2;
		    nnbindex[i]=index[j];
		}
	    }
	}
    }
}

double randx(){return drand48()*2-1;}

void init(int n)
{
    int i, j;
    volatile double  e, z;
    e = 1.0e-12;
    z = 0.0;
    for(i=0;i<n;i++){
	x[i][0]=randx();
	x[i][1]=randx();
	x[i][2]=randx();
	m[i]=1.0/n;
	eps2[i] =0.01+i*0.0001;
	jindex[i]=i;
    }
}

void print_result(int n)
{
    
    int i;
    double emax = 0;
    double pemax = 0;
    int nnberrs = 0;
    for(i=0;i<n;i++){
	printf("test i=%d, x, y, z, pot= %15.7e %15.7e %15.7e %15.7e %5d\n",
	       i, a[i][0], a[i][1], a[i][2], p[i], nnbindex[i]);
	printf("host i=%d, x, y, z, pot= %15.7e %15.7e %15.7e %15.7e %5d\n",
	       i, ahost[i][0], ahost[i][1], ahost[i][2], phost[i],
	       nnbindexhost[i]);
	double e2sum = 0;
	int k;
	for(k=0;k<3;k++){
	    double err = a[i][k]-ahost[i][k];
	    e2sum += err*err;
	}
	double perr = fabs(p[i]-phost[i]);
	if (perr > pemax) pemax = perr;
	if (e2sum > emax) emax = e2sum;
	printf("err = %g %g\n", sqrt(e2sum), perr);
	if (nnbindexhost[i] != nnbindex[i]){
	    printf("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n");
	    printf("nnb error: i=%d, nnbs = %d != %d\n",
		   i, nnbindexhost[i],nnbindex[i]);
	    printf("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n");
	    nnberrs++;
	}
    }
    printf("Max err = %g %g nnb errors = %d\n", sqrt(emax), pemax, nnberrs);
}    

double wallclock()
{
  struct timeval tv;
	
  gettimeofday(&tv, NULL);
  return tv.tv_sec + 1e-6*tv.tv_usec;
}

double curr_time;

double starttime()
{
  curr_time = wallclock();
}

double laptime()
{
  double lap;
  lap = wallclock() - curr_time;
  curr_time = wallclock();
  return lap;
}


typedef struct parmstruct{
    int firstchip;
    int nchips;
    int n;
    int kmax;
    int debug_level;
} PARMS, *PPARMS;


void usage()
{
    fprintf(stderr,"testg5 options:\n");
    fprintf(stderr,"  -h: This help\n");
    fprintf(stderr,"  -f: id of the first chip(default=0)\n");
    fprintf(stderr,"  -c: the number of chips(default=1)\n");
    fprintf(stderr,"  -n: # particles (default=128)\n");
    fprintf(stderr,"  -d: # debug_level (default=0)\n");
    fprintf(stderr,"  -k: # of repeat for debug (default=1\n");
}


extern char *optarg;
extern int optind;

void print_parms(FILE* stream, PPARMS parms)
{
    fprintf(stream,"firstchip=%d nchips=%d n=%d  kmax=%d debug_level=%d\n",
	    parms->firstchip, parms->nchips,  parms->n,
	    parms->kmax, parms->debug_level);
}

void read_parms(int argc, char * argv[], PPARMS parms)
{
    int ch;
    static struct option longopts[] = {
	{ "help",      no_argument,      0,           'h' },
	{ "firstchip",      optional_argument,      NULL,           'f' },
	{ "nchips",      optional_argument,      NULL,           'c' },
	{ "n",      optional_argument,      NULL,           'n' },
	{ "kmax",      optional_argument,      NULL,           'k' },
	{ "debug_level",    optional_argument,      NULL,           'd' },
	{ NULL,         0,                      NULL,           0 }
    };
    parms->firstchip=0;
    parms->nchips=1;
    parms->n=128;
    parms->kmax=1;
    parms->debug_level=0;
    while((ch=getopt_long(argc,argv,"c:d:f:hk:n:",longopts, NULL))!= -1){
	fprintf(stderr,"optchar = %c optarg=%s\n", ch,optarg);
	switch (ch) {
	    case 'c': parms->nchips = atoi(optarg); break;
	    case 'f': parms->firstchip = atoi(optarg); break;
	    case 'n': parms->n = atoi(optarg); break;
	    case 'd': parms->debug_level = atoi(optarg); break;
	    case 'k': parms->kmax = atoi(optarg); break;
	    case 'h': usage(); exit(1);
	    case '?':usage(); exit(1);
		break;
	    default:break;
	}
    }
    argc -= optind;
    argv += optind;
    if (parms->firstchip != 0){
	if (parms->nchips != 1){
	    fprintf(stderr,
		    "For non-zero firstchip, nchips must be 1, 1 forced\n");
	    parms-> nchips = 1;
	}
    }else{
	if ((parms->nchips > 1) && (parms->nchips != 4)){
	    fprintf(stderr, "nchips must be either 1 or 4, 4 forces\n");
	    parms-> nchips = 4;
	}
	
    }
    print_parms(stdout, parms);
}

int main(int argc, char **argv)
{
    PARMS parms;
    read_parms(argc, argv, &parms);
    double start, end, jtrans, itrans, run;
    int firstchip, nchips, chipid;
    firstchip=parms.firstchip;
    nchips=parms.nchips;
    int n = parms.n;
    init(n);
    g5_set_debug_level(parms.debug_level);
    if (nchips == 4){
	g5_open();
	g5sn_set_jp(0,  n, m, x, eps2, jindex);
	//	g5n_set_jp(0,  n, m, x, jindex);
	g5_set_n(n);
	g5sn_calculate_force_on_x(x, eps2, jindex, a, p, rnnb2, nnbindex,n);
    }else{
	g5_openMC(firstchip);
	g5sn_set_jpMC(firstchip, 0,  n, m, x, eps2, jindex);
	//	g5n_set_jpMC(firstchip, 0,  n, m, x,  jindex);
	g5_set_nMC(firstchip, n);
	g5sn_calculate_force_on_xMC(firstchip, x, eps2, jindex, a, p,
				    rnnb2, nnbindex,n);
    }
    calc_on_host(n, x, m, jindex, eps2, ahost, phost, rnnb2host, nnbindexhost);
    print_result(n);
}
