//
// testsimplegravity.c
// test driver for simplegravity.vsm-generated interface library
//
// J. Makino
//    Time-stamp: <10/11/23 22:57:48 makino>

#include <stdio.h>
#include <sys/time.h>
#include <stdlib.h>
#include <math.h>
#include <getopt.h>
#include "simplegravity-16x32.h"

#define NJ 32
#define NI 32
#define MAXNCHIPS 8
struct grape_j_particle_struct xj[32768];
struct grape_i_particle_struct xi[32768];
struct grape_result_struct result[MAXNCHIPS][32768];
struct grape_result_struct result_host[32768];

void calc_on_host(int ni, int nj, struct grape_j_particle_struct * jp0,
		  struct grape_i_particle_struct * ip0,
		  struct grape_result_struct * resp0)
{
    int i, j;
    for(i=0;i<ni; i++){
	struct grape_result_struct * p = resp0+i;
	p->accx = p->accy = p->accz = p->pot = 0;
    }
    for(i=0;i<ni; i++){
	struct grape_i_particle_struct * ip = ip0+i;
	struct grape_result_struct* resp  = resp0+i;
	for(j=0;j<nj; j++){
	    struct grape_j_particle_struct * jp = jp0+j;
	    double dx = jp->xj - ip->xi;
	    double dy = jp->yj - ip->yi;
	    double dz = jp->zj - ip->zi;
	    double r2inv = 1.0/(dx*dx+dy*dy+dz*dz+jp->eps2+ip->eps2i);
	    double mrinv = jp->mj*sqrt(r2inv);
	    double mr3inv = mrinv*r2inv;
	    //	    fprintf(stderr,"i, j, mrinv = %d %d %g %g %g\n",
	    //		    i, j, r2inv, mrinv, mr3inv);
	    resp->pot -= mrinv;
	    resp->accx += mr3inv * dx;
	    resp->accy += mr3inv * dy;
	    resp->accz += mr3inv * dz;
	}
    }
}

double randx(){return drand48()*2-1;}

void init(int ni, int nj)
{
  int i, j;
  volatile double m, e, z;
  m =1.0;
  e = 1.0e-12;
  z = 0.0;
  for(i=0;i<nj;i++){
      xj[i].xj=randx();
      xj[i].yj=randx();
      xj[i].zj=randx();
      xj[i].mj=1.0/nj;
      xj[i].eps2 =0.01;
  }

    for(i=0;i<ni;i++){
      xi[i].xi=randx();
      xi[i].yi=randx();
      xi[i].zi=randx();
      xi[i].eps2i =0.01;
  }

}

void init0(int ni, int nj)
{
  int i, j;
  volatile double m, e, z;
  m =1.0;
  e = 1.0e-12;
  z = 0.0;
  srand48(1);
  for(i=0;i<nj;i++){
      xj[i].xj=  xj[i].yj=  xj[i].zj= 0;
      xj[i].mj=1.0/nj;
      xj[i].eps2 =0.01;
  }
  
  for(i=0;i<ni;i++){
      xi[i].xi=0;
      xi[i].yi=1;
      xi[i].zi=0;
      xi[i].eps2i =0.01;
  }

  for(j=0;j<nj;j++)
    SING_send_elt_data0(xj+j, j);
}

    
print_result_old(struct grape_result_struct * result)
{
  struct grape_result_struct *p;
  int i;
  p=result;
  for(i=0;i<(NI*4);i++){
    printf("test result i=%d, x, y, z, pot= %e %e %e %e\n",
	   i,p->accx,p->accy,p->accz,p->pot);
    p++;
  }
}    
void print_result(struct grape_result_struct * result,
		  struct grape_result_struct * result2, int ni)
{

    int i;
    double emax = 0;
    double pemax = 0;
    for(i=0;i<ni;i++){
	struct grape_result_struct *p;
	struct grape_result_struct *pp;
	p=result+i;
	printf("test result i=%d, x, y, z, pot= %15.7e %15.7e %15.7e %15.7e ",
	       i,p->accx,p->accy,p->accz,p->pot);
	pp= result2+i;
	//	printf("host result i=%d, x, y, z, pot= %e %e %e %e\n",
	//	       i,pp->accx,pp->accy,pp->accz,pp->pot);
	double err = p->accx - pp->accx;
	double e2sum = err*err;
	 err = p->accy - pp->accy;
	e2sum += err*err;
	 err = p->accz - pp->accz;
	e2sum += err*err;
	double perr = fabs(p->pot+pp->pot);
	if (perr > pemax) pemax = perr;
	if (e2sum > emax) emax = e2sum;
	printf("err = %g %g\n", sqrt(e2sum), perr);
  }
    printf("Max err = %g %g\n", sqrt(emax), pemax);
}    

double wallclock()
{
  struct timeval tv;
	
  gettimeofday(&tv, NULL);
  return tv.tv_sec + 1e-6*tv.tv_usec;
}

double curr_time;

double starttime()
{
  curr_time = wallclock();
}

double laptime()
{
  double lap;
  lap = wallclock() - curr_time;
  curr_time = wallclock();
  return lap;
}

#if 0
int main()
{
  int i, niparallel, ni, nj, j;
  double start, end, jtrans;
  
  singchip_id = 0;
  ni=NI*4;
  nj=NJ;
  SING_grape_init();
  SING_grape_cleanup();
  SING_grape_init_noopen();
  init();
  niparallel = SING_number_of_pe()*4;
    
	    
  start = wallclock();
  for(j=0;j<10;j++){
    starttime();
    SING_send_elt_data0_m(xj, 0, nj);
    jtrans = laptime();
    for (i=0; i< ni;i+=niparallel){
      int njreal = nj/SING_number_of_bb();
      SING_send_i_particle(xi+i, niparallel);
      //usleep(1000);
      //printf("i, njreal=%d %d\n", i, njreal);
      SING_grape_run(njreal);
      //SING_grape_run(1);
      SING_get_result(result+i);
      //printf("%d \n", j);
    }
  }
  end = wallclock();
  print_result(result);
  printf("elapsed time %e [sec] jtrans %e [sec]. \n", end-start, jtrans);
}
#endif



typedef struct parmstruct{
    int firstchip;
    int nchips;
    int nj;
    int ni;
    int kmax;
} PARMS, *PPARMS;


void usage()
{
    fprintf(stderr,"testdgemm options:\n");
    fprintf(stderr,"  -h: This help\n");
    fprintf(stderr,"  -f: id of the first chip(default=0)\n");
    fprintf(stderr,"  -c: the number of chips(default=1)\n");
    fprintf(stderr,"  -i: # iparticles (default=128)\n");
    fprintf(stderr,"  -j: # jparticles (default=128)\n");
    fprintf(stderr,"  -k: # of repeat for debug (default=1\n");
}


extern char *optarg;
extern int optind;

void print_parms(FILE* stream, PPARMS parms)
{
    fprintf(stream,"firstchip=%d nchips=%d ni=%d nj=%d kmax=%d\n",
	    parms->firstchip, parms->nchips,  parms->ni,
	    parms->nj, parms->kmax);
}

void read_parms(int argc, char * argv[], PPARMS parms)
{
    int ch;
    static struct option longopts[] = {
	{ "help",      no_argument,      0,           'h' },
	{ "firstchip",      optional_argument,      NULL,           'f' },
	{ "nchips",      optional_argument,      NULL,           'c' },
	{ "ni",      optional_argument,      NULL,           'i' },
	{ "nj",      optional_argument,      NULL,           'j' },
	{ "kmax",      optional_argument,      NULL,           'k' },
	{ NULL,         0,                      NULL,           0 }
    };
    parms->firstchip=0;
    parms->nchips=1;
    parms->ni=128;
    parms->nj=128;
    parms->kmax=1;
    while((ch=getopt_long(argc,argv,"c:f:hi:j:k:",longopts, NULL))!= -1){
	fprintf(stderr,"optchar = %c optarg=%s\n", ch,optarg);
	switch (ch) {
	    case 'c': parms->nchips = atoi(optarg); break;
	    case 'f': parms->firstchip = atoi(optarg); break;
	    case 'i': parms->ni = atoi(optarg); break;
	    case 'j': parms->nj = atoi(optarg); break;
	    case 'k': parms->kmax = atoi(optarg); break;
	    case 'h': usage(); exit(1);
	    case '?':usage(); exit(1);
		break;
	    default:break;
	}
    }
    argc -= optind;
    argv += optind;
    print_parms(stdout, parms);
}

#define FORALL for(chipid=firstchip; chipid < firstchip+nchips; chipid++)
int main(int argc, char **argv)
{
    PARMS parms;
    read_parms(argc, argv, &parms);
    int i, niparallel, ni, nj, j;
    double start, end, jtrans, itrans, run;
    int firstchip, nchips, chipid;
    firstchip=parms.firstchip;
    nchips=parms.nchips;
    ni = parms.ni;
    nj = parms.nj;
    FORALL SING_grape_initMC(chipid);
    niparallel = SING_number_of_pe()*4;
    srand48(1);
    jtrans=itrans=run=0;
	    
    start = wallclock();

    init(ni,nj);
    for(j=0;j<1;j++){
	//#pragma omp parallel for private(chipid)
	FORALL {
	    int j;
	    for(j=0;j<nj;j++) {
		SING_send_j_particleMC(chipid,xj+j,j);
	    }
	}
	
#if 0
	starttime();
#pragma omp parallel for private(chipid)
	FORALL SING_send_elt_data0_mMC(chipid, xj, 0, nj);
	jtrans = laptime();
	for (i=0; i< ni;i+=niparallel){
	    int njreal = nj/SING_number_of_bb();
	    starttime();
#pragma omp parallel for private(chipid)
	    FORALL SING_send_i_particleMC(chipid, xi+i, niparallel);
	    itrans = laptime();
	    //usleep(1000);
	    //printf("i, njreal=%d %d\n", i, njreal);
	    starttime();
#pragma omp parallel for private(chipid)
	    FORALL SING_grape_runMC(chipid, njreal);
	    //SING_grape_run(1);
#pragma omp parallel for private(chipid)
	    FORALL SING_get_resultMC(chipid, result[chipid]+i);
	    run = laptime();
	    //printf("%d \n", j);
	}
#else

#pragma omp parallel for private(chipid)
	FORALL{
	    int i;
	    int kk;
	    for(kk=0;kk<parms.kmax;kk++){
		starttime();
		SING_send_elt_data0_mMC(chipid, xj, 0, nj);
		jtrans = laptime();
		//	    }
		//#pragma omp barrier
		//	    for(kk=0;kk<parms.kmax;kk++){
		for (i=0; i< ni;i+=niparallel){
		    int njreal = nj/SING_number_of_bb();
		    SING_send_i_particleMC(chipid, xi+i, niparallel);
		    //#pragma omp barrier
		    SING_grape_runMC(chipid, njreal);
		    SING_get_resultMC(chipid, result[chipid]+i);
		}
		     
	    }
	}
#endif	    
	calc_on_host(ni, nj, xj, xi, result_host);
	FORALL print_result(result[chipid], result_host,ni);
    }
    end = wallclock();
    FORALL SING_grape_closeMC(chipid);
    //    printf("elapsed time %e [sec] jtrans %e [sec] itrans %e [sec] run %e [sec]. \n", end-start, jtrans, itrans, run);
}
