//
// testsimplegravity.c
// test driver for simplegravity.vsm-generated interface library
//
// J. Makino
//    Time-stamp: <10/08/31 07:40:11 makino>

#include <stdio.h>
#include <sys/time.h>
#include <stdlib.h>
#include <math.h>
#include "simplegravity-16x32.h"

#define NJ 32
#define NI 32

struct grape_j_particle_struct xj[32768];
struct grape_i_particle_struct xi[32768];
struct grape_result_struct result[32768];
struct grape_result_struct result_host[32768];

void calc_on_host(int ni, int nj, struct grape_j_particle_struct * jp0,
		  struct grape_i_particle_struct * ip0,
		  struct grape_result_struct * resp0)
{
    int i, j;
    for(i=0;i<ni; i++){
	struct grape_result_struct * p = resp0+i;
	p->accx = p->accy = p->accz = p->pot = 0;
    }
    for(i=0;i<ni; i++){
	struct grape_i_particle_struct * ip = ip0+i;
	struct grape_result_struct* resp  = resp0+i;
	for(j=0;j<nj; j++){
	    struct grape_j_particle_struct * jp = jp0+j;
	    double dx = jp->xj - ip->xi;
	    double dy = jp->yj - ip->yi;
	    double dz = jp->zj - ip->zi;
	    double r2inv = 1.0/(dx*dx+dy*dy+dz*dz+jp->eps2+ip->eps2i);
	    double mrinv = jp->mj*sqrt(r2inv);
	    double mr3inv = mrinv*r2inv;
	    //	    fprintf(stderr,"i, j, mrinv = %d %d %g %g %g\n",
	    //		    i, j, r2inv, mrinv, mr3inv);
	    resp->pot -= mrinv;
	    resp->accx += mr3inv * dx;
	    resp->accy += mr3inv * dy;
	    resp->accz += mr3inv * dz;
	}
    }
}

double randx(){return drand48()*2-1;}

void init(int ni, int nj, int chipid)
{
  int i, j;
  volatile double m, e, z;
  m =1.0;
  e = 1.0e-12;
  z = 0.0;
  for(i=0;i<nj;i++){
      xj[i].xj=randx();
      xj[i].yj=randx();
      xj[i].zj=randx();
      xj[i].mj=1.0/nj;
      xj[i].eps2 =0.01;
  }

    for(i=0;i<ni;i++){
      xi[i].xi=randx();
      xi[i].yi=randx();
      xi[i].zi=randx();
      xi[i].eps2i =0.01;
  }

    for(j=0;j<nj;j++) {
	//	SING_send_elt_data0MC(chipid, xj+j, j);
	SING_send_j_particleMC(chipid,xj+j,j);
    }
}

void init0(int ni, int nj)
{
  int i, j;
  volatile double m, e, z;
  m =1.0;
  e = 1.0e-12;
  z = 0.0;
  srand48(1);
  for(i=0;i<nj;i++){
      xj[i].xj=  xj[i].yj=  xj[i].zj= 0;
      xj[i].mj=1.0/nj;
      xj[i].eps2 =0.01;
  }
  
  for(i=0;i<ni;i++){
      xi[i].xi=0;
      xi[i].yi=1;
      xi[i].zi=0;
      xi[i].eps2i =0.01;
  }

  for(j=0;j<nj;j++)
    SING_send_elt_data0(xj+j, j);
}

    
print_result_old(struct grape_result_struct * result)
{
  struct grape_result_struct *p;
  int i;
  p=result;
  for(i=0;i<(NI*4);i++){
    printf("test result i=%d, x, y, z, pot= %e %e %e %e\n",
	   i,p->accx,p->accy,p->accz,p->pot);
    p++;
  }
}    
void print_result(struct grape_result_struct * result,
		  struct grape_result_struct * result2, int ni)
{

    int i;
    double emax = 0;
    double pemax = 0;
    for(i=0;i<ni;i++){
	struct grape_result_struct *p;
	struct grape_result_struct *pp;
	p=result+i;
	printf("test result i=%d, x, y, z, pot= %15.7e %15.7e %15.7e %15.7e ",
	       i,p->accx,p->accy,p->accz,p->pot);
	pp= result2+i;
	//	printf("host result i=%d, x, y, z, pot= %e %e %e %e\n",
	//	       i,pp->accx,pp->accy,pp->accz,pp->pot);
	double err = p->accx - pp->accx;
	double e2sum = err*err;
	 err = p->accy - pp->accy;
	e2sum += err*err;
	 err = p->accz - pp->accz;
	e2sum += err*err;
	double perr = fabs(p->pot+pp->pot);
	if (perr > pemax) pemax = perr;
	if (e2sum > emax) emax = e2sum;
	printf("err = %g %g\n", sqrt(e2sum), perr);
  }
    printf("Max err = %g %g\n", sqrt(emax), pemax);
}    

double wallclock()
{
  struct timeval tv;
	
  gettimeofday(&tv, NULL);
  return tv.tv_sec + 1e-6*tv.tv_usec;
}

double curr_time;

double starttime()
{
  curr_time = wallclock();
}

double laptime()
{
  double lap;
  lap = wallclock() - curr_time;
  curr_time = wallclock();
  return lap;
}

#if 0
int main()
{
  int i, niparallel, ni, nj, j;
  double start, end, jtrans;
  
  singchip_id = 0;
  ni=NI*4;
  nj=NJ;
  SING_grape_init();
  SING_grape_cleanup();
  SING_grape_init_noopen();
  init();
  niparallel = SING_number_of_pe()*4;
    
	    
  start = wallclock();
  for(j=0;j<10;j++){
    starttime();
    SING_send_elt_data0_m(xj, 0, nj);
    jtrans = laptime();
    for (i=0; i< ni;i+=niparallel){
      int njreal = nj/SING_number_of_bb();
      SING_send_i_particle(xi+i, niparallel);
      //usleep(1000);
      //printf("i, njreal=%d %d\n", i, njreal);
      SING_grape_run(njreal);
      //SING_grape_run(1);
      SING_get_result(result+i);
      //printf("%d \n", j);
    }
  }
  end = wallclock();
  print_result(result);
  printf("elapsed time %e [sec] jtrans %e [sec]. \n", end-start, jtrans);
}
#endif

int main(int argc, char **argv)
{
    int i, niparallel, ni, nj, j;
    double start, end, jtrans, itrans, run;
    int chipid;

    if(argc>=2){
	chipid = atoi(argv[1]); //0-3
    } else {
	chipid = 0;
    }
    if(argc==4){
	ni = atoi(argv[2]);
	nj = atoi(argv[3]);
    } else {
	ni=NI*4;
	nj=NJ;
    }
    //  SING_grape_init(chipid);
    //  SING_use_default_singstruct(chipid);
    SING_grape_initMC(chipid);
    niparallel = SING_number_of_pe()*4;
    srand48(1);
    
	    
    start = wallclock();
    for(j=0;j<5;j++){
	init(ni,nj, chipid);
	starttime();
	SING_send_elt_data0_mMC(chipid, xj, 0, nj);
	jtrans = laptime();
	for (i=0; i< ni;i+=niparallel){
	    int njreal = nj/SING_number_of_bb();
	    starttime();
	    SING_send_i_particleMC(chipid, xi+i, niparallel);
	    itrans = laptime();
	    //usleep(1000);
	    //printf("i, njreal=%d %d\n", i, njreal);
	    starttime();
	    SING_grape_runMC(chipid, njreal);
	    //SING_grape_run(1);
	    SING_get_resultMC(chipid, result+i);
	    run = laptime();
	    //printf("%d \n", j);
	}
	calc_on_host(ni, nj, xj, xi, result_host);
	print_result(result, result_host,ni);
    }
    end = wallclock();
    SING_grape_closeMC(chipid);
    printf("elapsed time %e [sec] jtrans %e [sec] itrans %e [sec] run %e [sec]. \n", end-start, jtrans, itrans, run);
}
