//
// testsimplegravity.c
// test driver for simplegravity.vsm-generated interface library
//
// J. Makino
//    Time-stamp: <05/03/15 21:41:35 makino>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "simplegravity.h"

#define N 4096
#define NJ 64

struct grape_j_particle_struct xj[NJ];
struct grape_i_particle_struct xi[N*4];
struct grape_result_struct result[N*4];

struct grape_result_struct result_host[N*4];

double erand()
{
  double m, e;
  m = (double)rand() / (double)RAND_MAX;
  e = rand() % 256 - 128;
  return pow(2.0, e)*m;
}

void init()
{
    int i;
    xj[0].xj=0;
    xj[0].yj=0;
    xj[0].zj=0;
    xj[0].mj=0;
    xj[0].eps2=1.0e-12;
    xj[1]=xj[0];

    xj[1].xj=0;
    xj[1].mj=1.0;
    xj[2]=xj[0];
    xj[2].xj=0;
    xj[2].mj=0.0;
    for(i=3;i<NJ;i++){
      xj[i] = xj[0];
    }
      
    for(i=0;i<NJ;i++)SING_send_j_particle(xj+i, i);
    
    xi[0].xi=0;
    xi[0].yi=0;
    xi[0].zi=0;
    for(i=1;i<(N*4);i++)xi[i]=xi[0];
    for(i=0;i<(N*4);i++)xi[i].xi=erand();
    for(i=0;i<(N*4);i++)xi[i].yi=erand();
    for(i=0;i<(N*4);i++)xi[i].zi=erand();
    for(i=0;i<(N*4);i++){
      /*
	fprintf(stderr,"test init i=%d, x, y, z= %e %e %e\n",
		i,xi[i].xi,xi[i].yi,xi[i].zi);
      */
    }
}

void host_gravity(struct grape_result_struct *result, struct grape_i_particle_struct *parti, struct grape_j_particle_struct *partj, int ni, int nj)
{
  int i, j;
  double rinv, dx, dy, dz;

 

  for(i=0;i<ni;i++){
     result[i].accx = result[i].accy = result[i].accz = 0.0;
    for(j=0;j<nj;j++){
      dx = partj[j].xj - parti[i].xi;
      dy = partj[j].yj - parti[i].yi;
      dz = partj[j].zj - parti[i].zi;
      rinv = sqrt(dx*dx + dy*dy + dz*dz + partj[j].eps2);
      result[i].accx += partj[j].mj * dx / (rinv*rinv*rinv);
      result[i].accy += partj[j].mj * dy / (rinv*rinv*rinv);
      result[i].accz += partj[j].mj * dz / (rinv*rinv*rinv);
    }
  }
}


void print_result(struct grape_result_struct * result)
{
    struct grape_result_struct *p;
    int i;
    p=result;
    for(i=0;i<(N*4);i++){
	printf("test result i=%d, x, y, z, pot= %e %e %e %e\n",
		i,p->accx,p->accy,p->accz,p->pot);
	p++;
    }
}    

void compare_result(struct grape_result_struct *result_gdr, struct grape_result_struct *result_host)
{
    struct grape_result_struct *p;
    int i;
    double a1, a2, da, p1, p2;
    p=result;
    for(i=0;i<(N*4);i++){
      a1 = sqrt(result_gdr[i].accx*result_gdr[i].accx + result_gdr[i].accy*result_gdr[i].accy + result_gdr[i].accz*result_gdr[i].accz);
      a2 = sqrt(result_host[i].accx*result_host[i].accx + result_host[i].accy*result_host[i].accy + result_host[i].accz*result_host[i].accz);
      printf("%e %e \n", fabs((a1-a2)/a2), a2);
      //printf("%e %e \n", a1, a2);
    }
}    

int main()
{
  int i, niparallel, ni, nj, j;
    ni=N*4;
    nj=NJ;
    SING_grape_init(0);
    init();
    niparallel = SING_number_of_pe()*4;
 
    for (i=0; i< ni;i+=niparallel){
	int njreal = nj/SING_number_of_bb();
	SING_send_i_particle(xi+i, niparallel);
	//printf("i, njreal=%d %d\n", i, njreal);
	SING_grape_run(njreal);
	SING_get_result(result+i);
    }

    host_gravity(result_host, xi, xj, N*4, NJ);
    compare_result(result, result_host);
    
}
