//
// g6sim-gdr.c
//  Time-stamp: <10/09/13 14:46:33 makino>
// JM 

#include <stdio.h>
#include <sys/time.h>
#include <stdlib.h>
#include "hermite-16x32.h"
#include "g6sim-gdr.h"
#include <math.h>

#define MAXCHIPS 8
#define MAXPIPELINESPERCHIP 512


#define ERREXIT {fprintf(stderr,"%s: not implemented.\n", __func__);    exit(-1);}


//  first trial version: use clusterid as chipid

void g6_reset_(int * clusterid)  // in manual
{
    // currently do nothing
}

void g6_reset(int clusterid)
{
    int i = clusterid;
    g6_reset_(&i);
}

void g6_set_tunit_(int * newtunit) // in manual
{
    //  no need to do anything
}
void g6_set_tunit(int newtunit)
{
    int i = newtunit;
    g6_set_tunit_(&i);
}
    
void g6_set_xunit_(int * newxunit)  // in manual
{
    //  no need to do anything
}
void g6_set_xunit(int newxunit)
{
    int i = newxunit;
    g6_set_xunit_(&i);
}
    

int g6_open_(int *clusterid)  // in manual
{
        SING_grape_initMC(*clusterid);

}
int g6_open(int clusterid)
{
    int i = clusterid;
    g6_open_(&i);
}
int g6_close_(int *clusterid)
{
    SING_grape_closeMC(*clusterid);
    return 0;
}
int g6_close(int clusterid)
{
    int i = clusterid;
    return g6_close_(&i);
}

int g6_reinitialize_(int * clusterid)
{
    g6_close_(clusterid);
    return g6_open_(clusterid);
}

int g6_reinitialize(int clusterid)
{
    int i = clusterid;
    return g6_reinitialize_(&i);
}


static double ti_stored[MAXCHIPS];
void g6_set_ti_(int *clusterid, double *ti)  // in manual
{
    ti_stored[*clusterid]= *ti;
}

int g6_set_j_particle_(int * clusterid,        // in manual
		       int *address,
		       int *index,
		       double *tj, /* particle time */
		       double *dtj, /* particle time */
		       double *mass,
		       double a2by18[3], /* a2dot divided by 18 */
		       double a1by6[3], /* a1dot divided by 6 */
		       double aby2[3], /* a divided by 2 */
		       double v[3], /* velocity */
		       double x[3] /* position */)
{

    struct SING_elt_struct0 jp;
    jp.x0= x[0];
    jp.y0= x[1];
    jp.z0= x[2];
    jp.tj= *tj;
    jp.sx18= a2by18[0];
    jp.sy18= a2by18[1];
    jp.sz18= a2by18[2];
    jp.jx6= a1by6[0];
    jp.jy6= a1by6[1];
    jp.jz6= a1by6[2];
    jp.ax2= aby2[0];
    jp.ay2= aby2[1];
    jp.az2= aby2[2];
    jp.v0x= v[0];
    jp.v0y= v[1];
    jp.v0z= v[2];
    jp.mj= *mass;
    jp.eps2= 0.0;
    jp.idxj= *index;
    jp.jtmp= 0;
    SING_send_j_particle(&jp, *address);
    return 0;
}

int g6_set_j_particle(int clusterid,        // in manual
		       int*address,
		       int index,
		       double tj, /* particle time */
		       double dtj, /* particle time */
		       double mass,
		       double a2by18[3], /* a2dot divided by 18 */
		       double a1by6[3], /* a1dot divided by 6 */
		       double aby2[3], /* a divided by 2 */
		       double v[3], /* velocity */
		       double x[3] /* position */)
{
    return g6_set_j_particle(&clusterid, &address, &index, &tj, &dtj, & mass,
			     a2by18, a1by6, aby2, v, x);n
}



// 20100913---- written up to this point

static int ni_stored[MAXCHIPS];
static int nj_stored[MAXCHIPS];

void g6_set_nip_(int * clusterid, int * nip)  // in manual
{
    ni_stored[*clusterid] = *nip;
}

void g6_set_njp_(int * clusterid, int * njp)    // in manual
{
    nj_stored[*clusterid] = *njp;
}

void g6_set_i_particle_scales_from_real_value_(int * clusterid,   // in manual
					       int *address,
					       double acc[3],
					       double jerk[3],
					       double *phi,
					       double *jfactor,
					       double * ffactor)
{
    // do nothing
}

void g6_set_ip_scales_(int * clusterid,
		       int *address,
		       double acc[3],
		       double jerk[3],
		       double *phi,
		       double *jfactor,
		       double *ffactor)
{
    
}

void g6_set_i_particle_(int * clusterid,   // in manual
			int *address,
			int *index,
			double x[3], /* position */
			double v[3], /* velocity */
			double * eps2,
			double * h2)
{
    ERREXIT;
}

void g6_setfoerrchkmode_(int * clusterid, int * mode){}

int g6_get_force_etc_(int * clusterid,   // in manual
		      double acc[][3],
		      double jerk[][3],
		      double phi[],
		      int nnbindex[],
		      int flag[])
{
    ERREXIT;
}
    
int g6_get_force_(int * clusterid,    // in manual
		   double acc[][3],
	           double jerk[][3],
	           double phi[],
	           int flag[])
{
    ERREXIT;
}

int g6_test_flag_(int * clusterid, int * flagp){
    return 0;
}

void g6_adjust_ip_scales_(int * clusterid, int *address, int * flagp) //in manual
{
    ERREXIT;
}

int g6_npipes_()  // in manual
{
    return SING_npipes();
}

int g6_read_neighbour_list_(int * clusterid)     // in manual
{
    ERREXIT;
}

static struct  grape_i_particle_struct *ipbufp[MAXCHIPS]=
    {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
static struct  grape_result_struct *resbufp[MAXCHIPS]=
    {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
static int nia[MAXCHIPS];

static void setup_ipbufp(int devid)
{
    if (ipbufp[devid] == NULL){
	ipbufp[devid]= (struct grape_i_particle_struct *)
	    malloc(SING_number_of_pe()
		   *4* sizeof(struct grape_i_particle_struct));
	if (ipbufp[devid] == NULL){
	    fprintf(stderr,"g6: setup_ipbufp: buffer allocation failed\n");
	    exit(-1);
	}
    }
}

static void setup_resbufp(int devid)
{
    if (resbufp[devid] == NULL){
	resbufp[devid]= (struct grape_result_struct *)
	    malloc(SING_number_of_pe()
		   *4* sizeof(struct grape_result_struct));
	if (resbufp[devid] == NULL){
	    fprintf(stderr,"g5:setup_resbufp: buffer allocation failed\n");
	    exit(-1);
	}
    }
}

int  g6_get_neighbour_list_(int * clusterid, int * ipipe,  // in manual
			   int * maxlength, int * nblen,
			   int nbl[])
{
    ERREXIT;
}

int  clear_unused_i_particles(int devid,
			      struct grape_i_particle_struct * ip, 
			      int nused )
{
    int nend =g5_get_number_of_pipelinesMC(devid);
    int i;
    //    fprintf(stderr, "clear i: %d %d\n", nused, nend);
    for(i=nused; i<nend; i++){
	(ip+i)->xi = (ip+i)->yi = (ip+i)->zi = (ip+i)->eps2i=0.0;
	(ip+i)->vxi = (ip+i)->vyi = (ip+i)->vzi = 0.0;
	(ip+i)->idxi = -1;
    }
}
int  clear_unused_j_particles(int devid, int nused )
{
    int nend = ((nused + SING_number_of_bb()-1)/SING_number_of_bb())
	*SING_number_of_bb();
    int j;
    //    fprintf(stderr, "clear j: %d %d\n", nused, nend);
    struct grape_j_particle_struct  jp; 
    jp.x0 = jp.y0 = jp.z0 = jp.eps2=jp.mj=0.0;
    for(j=nused;j<nend; j++){
	SING_send_j_particleMC(devid,  &jp, j);
    }
}

void g6calc_firsthalf_(int * clusterid,   // in manual
		      int * nj,
		      int * ni,
		      int index[],
		      double xi[][3],
		      double vi[][3],
		      double fold[][3],
		      double j6old[][3],
		      double phiold[],
		      double *eps2,
		      double h2[])
{
    int devid = *clusterid;
    setup_ipbufp(devid);
    int i;
    struct grape_i_particle_struct * ip = ipbufp[devid];
    for(i=0;i<ni;i++){
	ip->xi = xi[i][0];
	ip->yi = xi[i][1];
	ip->zi = xi[i][2];
	ip->ti = ti_stored[devid];
	ip->eps2i = *eps2;
	ip->vxi = vi[i][0];
	ip->vyi = vi[i][1];
	ip->vzi = vi[i][2];
	ip->idxi = index[i];
	ip->itmp = 0:
	ip++;
    }
    nia[devid] = ni;
    nja[devid]=*nj;
    clear_unused_i_particles(devid, ipbufp[devid], nia[devid]);
    clear_unused_j_particles(devid, nja[devid]);
    SING_send_i_particleMC(devid,ipbufp[devid], nireal);
    int njreal = (nja[devid]+SING_number_of_bb()-1)/SING_number_of_bb();
    SING_grape_runMC(devid,njreal);
    
    
}

typedef union datapack{
    unsigned int i32[2];
    ULONG        i64;
    double       d;
}DATAPACK;


int g6calc_lasthalf2_(int * clusterid,   // in manual
		     int * nj,
		     int * ni,
		     int index[],
		     double xi[][3],
		     double vi[][3],
		     double *eps2,
		     double h2[],
		     double acc[][3],
		     double jerk[][3],
		     double pot[],
		      int nnbindex[])
{
    setup_resbufp(devid);
    struct grape_result_struct * rp;
    int k;
    SING_get_resultMC(devid,resbufp[devid]);
    rp = resbufp[devid];
    for(k=0;k<ni; k++){
	acc[k][0]=rp-> accx;
	acc[k][1]=rp->accy;
	acc[k][2]=rp->accz;
	jerk[k][0]=rp-> jerkx;
	jerk[k][1]=rp-> jerky;
	jerk[k][2]=rp-> jerkz;
	pot[k] = -rp->pot;
	DATAPACK pack;
	pack.d = rp->nnb;
	nnbindex =(int)( pack.i64 & 0xffffffL);
	rp++;
    }
    return 0;
}

int g6calc_lasthalf_(int * clusterid,   // in manual
		     int * nj,
		     int * ni,
		     int index[],
		     double xi[][3],
		     double vi[][3],
		     double *eps2,
		     double h2[],
		     double acc[][3],
		     double jerk[][3],
		     double pot[])
{
    int nnbindex[1024];
    return g6calc_lasthalf2_(clusterid, nj, ni, index, xi, vi, eps2, h2,
		      acc, jerk, pot, nnbindex);
}
    
int calculate_accel_by_grape6_separate_trial_noopen(int clusterid,   // in manual
					     int ni,
					     double xi[][3],
					     double vi[][3],
					     int nj,
					     double xj[][3],
					     double vj[][3],
					     double m[],
					     double a[][3],
					     double jerk[][3],
					     double pot[],
					     double eps2)
{
    double ajtmp[3];
    double jjtmp[3];
    double j2jtmp[3];
    double ti, tj, dtj;
    int j0, i0;
    int i,k,ii, iend;
    int flag[MAXPIPELINESPERCHIP+1];
    int index[MAXPIPELINESPERCHIP+1];
    double h2[MAXPIPELINESPERCHIP+1];
    double eps2array[MAXPIPELINESPERCHIP+1];
    int nharderror = 0;
    int ipmax = g6_npipes();
    ti = 0.0;tj =0.0; dtj = 0.0078125;
    for(k=0;k<3;k++){
	ajtmp[k] = 0.0;
	jjtmp[k] = 0.0;
	j2jtmp[k] = 0.0;
    }
 START:    
    g6_set_ti_(&clusterid, &ti);
    for(i=0;i<nj;i++){
	g6_set_j_particle_(&clusterid,&i,&i,&tj,&dtj,m+i,
			   j2jtmp,jjtmp,ajtmp,vj[i],xj[i]);
    }
    for(i=0;i<ipmax;i++){
	h2[i] = eps2;
	eps2array[i] = eps2;
    }
    for(i=0;i<ni;i+=ipmax){
	int error;
	iend = ipmax; if  (iend+i > ni) iend = ni-i;
	for(ii=0;ii<iend;ii++)index[ii]=i+ii;
	g6calc_firsthalf_(&clusterid, &nj, &iend,index,&(xi[i]),&(vi[i]),
			  &(a[i]),&(jerk[i]),
			  pot+i,eps2array, h2);
	if (error = g6calc_lasthalf_(&clusterid,&nj,&iend,index,&xi[i],&vi[i],
				     eps2array,h2, &a[i], &jerk[i], pot+i)){

#define	 ERRMAX 10
	    nharderror ++;
	    fprintf(stderr,"(calculate_accell_trial_noopen) hard error %d\n", error);
	    g6_close_(&clusterid);
	    fprintf(stderr,"(calculate_accell_trial_noopen) calling g6open\n");
	    g6_open_(&clusterid);
	    if (nharderror < ERRMAX){
		goto START;
	    }else{
		fprintf(stderr,"(calculate_accell_trial_noopen) too many errors... %d returning -1\n", nharderror);
		return -1;
	    }
	}
    return 0;
}

int calculate_accel_by_grape6_noopen(int clusterid,     // in manual
					     int n,
					     double x[][3],
					     double v[][3],
					     double m[],
					     double a[][3],
					     double jerk[][3],
					     double pot[],
					     double eps2)
{
    int error, error_prev;
    int errorcount = 0;
    error_prev = 0;
    do{
	error = calculate_accel_by_grape6_separate_trial_noopen(clusterid,n,x,
								v,n,x,v,
								m,a,jerk,
								pot, eps2);
	if (!error) error = check_newton_3rd_law(n,m,a,jerk);
	if (error) 	{
	    fprintf(stderr,"Force in Error\n");
	    errorcount ++;
	    clusters[clusterid]->errorcounters.cmerror_count++;
	    sleep(1);
	    if (errorcount > ERRMAX){
		fprintf(stderr,"Too many errors, exiting...\n");
		exit(-1);
	    }
	    error_prev = 1;
	}else{
	    if (error_prev){
		fprintf(stderr,"Force Corrected\n");
	    }
	}
    }while(error);
    return error;
}
 
void g6_set_ti(int clusterid, double ti)
{
    g6_set_ti_(&clusterid, &ti);
}
 
int g6_set_j_particle(int  clusterid,
		       int address,
		       int index,
		       double tj, /* particle time */
		       double dtj, /* particle time */
		       double mass,
		       double a2by18[3], /* a2dot divided by 18 */
		       double a1by6[3], /* a1dot divided by 6 */
		       double aby2[3], /* a divided by 2 */
		       double v[3], /* velocity */
		       double x[3] /* position */);
int g6_npipes()                                  // in manual
{
    return g6_npipes_();
}
int g6_read_neighbour_list(int clusterid)
{
    return g6_read_neighbour_list_(&clusterid);
}
int  g6_get_neighbour_list(int clusterid, int ipipe,
			   int maxlength, int * nblen,
			   int nbl[])
{
    return g6_get_neighbour_list(&clusterid, &ipipe,
				 &maxlength,  nblen, nbl);
}

int g6calc_lasthalf(int  clusterid,
		     int  nj,
		     int  ni,
		     int index[],
		     double xi[][3],
		     double vi[][3],
		     double eps2,
		     double h2[],
		     double acc[][3],
		     double jerk[][3],
		     double pot[])
{
    return     g6calc_lasthalf_(&clusterid, &nj, &ni, index, xi, vi,
			       &eps2, h2, acc, jerk, pot);
}

void g6calc_firsthalf(int  clusterid,
		      int  nj,
		      int  ni,
		      int index[],
		      double xi[][3],
		      double vi[][3],
		      double fold[][3],
		      double j6old[][3],
		      double phiold[],
		      double eps2,
		      double h2[])
{
    g6calc_firsthalf_(&clusterid, &nj,&ni,index, xi,vi,fold,j6old,phiold,
		     &eps2, h2);
}

int g6calc_lasthalf2(int  clusterid,
		     int  nj,
		     int  ni,
		     int index[],
		     double xi[][3],
		     double vi[][3],
		     double eps2,
		     double h2[],
		     double acc[][3],
		     double jerk[][3],
		     double pot[],
		    int nnbindex[])
{
    return     g6calc_lasthalf2_(&clusterid, &nj, &ni, index, xi, vi,
			       &eps2, h2, acc, jerk, pot,nnbindex);
}

int g6_get_nchips(int clusterid);

