//
// simgrape6.c
//
// GRAPE-6 compatibility interface for SING implementation of hermite
// pipeline
//
//
// J. Makino 2005/4/23
//
// do not support cluster_id (should be 0)
//g6_change_cbjpmode
//g6_change_clock
//g6_changed_index_
//g6_clock_state
//g6_close
//g6_dummywait
//g6_flush_jp_buffer_and_multisend
//g6_initialize_jp_buffer
//g6_npipes
//g6_open
//g6_print_chip_status
//g6_reset
//g6_reset_fofpga
//g6_set_ijp_mode
//g6_set_j_particle
//g6_set_j_particle_multisend
//g6_set_ti
//g6_set_xunit
//g6_setmccount
//g6_setup_njdata

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include "hermite-16x32.h"

#define NCHIPS 4
static int firstchip=0;
static int nchips=NCHIPS;
static int endchip=NCHIPS;
static int ipdegree = 1;
static int jpdegree = NCHIPS;

#define MAXCHIPS 8
#define MAXPIPES 256
#define JPBUFFSIZE 512
static struct grape_j_particle_struct xjbuf[MAXCHIPS][JPBUFFSIZE];
static int  xjindex[MAXCHIPS][JPBUFFSIZE];
static int xjbufloc[MAXCHIPS];

static inline UINT64 rdtsc2()
{
	UINT64 ret;
	asm __volatile__("rdtsc  \n\t" \
			 "shlq $32, %%rdx \n\t" \
			 "orq  %%rdx, %%rax \n\t" \
			 "movq %%rax, %0" : "=g"(ret) : :"rax", "rdx");
	return ret;
}

static UINT64 jptime = 0;
static UINT64 iptime = 0;
static UINT64 runtime = 0;
static int jpcalls = 0;
static int ipcalls = 0;
static int runcalls = 0;

static int debug_level=0;

static UINT64 last_time_for_jp;


void g6_set_debug_level(int level)
{
    debug_level = level;
}

void g6_set_debug_level_(int *level)
{
    g6_set_debug_level(*level);
}

void g6_report_timers()
{
    printf("JP= %e IP= %e RUN = %e\n", ((double)jptime)/jpcalls,
	   ((double)iptime)/ipcalls,	   ((double)runtime)/runcalls);
}

void g6_report_timers_()
{
    g6_report_timers();
}

void g6_gdr_set_mode(int ifirst, int np, int ipd)
{
    firstchip = ifirst;
    nchips    = np;
    endchip   = firstchip + nchips;
    ipdegree  = ipd;
    jpdegree = nchips/ipdegree;
}

int g6_gdr_firstchip(){return firstchip;}
int g6_gdr_nchips(){return nchips;}
int g6_gdr_ipdegree(){return ipdegree;}

#define FORALLCHIPS  for(ic=0; ic<nchips; ic++)

void g6_setup_njdata(int clusterid, int nbody)
{}
void g6_dummywait()
{}
void g6_setmccount(int dummy)
{}

    
void g6_set_ijp_mode(int clusterid,
		     int mode) /* mode=0: IP, 1: JP */
{
    // dummy
}
void g6_set_ijp_mode_(int * clusterid,
		      int * mode) /* mode=0: IP, 1: JP */
{
    g6_set_ijp_mode(*clusterid,* mode);
}




void g6_reset(int  clusterid)
{
    //dummy
}

void g6_reset_fofpga(int  clusterid)
{
    //dummy
}

	
void g6_change_cbjpmode_(unsigned int *	 clusterid,
			unsigned int *mode /*0: BCAST, 1: MCAST */)
{
    //dummy
}
void g6_change_cbjpmode(unsigned int 	 clusterid,
			unsigned int mode /*0: BCAST, 1: MCAST */)
{
    //dummy
}
	
static int chipid = 0;

void g6_change_clock(int clusterid, int mode)
     /* mode:0 for sleep, 1 for activate */
{
    //dummy
}

static int xunit = 51;

void g6_set_xunit_(int * newxunit)
{
  xunit = *newxunit;
}

void g6_set_xunit(int newxunit)
{
    g6_set_xunit_(&newxunit);
}

static int chipids[MAXCHIPS];
int g6open(int clusterid)
{
    if (debug_level){
	fprintf(stderr,"g6open called\n");
    }
    int ic;
    FORALLCHIPS {
	fprintf(stderr, "chip %d opened\n", ic+firstchip);
	SING_grape_initMC(ic+firstchip);
	chipids[ic]=ic+firstchip;
	xjbufloc[ic]=0;
    }
    return 0;
    
}



int g6_open_(int *clusterid)
{
    return g6open(*clusterid);
}
int g6_open(int clusterid)
{
    return g6open(clusterid);
}

int g6close(int clusterid)
{
    int ic;
    FORALLCHIPS SING_grape_closeMC(ic+firstchip);
    return 0;
}


int g6_close_(int *clusterid)
{
    return g6close(*clusterid);
}
int g6_close(int clusterid)
{
    return g6close(clusterid);
}


static double ti_set;
void g6_set_ti_(int *clusterid, double *ti)
{
    if(debug_level) fprintf(stderr,"g6_set_ti ti = %g\n", *ti);
    ti_set = *ti;
}

void g6_set_ti(int clusterid, double ti)
{
    g6_set_ti_(&clusterid, &ti);
}


int g6_initialize_jp_buffer(int clusterid, int size)
{
    int ic;
    FORALLCHIPS xjbufloc[ic]=0;
}
	
int g6_initialize_jp_buffer_(int* clusterid, int* size) 
{
    return g6_initialize_jp_buffer(*clusterid, *size);
}




static flush_jp_buffer(int chipid)
{
    int i;
    UINT64 start = rdtsc2();
    if (xjbufloc[chipid]==0)return;
    //    fprintf(stderr,"flush jp buffer %d %d\n", chipid, xjbufloc[chipid]);
    //    for(i=0; i<xjbufloc[chipid]; i++){
    //	SING_send_j_particleMC(chipid,  xjbuf[chipid]+i, xjindex[chipid][i]);
    //    }
    SING_send_elt_data0_vMC(chipid, xjbuf[chipid], xjindex[chipid], xjbufloc[chipid]);
    xjbufloc[chipid]=0;
    jptime += rdtsc2()-start;
    last_time_for_jp = rdtsc2();
}
    
int g6_flush_jp_buffer_and_multisend(int  clusterid, int nclusters)
{
    int ic;
    if(debug_level & 2) {
	fprintf(stderr,"g6_flush_jp_buffer_and_multisend called\n");
    }
    FORALLCHIPS{
	flush_jp_buffer(ic);
    }
    return 0;
}

int g6_set_j_particle_multisend_(int * clusterid,
				 int * nclusters,
				 int *address,
				 int *index,
				 double *tj, /* particle time */
				 double *dtj, /* particle time */
				 double *mass,
				 double a2by18[3], /* a2dot divided by 18 */
				 double a1by6[3], /* a1dot divided by 6 */
				 double aby2[3], /* a divided by 2 */
				 double v[3], /* velocity */
				 double x[3] /* position */)
{
    struct grape_j_particle_struct *xjp;
    if(debug_level & 2) {
	fprintf(stderr,"g6_set_jp address = %d index = %d x[0]  = %g\n",
		*address, *index, x[0]);
    }
    UINT64 start = rdtsc2();
    int ic = ((*address) % jpdegree); // assuming ipdegree=1;
    int realaddress= (*address)/jpdegree;
    if (xjbufloc[ic] >= JPBUFFSIZE) flush_jp_buffer(ic);
    xjp = xjbuf[ic]+xjbufloc[ic];
    xjindex[ic][xjbufloc[ic]]=realaddress;
    xjbufloc[ic]++;
    xjp->tj= *tj;
    xjp->x0=x[0]; 
    xjp->y0=x[1];
    xjp->z0=x[2];
    xjp->v0x= v[0];
    xjp->v0y= v[1];
    xjp->v0z= v[2];
    xjp->ax2= aby2[0];
    xjp->ay2= aby2[1];
    xjp->az2= aby2[2];
    xjp->jx6=a1by6[0];
    xjp->jy6=a1by6[1];
    xjp->jz6=a1by6[2];
    xjp->sx18=a2by18[0];
    xjp->sy18=a2by18[1];
    xjp->sz18=a2by18[2];
    xjp->mj= *mass;
    xjp->eps2=0;
    xjp->idxj=(*index)<<8;
    jptime += rdtsc2()-start;
    jpcalls ++;
    last_time_for_jp = rdtsc2();
    return 0;
}
int g6_set_j_particle_multisend_simple_(int * clusterid,
				 int * nclusters,
				 int *address,
				 int *index,
				 double *tj, /* particle time */
				 double *dtj, /* particle time */
				 double *mass,
				 double a2by18[3], /* a2dot divided by 18 */
				 double a1by6[3], /* a1dot divided by 6 */
				 double aby2[3], /* a divided by 2 */
				 double v[3], /* velocity */
				 double x[3] /* position */)
{
    struct grape_j_particle_struct xj;
    UINT64 start = rdtsc2();

    xj.tj= *tj;
    xj.x0=x[0]; 
    xj.y0=x[1];
    xj.z0=x[2];
    xj.v0x= v[0];
    xj.v0y= v[1];
    xj.v0z= v[2];
    xj.ax2= aby2[0];
    xj.ay2= aby2[1];
    xj.az2= aby2[2];
    xj.jx6=a1by6[0];
    xj.jy6=a1by6[1];
    xj.jz6=a1by6[2];
    xj.sx18=a2by18[0];
    xj.sy18=a2by18[1];
    xj.sz18=a2by18[2];
    xj.mj= *mass;
    xj.eps2=0;
    xj.idxj=(*index)<<8;
    int ic;
    int icfirst = ((*address) % jpdegree)*ipdegree;
    int realaddress= (*address)/jpdegree;
    for (ic = 0; ic < ipdegree; ic++){
//	fprintf(stderr,"jpset %d %d  %d %d\n", ipdegree, icfirst+ic+firstchip,
//		*address, realaddress);
	SING_send_j_particleMC(icfirst+ic+firstchip,  &xj, realaddress);
    }
    jptime += rdtsc2()-start;
    jpcalls ++;
    last_time_for_jp = rdtsc2();
    return 0;
}

int g6_set_j_particle_(int * clusterid,
		       int *address,
		       int *index,
		       double *tj, /* particle time */
		       double *dtj, /* particle time */
		       double *mass,
		       double a2by18[3], /* a2dot divided by 18 */
		       double a1by6[3], /* a1dot divided by 6 */
		       double aby2[3], /* a divided by 2 */
		       double v[3], /* velocity */
		       double x[3] /* position */)
{
    int one = 1;
    return g6_set_j_particle_multisend_(clusterid,&one,address,index,tj,dtj,mass,
			      a2by18, a1by6, aby2, v, x);
}

int g6_clear_one_j_particles(address)
{
    int clusterid = 0;
    int one = 1;
    int index = -1;
    double tj = 0;
    double dtj=0;
    double mass = 0;
    double x[3]={1e10,1e10,1e10};
    double v[3]={0,0,0};
    return g6_set_j_particle_multisend_(&clusterid,&one,&address,&index,
					&tj,&dtj,&mass,v,v,v,v,x);
}


int g6_set_j_particle(int  clusterid,
		       int address,
		       int index,
		       double tj, /* particle time */
		       double dtj, /* particle time */
		       double mass,
		       double a2by18[3], /* a2dot divided by 18 */
		       double a1by6[3], /* a1dot divided by 6 */
		       double aby2[3], /* a divided by 2 */
		       double v[3], /* velocity */
		       double x[3] /* position */)
{
    return g6_set_j_particle_(&clusterid,&address,&index,&tj,&dtj,&mass,
			      a2by18, a1by6, aby2, v, x);
}

int g6_set_j_particle_multisend(int  clusterid,
				int  nclusters,
		       int address,
		       int index,
		       double tj, /* particle time */
		       double dtj, /* particle time */
		       double mass,
		       double a2by18[3], /* a2dot divided by 18 */
		       double a1by6[3], /* a1dot divided by 6 */
		       double aby2[3], /* a divided by 2 */
		       double v[3], /* velocity */
		       double x[3] /* position */)
{
    return g6_set_j_particle_multisend_(&clusterid,&nclusters,&address,&index,&tj,&dtj,&mass,
			      a2by18, a1by6, aby2, v, x);
}

    

int g6_npipes()
{
    return g6_npipes_();
}

int g6_npipes_()
{
    return SING_number_of_pe()*4*ipdegree;
}

int g6_print_chip_status_(int * clusterid)
{
    // dummy
    return 0;
}

int g6_print_chip_status(int clusterid)
{
    return  g6_print_chip_status_(&clusterid);
}




void g6calc_firsthalf0_(int * clusterid,
			int * nj,
			int * ni,
			int index[],
			double x[][3],
			double v[][3],
		      double fold[][3],
			double j6old[][3],
			double phiold[],
			double eps2[],
			double h2[],
			int * mode)
     /* mode: 0 --- standard mode, regard both eps2 and h2 as array
	      1 --  GRAPE-4 compatibility mode, eps2 is scalar
	      2 (bit loc 1) skip send_njp... future extension */
// pit uses mode 2....
{

    static int nj_old = 0;
    int niparallel = SING_number_of_pe()*4;
    int njscale = SING_number_of_bb()*jpdegree;
    int njreal = (*nj + njscale-1)/njscale;
    int njchanged = 0;
    int njold2 = nj_old;
    if (nj_old != *nj){
	nj_old = *nj;
	//	fprintf(stderr, "njreal, n, njscale = %d %d %d\n",
	//		njreal, *nj, njscale);
	int j;
	for (j= (*nj); j< (njreal*njscale);j++)g6_clear_one_j_particles(j);
	njchanged = 1;
    }
    //    fprintf(stderr,"call flush buffer\n");
    g6_flush_jp_buffer_and_multisend(*clusterid, 1);
    //    fprintf(stderr,"call SING_wait_dmar\n");
    //    SING_wait_DMAR(nchips, chipids);
    int i;
    UINT64 start = rdtsc2();
    static struct grape_i_particle_struct xi[2048];
    if (start > last_time_for_jp){
	while(rdtsc2() < last_time_for_jp+10000);
    }

    if(debug_level & 4) {
	fprintf(stderr,"g6calc_firsthalf nj = %d ni = %d x[0][0]  = %g\n",
		*nj, *ni, x[0][0]);
    }
    
    for(i=0; i< *ni; i++){
	xi[i].ti= ti_set;
	xi[i].xi=x[i][0];
	xi[i].yi=x[i][1];
	xi[i].zi=x[i][2];
	xi[i].vxi=v[i][0];
	xi[i].vyi=v[i][1];
	xi[i].vzi=v[i][2];
	xi[i].idxi=(index[i])<<8;
	xi[i].eps2i= (*mode & 1) ? eps2[0]:eps2[i];
    }
    for(i= *ni; i< niparallel*ipdegree; i++) xi[i]=xi[0];

    // h2 ...
    //    fprintf(stderr,"call send_i_particle\n");

    int ic;
    if(njchanged){
	FORALLCHIPS{
	    int istart = ic % ipdegree;
	    SING_send_i_particleMC(ic+firstchip, xi+(istart*niparallel),
				   niparallel);
	}
    }else{
	SING_send_i_particleMCP(nchips, chipids, xi,  *ni);
    }
    //    fprintf(stderr,"end send_i_particle\n");
    iptime += rdtsc2()-start;
    ipcalls++;
    start = rdtsc2();
    //    FORALLCHIPS    
    //	SING_grape_runMC(ic+firstchip, njreal);
    SING_grape_runMCP(nchips, chipids, njreal);
    runtime += rdtsc2()-start;
    runcalls++;
}


void g6calc_firsthalf0(int clusterid,
		       int nj,
		       int ni,
		       int index[],
		       double xi[][3],
		       double vi[][3],
		       double fold[][3],
		       double j6old[][3],
		       double phiold[],
		       double eps2[],
		       double h2[],
		       int  mode)
{
     g6calc_firsthalf0_(&clusterid,&nj,&ni,index, xi, vi,fold,j6old,phiold,eps2,
			h2,&mode);
}

void g6calc_firsthalf_(int * clusterid,
		      int * nj,
		      int * ni,
		      int index[],
		      double xi[][3],
		      double vi[][3],
		      double fold[][3],
		      double j6old[][3],
		      double phiold[],
		      double *eps2,
		      double h2[])
{
    int mode = 1;
    g6calc_firsthalf0_(clusterid,nj,ni,index, xi, vi,fold,j6old,phiold,eps2,
			h2,&mode);
}
    


void g6calc_firsthalf(int  clusterid,
		      int  nj,
		      int  ni,
		      int index[],
		      double xi[][3],
		      double vi[][3],
		      double fold[][3],
		      double j6old[][3],
		      double phiold[],
		      double eps2,
		      double h2[])
{
    g6calc_firsthalf_(&clusterid, &nj,&ni,index, xi,vi,fold,j6old,phiold,
		     &eps2, h2);
}

static index_scalechanged = -2;

int g6_changed_index()
{
    return index_scalechanged;
}
int g6_changed_index_()
{
    return index_scalechanged;
}

typedef union datapack{
    unsigned int i32[2];
    UINT64        i64;
    double       d;
}DATAPACK;

int g6calc_lasthalf0a_(int * clusterid,
		     int * nj,
		     int * ni,
		     int index[],
		     double xi[][3],
		     double vi[][3],
		     double *eps2,
		     double h2[],
		     double acc[][3],
		     double jerk[][3],
		     double pot[],
		       int nnbindex[],
		       int *mode)
{
    int i,ip,jp;
    struct grape_result_struct results[MAXCHIPS][256];
    double rnnb[512];
    int ic;
    //#pragma omp parallel for private(ic)
    FORALLCHIPS{
	SING_get_resultMC(ic+firstchip, results[ic]);
    }
    
    for(ip=0;ip<ipdegree;ip++){
	
	int ifirst = ip * SING_number_of_pe()*4;
	int iend = (ip+1) * SING_number_of_pe()*4;
	if (iend > *ni) iend =*ni;
	struct grape_result_struct * result = results[ip];
	for (i=ifirst;i< iend; i++){
	    int ii = i-ifirst;
	    acc[i][0]=result[ii].accx;
	    acc[i][1]=result[ii].accy;
	    acc[i][2]=result[ii].accz;
	    jerk[i][0]=result[ii].jerkx;
	    jerk[i][1]=result[ii].jerky;
	    jerk[i][2]=result[ii].jerkz;
	    pot[i]    =result[ii].pot;
	    DATAPACK pack;
	    pack.d = result[ii].nnb;
	    rnnb[i]= pack.d;
	    nnbindex[i] =(int)( pack.i64 & 0xffffffL);
	    //	nnbindex[i]=result[i].nnb[1];
	}
	for(jp=1; jp<jpdegree; jp++){
	    result = results[ip+jp*ipdegree];
	    for (i=ifirst;i< iend; i++){
		int ii = i-ifirst;
		acc[i][0]+=result[ii].accx;
		acc[i][1]+=result[ii].accy;
		acc[i][2]+=result[ii].accz;
		jerk[i][0]+=result[ii].jerkx;
		jerk[i][1]+=result[ii].jerky;
		jerk[i][2]+=result[ii].jerkz;
		pot[i]    +=result[ii].pot;
		DATAPACK pack;
		pack.d = result[ii].nnb;
		if (result[ii].nnb < rnnb[i]){
		    nnbindex[i] =(int)( pack.i64 & 0xffffffL);
		    rnnb[i] = result[ii].nnb;
		}
	    }
	}
    }
    if(debug_level & 4) {
	fprintf(stderr,"g6calc_last nj = %d ni = %d acc[0][0]  = %g\n",
		*nj, *ni, acc[0][0]);
    }
    return 0;
}

int g6calc_lasthalf0(int clusterid,
		     int nj,
		     int ni,
		     int index[],
		     double xi[][3],
		     double vi[][3],
		     double *eps2,
		     double h2[],
		     double acc[][3],
		     double jerk[][3],
		     double pot[],
		     int mode)
{
    return   g6calc_lasthalf0_(&clusterid, &nj, &ni, index, xi, vi,eps2,h2,acc, jerk, pot,&mode);
}

int g6calc_lasthalf0a(int clusterid,
		      int nj,
		      int ni,
		      int index[],
		      double xi[][3],
		      double vi[][3],
		      double *eps2,
		      double h2[],
		      double acc[][3],
		      double jerk[][3],
		      double pot[],
		      int nnbindex[],
		      int mode)
{
    return   g6calc_lasthalf0a_(&clusterid, &nj, &ni, index, xi, vi,eps2,h2,acc, jerk, pot,
				nnbindex,&mode);
}


int g6calc_lasthalf2_(int * clusterid,
		     int * nj,
		     int * ni,
		     int index[],
		     double xi[][3],
		     double vi[][3],
		     double *eps2,
		     double h2[],
		     double acc[][3],
		     double jerk[][3],
		     double pot[],
		      int nnbindex[])
{
    int mode = 1;
    return   g6calc_lasthalf0a_(clusterid, nj, ni, index, xi, vi,eps2,h2,acc, jerk, pot,
				nnbindex,&mode);
}

int g6calc_lasthalf0_(int * clusterid,
		     int * nj,
		     int * ni,
		     int index[],
		     double xi[][3],
		     double vi[][3],
		     double *eps2,
		     double h2[],
		     double acc[][3],
		     double jerk[][3],
		     double pot[],
		     int * mode )
{
    int nnbindex[2048];
    return   g6calc_lasthalf0a_(clusterid, nj, ni, index, xi, vi,eps2,h2,acc, jerk, pot,
				nnbindex,mode);
}

int g6calc_lasthalf_(int * clusterid,
		     int * nj,
		     int * ni,
		     int index[],
		     double xi[][3],
		     double vi[][3],
		     double *eps2,
		     double h2[],
		     double acc[][3],
		     double jerk[][3],
		     double pot[])
{
    int mode = 1;
    return   g6calc_lasthalf0_(clusterid, nj, ni, index, xi, vi,eps2,h2,acc, jerk, pot,
				&mode);
}
    
int g6calc_lasthalf2(int  clusterid,
		     int  nj,
		     int  ni,
		     int index[],
		     double xi[][3],
		     double vi[][3],
		     double eps2,
		     double h2[],
		     double acc[][3],
		     double jerk[][3],
		     double pot[],
		    int nnbindex[])
{
    return     g6calc_lasthalf2_(&clusterid, &nj, &ni, index, xi, vi,
			       &eps2, h2, acc, jerk, pot,nnbindex);
}
int g6calc_lasthalf(int  clusterid,
		     int  nj,
		     int  ni,
		     int index[],
		     double xi[][3],
		     double vi[][3],
		     double eps2,
		     double h2[],
		     double acc[][3],
		     double jerk[][3],
		     double pot[])
{
    return     g6calc_lasthalf_(&clusterid, &nj, &ni, index, xi, vi,
			       &eps2, h2, acc, jerk, pot);
}
 
static double tol2a = 1e-20;
static double tol2j = 1e-0;
int check_newton_3rd_law(int n,
			  double m[],
			  double acc[][3],
			  double jerk[][3])
{
    int i, k, error;
    double acc2sum, jerk2sum, msum,accsum[3], jerksum[3];
    acc2sum = jerk2sum = msum = 0;
    for(k = 0;k<3;k++)accsum[k] = jerksum[k] = 0;
    for(i =0;i<n;i++){
	for(k = 0;k<3;k++){
	    double x;
	    x = acc[i][k]*m[i];
	    accsum[k] +=x;
	    acc2sum += x*x;
	    x = jerk[i][k]*m[i];

	    jerksum[k] +=x;
	    jerk2sum += x*x;

	}
	msum += m[i];
    }
    error = 0;
    for(k=0;k<3;k++){
	if (acc2sum > 0.0) if( accsum[k]*accsum[k]>=acc2sum*tol2a)error = 1;
	if (jerk2sum > 0.0) if( jerksum[k]*jerksum[k]>=jerk2sum*tol2j)error = 1;
    }
#ifdef INTERNAL_OUT
	fprintf(stderr,"check_3rd_law, a, j = %e %e %e %e\n       %e %e %e %e\n",
		accsum[0],accsum[1],accsum[2],acc2sum,
		jerksum[0],jerksum[1],jerksum[2],jerk2sum);
#endif
    if(error){
	fprintf(stderr,"check_3rd_law, a, j = %e %e %e %e %e %e %e %e\n",
		accsum[0],accsum[1],accsum[2],acc2sum,
		jerksum[0],jerksum[1],jerksum[2],jerk2sum);
    }
    return error;
}
    


int calculate_accel_by_grape6_separate_trial_noopen(int clusterid,   // in manual
						    int ni,
						    double xi[][3],
						    double vi[][3],
						    int nj,
						    double xj[][3],
						    double vj[][3],
						    double m[],
						    double a[][3],
						    double jerk[][3],
						    double pot[],
						    double eps2)
{
    double ajtmp[3];
    double jjtmp[3];
    double j2jtmp[3];
    double ti, tj, dtj;
    int j0, i0;
    int i,k,ii, iend;
    int flag[MAXPIPES+1];
    int index[MAXPIPES+1];
    double h2[MAXPIPES+1];
    double eps2array[MAXPIPES+1];
    int nharderror = 0;
    int ipmax = g6_npipes();
    ti = 0.0;tj =0.0; dtj = 0.0078125;
    for(k=0;k<3;k++){
	ajtmp[k] = 0.0;
	jjtmp[k] = 0.0;
	j2jtmp[k] = 0.0;
    }
 START:    
    g6_set_ti_(&clusterid, &ti);
    for(i=0;i<nj;i++){
	g6_set_j_particle_(&clusterid,&i,&i,&tj,&dtj,m+i,
			   j2jtmp,jjtmp,ajtmp,vj[i],xj[i]);
    }
    for(i=0;i<ipmax;i++){
	h2[i] = eps2;
	eps2array[i] = eps2;
    }
    for(i=0;i<ni;i+=ipmax){
	int error;
	iend = ipmax; if  (iend+i > ni) iend = ni-i;
	for(ii=0;ii<iend;ii++)index[ii]=i+ii;
	g6calc_firsthalf_(&clusterid, &nj, &iend,index,&(xi[i]),&(vi[i]),
			  &(a[i]),&(jerk[i]),
			  pot+i,eps2array, h2);
	if (error = g6calc_lasthalf_(&clusterid,&nj,&iend,index,&xi[i],&vi[i],
				     eps2array,h2, &a[i], &jerk[i], pot+i)){
	    
#define	 ERRMAX 10
	    nharderror ++;
	    fprintf(stderr,"(calculate_accell_trial_noopen) hard error %d\n", error);
	    g6_close_(&clusterid);
	    fprintf(stderr,"(calculate_accell_trial_noopen) calling g6open\n");
	    g6_open_(&clusterid);
	    if (nharderror < ERRMAX){
		goto START;
	    }else{
		fprintf(stderr,"(calculate_accell_trial_noopen) too many errors... %d returning -1\n", nharderror);
		return -1;
	    }
	}
    }
    return 0;
}

int calculate_accel_by_grape6_noopen(int clusterid,     // in manual
				     int n,
				     double x[][3],
				     double v[][3],
				     double m[],
				     double a[][3],
				     double jerk[][3],
				     double pot[],
				     double eps2)
{
    int error, error_prev;
    int errorcount = 0;
    error_prev = 0;
    do{
	error = calculate_accel_by_grape6_separate_trial_noopen(clusterid,n,x,
								v,n,x,v,
								m,a,jerk,
								pot, eps2);
	if (!error) error = check_newton_3rd_law(n,m,a,jerk);
	if (error) 	{
	    fprintf(stderr,"Force in Error\n");
	    errorcount ++;
	    sleep(1);
	    if (errorcount > ERRMAX){
		fprintf(stderr,"Too many errors, exiting...\n");
		exit(-1);
	    }
	    error_prev = 1;
	}else{
	    if (error_prev){
		fprintf(stderr,"Force Corrected\n");
	    }
	}
    }while(error);
    return error;
}
 
    // compatibility functions for KFCR-made library
    // added by JM 2010/9/18

void g6_open_all(void)
{
    g6_open(0);
}

void g6_close_all(void)
{
    g6_close(0);
}

int g6_set_j_particle_all(int address, int index, double tj, double dtj,
			      double mass, double a2by18[3], double a1by6[3],
			      double aby2[3], double v[3], double x[3])
{
    return  g6_set_j_particle(0, address, index, tj, dtj, mass,
			      a2by18, a1by6, aby2, v, x);
}

int g6_set_j_particle_mxonly_all(int address, int index, double mass,
				     double x[3])
{
    double tj = 0;
    double dtj =1;
    double dum[3]={0.0,0.0,0.0};
    return  g6_set_j_particle(0, address, index, tj, dtj, mass,
			      dum, dum, dum, dum,  x);
    
}
void g6_set_ti_all(double ti)
{
    g6_set_ti(0,  ti);
}

void g6calc_firsthalf_all(int nj, int ni, int index[], double xi[][3],
			      double vi[][3], double fold[][3], double jold[][3],
			      double phiold[], double eps2, double h2[])
{
    g6calc_firsthalf(0, nj, ni, index, xi, vi, fold, jold, phiold, eps2, h2);
}

int g6calc_lasthalf_all(int nj, int ni, int index[], double xi[][3],
			    double vi[][3], double eps2, double h2[],
			    double acc[][3], double jerk[][3], double pot[])
{
    return     g6calc_lasthalf(0, nj, ni, index, xi, vi, eps2, h2, acc, jerk, pot);
}

    int g6calc_lasthalf2_all(int nj, int ni, int index[], double xi[][3],
			     double vi[][3], double eps2, double h2[],
			     double acc[][3], double jerk[][3], double pot[],
			     int nnbindex[])
{
    g6calc_lasthalf2(0, nj, ni, index, xi, vi, eps2, h2,
		     acc, jerk, pot,nnbindex);
}

int g6_read_neighbour_list_all(void){}
int g6_get_neighbour_list_all(int ipipe, int maxlength, int *nblen, int nbl[]){}
void g6_set_nip_all(int nip){}
void g6_set_njp_all(int njp){}
void g6_set_i_particle_scales_from_real_value_all(int address, double acc[3],
						  double jerk[3], double phi,
						  double jfactor,
						  double ffactor){}
void g6_set_i_particle_all(int address, int index, double x[3],
			   double v[3], double eps2, double h2){}
int g6_get_force_all(double acc[][3], double jerk[][3], double phi[],
		     int flag[]){}

// functions proposed by Eiichiro Kokubo
//

#define M_1_6  (0.166666666666666666666666666666666666666)
#define NPIPE MAXPIPES

void g6_set_particle(int n, int ilist[],
		     double m[], double x[][3], double v[][3],
		     double a[][3], double j[][3],
		     double t[], double dt[])
{
    
    double aby2[3], a1by6[3];
    static double a2by18[3] = {0.0, 0.0, 0.0}; /* a'' not used in predictor */
    int i, id, k;
    for(i=0;i<n;i++){
	id = ilist[i];
	for(k=0;k<3;k++){
	    aby2[k] = a[id][k]*0.5;
	    a1by6[k] = j[id][k]*M_1_6;
	}
	g6_set_j_particle_all(id, id, t[id], dt[id], m[id], 
			      a2by18, a1by6, aby2, v[id], x[id]);
    }
}

void g6_calc_force(int nj, int ni, int ilist[], int nnb[],
		   double t,
		   double xp[][3], double vp[][3],
		   double a_old[][3], double j_old[][3],
		   double a_new[][3], double j_new[][3], double pot[],
		   double eps2, double h2[])
//
// in this implementation, a_old and j_old are not used, since they are not used in GDR lib 
//
{
    
    double xi[NPIPE][3], vi[NPIPE][3];
    double ai[NPIPE][3], j6i[NPIPE][3], poti[NPIPE], h2i[NPIPE];
    double a_tmp[NPIPE][3], j_tmp[NPIPE][3], pot_tmp[NPIPE];
    int nnb_tmp[NPIPE];
    int npipe = g6_npipes();
    int i, j, k, id, np;
    g6_set_ti_all(t);
    for(i=0;i<ni;i+=npipe){
	np = npipe;
	if(i+npipe>ni) np = ni-i;
	for(j=0;j<np;j++){
	    id = ilist[i+j];
	    for(k=0;k<3;k++){
		xi[j][k] = xp[id][k];
		vi[j][k] = vp[id][k];
		//		ai[j][k] = 0.0;
		//		j6i[j][k] = 0.0;
	    }
	    poti[j] = pot[id];
	    h2i[j] = h2[id];
	}
	g6calc_firsthalf_all(nj, np, ilist+i,
			     xi, vi, ai, j6i, poti, eps2, h2i);
	g6calc_lasthalf2_all(nj, np, ilist+i, xi, vi, eps2, h2i,
			     a_tmp, j_tmp, pot_tmp, nnb_tmp);
	for(j=0;j<np;j++){
	    id = ilist[i+j];
	    for(k=0;k<3;k++){
		a_new[id][k] = a_tmp[j][k];
		j_new[id][k] = j_tmp[j][k];
	    }
	    pot[id] = pot_tmp[j];
	    nnb[id] = nnb_tmp[j];
	}
    }
}



