//
// testshermite.c
// test driver for hermite.vsm-generated interface library
//
// J. Makino
//    Time-stamp: <2005/07/26 17:27:16 makino>

#include <stdio.h>
#include <stdlib.h>
#include "hermite.h"

#define N 2
#define NJ 32
struct grape_j_particle_struct xj[NJ];
struct grape_i_particle_struct xi[N*4];
struct grape_result_struct result[N*4];

double ran()
{
    return drand48()*2-1;
}
void init()
{
    int i;
    xj[0].tj=0;
    xj[0].x0=ran(); 
    xj[0].y0=ran();
    xj[0].z0=ran();
    xj[0].v0x=0;
    xj[0].v0y=0;
    xj[0].v0z=0;
    xj[0].ax2=0;
    xj[0].ay2=0.;
    xj[0].az2=0.;
    xj[0].jx6=0.;
    xj[0].jy6=0.;
    xj[0].jz6=0.;
    xj[0].sx18=0.0;
    xj[0].sy18=00.;
    xj[0].sz18=0.;
    xj[0].mj=1;
    xj[0].idxj=999;
    xj[1]=xj[0];
    for(i=1;i<NJ;i++)  xj[i]=xj[0];
    for(i=0;i<NJ;i++){
	SING_send_j_particle(xj+i, i);
    }
    for(i=0;i<(N*4);i++){
	xi[i].ti=0.0;
	xi[i].xi=ran();
	xi[i].yi=ran();
	xi[i].zi=ran();
	xi[i].vxi=0;
	xi[i].vyi=0;
	xi[i].vzi=0;
	xi[i].eps2i=(1.0/64)*(1.0/64);
	xi[i].idxi=i;
    }
}


struct grape_result_struct acc_and_jerk(int nj,
			     struct grape_j_particle_struct xj[],
			     struct grape_i_particle_struct *xi)

{
    int i, j, k;
    int minid = -1;
    double rmin = 1e300;
    struct grape_result_struct  rp;
    rp.accx=0;
    rp.accy=0;
    rp.accz=0;
    rp.jerkx=0;
    rp.jerky=0;
    rp.jerkz=0;
    rp.pot=0;
    for(j=0; j<nj; j++){
	double dx, dy, dz, r2, r3inv, fi;
	double dvx, dvy, dvz, rv, r5inv, r2inv, a3;
	if (xj[j].idxj != xi->idxi){
	    struct grape_j_particle_struct *p = xj+j; 
	    double s, s0, s1, s2;
	    double xxj,yj,zj,vxj,vyj,vzj;
	    s = xi->ti - xj[j].tj;
	    s0 = s*0.75;
	    s1 = s*1.5;
	    s2 = s*2;
	    xxj=(((p->sx18*s0+p->jx6)*s + p->ax2)*s +p->v0x)*s+ p->x0;
	    yj= (((p->sy18*s0+p->jy6)*s + p->ay2)*s +p->v0y)*s+ p->y0;
	    zj= (((p->sz18*s0+p->jz6)*s + p->az2)*s +p->v0z)*s+ p->z0;
            vxj =((p->sx18*s +p->jx6)*s1+ p->ax2)*s2+p->v0x;
            vyj =((p->sy18*s +p->jy6)*s1+ p->ay2)*s2+p->v0y;
            vzj =((p->sz18*s +p->jz6)*s1+ p->az2)*s2+p->v0z;
	    dx = xxj - xi->xi;
	    dy = yj - xi->yi;
	    dz = zj - xi->zi;
	    r2 = dx*dx + dy*dy + dz*dz + xi->eps2i;
	    if (r2 < rmin){
		rmin = r2;
		minid = xj[j].idxj;
	    }
	    r2inv = 1.0/r2;
    	    r3inv = sqrt(r2inv)*r2inv;
    	    fi = xj[j].mj*r3inv;
    	    rp.accx += fi*dx;
    	    rp.accy += fi*dy;
    	    rp.accz += fi*dz;
    	    rp.pot -= fi*r2;
	    dvx = vxj - xi->vxi;
	    dvy = vyj - xi->vyi;
	    dvz = vzj - xi->vzi;
	    //	    fprintf(stderr,"dx, dvx = %e %e\n",dx,dvx);
	    //	    fprintf(stderr,"xi,vxi = %e %e\n",xi->xi, xi->vxi);
	    //	    fprintf(stderr,"xj,vxj = %e %e\n",xj->x0, xj->v0x);
	    a3 = -3.0*(dx*dvx+dy*dvy+dz*dvz)*r2inv;
	    rp.jerkx += fi*(dvx+a3*dx);
	    rp.jerky += fi*(dvy+a3*dy);
	    rp.jerkz += fi*(dvz+a3*dz);
    	}
	rp.nnb[0]= double_to_sword(rmin);
	rp.nnb[1]= minid;
    }
    return rp;
}

void print_one_result(char *s, int i, struct grape_result_struct *p)
{
    fprintf(stderr,"%s %2d, acc, pot, jerk = %13e %13e %13e %13e ",
	    s,i,p->accx,p->accy,p->accz,p->pot);
    fprintf(stderr," %13e %13e %13e \n",
	    p->jerkx,p->jerky,p->jerkz);
}
void print_two_results(int i,
		      struct grape_result_struct *p1,
		      struct grape_result_struct *p2)
{
    static double e2sum = 0;
    static double esum = 0;
    static int count = 0;
    double err = p2->pot- p1->pot;
    printf("POTS: %d  %.18e %.18e\n", i,p1->pot, p2->pot, err,err/p1->pot);
    err = err/p1->pot;
    e2sum+= err*err;
    esum+= err;
    count += 1;
    printf("STATS:  %e %e\n", sqrt(e2sum/count), esum/count);
}

print_result(struct grape_result_struct * result)
{
    struct grape_result_struct *p;
    struct grape_result_struct r_comp;
    int i;
    p=result;
    for(i=0;i<(N*4);i++){
	r_comp = acc_and_jerk(NJ, xj,xi+i);
	print_two_results(i,p, &r_comp);
	p++;
    }
}    

int main()
{
    int i, niparallel, ni, nj, nloop,iloop;
    ni=N*4;
    nj=NJ;
    SING_grape_init();
    srand48(12345);
    nloop =2000;
    for(iloop=0;iloop<nloop;iloop++){
	init();
	niparallel = SING_number_of_pe()*4;
	for (i=0; i< ni;i+=niparallel){
	    int njreal = nj/SING_number_of_bb();
	    SING_send_i_particle(xi+i, niparallel);
	    printf("i, njreal=%d %d\n", i, njreal);
	    SING_grape_run(njreal);
	    SING_get_result(result+i);
	}
	print_result(result);
    }
}
