//
// testgrape6.c
// test driver for g6-compatibility library
//
// J. Makino
//    Time-stamp: <10/11/24 13:35:03 makino>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "hermite-16x32.h"
#include "grape6.h"


#define NMAX  100000

#define MAXPIPES 256


int sample_calculate_accel_by_grape6(int clusterid,   // in manual
				     int ni,
				     double xi[][3],
				     double vi[][3],
				     int nj,
				     double xj[][3],
				     double vj[][3],
				     double m[],
				     double a[][3],
				     double jerk[][3],
				     double pot[],
				     int nnbindex[],
				     double eps2)

{
    double ajtmp[3];
    double jjtmp[3];
    double j2jtmp[3];
    double ti, tj, dtj;
    int j0, i0;
    int i,k,ii, iend;
    int flag[MAXPIPES+1];
    int index[MAXPIPES+1];
    double h2[MAXPIPES+1];
    double eps2array[MAXPIPES+1];
    int nharderror = 0;
    int ipmax = g6_npipes();
    ti = 0.0;tj =0.0; dtj = 0.0078125;
    for(k=0;k<3;k++){
	ajtmp[k] = 0.0;
	jjtmp[k] = 0.0;
	j2jtmp[k] = 0.0;
    }
    g6_set_ti_(&clusterid, &ti);
    for(i=0;i<nj;i++){
	g6_set_j_particle_(&clusterid,&i,&i,&tj,&dtj,m+i,
			   j2jtmp,jjtmp,ajtmp,vj[i],xj[i]);
    }
    for(i=0;i<ipmax;i++){
	h2[i] = eps2;
	eps2array[i] = eps2;
    }
    for(i=0;i<ni;i+=ipmax){
	iend = ipmax; if  (iend+i > ni) iend = ni-i;
	for(ii=0;ii<iend;ii++)index[ii]=i+ii;
	g6calc_firsthalf_(&clusterid, &nj, &iend,index,&(xi[i]),&(vi[i]),
			  &(a[i]),&(jerk[i]),
			  pot+i,eps2array, h2);
	g6calc_lasthalf2_(&clusterid,&nj,&iend,index,&xi[i],&vi[i],
			  eps2array,h2, &a[i], &jerk[i], pot+i, nnbindex+i);
	
    }
    return 0;
}

typedef union datapack{
    unsigned int i32[2];
    UINT64        i64;
    double       d;
}DATAPACK;

#if 1
struct grape_result_struct acc_and_jerk(int nj,
			     struct grape_j_particle_struct xj[],
			     struct grape_i_particle_struct *xi)

{
    int i, j, k;
    int minid = -1;
    double rmin = 1e300;
    struct grape_result_struct  rp;
    rp.accx=0;
    rp.accy=0;
    rp.accz=0;
    rp.jerkx=0;
    rp.jerky=0;
    rp.jerkz=0;
    rp.pot=0;
    for(j=0; j<nj; j++){
	double dx, dy, dz, r2, r3inv, fi;
	double dvx, dvy, dvz, rv, r5inv, r2inv, a3;
	//	fprintf(stderr, "j, jid, iid = %d %d %d\n", j, xj[j].idxj,xi->idxi);
	if (xj[j].idxj != xi->idxi){
	    struct grape_j_particle_struct *p = xj+j; 
	    double s, s0, s1, s2;
	    double xxj,yj,zj,vxj,vyj,vzj;

	    s = xi->ti - xj[j].tj;
	    s0 = s*0.75;
	    s1 = s*1.5;
	    s2 = s*2;
	    xxj=(((p->sx18*s0+p->jx6)*s + p->ax2)*s +p->v0x)*s+ p->x0;
	    yj= (((p->sy18*s0+p->jy6)*s + p->ay2)*s +p->v0y)*s+ p->y0;
	    zj= (((p->sz18*s0+p->jz6)*s + p->az2)*s +p->v0z)*s+ p->z0;
            vxj =((p->sx18*s +p->jx6)*s1+ p->ax2)*s2+p->v0x;
            vyj =((p->sy18*s +p->jy6)*s1+ p->ay2)*s2+p->v0y;
            vzj =((p->sz18*s +p->jz6)*s1+ p->az2)*s2+p->v0z;

	    dx = xxj - xi->xi;
	    dy = yj - xi->yi;
	    dz = zj - xi->zi;
	    r2 = dx*dx + dy*dy + dz*dz + xi->eps2i;
	    //	    fprintf(stderr,"x,y,z, = %e %e %e %e %e %e\n",xxj, yj, zj,
	    //		    xi->xi,xi->yi,xi->zi);
	    //	    fprintf(stderr,"x,y,z, r2 = %e %e %e %e\n", dx, dy, dz, r2);
	    r2 = dx*dx + dy*dy + dz*dz;
	    if (r2 < rmin){
		rmin = r2;
		minid = xj[j].idxj;
		//		fprintf(stderr,"current minid = %d\n", minid);
	    }
	    r2inv = 1.0/r2;
    	    r3inv = sqrt(r2inv)*r2inv;
    	    fi = xj[j].mj*r3inv;
    	    rp.accx += fi*dx;
    	    rp.accy += fi*dy;
    	    rp.accz += fi*dz;
    	    rp.pot -= fi*r2;
	    dvx = vxj - xi->vxi;
	    dvy = vyj - xi->vyi;
	    dvz = vzj - xi->vzi;
	    //	    fprintf(stderr,"dx, dvx = %e %e\n",dx,dvx);
	    //	    fprintf(stderr,"xi,vxi = %e %e\n",xi->xi, xi->vxi);
	    //	    fprintf(stderr,"xj,vxj = %e %e\n",xj->x0, xj->v0x);
	    a3 = -3.0*(dx*dvx+dy*dvy+dz*dvz)*r2inv;
	    rp.jerkx += fi*(dvx+a3*dx);
	    rp.jerky += fi*(dvy+a3*dy);
	    rp.jerkz += fi*(dvz+a3*dz);
	    DATAPACK nnbd;
	    nnbd.i64 = minid ;
	    rp.nnb = nnbd.d;
	}
    }
    return rp;
}

void set_j_particle(struct grape_j_particle_struct xja[],
		    int address,
		    int index,
		    double tj, /* particle time */
		    double dtj, /* particle time */
		    double mass,
		    double a2by18[3], /* a2dot divided by 18 */
		    double a1by6[3], /* a1dot divided by 6 */
		    double aby2[3], /* a divided by 2 */
		    double v[3], /* velocity */
		    double x[3] /* position */)
{
    struct grape_j_particle_struct *xjp;
    xjp = xja + address;
    xjp->tj= tj;
    xjp->x0=x[0]; 
    xjp->y0=x[1];
    xjp->z0=x[2];
    xjp->v0x= v[0];
    xjp->v0y= v[1];
    xjp->v0z= v[2];
    xjp->ax2= aby2[0];
    xjp->ay2= aby2[1];
    xjp->az2= aby2[2];
    xjp->jx6=a1by6[0];
    xjp->jy6=a1by6[1];
    xjp->jz6=a1by6[2];
    xjp->sx18=a2by18[0];
    xjp->sy18=a2by18[1];
    xjp->sz18=a2by18[2];
    xjp->mj= mass;
    xjp->eps2=0;
    xjp->idxj=index;
}

static double ti_set;

void set_i_particle(struct grape_i_particle_struct * xi,
		    int index,
		    double x[3],
		    double v[3],
		    double eps2,
		    double h2)
{
    xi->ti= ti_set;
    xi->xi=x[0];
    xi->yi=x[1];
    xi->zi=x[2];
    xi->vxi=v[0];
    xi->vyi=v[1];
    xi->vzi=v[2];
    xi->idxi=index;
    xi->eps2i= eps2;
}


int host_nnb(double xi[3],
	     int iindex,
	     int nj,
	     double xj[][3])

{
    double rmin = 1e300;
    int imin =-1;
    int j;
    for(j=0;j<nj;j++){
	int k;
	double r2 = 0;
	for(k=0;k<3;k++) {
	    double dx = xi[k]-xj[j][k];
	    r2 += dx*dx;
	}
	if ((r2 < rmin ) &&(j != iindex)){
	    rmin = r2;
	    imin = j;
	}
    }
    return imin;
}

int host_calculate_accel_by_grape6(double xi[3],
				   double vi[3],
				   int iindex,
				   int nj,
				   double xj[][3],
				   double vj[][3],
				   double m[],
				   double a[3],
				   double jerk[3],
				   double *pot,
				   double *rnnb,
				   int *nnbindex,
				   double eps2,
				   struct grape_j_particle_struct *xjp)
{
    double ajtmp[3];
    double jjtmp[3];
    double j2jtmp[3];
    double ti, tj, dtj;
    int j0, i0;
    int i,k,ii, iend;
    int flag[MAXPIPES+1];
    int index[MAXPIPES+1];
    double h2;
    double eps2array[MAXPIPES+1];
    int nharderror = 0;
    int ipmax = g6_npipes();
    ti = 0.0;tj =0.0; dtj = 0.0078125;
    for(k=0;k<3;k++){
	ajtmp[k] = 0.0;
	jjtmp[k] = 0.0;
	j2jtmp[k] = 0.0;
    }
    ti_set = ti;
    for(i=0;i<nj;i++){
	set_j_particle(xjp,i,i,tj,dtj,m[i],
			   j2jtmp,jjtmp,ajtmp,vj[i],xj[i]);
    }
    h2 = eps2;
    struct grape_i_particle_struct xis;
    struct grape_result_struct result;
    set_i_particle(&xis,  iindex, xi, vi, eps2, h2);
    result =  acc_and_jerk( nj, xjp, &xis);
    a[0]=result.accx;
    a[1]=result.accy;
    a[2]=result.accz;
    jerk[0]=result.jerkx;
    jerk[1]=result.jerky;
    jerk[2]=result.jerkz;
    *pot    =result.pot;
    DATAPACK pack;
    pack.d = result.nnb;
    *rnnb= pack.d;
    *nnbindex =(int)( pack.i64 & 0xffffffL);
}


void print_one_result(char *s, int i, struct grape_result_struct *p)
{
    fprintf(stdout,"%s %2d, acc, pot, jerk = %13e %13e %13e %13e ",
	    s,i,p->accx,p->accy,p->accz,p->pot);
    //    fprintf(stdout," %13e %13e %13e \n",
    //	    p->jerkx,p->jerky,p->jerkz);
}
void print_two_results(int i,
		      struct grape_result_struct *p1,
		      struct grape_result_struct *p2)
{
    int nb1, nb2;
    DATAPACK nnbd1, nnbd2;
    nnbd1.d = p1->nnb;
    nb1 =nnbd1.i64;
    nnbd2.d = p2->nnb;
    nb2 =nnbd2.i64;
    fprintf(stdout,"gdr   %2d, acc, pot = %13e %13e %13e %13e\n",
	    i,p1->accx,p1->accy,p1->accz,p1->pot);
    fprintf(stdout,"check %2d, acc, pot = %13e %13e %13e %13e\n",
	    i,p2->accx,p2->accy,p2->accz,p2->pot);
    fprintf(stdout,"         gdr   jerk= %13e %13e %13e %8x %16Lx\n",
	    p1->jerkx,p1->jerky,p1->jerkz, nb1, nnbd1.i64   );
    fprintf(stdout,"         check jerk= %13e %13e %13e %8x %16Lx\n",
	    p2->jerkx,p2->jerky,p2->jerkz, nb2, nnbd2.i64);
}

#endif


int main(int argc, char **argv)
{
    static  double x[NMAX][3];
    static      double v[NMAX][3];
    static      double a2[NMAX][3];
    static      double j6[NMAX][3];
    static      double jd18[NMAX][3];
    static      double acc[NMAX][3];
    static      double jerk[NMAX][3];
    static      double pot[NMAX];
    static      int    nnbindex[NMAX];
    static      int    flag[NMAX];
    static      double m[NMAX];
    static      double t[NMAX];
    static      double dt[NMAX];
    static      struct grape_j_particle_struct xj[NMAX];

    int i,k,n,ii;
    int clusterid = 0;
    int mode = 0;
    double ti = 0;
    double eps2 = 0.01;
    double h2 = 2;
    n = 100;
    if (argc > 1){
	n = atoi(argv[1]);
    fprintf(stderr,"n = %d\n", n);
    }
    mode = 0;

    srand48(12345);
    for(i=0;i<n;i++){
	for(k=0;k<3;k++){
	    //	    x[i][k] = (n-i-1)*(k+2)/4.0;
	    x[i][k] = drand48();
	    v[i][k] = drand48();
	    a2[i][k] = j6[i][k] = jd18[i][k] = 0;
	    acc[i][k] = jerk[i][k] = 1;
	}
	m[i] = 1.0/n;
	t[i]  = 0;
	dt[i] = 0.125;
	pot[i] = 1;
    }
    g6_set_debug_level(255);
    g6_open_(&clusterid);
    //    fprintf(stderr," Nchips=%d\n",g6_getnchips(clusterid));
    //    fprintf(stderr," NJMAX=%d\n",g6_getnjmax(clusterid));
    int nerror = 0;

    if (1){
	double fscale = m[0]/eps2*100;
	double phiscale = m[0]/sqrt(eps2)*100;
	for(i=0;i<n;i++){
	    for(k=0;k<3;k++)acc[i][k] = fscale;
	    pot[i] = phiscale;
	}
	
	for(ii = 0;ii<10;ii++){
	    fprintf(stderr,"step %d\n", ii);
	    for(i=0;i<n;i++){
		for(k=0;k<3;k++){
		    x[i][k] = drand48();
		    //		    x[i][k]=i;
		    v[i][k] = drand48();
		}
	    }
	    //	    calculate_accel_by_grape6_noopen(clusterid,n,x,v,m,acc,jerk,pot,eps2);
	    sample_calculate_accel_by_grape6(clusterid, n, x, v, n, x, v, m,
					     acc, jerk, pot, nnbindex, eps2);
	}
	for(i=0;i<n;i++){
	    double ah[3], jh[3], ph, rnnbh;
	    int nnbh;
	    host_calculate_accel_by_grape6(x[i], v[i], i, n, x, v, m,
					   ah, jh, &ph, &rnnbh, &nnbh, eps2, xj);
	    int nnbh2 = host_nnb(x[i], i, n, x);
	    printf("i:%5d acc: %e  %e  %e \n          jerk %e  %e  %e \n pot, nnb: %e %d\n",
		   i, acc[i][0], acc[i][1], acc[i][2],
		   jerk[i][0], jerk[i][1], jerk[i][2],
		   pot[i], nnbindex[i]);
	    printf("i:%5d acc: %e  %e  %e \n          jerk %e  %e  %e \n pot, nnb: %e %d %d\n",
		   i, ah[0], ah[1], ah[2],
		   jh[0], jh[1], jh[2],
		   ph, nnbh, nnbh2 );
	    if ((nnbh != nnbindex[i])||(nnbh2 != nnbindex[i])){
		printf("NNB error: %d != %d, %d\n", nnbh, nnbh2, nnbindex[i]);
		nerror++;
	    }
	}
    }
    g6_close_(&clusterid);
    printf("Total NNB error = %d\n", nerror);
    return 0;
}
