//
// testgrape6_all.c
// test driver for g6-kfcr-compatibility library
//
// J. Makino
//    Time-stamp: <10/09/20 18:23:10 makino>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "grape6.h"


#define NMAX  100000

#define MAXPIPES 256


int sample_calculate_accel_by_grape6(int clusterid,   // in manual
						    int ni,
						    double xi[][3],
						    double vi[][3],
						    int nj,
						    double xj[][3],
						    double vj[][3],
						    double m[],
						    double a[][3],
						    double jerk[][3],
						    double pot[],
						    double eps2)
{
    double ajtmp[3];
    double jjtmp[3];
    double j2jtmp[3];
    double ti, tj, dtj;
    int j0, i0;
    int i,k,ii, iend;
    int flag[MAXPIPES+1];
    int index[MAXPIPES+1];
    double h2[MAXPIPES+1];
    double eps2array[MAXPIPES+1];
    int nharderror = 0;
    int ipmax = g6_npipes();
    ti = 0.0;tj =0.0; dtj = 0.0078125;
    for(k=0;k<3;k++){
	ajtmp[k] = 0.0;
	jjtmp[k] = 0.0;
	j2jtmp[k] = 0.0;
    }
    g6_set_ti_all(ti);
    for(i=0;i<nj;i++){
	g6_set_j_particle_all(i,i,tj,dtj,m[i],
			   j2jtmp,jjtmp,ajtmp,vj[i],xj[i]);
    }
    for(i=0;i<ipmax;i++){
	h2[i] = eps2;
	eps2array[i] = eps2;
    }
    for(i=0;i<ni;i+=ipmax){
	iend = ipmax; if  (iend+i > ni) iend = ni-i;
	for(ii=0;ii<iend;ii++)index[ii]=i+ii;
	g6calc_firsthalf_all(nj, iend,index,&(xi[i]),&(vi[i]),
			  &(a[i]),&(jerk[i]),
			  pot+i,eps2array[0], h2);
	g6calc_lasthalf_all(nj,iend,index,&xi[i],&vi[i],
			 eps2array[0],h2, &a[i], &jerk[i], pot+i);
	
    }
    return 0;
}

int main(int argc, char **argv)
{
    static  double x[NMAX][3];
    static      double v[NMAX][3];
    static      double a2[NMAX][3];
    static      double j6[NMAX][3];
    static      double jd18[NMAX][3];
    static      double acc[NMAX][3];
    static      double jerk[NMAX][3];
    static      double pot[NMAX];
    static      int    flag[NMAX];
    static      double m[NMAX];
    static      double t[NMAX];
    static      double dt[NMAX];
    int i,k,n,ii;
    int clusterid = 0;
    int mode = 0;
    double ti = 0;
    double eps2 = 0.01;
    double h2 = 2;
    n = 100;
    if (argc > 1){
	n = atoi(argv[1]);
    fprintf(stderr,"n = %d\n", n);
    }
    mode = 0;

    srand48(12345);
    for(i=0;i<n;i++){
	for(k=0;k<3;k++){
	    x[i][k] = (n-i-1)*(k+2)/4.0;
	    x[i][k] = drand48();
	    v[i][k] = drand48();
	    a2[i][k] = j6[i][k] = jd18[i][k] = 0;
	    acc[i][k] = jerk[i][k] = 1;
	}
	m[i] = 1.0/n;
	t[i]  = 0;
	dt[i] = 0.125;
	pot[i] = 1;
    }
    g6_open_all();
    //    fprintf(stderr," Nchips=%d\n",g6_getnchips(clusterid));
    //    fprintf(stderr," NJMAX=%d\n",g6_getnjmax(clusterid));

    if (1){
	double fscale = m[0]/eps2*100;
	double phiscale = m[0]/sqrt(eps2)*100;
	for(i=0;i<n;i++){
	    for(k=0;k<3;k++)acc[i][k] = fscale;
	    pot[i] = phiscale;
	}
	
	for(ii = 0;ii<10;ii++){
	    fprintf(stderr,"step %d\n", ii);
	    for(i=0;i<n;i++){
		for(k=0;k<3;k++){
		    x[i][k] = drand48();
		    v[i][k] = drand48();
		}
	    }
	    //	    calculate_accel_by_grape6_noopen(clusterid,n,x,v,m,acc,jerk,pot,eps2);
	    sample_calculate_accel_by_grape6(clusterid, n, x, v, n,
					     x, v, m,acc, jerk, pot, eps2);
	}
	for(i=0;i<n;i++){
	    printf("i:%5d acc: %e  %e  %e \n          jerk %e  %e  %e \n pot, flag: %e %x\n",
		   i, acc[i][0], acc[i][1], acc[i][2],
		   jerk[i][0], jerk[i][1], jerk[i][2],
		   pot[i], flag[i]);
	}
    }
    g6_close_(&clusterid);
    return 0;
}
