//
// testshermite.c
// test driver for hermite.vsm-generated interface library
//
// J. Makino
//    Time-stamp: <10/09/16 13:13:20 makino>

#include <stdio.h>
#include <sys/time.h>
#include <stdlib.h>
#include <math.h>
#include "hermite-16x32.h"

#define N  32
#define NJ 256
struct grape_j_particle_struct xj[NJ];
struct grape_i_particle_struct xi[N*4];
struct grape_result_struct result[N*4];

void init0()
{
    int i;
    xj[0].x0=1; 
    xj[0].y0=1;
    xj[0].z0=1;
#if 0
    xj[0].tj=0;
    xj[0].v0x=0.1;
    xj[0].v0y=0.2;
    xj[0].v0z=0.25;
    xj[0].ax2=1;
    xj[0].ay2=2.;
    xj[0].az2=3.;
    xj[0].jx6=40.;
    xj[0].jy6=50.;
    xj[0].jz6=60.;
    xj[0].sx18=1000.0;
    xj[0].sy18=00.;
    xj[0].sz18=0.;
    xj[0].mj=1;
    xj[0].idxj=8;
#endif    
#if 1
    xj[0].tj=0.;
    xj[0].v0x=1.0;
    xj[0].v0y=0.5;
    xj[0].v0z=-1;
    xj[0].ax2=-2;
    xj[0].ay2=3.;
    xj[0].az2=4.;
    xj[0].jx6=2.;
    xj[0].jy6=3.;
    xj[0].jz6=5.;
    xj[0].sx18=20.0;
    xj[0].sy18=30.;
    xj[0].sz18=40.;
    xj[0].mj=2;
    xj[0].idxj=8;
#endif    
    xj[1]=xj[0];
    for(i=1;i<NJ;i++)  xj[i]=xj[0];
    for(i=0;i<NJ;i++){
	//	xj[i].idxj=i;
	xj[i].x0=i-NJ/2;
	SING_send_j_particleMC(0,xj+i, i);
    }
    xi[0].ti=0.1;
    xi[0].xi=0.;
    xi[0].yi=1.;
    xi[0].zi=0.;
    xi[0].vxi=0;
    xi[0].vyi=0;
    xi[0].vzi=0;
    xi[0].eps2i=0;
    for(i=1;i<(N*4);i++)xi[i]=xi[0];
    for(i=0;i<(N*4);i++)xi[i].xi=1.0/(i+1.0);
    //    for(i=0;i<(N*4);i++)xi[i].yi=0.0;
    //    for(i=0;i<(N*4);i++)xi[i].vxi=0.0;
    //    for(i=0;i<(N*4);i++)xi[i].vyi=.0;
    //    for(i=0;i<(N*4);i++)xi[i].vzi=.0;
    //    for(i=0;i<(N*4);i++)xi[i].idxi=i;
    for(i=0;i<(N*4);i++){
	//	fprintf(stderr,"test init i=%d, x, y, z= %e %e %e\n",
	//		i,xi[i].xi,xi[i].yi,xi[i].zi);
    }
}
void init()
{
    int i;
    xj[0].x0=0; 
    xj[0].y0=1;
    xj[0].z0=0;
    xj[0].tj=0;
    xj[0].v0x=0;
    xj[0].v0y=0;
    xj[0].v0z=0;
    xj[0].ax2=0;
    xj[0].ay2=0.;
    xj[0].az2=0.;
    xj[0].jx6=0.;
    xj[0].jy6=0.;
    xj[0].jz6=0.;
    xj[0].sx18=000.0;
    xj[0].sy18=00.;
    xj[0].sz18=0.;
    xj[0].eps2 = 0;
    xj[0].mj=1;
    xj[0].idxj=8;
    for(i=1;i<NJ;i++)  xj[i]=xj[0];
    //    for(i=1;i<NJ;i++)  xj[i].mj=0;
    //   for(i=0;i<NJ;i++)  xj[i].y0=i;
   //   for(i=0;i<NJ;i++)  xj[i].v0x=i;
   for(i=0;i<NJ;i++)  xj[i].idxj= i<<8;
   for(i=0;i<NJ;i++)  xj[i].x0= i-NJ/2;
   for(i=0;i<NJ;i++)  xj[i].x0= i;
    //    for(i=0;i<NJ;i+=16)  xj[i].mj=1;
    
    for(i=0;i<(NJ);i++){
	SING_send_j_particleMC(0,xj+i, i);
    }
    
    xi[0].ti=0.0;
    xi[0].xi=0.;
    xi[0].yi=0.;
    xi[0].zi=0.;
    xi[0].vxi=0;
    xi[0].vyi=0;
    xi[0].vzi=0;
    xi[0].eps2i=0;
    xi[0].idxi=0;
    for(i=1;i<(N*4);i++)xi[i]=xi[0];
    for(i=0;i<(N*4);i++)xi[i].idxi = i<<8;
    for(i=0;i<(N*4);i++)xi[i].xi = i;
}


typedef union datapack{
    unsigned int i32[2];
    UINT64        i64;
    double       d;
}DATAPACK;


struct grape_result_struct acc_and_jerk(int nj,
			     struct grape_j_particle_struct xj[],
			     struct grape_i_particle_struct *xi)

{
    int i, j, k;
    int minid = -1;
    double rmin = 1e300;
    struct grape_result_struct  rp;
    rp.accx=0;
    rp.accy=0;
    rp.accz=0;
    rp.jerkx=0;
    rp.jerky=0;
    rp.jerkz=0;
    rp.pot=0;
    for(j=0; j<nj; j++){
	double dx, dy, dz, r2, r3inv, fi;
	double dvx, dvy, dvz, rv, r5inv, r2inv, a3;
	if (xj[j].idxj != xi->idxi){
	    struct grape_j_particle_struct *p = xj+j; 
	    double s, s0, s1, s2;
	    double xxj,yj,zj,vxj,vyj,vzj;
#if 1
	    s = xi->ti - xj[j].tj;
	    s0 = s*0.75;
	    s1 = s*1.5;
	    s2 = s*2;
	    xxj=(((p->sx18*s0+p->jx6)*s + p->ax2)*s +p->v0x)*s+ p->x0;
	    yj= (((p->sy18*s0+p->jy6)*s + p->ay2)*s +p->v0y)*s+ p->y0;
	    zj= (((p->sz18*s0+p->jz6)*s + p->az2)*s +p->v0z)*s+ p->z0;
            vxj =((p->sx18*s +p->jx6)*s1+ p->ax2)*s2+p->v0x;
            vyj =((p->sy18*s +p->jy6)*s1+ p->ay2)*s2+p->v0y;
            vzj =((p->sz18*s +p->jz6)*s1+ p->az2)*s2+p->v0z;
#endif
	    dx = xxj - xi->xi;
	    dy = yj - xi->yi;
	    dz = zj - xi->zi;
	    r2 = dx*dx + dy*dy + dz*dz + xi->eps2i;
	    //	    r2 = dx*dx + dy*dy + dz*dz;
	    if (r2 < rmin){
		rmin = r2;
		minid = xj[j].idxj;
	    }
	    r2inv = 1.0/r2;
    	    r3inv = sqrt(r2inv)*r2inv;
    	    fi = xj[j].mj*r3inv;
    	    rp.accx += fi*dx;
    	    rp.accy += fi*dy;
    	    rp.accz += fi*dz;
    	    rp.pot -= fi*r2;
	    dvx = vxj - xi->vxi;
	    dvy = vyj - xi->vyi;
	    dvz = vzj - xi->vzi;
	    //	    fprintf(stderr,"dx, dvx = %e %e\n",dx,dvx);
	    //	    fprintf(stderr,"xi,vxi = %e %e\n",xi->xi, xi->vxi);
	    //	    fprintf(stderr,"xj,vxj = %e %e\n",xj->x0, xj->v0x);
	    a3 = -3.0*(dx*dvx+dy*dvy+dz*dvz)*r2inv;
	    rp.jerkx += fi*(dvx+a3*dx);
	    rp.jerky += fi*(dvy+a3*dy);
	    rp.jerkz += fi*(dvz+a3*dz);
	    DATAPACK nnbd;
	    nnbd.i64 = minid >>8;
	    rp.nnb = nnbd.d;
	}
    }
    return rp;
}

void print_one_result(char *s, int i, struct grape_result_struct *p)
{
    fprintf(stdout,"%s %2d, acc, pot, jerk = %13e %13e %13e %13e ",
	    s,i,p->accx,p->accy,p->accz,p->pot);
    //    fprintf(stdout," %13e %13e %13e \n",
    //	    p->jerkx,p->jerky,p->jerkz);
}
void print_two_results(int i,
		      struct grape_result_struct *p1,
		      struct grape_result_struct *p2)
{
    int nb1, nb2;
    DATAPACK nnbd1, nnbd2;
    nnbd1.d = p1->nnb;
    nb1 =nnbd1.i64;
    nnbd2.d = p2->nnb;
    nb2 =nnbd2.i64;
    fprintf(stdout,"gdr   %2d, acc, pot = %13e %13e %13e %13e\n",
	    i,p1->accx,p1->accy,p1->accz,p1->pot);
    fprintf(stdout,"check %2d, acc, pot = %13e %13e %13e %13e\n",
	    i,p2->accx,p2->accy,p2->accz,p2->pot);
    fprintf(stdout,"         gdr   jerk= %13e %13e %13e %8x %16Lx\n",
	    p1->jerkx,p1->jerky,p1->jerkz, nb1, nnbd1.i64   );
    fprintf(stdout,"         check jerk= %13e %13e %13e %8x %16Lx\n",
	    p2->jerkx,p2->jerky,p2->jerkz, nb2, nnbd2.i64);
}

print_result(struct grape_result_struct * result)
{
    struct grape_result_struct *p;
    struct grape_result_struct r_comp;
    int i;
    p=result;
    for(i=0;i<(N*4);i++){
	r_comp = acc_and_jerk(NJ, xj,xi+i);
	//	print_one_result("test    result",i, p);
	//	print_one_result("compare result",i, &r_comp);
	print_two_results(i,p, &r_comp);
	p++;
    }
}    

int main()
{
    int i, niparallel, ni, nj;
    ni=N*4;
    nj=NJ;
    SING_grape_initMC(0);
    init();
    niparallel = SING_number_of_pe()*4;
    for (i=0; i< ni;i+=niparallel){
	int njreal = nj/SING_number_of_bb();
	SING_send_i_particleMC(0,xi+i, niparallel);
	printf("i, njreal=%d %d\n", i, njreal);
	SING_grape_runMC(0,njreal);
	SING_get_resultMC(0,result+i);
    }
    print_result(result);
    
}
