#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#ifndef DGEMMTEST
//#include <mpi.h>
#endif
#include "./dgemmconf.h"
#include "dgemmbody.h"
#include "matmul32-32.h"
#include "testmatmul32.h"

#include "./multiboard-control.h"

#define NCHIP 4
#define NPE   32
#define NBB   16

#define DIM0 (32*NBB*NCHIP) //=512*NCHIP
#define DIM1 (8*NPE)        //=25
#define USEGDR

#ifndef MPI_MAX_PROCESSOR_NAME
#define  MPI_MAX_PROCESSOR_NAME 1024
#endif

static int boardid = 0;
static int nboards = 1;

void gdrsetboardid(int id){boardid = id;}
int  gdrgetboardid(){return boardid;}
void gdrsetnboards(int nb)
{
    nboards = nb;
    matmul_setnboards(nb);
}
int  gdrgetnboards(){return nboards;}

void gdr_init()
{
  static int firstrun=1;
  if(firstrun==1){
      grape_init(boardid*NCHIP, nboards*NCHIP);
      gdr_multi_cntl_init(gdrgetboardid());
    set_coc_reg(2,(UINT64) 0);

    firstrun = 0;
  }
}

double wallclock();



void check_matrix_forced(int M, int N, int LDA, double A[][LDA], int LDB, double B[][LDB])
{
  int i, j;
  double err=0.0, da, damax=1e-15;

  for(i=0;i<M;i++){
    for(j=0;j<N;j++){
      da = fabs(A[j][i] - B[j][i]);
      if(da > damax){
	damax = da;
      }
      err += da;
      if(da > 1e-10){
	printf("(%d, %d) : err=%e \n", i, j, da);
	//exit(1);
      }
    }
    //printf("\n");
  }
  err = err / (double)M / (double)N;
  printf("error : %e errmax : %e \n", err, damax);
}

long long int calc_checksum(int DIMM, int DIMN, int LDC, double C[][LDC])
{
  long long int isum, ic;
  int i, j;

  isum=0;
  for(i=0;i<DIMN;i++){
    for(j=0;j<DIMM;j++){
      ic = *((long long int *)(&(C[i][j])));
      isum ^= ic;
    }
  }

  return isum;
}


void mapAB_ij_matrix(int M, int N, double *A, double *B, double **i, double **j, int *NOTI, int *NOTJ, int *NOTC)
{
  //sending of i matrix is faster than that of j matrix (overlap)
  if(M > N){
    (*i) = A;
    (*j) = B;
    (*NOTC) = 1;
    fprintf(stderr,"A->i B->j NOTC=1 \n");
  } else {
    (*i) = B;
    (*j) = A;
    (*NOTI) = (*NOTI)==1 ? 0 : 1;
    (*NOTJ) = (*NOTJ)==1 ? 0 : 1;
    (*NOTC) = 0;
    fprintf(stderr,"A->j B->i NOTC=0 \n");
  }
}

void print_matrix2(int M, int N, int LDA, double A[N][LDA])
{
  int i, j;
  int start=768;

  for(i=0;i<M;i++){
    for(j=0;j<N;j++){
      printf("%e ", A[j+250][i+250+256]);
    }
    printf("\n");
  }
}

static char procname[1024]="Uninitialized name";

void gdrdgemm_set_procname(int i)
{
    sprintf(procname, "Proc %d",i);
}


#if (defined CALCABCSUM && (!defined DGEMMTEST)) 
static void checkabcsum(int NOTA, int NOTB, int mblk, int nblk, int K, 
			int LDA, int LDB, int LDC, double A[][LDA], double B[][LDB],  double C[][LDC])

{
  unsigned long long ca, cb, cc;
  int len;
      
  if(NOTA==1){
    ca = calc_checksum(mblk, K, LDA, A);
  } else {
    ca = calc_checksum(K, mblk, LDA, A);
  }
  if(NOTB==1){
    cb = calc_checksum(K, nblk, LDB, B);
  } else {
    cb = calc_checksum(nblk, K, LDB, B);
  }
  cc = calc_checksum(mblk, nblk, LDC, C);
      
  fprintf(stderr,"%s : dgemmsum : mblk=%d nblk=%d K=%d A=%016llx B=%016llx C=%016llx  \n", procname, mblk, nblk, K, ca, cb, cc);
  if(cc==0){
    print_matrix2(10, 10, LDC, C);
  }
}

static void calcabsum(int NOTA, int NOTB, int mblk, int nblk, int K, 
		       int LDA, int LDB, double A[][LDA], double B[][LDB])

{
  unsigned long long ca, cb, cc;
  int len;
      
  if(NOTA==1){
    ca = calc_checksum(mblk, K, LDA, A);
  } else {
    ca = calc_checksum(K, mblk, LDA, A);
  }
  if(NOTB==1){
    cb = calc_checksum(K, nblk, LDB, B);
  } else {
    cb = calc_checksum(nblk, K, LDB, B);
  }
      
  fprintf(stderr,"%s : inputsum : A=%016llx B=%016llx \n", procname, ca, cb);
}
#endif


int dqsdelay=8;
int ch;
int dooverlap=1;

void setdqsdelayandch(int x, int y)
{
    dqsdelay=x;
    ch = y;
}

void setdooverlap(int x)
{
    dooverlap=x;
}

static int forceswapab = 0;

void gdrsetforceswapab(){forceswapab = 1;}
void gdrresetforceswapab(){forceswapab = 0;}

void gdr_dgemm_(int *NOTA, int *NOTB, int *M, int *N, int *K,
		double *ALPHA,
		double *BETA,
		int *LDA,
		int *LDB,
		int *LDC,
		double A[][(*LDA)],
		double B[][(*LDB)],
		double C[][(*LDC)],
		int *gdrdoneflag)
{
  static int initdone=0;
  int mblk, nblk, mrem, nrem, i, j;
  double *C2, start, end;
  char ta = ((*NOTA)==1) ? 'N' : 'T';
  char tb = ((*NOTB)==1) ? 'N' : 'T';
  long long int check, asum, bsum;

  if(initdone==0){
#ifdef USEGDR
    gdr_init();
#endif
    initdone=1;
  }
#ifdef VVERBOSELOG  
  fprintf(stderr,"gdr_dgemm_ forceswapab=%d DIM1=%d\n", forceswapab, DIM1);
#endif  
  //  fprintf(stderr,"gdr_dgemm_ MNK=%d %d %d DIM1 = %d\n", *M, *N, *K, DIM1);
#if MINK==256
  if(((*K==2048 || *K==1024 ) && *M>=DIM1 && *N>=DIM1)
     ||( *K==512 && *N>=1024)
     ||( *K==256 && *N>=8192))      {
#elif MINK==512
  //  if((*K==2048 || *K==1024 || *K==512) && *M>=DIM1*2 && *N>=DIM1*2){
  // why check the value of N>= 512? 256 should be fine.
  if((*K==2048 || *K==1024 || *K==512) && *M>=DIM1 && *N>=DIM1){
#elif MINK==1024
  if((*K==2048 || *K==1024 ) && *M>=DIM1*2 && *N>=DIM1*2){
#else
  if((*K==2048) && *M>=DIM1*2 && *N>=DIM1*2){
#endif
    start = wallclock();
#ifdef VVERBOSELOG
#ifdef DGEMMTEST
    fprintf(stderr,"NA:%d NB:%d (%d, %d, %d) a=%e b=%e LDA=%d LDB=%d LDC=%d \n", *NOTA, *NOTB, *M, *N, *K, *ALPHA, *BETA, *LDA, *LDB, *LDC);
#else
#ifdef USEMPILIB
    {
      int rank;
      MPI_Comm_rank(MPI_COMM_WORLD, &rank);
      if(rank==0){
	fprintf(stderr,"NA:%d NB:%d (%d, %d, %d) a=%e b=%e LDA=%d LDB=%d LDC=%d \n", *NOTA, *NOTB, *M, *N, *K, *ALPHA, *BETA, *LDA, *LDB, *LDC);
      }
    }
#else
    fprintf(stderr,"NA:%d NB:%d (%d, %d, %d) a=%e b=%e LDA=%d LDB=%d LDC=%d \n", *NOTA, *NOTB, *M, *N, *K, *ALPHA, *BETA, *LDA, *LDB, *LDC);
#endif
#endif
#endif

    mrem = (*M) % DIM1;
    nrem = (*N) % DIM1;
    mblk = (*M) - mrem;
    nblk = (*N) - nrem;

#ifdef VVERBOSELOG    
    fprintf(stderr,"mrem=%d, mblk=%d \n", mrem, mblk);
    fprintf(stderr,"nrem=%d, nblk=%d \n", nrem, nblk);
#endif

#ifdef CHECKGDR
    C2 = malloc(sizeof(double)*mblk*nblk); //C2[nblk][mblk]
    for(i=0;i<nblk;i++){
      for(j=0;j<mblk;j++){
	C2[mblk*i+j] = C[i][j];
      }
    }
#endif
    
#ifdef CHECKGDR
    dgemm_(&ta, &tb, &mblk, &nblk, K, ALPHA, A, LDA, B, LDB, BETA, C2, LDC);
    check_matrix_forced(nblk, mblk, *LDC, C, *LDC, C2);
    free(C2);
#endif

#if (defined CALCABCSUM && (!defined DGEMMTEST)) 
    calcabsum(*NOTA, *NOTB, mblk, nblk, *K, 
		*LDA, *LDB,  A, B);
#endif


#if 1
    if(mrem>0){
      double *c_ptr = &(C[0][mblk]);
      double al = *ALPHA;
      double bt = *BETA;
      int m = mrem;
      int n = (*N);
      int k = (*K);
      double *a_ptr;
      double *b_ptr;

      if((*NOTA)==1){
	a_ptr = &(A[0][mblk]);
      } else {
	a_ptr = &(A[mblk][0]);
      }

      b_ptr = &(B[0][0]);
      
      dgemm_(&ta, &tb, &m, &n, &k, &al, a_ptr, LDA, b_ptr, LDB, &bt, c_ptr, LDC);
    }

    if(nrem>0){
      double *c_ptr = &(C[nblk][0]);
      double al = *ALPHA;
      double bt = *BETA;
      double *a_ptr;
      double *b_ptr;
      int m = mblk;
      int n = nrem;
      int k = (*K);
      int i, j, l;
      a_ptr = &(A[0][0]);
      if((*NOTB)==1){
	b_ptr = &(B[nblk][0]);
      } else {
	b_ptr = &(B[0][nblk]);
      }
      dgemm_(&ta, &tb, &m, &n, &k, &al, a_ptr, LDA, b_ptr, LDB, &bt, c_ptr, LDC);
    }
#endif

#ifdef USEGDR
    if((*BETA)!=1.0){
      for(i=0;i<nblk;i++){
	for(j=0;j<mblk;j++){
	  C[i][j] *= (*BETA);
	}
      }
    }
    


#ifdef DGEMMTEST
    if(dooverlap){ 
#else
    if(*K==2048 || *K==1024 || *K==512 || *K==256){
#endif
#ifdef USEOVLDGEMM
#ifdef VVERBOSELOG
      fprintf(stderr,"overlapped dgemm \n"); // for update
#endif
      int bbcast = 0;
      if (forceswapab && gdrgetnboards()>1) bbcast = 1;
      //      bbcast =1;
      if((mblk >= nblk) || ( forceswapab)) { //A->i B->j C is not transposed
	  //	  fprintf(stderr,"usual call to ovl_dg with m, n=%d %d bbcast=%d\n", mblk, nblk, bbcast);
	  int ktry = *K;
	  matmul_tb4_mnk_ovl_dg_try(mblk, nblk, ktry, *NOTA, *NOTB, 0,bbcast,
				*LDA, *LDB, *LDC, A, B, C, *ALPHA, *BETA);

      } else { //A->j B->i C is transpose
#ifdef USEAUTOTRANS
	int NOTI = (*NOTA)==1 ? 0 : 1;
	int NOTJ = (*NOTB)==1 ? 0 : 1;
#ifdef VERBOSELOG
	fprintf(stderr,"A->j B->i NOTI=%d NOTJ=%d LDA=%d LDB=%d LDC=%d\n", NOTI, NOTJ, *LDA, *LDB , *LDC);
#endif
	matmul_tb4_mnk_ovl_dg_try(nblk, mblk, *K, NOTJ, NOTI,1,bbcast,
			      *LDB, *LDA, *LDC, B, A, C, *ALPHA, *BETA);

#else
	//	fprintf(stderr,"usual call to ovl_dg (should be transposed) with m, n=%d %d bbcast=%d\n", mblk, nblk, bbcast);

	matmul_tb4_mnk_ovl_dg_try(mblk, nblk, *K, *NOTA, *NOTB, 0,bbcast,
			      *LDA, *LDB, *LDC, A, B, C, *ALPHA, *BETA);
#endif
      }
#else

      //      fprintf(stderr,"usual call to nonovl_dg with m, n=%d %d bbcast=%d\n", mblk, nblk, bbcast);
      matmul_tb4_mnk_dg(mblk, nblk, *K, *NOTA, *NOTB, 0,
			*LDA, *LDB, *LDC, A, B, C, *ALPHA, *BETA);
#endif

    } else {
      fprintf(stderr,"dgemm \n"); // for dtrsm
      matmul_tb4_mnk_dg(mblk, nblk, *K, *NOTA, *NOTB, *LDA, *LDB, *LDC, A, B, C, *ALPHA, *BETA);
    }

#ifdef CALCCSUM
    check = calc_checksum(mblk, nblk, *LDC, C);
    fprintf(stderr,"C sum : %016llx \n", check);
    if(check==0){
      print_matrix2(32, 32, *LDC, C);
    }
#endif

#if (defined CALCABCSUM && (!defined DGEMMTEST)) 
    checkabcsum(*NOTA, *NOTB, mblk, nblk, *K, 
		*LDA, *LDB, *LDC, A, B, C);
#endif

#else
    dgemm_(&ta, &tb, &mblk, &nblk, K, ALPHA, A, LDA, B, LDB, BETA, C, LDC);
#endif
    
    end = wallclock();
#ifdef VVERBOSELOG
    fprintf(stderr,"gdr_dgemm elapsed : %e [sec] \n", end-start);
#endif

    *gdrdoneflag = 1;
  } else {
    *gdrdoneflag = 0;
  }
}


void set_matrix(int M, int N, int K, double A[K][M], double B[N][K], double C[N][M])
{
  int i, j;

  for(i=0;i<K;i++){
    for(j=0;j<M;j++){
      A[i][j] = drand48();
    }
  }

  for(i=0;i<N;i++){
    for(j=0;j<K;j++){
      B[i][j] = drand48();
    }
  }

  for(i=0;i<N;i++){
    for(j=0;j<M;j++){
      C[i][j] = drand48();
    }
  }
}

void set_matrix_2(int M, int N, int K, double A[K][M], double B[N][K], double C[N][M], double C2[N][M])
{
    int i, j;
    
    for(i=0;i<K;i++){
	for(j=0;j<M;j++){
	    A[i][j] = drand48();
	}
    }
    
    for(i=0;i<N;i++){
	for(j=0;j<K;j++){
	    B[i][j] = drand48();
	}
    }
    if (C2 != NULL){
	for(i=0;i<N;i++){
	    for(j=0;j<M;j++){
		C[i][j] = drand48();
		C2[i][j] = C[i][j];
	    }
	}
    }else{
	for(i=0;i<N;i++){
	    for(j=0;j<M;j++){
		C[i][j] = drand48();
	    }
	}
    }      
}
void set_matrix_2a(int M, int N, int K,
		   int lda, int ldb, int ldc,
		   double A[][lda], double B[][ldb],
		   double C[][ldc], double C2[][ldc])
{
    int i, j;
    
    for(i=0;i<K;i++){
	for(j=0;j<M;j++){
	    A[i][j] = drand48();
	}
    }
#if 0
    for(i=0;i<K;i++){
	for(j=0;j<M;j++){
	    //	    A[i][j] = (double)i+1;
	    A[i][j] = i*10000+j;
	}
    }
#endif    
#if 0
    for(i=0;i<K;i++){
	for(j=0;j<M;j++){
	    A[i][j] = 0;
	}
	A[i][i] = 0;
    }
#endif    
    for(i=0;i<N;i++){
	for(j=0;j<K;j++){
	    B[i][j] = drand48();
	}
    }
    if (C2 != NULL){
	for(i=0;i<N;i++){
	    for(j=0;j<M;j++){
		C[i][j] = drand48();
		C2[i][j] = C[i][j];
	    }
	}
    }else{
	for(i=0;i<N;i++){
	    for(j=0;j<M;j++){
		C[i][j] = drand48();
	    }
	}
    }      
}


void print_matrix(int M, int N, double A[N][M])
{
  int i, j;

  for(i=0;i<M;i++){
    for(j=0;j<N;j++){
      fprintf(stderr,"%e ", A[j][i]);
    }
    fprintf(stderr,"\n");
  }
}

double check_matrix(int M, int N, int LDA, int LDB, double A[][LDA], double B[][LDB])
{
  int i, j, imax=-1, jmax=-1;
  double err=0.0, da, damax=0.0e0;

  for(j=0;j<N;j++){
      for(i=0;i<M;i++){
	  da = fabs(A[j][i] - B[j][i]);
	  if(da > damax){
	      imax = i;
	      jmax = j;
	      damax = da;
	  }
	  err += da;
	  if(da > 1e-3){
	      //	printf("(%d, %d) : err=%e A=%e B=%e\n", i, j, da, A[j][i], B[i][j]);
	  }
	  if(isnan(A[j][i])){
	      printf("Nan detected mat A !\n");
	      exit(1);
	  }
	  if(isnan(B[j][i])){
	      printf("Nan detected mat B!\n");
	      exit(1);
	  }
      }
      //printf("\n");
  }
  err = err / (double)M / (double)N;
  fprintf(stderr,"error : %e errmax : %e  on i=%d , j=%d \n", err, damax, imax, jmax);
  return err;
}
double check_matrix_fast(int M, int N, int LDA, int LDB,
			 double A[][LDA], double B[][LDB],
			 int checkmode)
{
    int i, j, imax=-1, jmax=-1;
    double e=0.0, da, da2sum=0.0e0;
    
    for(j=0;j<N;j++){
	for(i=0;i<M;i++){
	    da = A[j][i] - B[j][i];
	    if (((checkmode == 2) && (da != 0.0))
		||(fabs(da) > 1e-9)){
		fprintf(stderr,"index: %d %d, data: %22.16e %22.16e %lx %lx\n",
			j,i, A[j][i], B[j][i],
			*((long*)(&(A[j][i]))),
			*((long*)(&(B[j][i]))));
	    }
	    da2sum += da*da;
	}
    }
    e = sqrt(da2sum/N/M);
    fprintf(stderr,"rms err = %e\n",e );
    return e;
}

#ifdef DGEMMTEST

extern int g_nloop;

#ifndef SIMPLETEST
int main(int argc, char **argv)
{
  int M=2048, N=2048, K=2048, LDA=2048, LDB=2048, LDC=2048;
  int NOTA=1, NOTB=1,  gdrdoneflag=0, check=1, i, seed;
  long long int asum, bsum, cnormal, ctest;
  double *A, *B, *C, *C2, alpha, beta;
  char ta = ((NOTA)==1) ? 'N' : 'T';
  char tb = ((NOTB)==1) ? 'N' : 'T';
  
  //  dqsdelay=0x8;
  //srand48((unsigned long)time(NULL));
  if(argc >= 5){
    M = atoi(argv[1]);
    N = atoi(argv[2]);
    K = atoi(argv[3]);
    check = atoi(argv[4]);
  }
  if(argc >=6){
    dooverlap = atoi(argv[5]);
  }
#if 0
  if(argc >=7){
    g_nloop = atoi(argv[6]);
    printf("NLOOP : %d \n", g_nloop);
  }
  if(argc >=8){
    seed = atoi(argv[7]);
    printf("SEED : %d \n", seed);
    srand48(seed);
  }
#endif
  alpha = 1.0;
  beta  = 1.0; 

  if(argc >= 7){
    alpha = atof(argv[6]);
  }
  if(argc >= 8){
    beta  = atof(argv[7]);
  }
  A = malloc(sizeof(double)*M*K);
  B = malloc(sizeof(double)*N*K);
  C = malloc(sizeof(double)*M*N);
  if(check==1 || check==2 || check==4){
    C2 = malloc(sizeof(double)*M*N);
  }
  if(A==NULL || B==NULL || C==NULL){
    printf("Alloc error for matrix data \n");
    return 1;
  }

  LDA = M; //Fortran order
  LDB = K;
  LDC = M;

  double (*pa)[];
  pa = (double(*)[])(&A);
  double (*pb)[];
  pb = (double(*)[])(&B);
  double (*pc)[];
  pc = (double(*)[])(&C);
  double (*pc2)[];
  pc2 = (double(*)[])(&C2);

  if(check==1 || check==2 || check==4){
      //    set_matrix_2(M, N, K, (double(*)[])A, (double(*)[])B,
      //		 (double(*)[])C, (double(*)[])C2);
      set_matrix_2(M, N, K,pa, pb, pc, pc2);
  } else {
    set_matrix(M, N, K, pa, pb, pc);
  }
  asum = calc_checksum(M, K, LDA, pa);
  bsum = calc_checksum(K, N, LDB, pb);
  printf("A sum : %016llx , B sum : %016llx \n");

  printf("A matrix => \n");
  //print_matrix(M, K, pa);
  printf("B matrix => \n");
  //print_matrix(K, N, pb);

  printf("GDR : \n");
  //  ch=4;
  //  dqsdelay=0x08;
  gdr_dgemm_(&NOTA, &NOTB, &M, &N, &K, &alpha, &beta, &LDA, &LDB, &LDC,
	     pa, pb, pc, &gdrdoneflag);
  cnormal =  calc_checksum(M, N, LDC, pc);
  switch(check){
  case 2:
    printf("GDR 2nd : \n");
    asum = calc_checksum(M, K, LDA, pa);
    bsum = calc_checksum(K, N, LDB, pb);
    printf("A sum : %016llx , B sum : %016llx \n");
    gdr_dgemm_(&NOTA, &NOTB, &M, &N, &K, &alpha, &beta, &LDA, &LDB, &LDC,
	       pa, pb, pc2, &gdrdoneflag);
    check_matrix(M, N, M, M, pc, pc2);
    for(i=0;i<3;i++){
      asum = calc_checksum(M, K, LDA, pa);
      bsum = calc_checksum(K, N, LDB, pb);
      printf("A sum : %016llx , B sum : %016llx \n");
      printf("GDR : \n");
      gdr_dgemm_(&NOTA, &NOTB, &M, &N, &K, &alpha, &beta, &LDA, &LDB, &LDC,
		 pa, pb, pc, &gdrdoneflag);
      asum = calc_checksum(M, K, LDA, pa);
      bsum = calc_checksum(K, N, LDB, pb);
      printf("A sum : %016llx , B sum : %016llx \n");
      check_matrix(M, N, M, M, pc, pc2);
      printf("GDR 2nd : \n");
      gdr_dgemm_(&NOTA, &NOTB, &M, &N, &K, &alpha, &beta, &LDA, &LDB, &LDC,
		 pa, pb, pc2, &gdrdoneflag);
      check_matrix(M, N, M, M, pc, pc2);
    } 
    break;

  case 1: //check result with  host dgemm
    printf("Host : \n");
    dgemm_(&ta, &tb, &M, &N, &K, &alpha, pa, &LDA, pb, &LDB, &beta, pc2, &LDC);
    check_matrix(M, N, LDC, LDC, pc, pc2);
    
    printf("GDR result : \n");
    print_matrix2(10, 10, LDC, pc);
    //print_matrix2(10, 10, M, pc);
      
    printf("\n\nHost result : \n");
    print_matrix2(10, 10, LDC, pc2);
    
    free(C2);
    break;

  case 3: //twice run only
    for(i=0;i<3;i++){
      printf("GDR : \n");
      gdr_dgemm_(&NOTA, &NOTB, &M, &N, &K, &alpha, &beta, &LDA, &LDB, &LDC,
		 pa, pb, pc, &gdrdoneflag);
      printf("GDR 2nd : \n");
      gdr_dgemm_(&NOTA, &NOTB, &M, &N, &K, &alpha, &beta, &LDA, &LDB, &LDC,
		 pa, pb, pc, &gdrdoneflag);
    }
    break;

  case 4:
    {
      int d, i, j;
      printf("input ch : ");
      scanf("%d", &ch);
      for(d=0;d<64;d++){
	//
	printf("GDR : delay=%d\n", d);
	for(i=0;i<M*N;i++){
	  C[i] = C2[i];
	}
	dqsdelay=d;
	gdr_dgemm_(&NOTA, &NOTB, &M, &N, &K, &alpha, &beta, &LDA, &LDB, &LDC,
		   pa, pb, pc, &gdrdoneflag);
	ctest = calc_checksum(M, N, LDC, pc);
	if(ctest == cnormal) {
	  printf("OK : ch=%d delay=%d cnormal=%016llx ctest=%016llx \n", ch, dqsdelay, cnormal, ctest); 
	} else {
	  printf("NG : ch=%d delay=%d cnormal=%016llx ctest=%016llx \n", ch, dqsdelay, cnormal, ctest); 
	}
      }
    }
  }

  free(A);
  free(B);
  free(C);
  printf("gdrdone : %d \n", gdrdoneflag);
#ifdef USEGDR
  grape_close();
#endif
  return 0;
}
#endif

 #ifdef SIMPLETEST
int main(int argc, char **argv)
{
  int M=2048, N=2048, K=2048, LDA=2048, LDB=2048, LDC=2048;
  int NOTA=1, NOTB=1,  gdrdoneflag=0, check=1, i, seed;
  long long int asum, bsum, cnormal, ctest;
  double *A, *B, *C, *C2, alpha, beta;
  char ta = ((NOTA)==1) ? 'N' : 'T';
  char tb = ((NOTB)==1) ? 'N' : 'T';
  
  //  dqsdelay=0x8;
  //srand48((unsigned long)time(NULL));
  if(argc >= 5){
    M = atoi(argv[1]);
    N = atoi(argv[2]);
    K = atoi(argv[3]);
    check = atoi(argv[4]);
  }

  alpha = 1.0;
  beta  = 1.0; 

  A = malloc(sizeof(double)*M*K);
  B = malloc(sizeof(double)*N*K);
  C = malloc(sizeof(double)*M*N);
  if(check==1 || check==2 || check==4){
    C2 = malloc(sizeof(double)*M*N);
  }
  if(A==NULL || B==NULL || C==NULL){
    printf("Alloc error for matrix data \n");
    return 1;
  }

  LDA = M; //Fortran order
  LDB = K;
  LDC = M;


  set_matrix_2(M, N, K, A, B, C, C2);

  asum = calc_checksum(M, K, LDA, A);
  bsum = calc_checksum(K, N, LDB, B);
  printf("A sum : %016llx , B sum : %016llx \n");

  printf("A matrix => \n");
  //print_matrix(M, K, A);
  printf("B matrix => \n");

  printf("GDR : \n");

  gdr_dgemm_(&NOTA, &NOTB, &M, &N, &K, &alpha, &beta, &LDA, &LDB, &LDC, A, B, C, &gdrdoneflag);
  cnormal =  calc_checksum(M, N, LDC, C);
  printf("Host : \n");
  dgemm_(&ta, &tb, &M, &N, &K, &alpha, A, &LDA, B, &LDB, &beta, C2, &LDC);
  check_matrix(M, N, LDC, LDC, C, C2);
  
  printf("GDR result : \n");
  print_matrix2(10, 10, LDC, C);
  //print_matrix2(10, 10, M, C);
  
  printf("\n\nHost result : \n");
  print_matrix2(10, 10, LDC, C2);
  
  grape_close();
  return 0;
}
#endif

 
#endif
