// multiboard-control.c
//
// J. Makino
//    Time-stamp: <10/05/22 11:39:35 makino>
//
/* provide three functions

   int gdr_multi_cntl_init(int myboardid)
   int gdr_multi_cntl_setstate(int state)
   int gdr_multi_cntl_getstate()

meaning of state:

bit0:  0: closed (if set correctly...)
       1: open
bit1:  0: not in Cmat dma
       1: in Cmat DMA
bit2:  0: not in dangerous state
       1: in dangerous state
       
*/       
   
#include <stdio.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
# include <math.h>
# include <float.h>
# include <limits.h>
# include <sys/time.h>


#include <emmintrin.h>
typedef double v2df __attribute__((vector_size(16)));
typedef union {v2df v; double s[2];}v2u;

#include "./dgemmbody.h"
#include "./multiboard-control.h"

static int multi_cntl_fd[MULTIBOARD_NBOARDS]={-1,-1,-1,-1};
static int * multi_cntl_mapp[MULTIBOARD_NBOARDS]={NULL,NULL,NULL,NULL};
static int nboards=1;
static int myboardid = -1;

static int shmeminit(int boardid, int writemode)
{
    int fd;
    int pagesize = getpagesize();
    size_t mapsize = pagesize;
    char fpath[255];
    sprintf(fpath,shmemfile_baseneme,boardid);
    //    fprintf(stderr,"shmeminit: pathname=%s\n", fpath);
    int flags = O_RDONLY;
    if (writemode)flags = O_RDWR|O_CREAT;
    if((fd = open(fpath,flags,0666))== -1){
	if (writemode){
	    perror("");
	    return -1;
	}
    }
    if (writemode){
	if (ftruncate(fd, mapsize) == -1) {
	    perror("ftruncate");
	    exit(-1);
	}
    }
    multi_cntl_fd[boardid]=fd;
    //    fprintf(stderr,"shmeminit, boardid, fd = %d, %d\n", boardid, fd);
    if (fd == -1) return 0;
    int prot = PROT_READ;
    if (writemode)prot = PROT_READ|PROT_WRITE;
    flags=MAP_SHARED;
    multi_cntl_mapp[boardid]=(int*)mmap(0,mapsize,prot,flags,fd,0);
    if (multi_cntl_mapp[boardid]==NULL){
	perror("");
	exit(-1);
    }
    if (writemode)multi_cntl_mapp[boardid][0]=1;
    return 0;
}

#if 0
int gdr_multi_cntl_init(int boardid)
{
    int i;
    myboardid=boardid;
    for(i=0;i<MULTIBOARD_NBOARDS; i++){
	multi_cntl_fd[i]= -1;
	multi_cntl_mapp[i]=NULL;
	shmeminit(i, (i==myboardid));
	//fprintf(stderr,"after shmeminit, boardid, map = %d, %d\n",
	//i, multi_cntl_mapp[i]);
    }
}

int gdr_multi_cntl_setstate(int state)
{
    if (multi_cntl_mapp[myboardid]!=NULL){
	multi_cntl_mapp[myboardid][0]=state;
	return 0;
    }else{
	return -1;
    }
}

int gdr_multi_cntl_getstate()
{
    int i;
    int retval=0;
    for(i=0;i<MULTIBOARD_NBOARDS; i++){
	if (multi_cntl_mapp[i]!=NULL){
	    if (i!=myboardid){
		retval |= multi_cntl_mapp[i][0];
	    }
	}
    }
    return retval;
}
#else
int gdr_multi_cntl_init(int boardid)
{
    return 0;
}

int gdr_multi_cntl_setstate(int state)
{
    return 0;
}

int gdr_multi_cntl_getstate()
{
    return 0;
}
#endif


