#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <getopt.h>
#include "./dgemmconf.h"

#define NCHIP 4
#define NPE   32
#define NBB   16

#define DIM0 (32*NBB*NCHIP) //=512*NCHIP
#define DIM1 (8*NPE)        //=25
#define USEGDR



typedef struct parmstruct{
    int m;
    int k;
    int n;
    int c;
    int l;
    int s;
    int usehugepage;
} PARMS, *PPARMS;


void usage()
{
    fprintf(stderr,"testdgemm options:\n");
    fprintf(stderr,"  -h: This help\n");
    fprintf(stderr,"  -m: # of cols for matrix A (default=2048)\n");
    fprintf(stderr,"  -k: # of rows for matrix A/cols for B (default=2048)\n");
    fprintf(stderr,"  -n: # of rows for matrix B (default=2048)\n");
    fprintf(stderr,"  -c: type of check (default=1)\n");
    fprintf(stderr,"  -l: number of loops (default=1,unused now)\n");
    fprintf(stderr,"  -s: seed (default=1)\n");
    fprintf(stderr,"  -g: usehugetlbfs (default=no)\n");
}


extern char *optarg;
extern int optind;

void print_parms(FILE* stream, PPARMS parms)
{
    fprintf(stream,"M=%d K=%d N=%d c=%d l=%d s=%d\n",
	    parms->m, parms->k,  parms->n,
	    parms->c,  parms->l,   parms->s);
}

void read_parms(int argc, char * argv[], PPARMS parms)
{
    int ch;
    static struct option longopts[] = {
	{ "help",      no_argument,      0,           'h' },
	{ "M",      optional_argument,      NULL,           'm' },
	{ "K",      optional_argument,      NULL,           'k' },
	{ "N",      optional_argument,      NULL,           'n' },
	{ "seed",      optional_argument,            NULL,           's' },
	{ "nloops",      optional_argument,      NULL,           'l' },
	{ "checktype",      optional_argument,      NULL,           'c' },
	{ "usehugepage",  no_argument,   0,     'g' },
	{ NULL,         0,                      NULL,           0 }
    };
    parms->s=1;
    parms->m=2048;
    parms->n=2048;
    parms->k=2048;
    parms->c=1;
    parms->l=1;
    parms->usehugepage=0;
    while((ch=getopt_long(argc,argv,"c:ghk:l:m:n:s:",longopts, NULL))!= -1){
	fprintf(stderr,"optchar = %c optarg=%s\n", ch,optarg);
	switch (ch) {
	    case 'c': parms->c = atoi(optarg); break;
	    case 'g': parms->usehugepage = 1; break;
	    case 'l': parms->l = atoi(optarg); break;
	    case 'k': parms->k = atoi(optarg); break;
	    case 'm': parms->m = atoi(optarg); break;
	    case 'n': parms->n = atoi(optarg); break;
	    case 's': parms->s = atoi(optarg); break;
	    case 'h': usage(); exit(1);
	    case '?':usage(); exit(1);
		break;
	    default:break;
	}
    }
    argc -= optind;
    argv += optind;
    print_parms(stderr, parms);
    print_parms(stdout, parms);
}

void* mymalloc(long size, int usehugepage)
{
    static int hugetlbinitialize=0;
    static off_t hugepageoffset = (off_t)0;
    static int fd;
    void * p;
    if (usehugepage){
	if (hugetlbinitialize==0){
	    fd = open("/mnt/huge/aaa", O_RDWR|O_CREAT, 0777);
	}
	size_t mapsize = (size+0x400000)&0xffffffffffc00000L;
	p = mmap(0, size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, hugepageoffset);
	hugepageoffset += (off_t)mapsize;
    }else{
	p=malloc(size);
    }
    return p;
}


int main(int argc, char **argv)
{
  int M=8192, N=2048, K=2048, LDA=2048, LDB=2048, LDC=2048;
  int NOTA=1, NOTB=1,  gdrdoneflag=0, check=1, i, seed;
  long long int asum, bsum, cnormal, ctest;
  int usehugepage=0;
  double *A, *B, *C, *C2, alpha, beta;
  char ta = ((NOTA)==1) ? 'N' : 'T';
  char tb = ((NOTB)==1) ? 'N' : 'T';
  
  //  dqsdelay=0x8;
  //srand48((unsigned long)time(NULL));
  PARMS parms;
  read_parms(argc, argv, &parms);
  M=parms.m;
  N=parms.n;
  K=parms.k;
  seed=parms.s;
  check=parms.c;
  //  g_nloop=parms.l;
  usehugepage= parms.usehugepage;

  alpha = 1.0;
  beta  = 1.0; 

  A = malloc(sizeof(double)*M*K);
  B = malloc(sizeof(double)*N*K);
  C = malloc(sizeof(double)*M*N);
  if(check==1 || check==2 || check==4){
    C2 = malloc(sizeof(double)*M*N);
  }
  if(A==NULL || B==NULL || C==NULL){
    printf("Alloc error for matrix data \n");
    return 1;
  }

  LDA = M; //Fortran order
  LDB = K;
  LDC = M;


  set_matrix_2(M, N, K, A, B, C, C2);

  asum = calc_checksum(M, K, LDA, A);
  bsum = calc_checksum(K, N, LDB, B);
  printf("A sum : %016llx , B sum : %016llx \n");

  printf("A matrix => \n");
  //print_matrix(M, K, A);
  printf("B matrix => \n");

  printf("GDR : \n");

  gdr_dgemm_(&NOTA, &NOTB, &M, &N, &K, &alpha, &beta, &LDA, &LDB, &LDC, A, B, C, &gdrdoneflag);
  cnormal =  calc_checksum(M, N, LDC, C);
  printf("Host : \n");
  dgemm_(&ta, &tb, &M, &N, &K, &alpha, A, &LDA, B, &LDB, &beta, C2, &LDC);
  check_matrix(M, N, LDC, LDC, C, C2);
  
  printf("GDR result : \n");
  print_matrix2(10, 10, LDC, C);
  //print_matrix2(10, 10, M, C);
  
  printf("\n\nHost result : \n");
  print_matrix2(10, 10, LDC, C2);
  
  grape_close();
  return 0;
}
