#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <getopt.h>

#include "dgemmbody.h"

#define NCHIP 4
#define NPE   32
#define NBB   16

#define DIM0 (32*NBB*NCHIP) //=512*NCHIP
#define DIM1 (8*NPE)        //=25
#define USEGDR
//#define USEOVLDGEMM
#define CALCCSUM
//#define CHECKGDR
//#define CALCABCSUM
#define MINK 512
#define USEAUTOTRANS

void setdeqsdelay(int x);
void setdoverlap(int x);

extern int g_nloop;


typedef struct parmstruct{
    int m;
    int k;
    int n;
    int boardid;
    int c;
    int l;
    int unit_bmat;
    int s;
    int lda;
    int ldb;
    int ldc;
    int nota;
    int notb;
    int coc_mask;
    int usehugepage;
    int forceswapab;
    int nboards;
    double beta;
    double alpha;
} PARMS, *PPARMS;


void usage()
{
    fprintf(stderr,"testdgemm options:\n");
    fprintf(stderr,"  -h: This help\n");
    fprintf(stderr,"  -m: # of cols for matrix A (default=2048)\n");
    fprintf(stderr,"  -k: # of rows for matrix A/cols for B (default=2048)\n");
    fprintf(stderr,"  -n: # of rows for matrix B (default=2048)\n");
    fprintf(stderr,"  -b: boardid (default=0)\n");
    fprintf(stderr,"  -N: number of boards to use(default=1)\n");
    fprintf(stderr,"  -e: DGEMM beta (default=1)\n");
    fprintf(stderr,"  -f: DGEMM alpha (default=1)\n");
    fprintf(stderr,"  -c: type of check (default=1)\n");
    fprintf(stderr,"      1: normal compare with host DGEMM\n");
    fprintf(stderr,"      2: compare with first try\n");
    fprintf(stderr,"      4: no compare (speed-only test)\n");
    fprintf(stderr,"  -l: number of loops (default=1)\n");
    fprintf(stderr,"  -s: seed (default=1)\n");
    fprintf(stderr,"  -i: set B to unit matrix\n");
    fprintf(stderr,"  -g: usehugetlbfs (default=no)\n");
    fprintf(stderr,"  -A: transpose A  (default=no trans)\n");
    fprintf(stderr,"  -B: transpose B  (default=no trans)\n");
    fprintf(stderr,"  -C: value to set to COC AUX reg  (default=15)\n");
    fprintf(stderr,"  -F: forceswapab flag(default=no)\n");
   fprintf(stderr,"  -x: # LDA (default=2048)\n");
   fprintf(stderr,"  -y: # LDB (default=2048)\n");
   fprintf(stderr,"  -z: # LDC (default=2048)\n");
 
}


extern char *optarg;
extern int optind;

void print_parms(FILE* stream, PPARMS parms)
{
    fprintf(stream,"M=%d K=%d N=%d c=%d l=%d s=%d boardid=%d nboards=%d ",
	    parms->m, parms->k,  parms->n,
	    parms->c,  parms->l,   parms->s, parms->boardid,
	    parms->nboards);
    fprintf(stream," unit_b=%d alpha=%e beta=%e cmask=%x\n",
	    parms->unit_bmat, parms->alpha, parms->beta, parms->coc_mask);
    fprintf(stream,"NOTA=%d NOTB=%d fswapab=%d\n",parms->nota, parms->notb,
	    parms->forceswapab);
}

void read_parms(int argc, char * argv[], PPARMS parms)
{
    int ch;
    static struct option longopts[] = {
	{ "help",      no_argument,      0,           'h' },
	{ "M",      optional_argument,      NULL,           'm' },
	{ "K",      optional_argument,      NULL,           'k' },
	{ "N",      optional_argument,      NULL,           'n' },
	{ "LDA",      optional_argument,      NULL,           'x' },
	{ "LDB",      optional_argument,      NULL,           'y' },
	{ "LDC",      optional_argument,      NULL,           'z' },
	{ "seed",      optional_argument,            NULL,           's' },
	{ "nloops",      optional_argument,      NULL,           'l' },
	{ "checktype",      optional_argument,      NULL,           'c' },
	{ "boardid",      optional_argument,      NULL,           'b' },
	{ "nboard",      optional_argument,      NULL,           'N' },
	{ "beta",      optional_argument,      NULL,           'e' },
	{ "alpha",      optional_argument,      NULL,           'f' },
	{ "NOTA", no_argument,   0,     'A' },
	{ "NOTB", no_argument,   0,     'B' },
	{ "unit_bmatrix", no_argument,   0,     'i' },
	{ "usehugepage",  no_argument,   0,     'g' },
	{ "forceswapab",  no_argument,   0,     'f' },
	{ "coc_mask",      optional_argument,      NULL,           'C' },
	{ NULL,         0,                      NULL,           0 }
    };
    parms->s=1;
    parms->m=2048;
    parms->n=2048;
    parms->k=2048;
    parms->lda=2048;
    parms->ldb=2048;
    parms->ldc=2048;
    parms->boardid=0;
    parms->unit_bmat=0;
    parms->c=1;
    parms->l=1;
    parms->usehugepage=0;
    parms->beta=1.0;
    parms->alpha=1.0;
    parms->coc_mask = 15;
    parms->nota=1;
    parms->notb=1;
    parms->nboards=1;
    parms->forceswapab=0;
    while((ch=getopt_long(argc,argv,"ABC:FN:b:c:e:f:ghik:l:m:n:s:x:y:z:",longopts, NULL))!= -1){
	fprintf(stderr,"optchar = %c optarg=%s\n", ch,optarg);
	switch (ch) {
	    case 'A': parms->nota = 0; break;
	    case 'B': parms->notb = 0; break;
	    case 'C': parms->coc_mask = atoi(optarg); break;
	    case 'F': parms->forceswapab=1; break;
	    case 'N': parms->nboards = atoi(optarg); break;
	    case 'b': parms->boardid = atoi(optarg); break;
	    case 'c': parms->c = atoi(optarg); break;
	    case 'e': parms->beta = atof(optarg); break;
	    case 'f': parms->alpha = atof(optarg); break;
	    case 'g': parms->usehugepage = 1; break;
	    case 'i': parms->unit_bmat = 1; break;
	    case 'l': parms->l = atoi(optarg); break;
	    case 'k': parms->k = atoi(optarg); break;
	    case 'm': parms->m = atoi(optarg); break;
	    case 'n': parms->n = atoi(optarg); break;
	    case 's': parms->s = atoi(optarg); break;
	    case 'x': parms->lda = atoi(optarg); break;
	    case 'y': parms->ldb = atoi(optarg); break;
	    case 'z': parms->ldc = atoi(optarg); break;
	    case 'h': usage(); exit(1);
	    case '?':usage(); exit(1);
		break;
	    default:break;
	}
    }
    argc -= optind;
    argv += optind;
    print_parms(stderr, parms);
    print_parms(stdout, parms);
}

void* mymalloc(long size, int usehugepage, int boardid)
{
    static int hugetlbinitialize=0;
    static off_t hugepageoffset = (off_t)0;
    static int fd;
    void * p;
    if (usehugepage){
	if (hugetlbinitialize==0){
	    char fname[256];
	    sprintf(fname,"/mnt/huge/aaa-%d",boardid);
	    fd = open(fname, O_RDWR|O_CREAT, 0777);
	}
	size_t mapsize = (size+0x400000)&0xffffffffffc00000L;
	p = mmap(0, size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, hugepageoffset);
	hugepageoffset += (off_t)mapsize;
    }else{
	p=malloc(size);
    }
    return p;
}

    
void copymat(int M, int N, int nc, double C[][nc], double C2[][nc])
{
    int i, j;
    
#pragma omp parallel for private(i,j)
    for(i=0;i<N;i++){
	for(j=0;j<M;j++){
	    C2[i][j] = C[i][j];
	}
    }
}

void set_unit_matrix(int N, int K, int ldb, double B[][ldb])
{
    int i, j;
    fprintf(stderr,"set_unit mat called with %d %d\n", N, K);
    for(i=0;i<N;i++){
	for(j=0;j<K;j++){
	    B[i][j] = 0;
	}
    }
    int imax = N;
    if (imax < K) imax = K;
    int imin =0;
    imax = K;
    for(i=imin;i<imax;i++)B[i%N][i%K]=1;
}



int main(int argc, char **argv)
{
    int M=8192, N=2048, K=2048, LDA=2048, LDB=2048, LDC=2048;
    int NOTA=1, NOTB=1,  gdrdoneflag=0, check=1, i, seed;
    long long int asum, bsum, cnormal, ctest;
    int usehugepage;
    int bid;
    int nloop;
    double (*A)[], (*B)[], (*C)[], (*C2)[],(*C3)[], alpha, beta;
    PARMS parms;
    read_parms(argc, argv, &parms);
    M=parms.m;
    N=parms.n;
    K=parms.k;
    LDA=parms.lda;
    LDB=parms.ldb;
    LDC=parms.ldc;
    seed=parms.s;
    check=parms.c;
    nloop=parms.l;
    bid = parms.boardid;
    NOTA=parms.nota;
    NOTB=parms.notb;
    char ta = ((NOTA)==1) ? 'N' : 'T';
    char tb = ((NOTB)==1) ? 'N' : 'T';
    gdrsetboardid(parms.boardid);
    gdrsetnboards(parms.nboards);
    if (parms.forceswapab)gdrsetforceswapab();
    if (LDA < M) LDA = M;
    if (LDC < M) LDC = M;
    if (LDB < K) LDB = K;
    
    usehugepage= parms.usehugepage;
    //    setdqsdelayandch(0x8,4);

    //    LDC=LDA;
    A = mymalloc(sizeof(double)*((long)LDA)*K,usehugepage,bid);
    B = mymalloc(sizeof(double)*((long)LDB)*N,usehugepage,bid);
    C = mymalloc(sizeof(double)*((long)N)*LDC,usehugepage,bid);

    if(A==NULL || B==NULL || C==NULL){
	printf("Alloc error for matrix data \n");
	return 1;
    }
#if 0
    LDA = M; //Fortran order
    LDB = K;
    LDC = M;
#endif    
    alpha = parms.alpha;
    beta  = parms.beta;
    srand48(seed);
    if(check != 4){
	C2 = mymalloc(sizeof(double)*((long)N)*LDC,usehugepage,bid);
    }else{
	C2 = C;
    }
    set_matrix_2a(M, N, K, LDA, LDB, LDC, A, B, C, C2);
    if (nloop>1){
	if(check != 4){
	    C3 = mymalloc(sizeof(double)*N*LDC,usehugepage,bid);
	    copymat(M,N,LDC,C2,C3);
	}else{
	    C3 = C;
	}
    }

    if (parms.unit_bmat)set_unit_matrix(N,K,LDB,B);
	
    
    asum = calc_checksum(M, K, LDA, A);
    bsum = calc_checksum(K, N, LDB, B);
    printf("A sum : %016llx , B sum : %016llx \n");

    printf("A matrix => \n");
    //print_matrix(M, K, A);
    printf("B matrix => \n");
    //print_matrix(K, N, B);

    gdr_init();
    //    hib_set_dmacheck_mode((N>2048)? 1:0);
    hib_set_dmacheck_mode(0);
    set_coc_reg(2,(UINT64)( parms.coc_mask|0xff00));
    int NDGEMM = N;
    int NDGEMM2 = N;
    int KDGEMM = K;
    int KDGEMM2 = K;
    switch(check){
	case 2:
	    printf("GDR : \n");
	    asum = calc_checksum(M, K, LDA, A);
	    bsum = calc_checksum(K, N, LDB, B);
	    printf("A sum : %016llx , B sum : %016llx \n");
	    gdr_dgemm_(&NOTA, &NOTB, &M, &N, &K, &alpha, &beta, &LDA, &LDB, &LDC, A, B, C, &gdrdoneflag);
	    break;

	case 1: //check result with  host dgemm
	    printf("Host : \n");
#ifndef NOBLAS	    
	    dgemm_(&ta, &tb, &M, &NDGEMM, &KDGEMM, &alpha, A, &LDA, B, &LDB,
		   &beta, C, &LDC);
#endif	    
	    break;

    }
    
    int kk;
    double emax = 0.0;
    int nerrors = 0;
    for (kk=0;kk<nloop; kk++){
	printf("GDR test loop %d\n",kk);
	//	setdqsdelayandch(0x08,4);
	srand48(seed);
	if (nloop > 1){
	    copymat(M,N,LDC,C3,C2);
	    //	set_matrix_2a(M, N, K, LDA, LDB, LDC, A, B, C2,NULL);
	}
	fprintf(stderr,"end mat reinitial\n");
	gdr_dgemm_(&NOTA, &NOTB, &M, &NDGEMM2, &KDGEMM2, &alpha, &beta,
		   &LDA, &LDB, &LDC, A, B, C2, &gdrdoneflag);
	
	//	cnormal =  calc_checksum(M, N, LDC, C);
	int cmode = check;
	if (parms.unit_bmat) cmode = 2;
	if (check && (check != 4)){
	    double err = check_matrix_fast(M, NDGEMM2, LDC, LDC, C, C2, cmode);
	    if (emax < err) emax = err;
	    if (err > 0.0) nerrors ++;
	    printf(" err = %e %e\n", err, emax);
	}
    }

    printf(" maxerr = %e nerrors=%d\n", emax, nerrors);
#ifdef USEGDR
    set_coc_reg(2,(UINT64) parms.coc_mask);
    grape_close();
#endif
    return 0;

}
