//
// testmatmul32.c
// test driver for matmul32.vsm-generated interface library
//
// J. Makino
//    Time-stamp: <11/01/20 17:39:16 makino>

#define DONOTUSEPTHREADS

// patch for strange bug in two-board operation
// JM 2009/12/30
// currently turned off -- AS OF 2010/4/26
//#define SET_BB0_MANUALLY

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <unistd.h>
#ifndef DONOTUSEPTHREADS
#include <pthread.h>
#endif
#include <xmmintrin.h>
#include <sched.h>
#include <sys/types.h>

#include <emmintrin.h>
typedef double v2df __attribute__((vector_size(16)));
typedef union {v2df v; double s[2];}v2u;

#include "matmul32-32.h"
#include "dgemmconf.h"
#include <sing_chip.h>

#ifndef NPE
#define NPE 32
#endif
#ifndef NBB
#define NBB 16
#endif

#define DIM0 (32*NBB*NCHIP)
#define DIM1 (8*NPE)

void gdrinit_threads(void);

SINGinfo *s[NCHIP];

static char myhostname[257];



static inline UINT64 rdtsc2()
{
	UINT64 ret;
	asm __volatile__("rdtsc  \n\t" \
			 "shlq $32, %%rdx \n\t" \
			 "orq  %%rdx, %%rax \n\t" \
			 "movq %%rax, %0" : "=g"(ret) : :"rax", "rdx");
	return ret;
}

void setdqsdelay2(int ch, int delay)
{
  UINT64 dqs , delayctl[5];
  int i;

  printf("setting dqs%d skew=%d \n", ch, delay);
  //printf("setting all skew=0x18 \n");
  for(i=0;i<NCHIP;i++){
#if 0
    delayctl[0] = 0x10;  //delay; //0x10;//0x10; //0x10;    /* 6bit */  
    delayctl[1] = 0x10;      //0x10; //0x10; //0x10;    /* 6bit */
    delayctl[2] = 0x10;  //0x10; //0x10;    /* 6bit */
    delayctl[3] = 0x12; //0x12; //delay; //0x12; //0x5; //0x5   /* 6bit */
    delayctl[4] = 0x08; // //0x08; //0x4; //0x00;    /* 6bit */
#endif
    delayctl[0] = 0x18; //0x18;  //delay; //0x10;//0x10; //0x10;    /* 6bit */  
    delayctl[1] = 0x18; //0x18;      //0x10; //0x10; //0x10;    /* 6bit */
    delayctl[2] = 0x18; //0x18;  //0x10; //0x10;    /* 6bit */
    delayctl[3] = 0x18; //0x18; //0x12; //delay; //0x12; //0x5; //0x5   /* 6bit */
    delayctl[4] = 0x18; //0x18; // //0x08; //0x4; //0x00;    /* 6bit */
    delayctl[ch] = delay;
    dqs = (delayctl[0] << 24) |  (delayctl[1] << 18) |(delayctl[2] << 12) | (delayctl[3] << 6) | (delayctl[4] );
    CTRL_config_write(s[i]->c, 9, &dqs, 1);
  }
}

static void xwait(int wait)
{
    UINT64 st = rdtsc2();
    while((rdtsc2()-st)<wait) {}
}


double getclock()
{
    FILE * fd;
    fd = fopen("/proc/cpuinfo", "r");
    double x=-1;
    char s[100];
    int success = 0;
    while (!success){
	fgets(s,99,fd);
	//	fprintf(stderr, "string=%s\n",s);
	if(strncmp(s,"cpu MHz",7)==0){
	    sscanf(s+10,"%lf",&x);
	    //	    fprintf(stderr, "string=%s clock=%e\n",s+10,x);
	    fclose(fd);
    	    return x*1e6;
	}
    }
    return x;
}
	    
double wallclock()
{
    static double clock=0;
    if (clock==0)clock=getclock();
    return ((double)rdtsc2()/clock);
}

static void readmat(double a[DIM1][DIM0])
{
    int i,j;
    for(i=0;i<DIM1;i++){
	for(j=0;j<DIM0/NCHIP;j++){
	    int k;
	    double x;
	    k=scanf("%le",&x);
	    a[i][j]=x;
	    a[i][j+(DIM0/NCHIP)*1]=x;
	    a[i][j+(DIM0/NCHIP)*2]=x;
	    a[i][j+(DIM0/NCHIP)*3]=x;
	    if (k != 1){
		fprintf(stderr,"readmat scanf returns %d %d %d\n", k,i,j);
	    }
	}
    }
}

static void printmat(double a[DIM1][DIM0])
{
    int i,j;
    for(i=0;i<DIM1;i++){
	printf("%2d:",i);
	for(j=0;j<DIM0;j++){
	    printf(" %1d",(int)(a[i][j]));
	}
	printf("\n");
    }
}
    
static void printcmat(double a[DIM1][DIM1])
{
    int i,j;
    for(i=0;i<DIM1;i++){
	printf("%2d:",i);
	for(j=0;j<DIM1;j++){
	    printf(" %1d",(int)(a[i][j]));
	}
	printf("\n");
    }
}

static void printcmatdiff(double a[DIM1][DIM1],double b[DIM1][DIM1] )
{
    int i,j;
    double emax, e2sum,esum, remax, re2sum,resum;
    int ndata=0;
    emax= e2sum= remax= re2sum=esum=resum=0;
    printf("ERRORS:\n");
    for(i=0;i<DIM1;i++){
	printf("%2d:",i);
	for(j=0;j<DIM1;j++){
	    printf(" %.5e",a[i][j]-b[i][j]);
	    if (b[i][j] != 0.0){
		double e, re;
		ndata++;
		e = fabs(a[i][j]-b[i][j]);
		re =e/b[i][j];
		if (e > emax) emax=e;
		esum+= e;
		e2sum+= e*e;
		if (re > remax) remax=re;
		resum+= re;
		re2sum+= re*re;
	    }
	}
	printf("\n");
    }
    printf("Abs errors = %16.7e %16.7e %16.7e\n",
	   emax, esum/ndata, sqrt(e2sum/ndata));
    printf("Rel errors = %16.7e %16.7e %16.7e\n",
	   remax, resum/ndata, sqrt(re2sum/ndata));
}

static void printcmatdiff_mn(int DIMM, int DIMN, double a[DIMM][DIMN],double b[DIMM][DIMN] )
{
    int i,j;
    double emax, e2sum,esum, remax, re2sum,resum;
    int ndata=0;
    emax= e2sum= remax= re2sum=esum=resum=0;
    printf("ERRORS:\n");
    for(i=0;i<DIMM;i++){
      //printf("%2d:",i);
	for(j=0;j<DIMN;j++){
	  //printf(" %.5e",a[i][j]-b[i][j]);
	    if (b[i][j] != 0.0){
		double e, re;
		ndata++;
		e = fabs(a[i][j]-b[i][j]);
		re =e/b[i][j];
		if (e > emax) emax=e;
		esum+= e;
		e2sum+= e*e;
		if (re > remax) remax=re;
		resum+= re;
		re2sum+= re*re;
	    }
	}
	//printf("\n");
    }
    printf("Abs errors = %16.7e %16.7e %16.7e\n",
	   emax, esum/ndata, sqrt(e2sum/ndata));
    printf("Rel errors = %16.7e %16.7e %16.7e\n",
	   remax, resum/ndata, sqrt(re2sum/ndata));
}
    

// LM matrix address etc
//
//    0       4    ................  124
//    1       5    ................  125
//    2       6    ................  126
//    3       7    ................  127
//    128     132  ................  252
//    129     133  ................  253
//    130     134  ................  254
//    131     135  ................  255
void copytomata(double a[DIM1][DIM0],
	     int ibb,
	     int ipe,
	     double alm[NBB][64][4])
{
    int i,j;
    for (i=0;i<64; i++){
	int k, l;
	k = i % 32;
	l = (i / 32)*4;
	for (j=0;j<4; j++){
	    alm[ibb][i][j] = a[ipe*8+l+j][ibb*32+k];
	}
    }
}

void copytomata_tb4(double a[DIM1][DIM0],
		    int ichip,
		    int ibb,
		    int ipe,
		    double alm[NBB][64][4])
{
    int i,j;
    for (i=0;i<64; i++){
	int k, l;
	k = i % 32;
	l = (i / 32)*4;
	for (j=0;j<4; j++){
	    alm[ibb][i][j] = a[ipe*8+l+j][ichip*512+ibb*32+k];
	}
    }
}


// LM matrix address etc
//
//    0       4    ................  124
//    1       5    ................  125
//    2       6    ................  126
//    3       7    ................  127
//    128     132  ................  252
//    129     133  ................  253
//    130     134  ................  254
//    131     135  ................  255
void copytomata_tb4_dg(int LDA, int NOTA, double a[][LDA], //fortran order a[col][row]
		       int start,
		       int ichip,
		       int ibb,
		       int ipe,
		       int DIMK,
		       double alm[NBB][64][4])
{
  int i,j, x, chipstep, bb;
  
  chipstep = 512*DIMK/2048; //DIMK==2048 => chipstep=512 ,  DIMK==1024 => chipstep=256 , DIMK==512 => chipstep=128
  bb = NBB*DIMK/2048;
  
  if(ibb < bb){
    // data is written to alm
    if(NOTA==1){
      for(i=0;i<32; i++){
	x = ichip*chipstep + ibb*32 + i;
	alm[ibb][i][0]    = a[x][ipe*8+start+0];
	alm[ibb][i][1]    = a[x][ipe*8+start+1];
	alm[ibb][i][2]    = a[x][ipe*8+start+2];
	alm[ibb][i][3]    = a[x][ipe*8+start+3];
	alm[ibb][i+32][0] = a[x][ipe*8+start+4];
	alm[ibb][i+32][1] = a[x][ipe*8+start+5];
	alm[ibb][i+32][2] = a[x][ipe*8+start+6];
	alm[ibb][i+32][3] = a[x][ipe*8+start+7];
      }
    } else {
      for(i=0;i<32; i++){
	x = ichip*chipstep +ibb*32 + i;
	alm[ibb][i][0]    = a[ipe*8+start+0][x];
	alm[ibb][i][1]    = a[ipe*8+start+1][x];
	alm[ibb][i][2]    = a[ipe*8+start+2][x];
	alm[ibb][i][3]    = a[ipe*8+start+3][x];
	alm[ibb][i+32][0] = a[ipe*8+start+4][x];
	alm[ibb][i+32][1] = a[ipe*8+start+5][x];
	alm[ibb][i+32][2] = a[ipe*8+start+6][x];
	alm[ibb][i+32][3] = a[ipe*8+start+7][x];
      }
    }
  } else {
    //padding with 0
#if 0
    for(i=0;i<32; i++){
      alm[ibb][i][0]    = 0.0;
      alm[ibb][i][1]    = 0.0;
      alm[ibb][i][2]    = 0.0;
      alm[ibb][i][3]    = 0.0;
      alm[ibb][i+32][0] = 0.0;
      alm[ibb][i+32][1] = 0.0;
      alm[ibb][i+32][2] = 0.0;
      alm[ibb][i+32][3] = 0.0;
    }
#endif
  }

}

void copytomata_tb4_dg_onebb(int LDA, int NOTA, double a[][LDA], //fortran order a[col][row]
		       int start,
		       int ichip,
		       int ibb,
		       int ipe,
		       int DIMK,
		       double alm[64][4])
{
  int i,j, x, chipstep, bb;
  
  chipstep = 512*DIMK/2048; //DIMK==2048 => chipstep=512 ,  DIMK==1024 => chipstep=256 , DIMK==512 => chipstep=128
  bb = NBB*DIMK/2048;
  
  if(ibb < bb){
    // data is written to alm
    if(NOTA==1){
      for(i=0;i<32; i++){
	x = ichip*chipstep + ibb*32 + i;
	alm[i][0]    = a[x][ipe*8+start+0];
	alm[i][1]    = a[x][ipe*8+start+1];
	alm[i][2]    = a[x][ipe*8+start+2];
	alm[i][3]    = a[x][ipe*8+start+3];
	alm[i+32][0] = a[x][ipe*8+start+4];
	alm[i+32][1] = a[x][ipe*8+start+5];
	alm[i+32][2] = a[x][ipe*8+start+6];
	alm[i+32][3] = a[x][ipe*8+start+7];
      }
    } else {
      for(i=0;i<32; i++){
	x = ichip*chipstep +ibb*32 + i;
	alm[i][0]    = a[ipe*8+start+0][x];
	alm[i][1]    = a[ipe*8+start+1][x];
	alm[i][2]    = a[ipe*8+start+2][x];
	alm[i][3]    = a[ipe*8+start+3][x];
	alm[i+32][0] = a[ipe*8+start+4][x];
	alm[i+32][1] = a[ipe*8+start+5][x];
	alm[i+32][2] = a[ipe*8+start+6][x];
	alm[i+32][3] = a[ipe*8+start+7][x];
      }
    }
  } else {
    //padding with 0
#if 1

    for(i=0;i<32; i++){
      alm[i][0]    = 0.0;
      alm[i][1]    = 0.0;
      alm[i][2]    = 0.0;
      alm[i][3]    = 0.0;
      alm[i+32][0] = 0.0;
      alm[i+32][1] = 0.0;
      alm[i+32][2] = 0.0;
      alm[i+32][3] = 0.0;
    }
#endif
  }

}




void copytomata_em_1024_tb4(double a[DIM1][DIM0],
			    int ichip,
			    double aem[NPE][4][NBB][64])
{
  int i, j, k, l, m, n, o;
  UINT64 start, end;

  start = rdtsc2();
  for(i=0;i<NPE;i++){
    for(j=0;j<2;j++){
      for(k=0;k<2;k++){
	for(l=0;l<NBB;l++){
	  for(m=0;m<4;m++){
	    for(n=0;n<16;n++){
	      aem[i][2*j+k][l][4*n+m] = a[i*8+4*j+m][512*ichip+32*l+16*k+n];
	    }
	  }
	}
      }
    }
  }
  end = rdtsc2();
  //printf("copymata_em elapsed : %lld [clk] \n", end-start);
}

typedef double *aem_type[NPE][4][NBB][64];


void copytomata_em_1024_tb4_dg_partial(int pestart, int peend,
				       int LDA, int NOTA, double a[][LDA],
				       int start, int ichip, int DIMK,
				       double aem[NPE][4][NBB][64])
//fortran order a[col][row]
{
  int i, j, k, x, y, bb, chipstep;
  UINT64 startt, end;

  startt = rdtsc2();
  bb = NBB*DIMK/2048; //DIMK==2048 => 16bb ,  DIMK==1024 => 8bb , DIMK==512 => 4bb
  chipstep = 512*DIMK/2048; //DIMK==2048 => chipstep=512 ,  DIMK==1024 => chipstep=256 , DIMK==512 => chipstep=128
  //  fprintf(stderr,"copymata1024....nota=%d\n",NOTA);
  if(NOTA==1){
      for(j=0;j<bb;j++){
	  for(i=pestart;i<peend;i++){
	      x = chipstep*ichip+j*32;
	      y = i*8 + start;
	      for(k=0;k<16;k++){
		  v2df * dest1 = (v2df*)&aem[i][0][j][k*4+0];
		  v2df * dest2 = (v2df*)&aem[i][2][j][k*4+0];
		  v2df * dest3 = (v2df*)&aem[i][1][j][k*4+0];
		  v2df * dest4 = (v2df*)&aem[i][3][j][k*4+0];
		  v2df * src1 = (v2df*)&a[k+x][0+y];
		  v2df * src2 = (v2df*)&a[k+x+16][0+y];
		  
		  /*
		    __builtin_prefetch(&(a[k+x+1][0+y]), 0, 2);
		    __builtin_prefetch(&(a[k+x+17][0+y]), 0, 2);
		  */
		  dest1[0]=src1[0];
		  dest1[1]=src1[1];
		  dest2[0]=src1[2];
		  dest2[1]=src1[3];
		  dest3[0]=src2[0];
		  dest3[1]=src2[1];
		  dest4[0]=src2[2];
		  dest4[1]=src2[3];
	      }
	  }
      }
  } else {
      for(i=pestart;i<peend;i++){
	  for(j=0;j<bb;j++){
	      for(k=0;k<16;k++){
		  x = 512*ichip+j*32;
		  y = i*8 + start;
		  /*
		    __builtin_prefetch(&(a[0+y][k+x+1]), 0, 0);
		    __builtin_prefetch(&(a[0+y][k+x+17]), 0, 0);
		  */
		  aem[i][0][j][k*4+0] = a[0+y][k+x];
		  aem[i][0][j][k*4+1] = a[1+y][k+x];
		  aem[i][0][j][k*4+2] = a[2+y][k+x];
		  aem[i][0][j][k*4+3] = a[3+y][k+x];
		  
		  aem[i][2][j][k*4+0] = a[4+y][k+x];
		  aem[i][2][j][k*4+1] = a[5+y][k+x];
		  aem[i][2][j][k*4+2] = a[6+y][k+x];
		  aem[i][2][j][k*4+3] = a[7+y][k+x];
		  
		  aem[i][1][j][k*4+0] = a[0+y][k+x+16];
		  aem[i][1][j][k*4+1] = a[1+y][k+x+16];
		  aem[i][1][j][k*4+2] = a[2+y][k+x+16];
		  aem[i][1][j][k*4+3] = a[3+y][k+x+16];
		  
		  aem[i][3][j][k*4+0] = a[4+y][k+x+16];
		  aem[i][3][j][k*4+1] = a[5+y][k+x+16];
		  aem[i][3][j][k*4+2] = a[6+y][k+x+16];
		  aem[i][3][j][k*4+3] = a[7+y][k+x+16];
	      }
	  }
      }
  }
  
  //padding with 0
  for(i=pestart;i<peend;i++){
      for(j=bb;j<NBB;j++){
	  for(k=0;k<16;k++){
	      x = chipstep*ichip+j*32;
	      y = i*8 + start;
	      aem[i][0][j][k*4+0] = 0.0;
	      aem[i][0][j][k*4+1] = 0.0;
	      aem[i][0][j][k*4+2] = 0.0;
	      aem[i][0][j][k*4+3] = 0.0;
	      aem[i][2][j][k*4+0] = 0.0;
	      aem[i][2][j][k*4+1] = 0.0;
	      aem[i][2][j][k*4+2] = 0.0;
	      aem[i][2][j][k*4+3] = 0.0;
	      aem[i][1][j][k*4+0] = 0.0;
	      aem[i][1][j][k*4+1] = 0.0;
	      aem[i][1][j][k*4+2] = 0.0;
	      aem[i][1][j][k*4+3] = 0.0;
	      aem[i][3][j][k*4+0] = 0.0;
	      aem[i][3][j][k*4+1] = 0.0;
	      aem[i][3][j][k*4+2] = 0.0;
	      aem[i][3][j][k*4+3] = 0.0;
	  }
      }
  }
  
  end = rdtsc2();
#ifdef BENCHSENDI
  //printf("copymata_em elapsed : %lld [clk] %e[s] \n", end-startt, (end-startt)/2.6e9);
#endif
}

void copytomata_em_1024_tb4_dg_singlepe_old(int pestart, 
				       int LDA, int NOTA, double a[][LDA],
				       int start, int ichip, int DIMK,
				       double aem[NPE][4][NBB][64])
//fortran order a[col][row]
{
  int i, j, k, x, y, bb, chipstep;
  UINT64 startt, end;
  i = pestart;
  startt = rdtsc2();
  bb = NBB*DIMK/2048; //DIMK==2048 => 16bb ,  DIMK==1024 => 8bb , DIMK==512 => 4bb
  chipstep = 512*DIMK/2048; //DIMK==2048 => chipstep=512 ,  DIMK==1024 => chipstep=256 , DIMK==512 => chipstep=128
  if(NOTA==1){
      for(j=0;j<bb;j++){
	  x = chipstep*ichip+j*32;
	  y = i*8 + start;
	  for(k=0;k<16;k++){
	      v2df * dest1 = (v2df*)&aem[i][0][j][k*4+0];
	      v2df * dest2 = (v2df*)&aem[i][2][j][k*4+0];
	      v2df * dest3 = (v2df*)&aem[i][1][j][k*4+0];
	      v2df * dest4 = (v2df*)&aem[i][3][j][k*4+0];
	      v2df * src1 = (v2df*)&a[k+x][0+y];
	      v2df * src2 = (v2df*)&a[k+x+16][0+y];

	      dest1[0]=src1[0];
	      dest1[1]=src1[1];
	      dest2[0]=src1[2];
	      dest2[1]=src1[3];
	      dest3[0]=src2[0];
	      dest3[1]=src2[1];
	      dest4[0]=src2[2];
	      dest4[1]=src2[3];
	  }
      }
  } else {
      for(j=0;j<bb;j++){
	  for(k=0;k<16;k++){
	      x = 512*ichip+j*32;
	      y = i*8 + start;

	      aem[i][0][j][k*4+0] = a[0+y][k+x];
	      aem[i][0][j][k*4+1] = a[1+y][k+x];
	      aem[i][0][j][k*4+2] = a[2+y][k+x];
	      aem[i][0][j][k*4+3] = a[3+y][k+x];
	      
	      aem[i][2][j][k*4+0] = a[4+y][k+x];
	      aem[i][2][j][k*4+1] = a[5+y][k+x];
	      aem[i][2][j][k*4+2] = a[6+y][k+x];
	      aem[i][2][j][k*4+3] = a[7+y][k+x];
	      
	      aem[i][1][j][k*4+0] = a[0+y][k+x+16];
	      aem[i][1][j][k*4+1] = a[1+y][k+x+16];
	      aem[i][1][j][k*4+2] = a[2+y][k+x+16];
	      aem[i][1][j][k*4+3] = a[3+y][k+x+16];
	      
	      aem[i][3][j][k*4+0] = a[4+y][k+x+16];
	      aem[i][3][j][k*4+1] = a[5+y][k+x+16];
	      aem[i][3][j][k*4+2] = a[6+y][k+x+16];
	      aem[i][3][j][k*4+3] = a[7+y][k+x+16];
	  }
      }
  }
  
  //padding with 0
  //  for(j=bb;j<NBB;j++){
      // skip padding for test
  for(j=NBB;j<NBB;j++){
      for(k=0;k<16;k++){
	  x = chipstep*ichip+j*32;
	  y = i*8 + start;
	  aem[i][0][j][k*4+0] = 0.0;
	  aem[i][0][j][k*4+1] = 0.0;
	  aem[i][0][j][k*4+2] = 0.0;
	  aem[i][0][j][k*4+3] = 0.0;
	  aem[i][2][j][k*4+0] = 0.0;
	  aem[i][2][j][k*4+1] = 0.0;
	  aem[i][2][j][k*4+2] = 0.0;
	  aem[i][2][j][k*4+3] = 0.0;
	  aem[i][1][j][k*4+0] = 0.0;
	  aem[i][1][j][k*4+1] = 0.0;
	  aem[i][1][j][k*4+2] = 0.0;
	  aem[i][1][j][k*4+3] = 0.0;
	  aem[i][3][j][k*4+0] = 0.0;
	  aem[i][3][j][k*4+1] = 0.0;
	  aem[i][3][j][k*4+2] = 0.0;
	  aem[i][3][j][k*4+3] = 0.0;
      }
  }
  
  end = rdtsc2();
#ifdef BENCHSENDI
  //printf("copymata_em elapsed : %lld [clk] %e[s] \n", end-startt, (end-startt)/2.6e9);
#endif
}


#define MATACACHEK 2048
#define MATACACHEN 16
#define MATACACHEMASK (0xffffffff0)
static double matacache[MATACACHEK*MATACACHEN];

void copytomatacache(int LDA, int NOTA, double a[][LDA],
		     int jstart,
		     int DIMK)
{
    static int jrecord=-1;
    int j,k;
    jstart &= MATACACHEMASK;
    if (jrecord != jstart){
	jrecord = jstart;
	//	fprintf(stderr,"copy to matacache, LDA=%d, DIMK=%d nota=%d jstart=%d\n",
	//		LDA, DIMK,NOTA, jstart);
	if(NOTA==1){
	    for(j=0;j<DIMK; j++){
		v2df* dest = (v2df*)(matacache+j*MATACACHEN);
		v2df* src = (v2df*)(a[j]+jstart);
		for(k=0;k<MATACACHEN/2;k+=4){
		    dest[k]=src[k];
		    dest[k+1]=src[k+1];
		    dest[k+2]=src[k+2];
		    dest[k+3]=src[k+3];
		}
	    }
	}
    }
}


void copytomata_em_1024_tb4_dg_singlepe_original(int pestart, 
				       int LDA, int NOTA, double a[][LDA],
				       int start, int ichip, int DIMK,
				       double aem[NPE][4][NBB][64])
//fortran order a[col][row]
{
  int i, j, k, x, y, bb, chipstep;
  UINT64 startt, end;
  i = pestart;
  startt = rdtsc2();
  bb = NBB*DIMK/2048; //DIMK==2048 => 16bb ,  DIMK==1024 => 8bb ,
                      //DIMK==512 => 4bb
  chipstep = 512*DIMK/2048; // DIMK==2048 => chipstep=512,
                            // DIMK==1024 => chipstep=256 ,
                            // DIMK==512 => chipstep=128
  
  y = i*8 + start;
  
  if (NOTA==1){
      //      fprintf(stderr,"non-trans version of copytomata called\n");
      copytomatacache(LDA, NOTA, a, y, DIMK);
      int ycache = y % MATACACHEN;
      for(j=0;j<bb;j++){
	  x = chipstep*ichip+j*32;
	  int DIMK2 = DIMK/2;
	  for(k=0;k<16;k++){
	      //      fprintf(stderr,"j,k=%d %d a, achache = %e %e\n",j,k,
	      //		      a[k+x][y],matacache[ycache][k+x]);
	      //	      fprintf(stderr,"index =%d %d %d %d\n",k,x,y, k+x+ycache*DIMK);
	      v2df * src1 = (v2df*)&matacache[(k+x)*MATACACHEN+ycache];
	      v2df * src2 = (v2df*)&matacache[(k+x+16)*MATACACHEN+ycache];
	      v2df * dest1 = (v2df*)&aem[i][0][j][k*4+0];
	      v2df * dest2 = (v2df*)&aem[i][2][j][k*4+0];
	      v2df * dest3 = (v2df*)&aem[i][1][j][k*4+0];
	      v2df * dest4 = (v2df*)&aem[i][3][j][k*4+0];
	      
	      dest1[0]=src1[0];
	      dest1[1]=src1[1];
	      dest2[0]=src1[2];
	      dest2[1]=src1[3];
	      dest3[0]=src2[0];
	      dest3[1]=src2[1];
	      dest4[0]=src2[2];
	      dest4[1]=src2[3];
	  }
      }
      return;
  } else {
      for(j=0;j<bb;j++){
	  for(k=0;k<16;k++){
	      x = 512*ichip+j*32;
	      aem[i][0][j][k*4+0] = a[0+y][k+x];
	      aem[i][0][j][k*4+1] = a[1+y][k+x];
	      aem[i][0][j][k*4+2] = a[2+y][k+x];
	      aem[i][0][j][k*4+3] = a[3+y][k+x];
	      
	      aem[i][2][j][k*4+0] = a[4+y][k+x];
	      aem[i][2][j][k*4+1] = a[5+y][k+x];
	      aem[i][2][j][k*4+2] = a[6+y][k+x];
	      aem[i][2][j][k*4+3] = a[7+y][k+x];
	      
	      aem[i][1][j][k*4+0] = a[0+y][k+x+16];
	      aem[i][1][j][k*4+1] = a[1+y][k+x+16];
	      aem[i][1][j][k*4+2] = a[2+y][k+x+16];
	      aem[i][1][j][k*4+3] = a[3+y][k+x+16];
	      
	      aem[i][3][j][k*4+0] = a[4+y][k+x+16];
	      aem[i][3][j][k*4+1] = a[5+y][k+x+16];
	      aem[i][3][j][k*4+2] = a[6+y][k+x+16];
	      aem[i][3][j][k*4+3] = a[7+y][k+x+16];
	  }
      }
  }
  
  //padding with 0
  //  for(j=bb;j<NBB;j++){
      // skip padding for test
  for(j=NBB;j<NBB;j++){
      for(k=0;k<16;k++){
	  x = chipstep*ichip+j*32;
	  y = i*8 + start;
	  aem[i][0][j][k*4+0] = 0.0;
	  aem[i][0][j][k*4+1] = 0.0;
	  aem[i][0][j][k*4+2] = 0.0;
	  aem[i][0][j][k*4+3] = 0.0;
	  aem[i][2][j][k*4+0] = 0.0;
	  aem[i][2][j][k*4+1] = 0.0;
	  aem[i][2][j][k*4+2] = 0.0;
	  aem[i][2][j][k*4+3] = 0.0;
	  aem[i][1][j][k*4+0] = 0.0;
	  aem[i][1][j][k*4+1] = 0.0;
	  aem[i][1][j][k*4+2] = 0.0;
	  aem[i][1][j][k*4+3] = 0.0;
	  aem[i][3][j][k*4+0] = 0.0;
	  aem[i][3][j][k*4+1] = 0.0;
	  aem[i][3][j][k*4+2] = 0.0;
	  aem[i][3][j][k*4+3] = 0.0;
      }
  }
  
  end = rdtsc2();
#ifdef BENCHSENDI
  //printf("copymata_em elapsed : %lld [clk] %e[s] \n", end-startt, (end-startt)/2.6e9);
#endif
}
void copytomata_em_1024_tb4_dg_singlepe(int pestart, 
				       int LDA, int NOTA, double a[][LDA],
				       int start, int ichip, int DIMK,
				       double aem[NPE][4][NBB][64])
//fortran order a[col][row]
{
  int i, j, k, x, y, bb, chipstep;
  UINT64 startt, end;
  i = pestart;
  startt = rdtsc2();
  bb = NBB*DIMK/2048; //DIMK==2048 => 16bb ,  DIMK==1024 => 8bb ,
                      //DIMK==512 => 4bb
  chipstep = 512*DIMK/2048; // DIMK==2048 => chipstep=512,
                            // DIMK==1024 => chipstep=256 ,
                            // DIMK==512 => chipstep=128
  
  y = i*8 + start;
  int ibuff = i %1;
  if (NOTA==1){
//    fprintf(stderr,"non-trans version of copytomata called\n");
//    copytomatacache(LDA, NOTA, a, y, DIMK);
      int ycache = y % MATACACHEN;
      for(j=0;j<bb;j++){
	  x = chipstep*ichip+j*32;
	  int bid;
	  for(bid=0;bid<4;bid++){
	      int x0 = chipstep*ichip+j*32+bid*8;
	      for(k=0;k<8;k++){
		  x = x0+k;
		  v2df *src = (v2df*)(&a[x][y]);
  //  		  v2df * src = (v2df*)&matacache[(k+x+bid*8)*MATACACHEN+ycache];
		  v2df * dest = (v2df*)&aem[ibuff][bid][j][k*8];
		  dest[0]=src[0];
		  dest[1]=src[1];
		  dest[2]=src[2];
		  dest[3]=src[3];
		  
	      }
	  }
      }
      return;
  } else {
      for(j=0;j<bb;j++){
	  int bid;
	  for(bid=0;bid<4;bid++){
	      int x0 = chipstep*ichip+j*32+bid*8;
	      for(k=0;k<8;k++){
		  x = x0+k;
		  int k8=k*8;
		  aem[ibuff][bid][j][k8] = a[y][x];
		  aem[ibuff][bid][j][k8+1] = a[y+1][x];
		  aem[ibuff][bid][j][k8+2] = a[y+2][x];
		  aem[ibuff][bid][j][k8+3] = a[y+3][x];
		  aem[ibuff][bid][j][k8+4] = a[y+4][x];
		  aem[ibuff][bid][j][k8+5] = a[y+5][x];
		  aem[ibuff][bid][j][k8+6] = a[y+6][x];
		  aem[ibuff][bid][j][k8+7] = a[y+7][x];
	      }
	  }
      }
  }
  
  //padding with 0
  //  for(j=bb;j<NBB;j++){
      // skip padding for test
  for(j=NBB;j<NBB;j++){
      for(k=0;k<16;k++){
	  x = chipstep*ichip+j*32;
	  y = i*8 + start;
	  aem[ibuff][0][j][k*4+0] = 0.0;
	  aem[ibuff][0][j][k*4+1] = 0.0;
	  aem[ibuff][0][j][k*4+2] = 0.0;
	  aem[ibuff][0][j][k*4+3] = 0.0;
	  aem[ibuff][2][j][k*4+0] = 0.0;
	  aem[ibuff][2][j][k*4+1] = 0.0;
	  aem[ibuff][2][j][k*4+2] = 0.0;
	  aem[ibuff][2][j][k*4+3] = 0.0;
	  aem[ibuff][1][j][k*4+0] = 0.0;
	  aem[ibuff][1][j][k*4+1] = 0.0;
	  aem[ibuff][1][j][k*4+2] = 0.0;
	  aem[ibuff][1][j][k*4+3] = 0.0;
	  aem[ibuff][3][j][k*4+0] = 0.0;
	  aem[ibuff][3][j][k*4+1] = 0.0;
	  aem[ibuff][3][j][k*4+2] = 0.0;
	  aem[ibuff][3][j][k*4+3] = 0.0;
      }
  }
  
  end = rdtsc2();
#ifdef BENCHSENDI
  //printf("copymata_em elapsed : %lld [clk] %e[s] \n", end-startt, (end-startt)/2.6e9);
#endif
}

void copytomata_em_1024_tb4_dg_singlepe_secondhalf(int pestart, 
				       int LDA, int NOTA, double a[][LDA],
				       int start, int ichip, int DIMK,
				       double aem[NPE][4][NBB][64])
//fortran order a[col][row]
{
  int i, j, k, x, y, bb, chipstep;
  UINT64 startt, end;
  i = pestart;
  startt = rdtsc2();
  bb = NBB*DIMK/2048; //DIMK==2048 => 16bb ,  DIMK==1024 => 8bb , DIMK==512 => 4bb
  chipstep = 512*DIMK/2048; //DIMK==2048 => chipstep=512 ,  DIMK==1024 => chipstep=256 , DIMK==512 => chipstep=128
  //  fprintf(stderr,"copymata1024....nota=%d\n",NOTA);
  if(NOTA==1){
      for(j=0;j<bb;j++){
	  x = chipstep*ichip+j*32;
	  y = i*8 + start;
	  for(k=0;k<16;k++){
	      v2df * dest2 = (v2df*)&aem[i][2][j][k*4+0];
	      v2df * dest4 = (v2df*)&aem[i][3][j][k*4+0];
	      v2df * src1 = (v2df*)&a[k+x][0+y];
	      v2df * src2 = (v2df*)&a[k+x+16][0+y];

	      dest2[0]=src1[2];
	      dest2[1]=src1[3];
	      dest4[0]=src2[2];
	      dest4[1]=src2[3];
	  }
      }
  } else {
      for(j=0;j<bb;j++){
	  for(k=0;k<16;k++){
	      x = 512*ichip+j*32;
	      y = i*8 + start;
	      
	      aem[i][2][j][k*4+0] = a[4+y][k+x];
	      aem[i][2][j][k*4+1] = a[5+y][k+x];
	      aem[i][2][j][k*4+2] = a[6+y][k+x];
	      aem[i][2][j][k*4+3] = a[7+y][k+x];
	      
	      aem[i][3][j][k*4+0] = a[4+y][k+x+16];
	      aem[i][3][j][k*4+1] = a[5+y][k+x+16];
	      aem[i][3][j][k*4+2] = a[6+y][k+x+16];
	      aem[i][3][j][k*4+3] = a[7+y][k+x+16];
	  }
      }
  }
  
  //padding with 0
  for(j=bb;j<NBB;j++){
      for(k=0;k<16;k++){
	  x = chipstep*ichip+j*32;
	  y = i*8 + start;
	  aem[i][2][j][k*4+0] = 0.0;
	  aem[i][2][j][k*4+1] = 0.0;
	  aem[i][2][j][k*4+2] = 0.0;
	  aem[i][2][j][k*4+3] = 0.0;
	  aem[i][3][j][k*4+0] = 0.0;
	  aem[i][3][j][k*4+1] = 0.0;
	  aem[i][3][j][k*4+2] = 0.0;
	  aem[i][3][j][k*4+3] = 0.0;
      }
  }
  
  end = rdtsc2();
#ifdef BENCHSENDI
  //printf("copymata_em elapsed : %lld [clk] %e[s] \n", end-startt, (end-startt)/2.6e9);
#endif
}

void copytomata_em_1024_tb4_dg_singlepe_firsthalf(int pestart, 
				       int LDA, int NOTA, double a[][LDA],
				       int start, int ichip, int DIMK,
				       double aem[NPE][4][NBB][64])
//fortran order a[col][row]
{
  int i, j, k, x, y, bb, chipstep;
  UINT64 startt, end;
  i = pestart;
  startt = rdtsc2();
  bb = NBB*DIMK/2048; //DIMK==2048 => 16bb ,  DIMK==1024 => 8bb , DIMK==512 => 4bb
  chipstep = 512*DIMK/2048; //DIMK==2048 => chipstep=512 ,  DIMK==1024 => chipstep=256 , DIMK==512 => chipstep=128
  //  fprintf(stderr,"copymata1024....nota=%d\n",NOTA);
  if(NOTA==1){
      for(j=0;j<bb;j++){
	  x = chipstep*ichip+j*32;
	  y = i*8 + start;
	  for(k=0;k<16;k++){
	      v2df * dest1 = (v2df*)&aem[i][0][j][k*4+0];
	      v2df * dest3 = (v2df*)&aem[i][1][j][k*4+0];
	      v2df * src1 = (v2df*)&a[k+x][0+y];
	      v2df * src2 = (v2df*)&a[k+x+16][0+y];

	      dest1[0]=src1[0];
	      dest1[1]=src1[1];
	      dest3[0]=src2[0];
	      dest3[1]=src2[1];
	  }
      }
  } else {
      for(j=0;j<bb;j++){
	  for(k=0;k<16;k++){
	      x = 512*ichip+j*32;
	      y = i*8 + start;

	      aem[i][0][j][k*4+0] = a[0+y][k+x];
	      aem[i][0][j][k*4+1] = a[1+y][k+x];
	      aem[i][0][j][k*4+2] = a[2+y][k+x];
	      aem[i][0][j][k*4+3] = a[3+y][k+x];
	      
	      aem[i][1][j][k*4+0] = a[0+y][k+x+16];
	      aem[i][1][j][k*4+1] = a[1+y][k+x+16];
	      aem[i][1][j][k*4+2] = a[2+y][k+x+16];
	      aem[i][1][j][k*4+3] = a[3+y][k+x+16];
	  }
      }
  }
  
  //padding with 0
  for(j=bb;j<NBB;j++){
      for(k=0;k<16;k++){
	  x = chipstep*ichip+j*32;
	  y = i*8 + start;
	  aem[i][0][j][k*4+0] = 0.0;
	  aem[i][0][j][k*4+1] = 0.0;
	  aem[i][0][j][k*4+2] = 0.0;
	  aem[i][0][j][k*4+3] = 0.0;
	  aem[i][1][j][k*4+0] = 0.0;
	  aem[i][1][j][k*4+1] = 0.0;
	  aem[i][1][j][k*4+2] = 0.0;
	  aem[i][1][j][k*4+3] = 0.0;
      }
  }
  
  end = rdtsc2();
#ifdef BENCHSENDI
  //printf("copymata_em elapsed : %lld [clk] %e[s] \n", end-startt, (end-startt)/2.6e9);
#endif
}


void copytomata_em_1024_tb4_dg_singlepe_mc(int pestart, int mode,
					  int LDA, int NOTA, double a[][LDA],
					  int start, int nchip, int DIMK,
					  double *aem[])
// mode=0: send single PE
//     =1: send first half of PE
//     =2: send second half of PE    
{
    int ichip;
    double (*pa)[NPE][4][NBB][64];
    for(ichip=0;ichip<nchip; ichip++){
	pa = (double (*)[NPE][4][NBB][64])aem[ichip];
	if (mode == 0){
	    copytomata_em_1024_tb4_dg_singlepe(pestart,  LDA, NOTA, a,
					       start, ichip,DIMK,*pa);
	}else if (mode == 1){
	    copytomata_em_1024_tb4_dg_singlepe_firsthalf(pestart,  LDA, NOTA, a,
							 start, ichip,DIMK,*pa);
	}else {
	    copytomata_em_1024_tb4_dg_singlepe_secondhalf(pestart,  LDA, NOTA,
							  a, start, ichip,
							  DIMK,*pa);
	}
    }
}

void copytomata_em_1024_tb4_dg_partial_mc(int pestart, int peend,
					  int LDA, int NOTA, double a[][LDA],
					  int start, int nchip, int DIMK,
					  double *aem[])
{
    int ichip;
    double (*pa)[NPE][4][NBB][64];
    int ipe;
    //    fprintf(stderr,"copytomata... pestart = %d\n", pestart);
    for (ipe=pestart; ipe<peend; ipe++){
	//	fprintf(stderr,"ipe = %d\n", ipe);
	copytomata_em_1024_tb4_dg_singlepe_mc(ipe,0,LDA, NOTA, a,
					      start, nchip, DIMK,aem);
    }
}

void copytomata_partial_mc(int iter, int nloop,
			   int LDA, int NOTA, double a[][LDA],
			   int start, int nchip, int DIMK,
			   double *aem[])
{

    int ipe;
    if (nloop==64){
	for (ipe=iter*2; ipe<iter*2+2; ipe++){
	    //	fprintf(stderr,"ipe = %d\n", ipe);
	    copytomata_em_1024_tb4_dg_singlepe_mc(ipe,0,LDA, NOTA, a,
						  start, nchip, DIMK,aem);
	}
    }else if (nloop==32){
	    copytomata_em_1024_tb4_dg_singlepe_mc(iter,0,LDA, NOTA, a,
						  start, nchip, DIMK,aem);
    }else if (nloop==16){
	ipe = iter/2;
	int mode = (iter %2 )+1;
	copytomata_em_1024_tb4_dg_singlepe_mc(ipe,mode,LDA, NOTA, a,
					      start, nchip, DIMK,aem);
    }else if (nloop==8){
	ipe = iter/4;
	int mode = ((iter/2) %2 )+1;
	copytomata_em_1024_tb4_dg_singlepe_mc(ipe,mode,LDA, NOTA, a,
					      start, nchip, DIMK,aem);
    }else{
	fprintf(stderr,"copytomata_partial_mc: nloop=%d not supported\n");
    }
}

void copytomata_em_1024_tb4_dg(int LDA, int NOTA, double a[][LDA], int start, int ichip, int DIMK, double aem[NPE][4][NBB][64]) //fortran order a[col][row]
{
  int i, j, k, x, y, bb, chipstep;
  UINT64 startt, end;
  copytomata_em_1024_tb4_dg_partial(0, NPE, LDA, NOTA, a, start, ichip,DIMK, aem);
  return;

  startt = rdtsc2();
  bb = NBB*DIMK/2048; //DIMK==2048 => 16bb ,  DIMK==1024 => 8bb , DIMK==512 => 4bb
  chipstep = 512*DIMK/2048; //DIMK==2048 => chipstep=512 ,  DIMK==1024 => chipstep=256 , DIMK==512 => chipstep=128
  //  fprintf(stderr,"copymata1024....nota=%d\n",NOTA);
  if(NOTA==1){
      for(j=0;j<bb;j++){
	  for(i=0;i<NPE;i++){
	  x = chipstep*ichip+j*32;
	  y = i*8 + start;
#if 1
	for(k=0;k<16;k++){
	  v2df * dest1 = (v2df*)&aem[i][0][j][k*4+0];
	  v2df * dest2 = (v2df*)&aem[i][2][j][k*4+0];
	  v2df * dest3 = (v2df*)&aem[i][1][j][k*4+0];
	  v2df * dest4 = (v2df*)&aem[i][3][j][k*4+0];
	  v2df * src1 = (v2df*)&a[k+x][0+y];
	  v2df * src2 = (v2df*)&a[k+x+16][0+y];
	  
	  /*
	  __builtin_prefetch(&(a[k+x+1][0+y]), 0, 2);
	  __builtin_prefetch(&(a[k+x+17][0+y]), 0, 2);
	  */
	  dest1[0]=src1[0];
	  dest1[1]=src1[1];
	  dest2[0]=src1[2];
	  dest2[1]=src1[3];
	  dest3[0]=src2[0];
	  dest3[1]=src2[1];
	  dest4[0]=src2[2];
	  dest4[1]=src2[3];
	}
#endif
#if 0	
	for(k=0;k<16;k++){
	  v2df * dest1 = (v2df*)&aem[i][0][j][k*4+0];
	  v2df * dest2 = (v2df*)&aem[i][2][j][k*4+0];
	  v2df * src1 = (v2df*)&a[k+x][0+y];
	  dest1[0]=src1[0];
	  dest1[1]=src1[1];
	  dest2[0]=src1[2];
	  dest2[1]=src1[3];
	}
	for(k=0;k<16;k++){
	  v2df * dest3 = (v2df*)&aem[i][1][j][k*4+0];
	  v2df * dest4 = (v2df*)&aem[i][3][j][k*4+0];
	  v2df * src2 = (v2df*)&a[k+x+16][0+y];
	  dest3[0]=src2[0];
	  dest3[1]=src2[1];
	  dest4[0]=src2[2];
	  dest4[1]=src2[3];
	}
#endif	
      }
    }
  } else {
    for(i=0;i<NPE;i++){
      for(j=0;j<bb;j++){
	for(k=0;k<16;k++){
	  x = 512*ichip+j*32;
	  y = i*8 + start;
	  /*
	  __builtin_prefetch(&(a[0+y][k+x+1]), 0, 0);
	  __builtin_prefetch(&(a[0+y][k+x+17]), 0, 0);
	  */
	  aem[i][0][j][k*4+0] = a[0+y][k+x];
	  aem[i][0][j][k*4+1] = a[1+y][k+x];
	  aem[i][0][j][k*4+2] = a[2+y][k+x];
	  aem[i][0][j][k*4+3] = a[3+y][k+x];

	  aem[i][2][j][k*4+0] = a[4+y][k+x];
	  aem[i][2][j][k*4+1] = a[5+y][k+x];
	  aem[i][2][j][k*4+2] = a[6+y][k+x];
	  aem[i][2][j][k*4+3] = a[7+y][k+x];

	  aem[i][1][j][k*4+0] = a[0+y][k+x+16];
	  aem[i][1][j][k*4+1] = a[1+y][k+x+16];
	  aem[i][1][j][k*4+2] = a[2+y][k+x+16];
	  aem[i][1][j][k*4+3] = a[3+y][k+x+16];
	    
	  aem[i][3][j][k*4+0] = a[4+y][k+x+16];
	  aem[i][3][j][k*4+1] = a[5+y][k+x+16];
	  aem[i][3][j][k*4+2] = a[6+y][k+x+16];
	  aem[i][3][j][k*4+3] = a[7+y][k+x+16];
	}
      }
    }
  }
  
  //padding with 0
  for(i=0;i<NPE;i++){
    for(j=bb;j<NBB;j++){
      for(k=0;k<16;k++){
	x = chipstep*ichip+j*32;
	y = i*8 + start;
	aem[i][0][j][k*4+0] = 0.0;
	aem[i][0][j][k*4+1] = 0.0;
	aem[i][0][j][k*4+2] = 0.0;
	aem[i][0][j][k*4+3] = 0.0;
      	aem[i][2][j][k*4+0] = 0.0;
	aem[i][2][j][k*4+1] = 0.0;
	aem[i][2][j][k*4+2] = 0.0;
	aem[i][2][j][k*4+3] = 0.0;
	aem[i][1][j][k*4+0] = 0.0;
	aem[i][1][j][k*4+1] = 0.0;
	aem[i][1][j][k*4+2] = 0.0;
	aem[i][1][j][k*4+3] = 0.0;
	aem[i][3][j][k*4+0] = 0.0;
	aem[i][3][j][k*4+1] = 0.0;
	aem[i][3][j][k*4+2] = 0.0;
	aem[i][3][j][k*4+3] = 0.0;
      }
    }
  }
	    
  end = rdtsc2();
#ifdef BENCHSENDI
  //printf("copymata_em elapsed : %lld [clk] %e[s] \n", end-startt, (end-startt)/2.6e9);
#endif
}


void copytomatb(double b[DIM1][DIM0],
		int k,
		double brf[NBB][32]) 
{
    int i;
    double * pb =  brf[0];
	
    for (i=0;i<DIM0; i++){
	pb[i]=b[k][i];
    }
}

void copytomatb_tb4(double b[][DIM0],
		    int ichip,
		    int k,
		    double brf[NBB][32]) 
{
    int i;
    double * pb =  brf[0];
	
    for (i=0;i<DIM0/NCHIP; i++){
	pb[i]=b[k][512*ichip+i];
    }
}

void copytomatb_tb4_dg_old(int LDB, int NOTB, double b[][LDB],
		       //fortran order b[col][row] <= b[DIM1(col)][DIM0(row)]
		       int ichip,
		       int k,
		       int DIMK,
		       double brf[NBB][32]) 
{
    int i,j,x, bb, jstep;
    //double * pb =  brf[0];
  
    bb = NBB*DIMK/2048;
    jstep = 512*DIMK/2048;
    if(NOTB==1){
	for (i=0;i<bb;i++){ //BB
	    v2df * dest = (v2df*)brf[i];
	    v2df * src = (v2df*)(b[k]+jstep*ichip+i*32);
	    for(j=0;j<16;j+=2){
		dest[j]=src[j];
		dest[j+1]=src[j+1];
	    }
	}
    } else {
	for (i=0;i<bb;i++){ //BB
	    for(j=0;j<32;j++){ //reg
		x = jstep*ichip+i*32+j;
		brf[i][j]=b[x][k]; 
	    }
	}
    }
    if (0){
	// Padding with 0
	
	for(i=bb;i<NBB;i++){
	    for(j=0;j<32;j++){
		brf[i][j] = 0.0;
	    }
	}
	for(i=0;i<NBB;i++){
	    fprintf(stderr, "cpbm i=%2d ", i);
	    for(j=0;j<32;j++){
		fprintf(stderr," %2d", (int)(brf[i][j]*100));
	    }
	    fprintf(stderr, "\n");
	}
    }
}

#define MATBCACHEK 2048
#define MATBCACHEN 16
#define MATBCACHEMASK (0xfffffff0)
static double matbcache[MATBCACHEN*MATBCACHEK];

static void transpose_rowtocol_simple(int n, int RDIM, double a[][RDIM],int m,
			       double at[][n],int istart)
{
    int i,j,k;
    for(i=0;i<n;i++){
	for(k=0;k<m;k++){
	    at[k][i]=a[i][k+istart];
	}
    }
}


static void transpose_rowtocol(int n, int RDIM, double a[][RDIM],int m,
			       double at[][n],int jstart)
{
    int i,j,k;
    double atmp[m][m];
    for(i=0;i<n;i+=m){
	for(k=0;k<m;k++){
	    for(j=0;j<m;j++){
		atmp[j][k]  =a[i+k][j+jstart];
	    }
	}
	for(j=0;j<m;j++){
	    for(k=0;k<m;k++){
		at[j][i+k]=atmp[j][k];
	    }
	}
    }
}


static void transpose_rowtocol16(int n, int RDIM, double a[][RDIM],
				 double at[][n], int jstart)
{
    int i,j,k;
    const int m=16;
    int mend;
#pragma omp parallel for private(i,j,k)
    for(i=0;i<n;i+=m){
	double atmp[m][m];
	//	BEGIN_TSC;
	for(k=0;k<m;k++){
	    v2df * ak = (v2df*) (a[i+k]+jstart);
	    v2df * akk = (v2df*) atmp[k];
	    asm("prefetchnta %0"::"m"(a[i+k+m*2][0]):"memory");
	    asm("prefetchnta %0"::"m"(a[i+k+m*2][8]):"memory");
	    //	    __builtin_prefetch(a[i+k+m*2],0,0);
	    //	    __builtin_prefetch(a[i+k+m*2]+8,0,0);
	    akk[0]  =ak[0];
	    akk[1]  =ak[1];
	    akk[2]  =ak[2];
	    akk[3]  =ak[3];
	    akk[4]  =ak[4];
	    akk[5]  =ak[5];
	    akk[6]  =ak[6];
	    akk[7]  =ak[7];
	}
	//	END_TSC(t,17);
	//	{
	//	BEGIN_TSC;
	for(j=0;j<m;j++){
	    v2df * atk= (v2df*)(at[j]+i);
	    atk[0]=(v2df){atmp[0][j],atmp[1][j]};
	    atk[1]=(v2df){atmp[2][j],atmp[3][j]};
	    atk[2]=(v2df){atmp[4][j],atmp[5][j]};
	    atk[3]=(v2df){atmp[6][j],atmp[7][j]};
	    atk[4]=(v2df){atmp[8][j],atmp[9][j]};
	    atk[5]=(v2df){atmp[10][j],atmp[11][j]};
	    atk[6]=(v2df){atmp[12][j],atmp[13][j]};
	    atk[7]=(v2df){atmp[14][j],atmp[15][j]};
	}
	//	END_TSC(t2,18);
	//	}			   int istart)
    }

}

void copytomatbcache(int LDB, int NOTB, double b[][LDB],
		     int jstart,
		     int DIMK)
{
    static int jrecord=-1;
    int j,k;
    jstart &= MATBCACHEMASK;
    if (jrecord != jstart){
	jrecord = jstart;
	//	fprintf(stderr,"copy to matbcache, notb=%d jstart=%d\n", NOTB, jstart);
	if(NOTB==1){
	    
	    for(j=0;j<MATBCACHEN; j++){
		int jsrc = j+jstart;
		int jdest = j*DIMK;
		v2df* dest = (v2df*)(matbcache+jdest);
		v2df* src = (v2df*)(b[jsrc]);
		for(k=0;k<DIMK/2;k+=4){
		    dest[k]=src[k];
		    dest[k+1]=src[k+1];
		    dest[k+2]=src[k+2];
		    dest[k+3]=src[k+3];
		}
	    }
	}else{
	    transpose_rowtocol16(DIMK, LDB, b,
				 (double(*)[])matbcache, jstart);
	    return;
	    for(k=0;k<DIMK;k++){
		for(j=0;j<MATBCACHEN; j++){
		    int jsrc = j+jstart;
		    int jdest = j*DIMK;
		    matbcache[jdest+k] = b[k][jsrc];
		}
	    }
	}
    }
}




void copytomatb_tb4_dg(int LDB, int NOTB, double b[][LDB], //fortran order b[col][row] <= b[DIM1(col)][DIM0(row)]
		       int ichip,
		       int k,
		       int DIMK,
		       double brf[NBB][32]) 
{
    int i,j,x, bb, jstep;
    //double * pb =  brf[0];
  
    bb = NBB*DIMK/2048;
    jstep = 512*DIMK/2048;

    //
    // this function can be and should be simplified a lot -- JM
    //
    if (NOTB!=1){
	copytomatbcache(LDB, NOTB, b, k,DIMK);
	
	int kcache = k % MATBCACHEN;
	v2df * src = (v2df*)(matbcache+kcache*DIMK+jstep*ichip);
	
	for (i=0;i<bb;i++){ //BB
	    v2df * dest = (v2df*)brf[i];
	    for(j=0;j<16;j+=8){
		dest[j]=src[j];
		dest[j+1]=src[j+1];
		dest[j+2]=src[j+2];
		dest[j+3]=src[j+3];
		dest[j+4]=src[j+4];
		dest[j+5]=src[j+5];
		dest[j+6]=src[j+6];
		dest[j+7]=src[j+7];
	    }
	    src += 16;
	}
	return ;
    }
#if 0
    for (i=0;i<bb;i++){ //BB
	v2df * dest = (v2df*)brf[i];
	v2df * src = (v2df*)(b[k]+jstep*ichip+i*32);
	for(j=0;j<16;j+=2){
	    dest[j]=src[j];
	    dest[j+1]=src[j+1];
	}
    }
#else
    v2df * dest = (v2df*)brf[0];
    v2df * src = (v2df*)(b[k]+jstep*ichip);
    for(j=0;j<16*bb;j+=2){
	dest[j]=src[j];
	dest[j+1]=src[j+1];
    }
#endif    
}


void grape_init(int chipid, int nchips)
{
  int i;

  for(i=0;i<nchips;i++){
    s[i] = SING_openMC(i+chipid);
    matmul_grape_init(s[i]);
  }
  matmul_set_2stage(s, nchips, 1);
  CTRL_write_setflushmode(0);// workaround for
                             // problem found in 2011/1
  gethostname(myhostname,256);
  fprintf(stderr,"%s grape_init with %d chips\n", myhostname,nchips);
}

void grape_close()
{
  int i;
  matmul_set_2stage(s, NCHIP, 0);
  for(i=0;i<NCHIP;i++){
    SING_closeMC(s[i]);
  }
}

void set_coc_reg(int address, UINT64 data)
{
  int i;
  int nchips=matmul_getnchips();
  for(i=0;i<nchips;i++){
      SING_coc_reg_write(s[i],  address, data);
  }
}

void dumpbr(int k,
	    double br[NBB][32])
{
    int i,j;
    printf("Dumpbr for %d\n", k);
    for(i=0;i<NBB;i++){
	for(j=0;j<32;j++){
	    printf(" %.1f", br[i][j]);
	}
	printf("\n");
    }
}



void matmul_host(double a[DIM1][DIM0],
	    double b[DIM1][DIM0],
	    double c[DIM1][DIM1])
{
    int i, j, k;

    for(i=0;i<DIM1;i++){
      for(j=0;j<DIM1;j++){
	    c[i][j]=0;
	    for(k=0;k<DIM0;k++) c[i][j]+= a[j][k]*b[i][k];
	}
    }
}

void matmul_host_mnk(int DIMM, int DIMN, int DIMK,
		 double a[DIMM][DIMK],
		 double b[DIMN][DIMK],
		 double c[DIMN][DIMM])
{
  int i, j, k;
  for(i=0;i<DIMN;i++){
    for(j=0;j<DIMM;j++){
      c[i][j]=0;
      for(k=0;k<DIMK;k++) c[i][j]+= a[j][k]*b[i][k];
    }
  }
}



#define OVERLAP_ITRANS



#ifndef NTHREAD
#define NTHREAD 4
#endif

struct gdrthread_arg {
  void *arg;
  int tid;
  enum {run, finish, stop} running;
    //#ifndef DONOTUSEPTHREADS    
  pthread_cond_t start , end, *share_start;
  pthread_mutex_t mstart, mend;
    //#endif    
  void (*work)(void *);
} gdr_Thread_arg[NTHREAD];

//#ifndef DONOTUSEPTHREADS    
static pthread_t thread_handle[NTHREAD];
static pthread_cond_t master_start;
//#endif    

#ifndef DONOTUSEPTHREADS
void* gdrthread_loop(void *ptr)
{
  pid_t pid;
  cpu_set_t cpu;
  struct gdrthread_arg *arg = (struct thread_arg*)ptr;

#ifdef SETTHREADAFFINITY
  __CPU_ZERO(&cpu);
  __CPU_SET(arg->tid, &cpu);
  sched_setaffinity(0, sizeof(cpu_set_t), &cpu);
#endif

  while(1){
    pthread_mutex_lock(&(arg->mstart));
    while(arg->running != run){
      pthread_cond_wait(&(arg->start), &(arg->mstart));
    }
    
    if(arg->work)
      (*(arg->work))(arg->arg);

    arg->running = finish;
    pthread_mutex_unlock(&(arg->mstart));
    pthread_cond_signal(&(arg->end));
  }

}
#endif 
void gdrinit_threads(void)
{
  int i;
  static int firstcall=1;
  pid_t pid;
  cpu_set_t cpu;

  if(firstcall==0){
    return;
  }

  for(i=0;i<NTHREAD;i++){
    gdr_Thread_arg[i].running = stop;
    gdr_Thread_arg[i].tid = i;
    gdr_Thread_arg[i].work = NULL;
    gdr_Thread_arg[i].share_start = &master_start;
  }
  
#ifndef DONOTUSEPTHREADS
  
#ifdef SETTHREADAFFINITY
  __CPU_ZERO(&cpu);
  __CPU_SET(NTHREAD-1, &cpu);
  sched_setaffinity(0, sizeof(cpu_set_t), &cpu);
#endif

  pthread_cond_init(&master_start, NULL);
  for(i=0;i<NTHREAD;i++){
    gdr_Thread_arg[i].running = stop;
    gdr_Thread_arg[i].tid = i;
    gdr_Thread_arg[i].work = NULL;
    gdr_Thread_arg[i].share_start = &master_start;
    pthread_cond_init(&(gdr_Thread_arg[i].start), NULL);
    pthread_mutex_init(&(gdr_Thread_arg[i].mstart), NULL);
    pthread_cond_init(&(gdr_Thread_arg[i].end), NULL);
    pthread_mutex_init(&(gdr_Thread_arg[i].mend), NULL);
    
    pthread_mutex_lock(&(gdr_Thread_arg[i].mstart));

    pthread_create(&(thread_handle[i]), NULL, &gdrthread_loop, &(gdr_Thread_arg[i]));
  }
#endif  
  firstcall=0;
}


void gdrrun_threads_serialized(void (*run_func)(void *), void *param[NTHREAD])
{
  int i;
  UINT64 start, run, end;

  for(i=0;i<NTHREAD;i++){
      if(run_func){
	  (*run_func)(param[i]);
      }
  }
}
void gdrrun_threads_omp(void (*run_func)(void *), void *param[NTHREAD])
{
  int i;
  UINT64 start, run, end;
  //#define DONOTUSEOMP
#ifdef DONOTUSEOMP
   (*run_func)(param[0]); 
   (*run_func)(param[1]); 
   (*run_func)(param[2]); 
   (*run_func)(param[3]);
#else   
  
   //    fprintf(stderr,"gdrrun_threads omp_max_threads=%d procs=%d\n",
   //	    omp_get_max_threads(),omp_get_num_procs());

#pragma omp parallel
#pragma omp sections
  {
#pragma omp section
      {
	  //	  print_current_time("thread1 start");
	  //	  fprintf(stderr,"thread1 threadnum=%d\n", omp_get_thread_num());
	  (*run_func)(param[0]);
	  //	  print_current_time("thread1 end");
      }
#pragma omp section
      {
	  //	  print_current_time("thread2");
	  //	  fprintf(stderr,"thread2 threadnum=%d\n", omp_get_thread_num());
	  (*run_func)(param[1]);
	  //	  print_current_time("thread2 end");
      }
#pragma omp section
      {
	  //	  print_current_time("thread3");
	  //	  fprintf(stderr,"thread3 threadnum=%d\n", omp_get_thread_num());
	  (*run_func)(param[2]);
	  //	  print_current_time("thread3 end");
      }
#pragma omp section
      {
	  //	  print_current_time("thread4");
	  //	  fprintf(stderr,"thread4 threadnum=%d\n", omp_get_thread_num());
	  (*run_func)(param[3]);
	  //	  print_current_time("thread4 end");

      }
  }
#endif  
}


void gdrrun_threads(void (*run_func)(void *), void *param[NTHREAD])
{
  int i;
  UINT64 start, run, end;
#ifdef DONOTUSEPTHREADS
  gdrrun_threads_omp(run_func, param);
  return;
#else

  for(i=0;i<NTHREAD-1;i++){
    gdr_Thread_arg[i].work = run_func;
    gdr_Thread_arg[i].arg  = param[i];
    gdr_Thread_arg[i].running = run;
  }
  start = rdtsc2();
  for(i=0;i<NTHREAD-1;i++){
    pthread_mutex_unlock(&(gdr_Thread_arg[i].mstart));
    pthread_cond_signal(&(gdr_Thread_arg[i].start));
  }
  /* WORKING IN WORKER THREAD except last thread */
  if(run_func){
    (*run_func)(param[NTHREAD-1]);
  }

  run = rdtsc2();
  for(i=0;i<NTHREAD-1;i++){
    pthread_mutex_lock(&(gdr_Thread_arg[i].mstart));
    while(gdr_Thread_arg[i].running != finish){
      pthread_cond_wait(&(gdr_Thread_arg[i].end), &(gdr_Thread_arg[i].mstart));
    } 
  }
  end = rdtsc2();
#if 0 //def BENCHRUNTHREADS
  printf("runthreads : run=%lld , wait=%lld \n", run-start, end-run);
#endif
#endif  
}

struct gdr_send_i_param
{
  int DIMM;
  int DIMK;
  int ichip;
  int i;
  int NOTA;
  int LDA;
  double *a;
  double *aem;
  int overlap;
} gdr_send_i_data[NTHREAD];

void send_i_matrix_mtwork(void *ptr)
{
    double data[16][64][4];
    int ipe, ibb;
  
    struct gdr_send_i_param *p = (struct gdr_send_i_param*)ptr;
    int DIMM  = p->DIMM;
    int DIMK  = p->DIMK;
    int ichip = p->ichip;
    int i     = p->i;
    int NOTA  = p->NOTA;
    int LDA   = p->LDA;
    double (*a)[LDA];
    a = (double(*)[LDA])(p->a);
    double *aem = p->aem;
    int overlap = p->overlap;
    UINT64 start, copy, end;
    int j, nbb=NBB;

    if(overlap){
	/* Overlap case */
	if(DIMK==1024 || DIMK==512){
	    nbb = 16 * DIMK / 2048; 
	    /* if i==0, normal piow write i matdata for LM, and prepare for the next matrixdata */
	    if(i==0){
		double *anext = aem;
		for(ipe=0;ipe < NPE; ipe++){
		    clearBM_i_matrix_16bbMC(s[ichip], ipe);
		    for(ibb=0;ibb < nbb; ibb++){
			copytomata_tb4_dg(LDA, NOTA, a, i, ichip, ibb, ipe, DIMK, data);
		    }
		    send_i_matrix_nbbMC(s[ichip], data, ipe, nbb);
		}
		for(ipe=0;ipe<NPE; ipe++){
		    for(ibb=0;ibb<nbb; ibb++){
#if 1
			copytomata_tb4_dg_onebb(LDA, NOTA, a, i+DIM1, ichip, ibb, ipe, DIMK, (double (*)[4])anext);
#else
			for(j=0;j<256;j++){
			    anext[j] = 0.0;
			}
#endif

			anext += 256;
		    }
		}
	    } else {
		/* if i!=0, invoked EM->BM trans and BM->LM in partially 8BB(K==1024), 4BB(K==512), and prepare for the next matrixdata */
		double *anext = aem;
		setsend_i_elt2_16bbMC(s[ichip], 0); //PEID
		setsend_i_elt2_BBid_MC(s[ichip], nbb); //BBID < 8 is enable
		clearBM_i_matrix_16bbMC(s[ichip], 0);
		if(DIMK==512) {
		    matmul_move_elt_data2_m1024_npe_nbb_k512(s[ichip], 0, 0, 32);
		} else {
		    matmul_move_elt_data2_m1024_npe_nbb(s[ichip], 0, 0, 32);
		}

		if((i+DIM1)<DIMM){
		    for(ipe=0;ipe<NPE; ipe++){
			for(ibb=0;ibb<nbb; ibb++){
#if 1
			    copytomata_tb4_dg_onebb(LDA, NOTA, a, i+DIM1,
						    ichip, ibb, ipe, DIMK,
						    (double(*)[4])anext);
#else
			    for(j=0;j<256;j++){
				anext[j] = 0.0;
			    }
#endif

			    anext += 256;
			}
		    }
		} else {
#ifdef MATMULELT2ASYNC
		    while(CTRL_coc_busy(s[ichip]->c)) {}
#endif
		}
		matmul_move_elt_data2_m1024_npe_finish(s[ichip]);
	    }
	} else {
	    /* overlap case K==2048*/
	    if(i != 0){
		//fprintf(stderr,"send_i_matrix_mtwork, i=%d, ichip=%d\n",i,ichip);
		start = rdtsc2();
		setsend_i_elt2_16bbMC(s[ichip], 0);
		matmul_move_elt_data2_m1024_npe(s[ichip], 0, 0, 32);
		copy = rdtsc2();
		if((i+DIM1)<DIMM){
		    copytomata_em_1024_tb4_dg(LDA, NOTA, a, i+DIM1,
					      ichip, DIMK,
					      (double(*)[4][NBB][64])aem);
		} else {
#ifdef MATMULELT2ASYNC
		    while(CTRL_coc_busy(s[ichip]->c)) {}
#endif
		}
		matmul_move_elt_data2_m1024_npe_finish(s[ichip]);
		end = rdtsc2();
#ifdef BENCHSENDI
		printf("ovl send %lld [clk] , copy %lld [clk] \n", copy-start, end-copy);
#endif
	    } else {
		start = rdtsc2();
		for(ipe=0;ipe < NPE; ipe++){
		    for(ibb=0;ibb < NBB; ibb++){
			copytomata_tb4_dg(LDA, NOTA, a, i, ichip, ibb, ipe, DIMK, data);
		    }
		    send_i_matrix_16bbMC(s[ichip], data, ipe);
		}
		//transfer next a matrix to buffer
		copy = rdtsc2();
		copytomata_em_1024_tb4_dg(LDA, NOTA, a, i+DIM1, ichip, DIMK,
					  (double(*)[4][NBB][64])aem);
		end = rdtsc2();
#ifdef BENCHSENDI
		printf("nonovl send %lld [clk] , copy %lld [clk] \n", copy-start, end-copy);
#endif
	    }
	}
    } else {
	/* overlap case K==1024*/
	start = rdtsc2();
	if(DIMK<2048) {
	    clearBM_i_matrix_16bbMC(s[ichip], ipe);
	}
	nbb = NBB*DIMK/2048;
	copy = rdtsc2();
	for(ipe=0;ipe < NPE; ipe++){
	    for(ibb=0;ibb < nbb; ibb++){
		copytomata_tb4_dg(LDA, NOTA, a, i, ichip, ibb, ipe, DIMK, data);
	    }
	    send_i_matrix_nbbMC(s[ichip], data, ipe, nbb);
	}
	end = rdtsc2();
#ifdef BENCHSENDI
	printf("ipe %d send : clear %lld[clk] send %lld[clk] \n", ipe, copy-start, end-copy); 
#endif
	while(CTRL_coc_busy(s[ichip]->c)) {}
    }
}
void send_i_matrix_mtwork_overlap_inoz_k2k(void *ptr)
{
    double data[16][64][4];
    int ipe, ibb;
  
    struct gdr_send_i_param *p = (struct gdr_send_i_param*)ptr;
    int DIMM  = p->DIMM;
    int DIMK  = p->DIMK;
    int ichip = p->ichip;
    int i     = p->i;
    int NOTA  = p->NOTA;
    int LDA   = p->LDA;
    double *a = p->a;
    double *aem = p->aem;
    int overlap = p->overlap;
    UINT64 start, copy, end;
    UINT64 wrap1, wrap2;
    int j, nbb=NBB;

    //fprintf(stderr,"send_i_matrix_mtwork, i=%d, ichip=%d\n",i,ichip);
    start = rdtsc2();
    setsend_i_elt2_16bbMC(s[ichip], 0);
    wrap1 = rdtsc2();
    matmul_move_elt_data2_m1024_npe(s[ichip], 0, 0, 32);
    wrap2 = rdtsc2();
    while(rdtsc2()< wrap2 + 10000){}
    matmul_move_elt_data2_m1024_npe_finish(s[ichip]);
    copy = rdtsc2();
    if((i+DIM1)<DIMM){
	copytomata_em_1024_tb4_dg(LDA, NOTA, (double(*)[LDA])a, i+DIM1,
				  ichip, DIMK,
				  (double(*)[4][NBB][64])aem);
    }
    end = rdtsc2();
    fprintf(stderr,"ovl send %lld %lld %lld [clk] , copy %lld [clk] \n", wrap1-start,
	    wrap2-wrap1, copy-wrap2, end-copy);
}
void send_i_matrix_overlap_inoz_k2k(int DIMM,
				    int DIMK,
				    int ichip,
				    int i,
				    int NOTA,
				    int LDA,
				    double *a,
				    double *aem)
{
    double data[16][64][4];
    int ipe, ibb;
  
    UINT64 start, copy, end;
    UINT64 wrap1, wrap2;
    int j, nbb=NBB;

    //fprintf(stderr,"send_i_matrix_mtwork, i=%d, ichip=%d\n",i,ichip);
    start = rdtsc2();
    setsend_i_elt2_16bbMC(s[ichip], 0);
    wrap1 = rdtsc2();
    matmul_move_elt_data2_m1024_npe(s[ichip], 0, 0, 32);
    wrap2 = rdtsc2();
    copy = rdtsc2();
    if((i+DIM1)<DIMM){
	copytomata_em_1024_tb4_dg(LDA, NOTA, (double(*)[LDA])a, i+DIM1, ichip,
					      DIMK, (double(*)[4][NBB][64])aem);
    }else{
	while(rdtsc2()< wrap2 + 10000){}
    }
    matmul_move_elt_data2_m1024_npe_finish(s[ichip]);
    end = rdtsc2();
    fprintf(stderr,"ovl send %lld %lld %lld [clk] , copy %lld [clk] \n", wrap1-start,
	    wrap2-wrap1, copy-wrap2, end-copy);
}

void send_i_matrix_overlap_inoz_k2k_mc(int DIMM,
				    int DIMK,
				    int nc,
				    int i,
				    int NOTA,
				    int LDA,
				    double *a,
				    double *aem[nc])
{
    double data[16][64][4];
    int ipe, ibb;
    int ic;
    UINT64 start, copy, end;
    UINT64 wrap1, wrap2;
    int j, nbb=NBB;
    //    fprintf(stderr, "send_i_matrix_overlap_inoz_k2k_mc, i=%d\n",i);
	
    start = rdtsc2();
    for(ic=0;ic<nc;ic++)setsend_i_elt2_16bbMC(s[ic], 0);
    wrap1 = rdtsc2();
    for(ic=0;ic<nc;ic++)matmul_move_elt_data2_m1024_npe(s[ic], 0, 0, 32);
    wrap2 = rdtsc2();
    copy = rdtsc2();
    while(rdtsc2()< wrap2 + 10000){}
    for(ic=0;ic<nc;ic++)matmul_move_elt_data2_m1024_npe_finish(s[ic]);
    xwait(10000);// this wait seems necessary here -- JM 2010/5/4
                  // might need larger value for clocks below 400MHz
    end = rdtsc2();
    //     fprintf(stderr,"ovl send %lld %lld %lld [clk] , copy %lld [clk] \n",
    //          wrap1-start,  wrap2-wrap1, copy-wrap2, end-copy);
}
void move_i_matrix_from_em(int firstchip, int nc)
{
    int ipe, ibb;
    int ic;
    UINT64 start, copy, end;
    UINT64 wrap1, wrap2;
    int j, nbb=NBB;
    //    fprintf(stderr, "send_i_matrix_overlap_inoz_k2k_mc, i=%d\n",i);
	
    start = rdtsc2();
    for(ic=0;ic<nc;ic++)setsend_i_elt2_16bbMC(s[ic+firstchip], 0);
    wrap1 = rdtsc2();
    for(ic=0;ic<nc;ic++){
	matmul_move_elt_data2_m1024_npe(s[ic+firstchip], 0, 0, 32);
    }
    wrap2 = rdtsc2();
    copy = rdtsc2();
    while(rdtsc2()< wrap2 + 20000){}
    for(ic=0;ic<nc;ic++){
	matmul_move_elt_data2_m1024_npe_finish(s[ic+firstchip]);
    }
    xwait(10000);// this wait seems necessary here -- JM 2010/5/4
                  // might need larger value for clocks below 400MHz
    end = rdtsc2();
    //     fprintf(stderr,"ovl send %lld %lld %lld [clk] , copy %lld [clk] \n",
    //          wrap1-start,  wrap2-wrap1, copy-wrap2, end-copy);
}

void send_i_matrix_mt(int DIMM, int DIMK, int i, int NOTA, int LDA, double *a, double aem[NCHIP][NPE][4][NBB][64], int overlap)
{
  int ichip;
  void *p[NCHIP];

  for(ichip=0;ichip<NCHIP;ichip++){
    gdr_send_i_data[ichip].DIMM = DIMM;
    gdr_send_i_data[ichip].DIMK = DIMK;
    gdr_send_i_data[ichip].i    = i;
    gdr_send_i_data[ichip].ichip = ichip;
    gdr_send_i_data[ichip].NOTA = NOTA;
    gdr_send_i_data[ichip].LDA  = LDA;
    gdr_send_i_data[ichip].a    = a;
    gdr_send_i_data[ichip].aem  = (double*)aem[ichip];
    gdr_send_i_data[ichip].overlap = overlap;
    p[ichip] = &(gdr_send_i_data[ichip]);
  }
  gdrrun_threads(send_i_matrix_mtwork, p);
}


struct gdr_send_j_param
{
  int LDB;
  int NOTB;
  double *b;
  int ichip;
  int DIMN;
  int DIMK;
} gdr_send_j_data[NTHREAD];

void send_j_matrix_mtwork(void *ptr)
{
  int m, i, size;
  double br0[NBB][32];
  double br1[NBB][32];
  //UINT64 start, copy, send;
  double start, copy, send;
  
  struct gdr_send_j_param *p = (struct gdr_send_j_param *)ptr;
  int LDB   = p->LDB;
  int NOTB  = p->NOTB;
  double (*b)[LDB];
  b = (double(*)[LDB])p->b;
  int ichip = p->ichip;
  int DIMN  = p->DIMN;
  int DIMK  = p->DIMK;


  m=0;
#ifdef SAFEEMSEND
  size = NBB; //Failsafe
#else
  size = NBB*DIMK/2048; //DIMK==2048 => size=16 , DIMK==1024 => size=8, DIMK==512 => size=4
#endif

  start = wallclock(); //rdtsc2();
  for(i=0;i< DIMN ; i+= (NPE*2)){ //Asuumed DIMN%(NPE*2)==0
    int j;
    for(j=0;j<(NPE*2); j+= 2){
      int k = i + j;
      int l;
  
      copytomatb_tb4_dg(LDB, NOTB, b, ichip, k, DIMK, br0);
      copytomatb_tb4_dg(LDB, NOTB, b, ichip, k+1, DIMK, br1);
      copy = rdtsc2();
      matmul_send_elt_data0MC_m(s[ichip], (struct matmul_elt_struct0 *)(br0),m, size);
      matmul_send_elt_data1MC_m(s[ichip], (struct matmul_elt_struct1 *)(br1),m, size);


      m+=NBB;
    }
  }
  send = wallclock(); //rdtsc2();  
#ifdef BENCHSENDJ
  //printf("send j matrix : copy=%lld send=%lld \n", copy-start, send-copy);
  printf("send j matrix : send=%lf \n", send-start);
#endif
}



#define SENDJNROLL 1 // used to be 8, 1 is for testing

inline void prefetch_vector(double *x, int len)
{
    int i;
    for (i=0;i<len;i+= 4){
	__builtin_prefetch(x+i, 0, 0);
	// here, 1, 0 gives slightly better performance than 0,0
	// for small matrix, but not for very large ones?
    }
}
	
	

void send_j_matrix_st(int firstchip, int bcast, int DIMN, int DIMK,
		      int LDB, int NOTB, double *b)
{
  int m, i, size;
  double br0[NCHIP][NBB*SENDJNROLL][32];
  double br1[NCHIP][NBB*SENDJNROLL][32];
  
  UINT64  t0,t1, t2, t3,t1a, tcopy, tcopya,tdma, start,  send;;
  
  int ichip;
  m=0;
  int nchips = bcast? matmul_getnchips():NCHIP;
  //  fprintf(stderr,"send_j nchips=%d bcast=%d\n", nchips, bcast);
  size = NBB*DIMK/2048; //DIMK==2048 => size=16 ,
                        //DIMK==1024 => size=8, DIMK==512 => size=4

  //  int ndmablock = 32/(DIMK/512);
  //  int ndmablock = 32768/DIMK;
  //  if (DIMK == 256) ndmablock /= 2;
 int ndmablock = 16384/DIMK;
  //  int ndmablock = 64/(DIMK/512); // size increased. Should be okay.
  
 // ndmablock = 14;// testing -- JM
  //  start = rdtsc2();

  //  fprintf(stderr,"send_j_matrix_st: ndmablock=%d\n", ndmablock);

  SING_EM_set_areaid_multichip(nchips, s+firstchip, 0);
  SING_EM_set_skip_areaid(1);

  CTRL_write_set_dmar_buffered_mode(1);
  tcopy = tcopya= tdma=0;

  UINT64 lldb = LDB;
  int dmar_started = 0;
  int dmar_bcastmode = bcast? NCHIP:0;
  for(i=0;i< DIMN ; i+= (NPE*2)){ //Asuumed DIMN%(NPE*2)==0
      int j;
      //      fprintf(stderr,"send_j_matrix_st: i=%d\n",i);
      
      for(j=0;j<(NPE*2); j+= 2){
	  int k = i + j;
	  //	  t0 = rdtsc2();
	  if (!NOTB){
	      //	  fprintf(stderr,"send_j_matrix_st: j=%d\n",j);
	      for(ichip=0;ichip<NCHIP;ichip++){
		  copytomatb_tb4_dg(LDB, NOTB, (double(*)[LDB])b, ichip,
				    k, DIMK,
				    (double(*)[32])(((double*)br0[ichip])));
	      }
	      
	      for(ichip=0;ichip<NCHIP;ichip++){
		  copytomatb_tb4_dg(LDB, NOTB, (double(*)[LDB])b, ichip,
				    k+1, DIMK,
				    (double(*)[32])(((double*)br1[ichip])));
	      }
	  }
	  //	  t1a = rdtsc2();
	  //	  tcopya += t1a-t0;
	  
	  int increment = NBB*32;
	  int sendsize = size;
	  //	  fprintf(stderr,"send_j_matrix_st:call send_elt_data0mc\n");
	  if(NOTB){
	      increment = DIMK/4;
	      prefetch_vector(b+lldb*(k),DIMK);
	      matmul_send_elt_data0MC_m_singlethread(NCHIP,s+firstchip,
						     (struct
						      matmul_elt_struct0 *)
						     (b+lldb*k),m,
						     sendsize,increment);
	      prefetch_vector(b+lldb*(k+1),DIMK);
	      matmul_send_elt_data1MC_m_singlethread(NCHIP,s+firstchip,
						     (struct
						      matmul_elt_struct1 *)
						     (b+lldb*(k+1)),m,
						     sendsize,increment);
	  }else{
	      matmul_send_elt_data0MC_m_singlethread(NCHIP,s+firstchip,
						     (struct
						      matmul_elt_struct0 *)
						     (br0), m,
						     sendsize,increment);
	      
	      matmul_send_elt_data1MC_m_singlethread(NCHIP,s+firstchip,
						     (struct
						      matmul_elt_struct1 *)
						     (br1),m,
						     sendsize,increment);
	  }
	  m+=NBB;
	  //	  t1 = rdtsc2();
	  //	  tcopy += t1-t1a;
	  dmar_started = 0;
	  if ((j % ndmablock) == (ndmablock-2)){
	      for(ichip=firstchip;ichip<firstchip+nchips;ichip++){
		  hib_start2_dmarMC_multiplane_bcast(s[ichip]->c->hib_id,dmar_bcastmode);
	      }
	      //	      SING_DMAR_wait(nchips, s+firstchip);
	      dmar_started = 1;
	  }
	  //	  t2 = rdtsc2();
	  //	  tdma += t2-t1;
      }
  }
  if (dmar_started == 0){
      for(ichip=firstchip;ichip<firstchip+nchips;ichip++){
	  hib_start2_dmarMC_multiplane_bcast(s[ichip]->c->hib_id,dmar_bcastmode);
      }
  }      
  for(ichip=firstchip;ichip<firstchip+nchips;ichip++){
      hib_clear_dmarMC_multiplane(s[ichip]->c->hib_id);
  }
  //  tdma += rdtsc2()-t2;
  
  
  SING_DMAR_wait(nchips, s+firstchip);
  //  send = rdtsc2();
  //  fprintf(stderr,"send j matrix : fc=%d tcopy=%ld %ld tdma=%ld total=%ld\n",
  //	  firstchip, tcopya, tcopy, tdma, send-start);
#ifdef VERBOSELOG  
  //  fprintf(stderr,"send j matrix : fc=%d tcopy=%ld tdma=%ld total=%ld\n",
  //	  firstchip, tcopy, tdma, send-start);
#endif  
#ifdef BENCHSENDJ
  //printf("send j matrix : copy=%lld send=%lld \n", copy-start, send-copy);
  //  printf("send j matrix : send=%lf \n", send-start);
#endif
}

void send_j_matrix_st_old(int firstchip, int bcast, int DIMN, int DIMK,
		      int LDB, int NOTB, double *b)
{
  int m, i, size;
  double br0[NCHIP][NBB*SENDJNROLL][32];
  double br1[NCHIP][NBB*SENDJNROLL][32];
  
  UINT64  t0,t1, t2, t3, tcopy, tdma, start,  send;;
  
  int ichip;
  m=0;
  int nchips = bcast? matmul_getnchips():NCHIP;
  //  fprintf(stderr,"send_j nchips=%d bcast=%d\n", nchips, bcast);
  size = NBB*DIMK/2048; //DIMK==2048 => size=16 ,
                        //DIMK==1024 => size=8, DIMK==512 => size=4

  int ndmablock = 32/(DIMK/512);
  
  //  size = NBB; ndmablock = 8;// testing -- JM
  //  ndmablock=14;
  start = rdtsc2();

  //  fprintf(stderr,"send_j_matrix_st: ndmablock=%d\n", ndmablock);

  SING_EM_set_areaid_multichip(nchips, s+firstchip, 0);
  SING_EM_set_skip_areaid(1);

  CTRL_write_set_dmar_buffered_mode(1);
  tcopy = tdma=0;
  for(i=0;i< DIMN ; i+= (NPE*2)){ //Asuumed DIMN%(NPE*2)==0
      int j;
      //      fprintf(stderr,"send_j_matrix_st: i=%d\n",i);
      
      for(j=0;j<(NPE*2); j+= 2){
	  int k = i + j;
	  t0 = rdtsc2();
	  //	  fprintf(stderr,"send_j_matrix_st: j=%d\n",j);
	  for(ichip=0;ichip<NCHIP;ichip++){
	      copytomatb_tb4_dg(LDB, NOTB, (double(*)[LDB])b, ichip,
				k, DIMK,
				(double(*)[32])(((double*)br0[ichip])));
	      copytomatb_tb4_dg(LDB, NOTB, (double(*)[LDB])b, ichip,
				k+1, DIMK,
				(double(*)[32])(((double*)br1[ichip])));
	  }
	  
	  
	  int increment = NBB*32;
	  int sendsize = size;
	  //	  fprintf(stderr,"send_j_matrix_st:call send_elt_data0mc\n");
	  matmul_send_elt_data0MC_m_singlethread(NCHIP,s+firstchip,
						 (struct matmul_elt_struct0 *)
						 (br0),m, sendsize,increment);
	  
	  matmul_send_elt_data1MC_m_singlethread(NCHIP,s+firstchip,
						 (struct matmul_elt_struct1 *)
						 (br1),m, sendsize,increment);
	  m+=NBB;
	  t1 = rdtsc2();
	  tcopy += t1-t0;
	  int dmar_bcastmode = bcast? NCHIP:0;
	  if ((j % ndmablock) == (ndmablock-2)){
	      for(ichip=firstchip;ichip<firstchip+nchips;ichip++){
		  hib_start2_dmarMC_multiplane_bcast(s[ichip]->c->hib_id,dmar_bcastmode);
	      }
	  }
	  t2 = rdtsc2();
	  tdma += t2-t1;
      }
  }
  for(ichip=firstchip;ichip<firstchip+nchips;ichip++){
      hib_clear_dmarMC_multiplane(s[ichip]->c->hib_id);
  }
  tdma += rdtsc2()-t2;
  
  
  SING_DMAR_wait(nchips, s+firstchip);
  send = rdtsc2();
#ifdef VERBOSELOG  
  fprintf(stderr,"send j matrix : fc=%d tcopy=%ld tdma=%ld total=%ld\n",
	  firstchip, tcopy, tdma, send-start);
#endif  
#ifdef BENCHSENDJ
  //printf("send j matrix : copy=%lld send=%lld \n", copy-start, send-copy);
  printf("send j matrix : send=%lf \n", send-start);
#endif
}


void send_j_matrix_mt(int DIMN, int DIMK, int LDB, int NOTB, double *b)
{
  int ichip;
  void *p[NCHIP];

  for(ichip=0;ichip<NCHIP;ichip++){
    gdr_send_j_data[ichip].LDB  = LDB;
    gdr_send_j_data[ichip].NOTB = NOTB;
    gdr_send_j_data[ichip].b    = b;
    gdr_send_j_data[ichip].ichip = ichip;
    gdr_send_j_data[ichip].DIMN = DIMN;
    gdr_send_j_data[ichip].DIMK = DIMK;
    p[ichip] = &(gdr_send_j_data[ichip]);
    //    send_j_matrix_mtwork(p[ichip]); //for debug
  }

  gdrrun_threads(send_j_matrix_mtwork, p);
}

static int skip_sending_jmatrix = 0;
static int use_multithreads = 0;

void gdrsetskipsendjmat(){skip_sending_jmatrix = 1;}
void gdrresetskipsendjmat(){skip_sending_jmatrix = 0;}
void gdrsetusemultithread(){use_multithreads = 1;}
void gdrresetusemultithread(){use_multithreads = 0;}


static void fix_bb0(int DIMK, int NOTA,  int LDA,  double a[][LDA],
		    int i)
{
    int ichip;
#pragma omp parallel for private(ichip)
    for(ichip=0;ichip<NCHIP;ichip++){
	int ipe,ibb;
	ibb = 0;
	double data[256];
	set_bbid_for_send_imatrixMC(s[ichip],  0);
	for(ipe=1;ipe < NPE; ipe+=2){
	    copytomata_tb4_dg(LDA, NOTA, a, i, ichip,
			      ibb, ipe, DIMK,
			      (double (*)[64][4]) data);
	    send_i_matrix_singlebbMC(s[ichip], data, ipe,ibb);
	}
	set_bbid_for_send_imatrixMC(s[ichip],  -1);
    }
}

static int overlapa=0;

void matmul_tb4_sub(int DIMM, int DIMN, int DIMK,
		    int NOTA, int ctrans, int abcast,
		    int LDA,  int LDC, 
		    double a[][LDA],
		    double c[][LDC], double alpha, double beta,
		    double *isend, double * calc)
{
    int ipe, ibb, i, m,n, ichip, j, k, sent128, sendsize;
    static double data[16][64][4], aem[NCHIP][NPE][4][NBB][64];
    static double aem2[NCHIP][NPE][4][NBB][64];
    double *p, *out, *next_a, start;
    double  currt, gflops, *aptr[NCHIP];
    int prev_dimk = 4096;
    int nloop = 64;
    int nsend = nloop*4;
    int sendc = nsend*32;
    int nc = abcast? (matmul_getnchips()):NCHIP;
    start = wallclock();
    //    fprintf(stderr,"nloop = %d\n", nloop);
    if (prev_dimk > DIMK){
	// zero-fill aem;
	int len = NCHIP*NPE*4*NBB*64;
	double * ap = (double*) aem;
	for(i=0;i<len;i++)ap[i]=0.0;
    }

    //   fprintf(stderr,"ctrans=%d\n", ctrans);
    
    overlapa=0;
    
    next_a = (double*)a;
    int di = DIM1;
    if ((abcast==0) && (matmul_getnboards()>1)) di = DIM1 * matmul_getnboards();
    for(i=0;i<DIMM;i+=di){ 
	
	//	fprintf(stderr,"testmatmul i=%d\n",i);

	currt = wallclock();
	int j;
	
	if (abcast || matmul_getnboards()==1){
	    //fprintf(stderr,"sendoneamat abcast mode\n");
	    sendoneamat_initialize(s,0, abcast);
	    sendoneamat(s, 0, abcast,
			(double(*)[NPE*NBB*256]) aem,
			DIMK,i,  NOTA, LDA, a);
	}else if(matmul_getnboards()==2){
	    sendoneamat_initialize(s,0, abcast);
	    sendoneamat_initialize(s,NCHIP, abcast);

#pragma omp parallel
#pragma omp sections
	    {
#pragma omp section
		sendoneamat(s, 0, abcast,
			    (double(*)[NPE*NBB*256]) aem,
			    DIMK,i,  NOTA, LDA, a);
#pragma omp section
		sendoneamat(s, NCHIP, abcast,
			    (double(*)[NPE*NBB*256]) aem2,
			    DIMK,i+DIM1,  NOTA, LDA, a);
	    }
	}else{
	    fprintf(stderr,"sendoneamat: nboards=%d not implemented yet\n",
		    matmul_getnboards());
	}
	sendoneamat_finalize();
	move_i_matrix_from_em(0, matmul_getnchips());
	double disend = wallclock()-currt;
	*isend += disend;
	//printf("matrix i send..\n");
	currt = wallclock();
	n = DIMN / 2; //one loop contains b0 and b1 for double buffering.
	out = ctrans? ((double*)(&c[i][0])):( (double*)c + i);

	// multi-board version under development --- code
	// changed up to here -- JM 2010/5/19

	//	fprintf(stderr,"omp_max_threads=%d procs=%d\n",
	//		omp_get_max_threads(),omp_get_num_procs());
	
	if (matmul_getnboards()==1){
	    
	    matmul_type2_runMC_tb4_2_ovl_dg(s, NCHIP, n, out, LDC,
					    (double(*)[NPE*NBB*256])aem,
					    alpha, beta, ctrans, 
					    DIMK,i+DIM1, nloop,
					    NOTA, LDA, a,overlapa);
	}else if (matmul_getnboards()==2){
	    int ngdr = abcast? n/2 : n;
	    double * out2;
	    if (ctrans){
		out2 = abcast? (&(c[n][i])) : (out+DIM1);// this must be wrong...
	    }else{
		out2 = abcast? (&(c[n][i])) : (out+DIM1);
	    }
	    
#pragma omp parallel
#pragma omp sections
	    {
#pragma omp section
		matmul_type2_runMC_tb4_2_ovl_dg(s, NCHIP, ngdr, out, LDC,
						(double(*)[NPE*NBB*256])aem,
						alpha, beta, ctrans, 
						DIMK,i+DIM1, nloop,
						NOTA, LDA, a,overlapa);
#pragma omp section
		matmul_type2_runMC_tb4_2_ovl_dg(s+NCHIP, NCHIP, ngdr, out2, LDC,
						(double(*)[NPE*NBB*256])aem,
						alpha, beta, ctrans, 
						DIMK,i+DIM1, nloop,
						NOTA, LDA, a,overlapa);
	    }
	    
	}else{
	    fprintf(stderr,"matmul_...t: nboards=%d not implemented yet\n",
		    matmul_getnboards());
	}
    //	fprintf(stderr,"matmul_type2... end\n");
    
	*calc += wallclock()-currt;
	//	fprintf(stderr,"loop end\n");
    }
    
}
void matmul_tb4_mnk_ovl_dg(int DIMM, int DIMN, int DIMK,
			   int NOTA, int NOTB, int ctrans, int bbcast,
			   int LDA, int LDB, int LDC, 
			   double a[][LDA],
			   double b[][LDB],
			   double c[][LDC], double alpha, double beta)
{
    int ipe, ibb, i, m,n, ichip, j, k, sent128, sendsize;
    static double data[16][64][4], aem[NCHIP][NPE][4][NBB][64];
    double *p, *out, *next_a, start;
    double isend=0.0, jsend=0.0, calc=0.0, currt, gflops, *aptr[NCHIP];
    int prev_dimk = 4096;
    int nloop = 32;
    int nsend = nloop*4;
    int sendc = nsend*32;
    
    start = wallclock();

    if (prev_dimk > DIMK){
	// zero-fill aem;
	int len = NCHIP*NPE*4*NBB*64;
	double * ap = (double*) aem;
	for(i=0;i<len;i++)ap[i]=0.0;
    }
    
    //    fprintf(stderr,"matmul_tb4_mnk_ovl_dg MNK= %d %d %d\n",DIMM, DIMN, DIMK);

    if (skip_sending_jmatrix ==0){
	//	fprintf(stderr,"send_j_matrix_st called with bbcast=%d\n", bbcast);
	if (bbcast || matmul_getnboards()==1){
	    // send same b to all boards
	    //	    fprintf(stderr,"send_jmat bcast/1card %d %d\n",
	    //		    bbcast, matmul_getnboards());
	    send_j_matrix_st(0, bbcast, DIMN, DIMK, LDB, NOTB, (double*)b);
	    double jsend1 = wallclock();
	    //	    fprintf(stderr, "send_j_mat time = %e\n", jsend1-start);
	}else if(matmul_getnboards()==2){ 
	    // send part of B to multiple boards
	    // pragma for OMP is necessary here
	    //	    fprintf(stderr,"send_jmat nobcast 1st\n");
	    double jsend1, jsend2;
#pragma omp parallel
#pragma omp sections
	    {
#pragma omp section
		{
		    send_j_matrix_st(0, 0, DIMN/2, DIMK, LDB, NOTB,
				     (double*)b);
		    jsend1 = wallclock();
		}
#pragma omp section
		{
		    if (NOTB){
			//fprintf(stderr,"send_jmat nobcast 2nd\n");
			send_j_matrix_st(NCHIP, 0, DIMN/2, DIMK, LDB, NOTB,
					 (double*)(&b[DIMN/2][0]));
		    }else{
			//fprintf(stderr,"send_jmat nobcast 2nd\n");
			send_j_matrix_st(NCHIP, 0, DIMN/2, DIMK, LDB, NOTB,
					 (double*)(&b[0][DIMN/2]));
		    }
		    jsend2 = wallclock();
		}
	    }
	    //	    fprintf(stderr, "send_j_mat times = %e %e\n", jsend1-start,
	    //		    jsend2-start);
	}else{
	    fprintf(stderr,"send_j_matrix_st: nboards=%d not implemented yet\n",
		    matmul_getnboards());
	}
	CTRL_write_set_dmar_buffered_mode(0);
	SING_EM_set_skip_areaid(0);
    }
    
    jsend = wallclock();
    overlapa=0;
    int abcast = 0;
    if (bbcast==0 && matmul_getnboards()>1) abcast = 1;
    matmul_tb4_sub(DIMM, DIMN, DIMK, NOTA, ctrans, abcast, LDA, LDC, 
		   a, c, alpha, beta, &isend, &calc);
    gflops = ((double)DIMM*(double)DIMN*(double)DIMK*2.0 / (isend+jsend+calc-start))/1e9;
#ifdef VERBOSELOG
    if (get_matmul_msg_level()){
	fprintf(stderr,"%s M=%d N=%d K=%d ip=%.4e jp=%.4e fo=%.4e all=%.4e %.2lf GFlops \n",
		myhostname,DIMM, DIMN, DIMK, isend,  jsend-start, calc,
		(isend+jsend+calc-start), gflops);
    }
#endif
}


void matmul_tb4_mnk_dg(int DIMM, int DIMN, int DIMK, int NOTA, int NOTB, int LDA, int LDB, int LDC, 
			   double a[][LDA],
			   double b[][LDB],
			   double c[][LDC], double alpha, double beta)
{
  int ipe, ibb, i, m,n, ichip, j, k, sent128, sendsize;
  static double data[16][64][4], aem[NCHIP][NPE][4][NBB][64];
  double *p, *out, *next_a, start, isend=0.0, jsend=0.0, calc=0.0, currt, gflops, *aptr[NCHIP];
  int sendc = SENDCHUNK*32;

  
  start = wallclock();

  send_j_matrix_mt(DIMN, DIMK, LDB, NOTB, (double*)b);

  jsend = wallclock();

  next_a = (double*) a;
  for(i=0;i<DIMM;i+=DIM1){ // Assumed i%DIM1==0
    currt = wallclock();

    send_i_matrix_mt(DIMM, DIMK, i, NOTA, LDA, (double*)a, aem, 0);

    isend += wallclock()-currt;
    //printf("matrix i send..\n");
    currt = wallclock();
    n = DIMN / 2; //one loop contains b0 and b1 for double buffering.
    out = (double*)c + i;

    sent128 = 0;
    matmul_type2_runMC_tb4_2dg(s, NCHIP, n, out, LDC, alpha, beta);

    calc += wallclock()-currt;
  }

  gflops = ((double)DIMM*(double)DIMN*(double)DIMK*2.0 / (isend+jsend+calc-start))/1e9;
#ifdef VERBOSELOG
  fprintf(stderr,"M=%d N=%d K=%d ip=%e jp=%e fo=%e all=%e %.2lf GFlops \n", DIMM, DIMN, DIMK, isend, jsend-start, calc, (isend+jsend+calc-start), gflops);
#endif
}

static int matmul_stress_factor = 0;

void gdrdgemm_set_stress_factor(int factor)
{
    matmul_stress_factor = factor;
}

void matmul_tb4_mnk_ovl_dg_try(int DIMM, int DIMN, int DIMK,
			   int NOTA, int NOTB, int ctrans, int bbcast,
			   int LDA, int LDB, int LDC, 
			   double a[][LDA],
			   double b[][LDB],
			   double c[][LDC], double alpha, double beta)
{
    int i;
    for(i=matmul_stress_factor; i>= 0; i--){
	gdrdgemm_set_skip_cmat_mode(i);
	matmul_tb4_mnk_ovl_dg(DIMM, DIMN, DIMK, NOTA, NOTB, ctrans, bbcast,
			      LDA, LDB, LDC, a,b,c, alpha,  beta);
    }
    
}
    

#define MLOOP 32

#ifdef TESTMATMUL

int main()
{
  #if 0
  static double a[DIM1*MLOOP][DIM0];  
  static double b[DIM1*MLOOP][DIM0];  //M=2048, N=8192
  static double c[DIM1*MLOOP][DIM1*MLOOP];
  static double c2[DIM1*MLOOP][DIM1*MLOOP];
  #else
  double *a, *b, *c, *c2;
  a = malloc(DIM1*MLOOP*DIM0*sizeof(double));
  b = malloc(DIM1*MLOOP*DIM0*sizeof(double));
  c = malloc(DIM1*MLOOP*DIM1*MLOOP*sizeof(double));
  c2 = malloc(DIM1*MLOOP*DIM1*MLOOP*sizeof(double));
  #endif
    double start, end, gdr, host;
    int i,j,k;

    readmat(a);
    readmat(b);
    for(i=0;i<DIM1;i++){
      for(j=0;j<DIM0;j++){
	for(k=1;k<MLOOP;k++){
#if 0
	  a[i+DIM1*k][j] = a[i][j];
#else
	  a[(i+DIM1*k)*DIM0 + j] = a[i*DIM0 + j];
#endif

	}
	for(k=1;k<MLOOP;k++){
#if 0
	  b[i+DIM1*k][j] = b[i][j];
#else 
	  b[(i+DIM1*k)*DIM0 + j] = b[i*DIM0 + j];
#endif
	}
      }
    }
    //printmat(a);
    //printmat(b);
    grape_init();
    start = wallclock();
    
    printf("check with GDR <-> GDR \n");
    for(i=0;i<10;i++){

	matmul_tb4_mnk_ovl_dg(DIM1*MLOOP, DIM1*MLOOP, DIM0, 0, 1, 0,
			      DIM0, DIM0, DIM1*MLOOP, a, b, c, 1.0, 0.0);
#if 1
	matmul_tb4_mnk_ovl_dg(DIM1*MLOOP, DIM1*MLOOP, DIM0, 0, 1, 0,
			      DIM0, DIM0, DIM1*MLOOP, a, b, c2, 1.0, 0.0);
      printcmatdiff_mn(DIM1*MLOOP, DIM1*MLOOP, c,c2);
#endif
    }
    gdr = wallclock() - start;
    //printcmat(c);
#if 1
    printf("check with GDR <-> Host \n");
    start = wallclock();
    for(i=0;i<1;i++){
      /*
      matmul_host(a,b,c2);
      */
      matmul_host_mnk(DIM1*MLOOP, DIM1*MLOOP, DIM0, a,b,c2);
    }
    //printcmat(c2);
    host = wallclock() - start;

    printcmatdiff_mn(DIM1*MLOOP, DIM1*MLOOP, c,c2);
#endif
    printf("elapse time [s] gdr=%e , host=%e \n", gdr, host);
    grape_close();
}

#endif
