//
// simgrape6.c
//
// GRAPE-6 compatibility interface for SING implementation of hermite
// pipeline
//
//
// J. Makino 2005/4/23
//
// do not support cluster_id (should be 0)
//g6_change_cbjpmode
//g6_change_clock
//g6_changed_index_
//g6_clock_state
//g6_close
//g6_dummywait
//g6_flush_jp_buffer_and_multisend
//g6_initialize_jp_buffer
//g6_npipes
//g6_open
//g6_print_chip_status
//g6_reset
//g6_reset_fofpga
//g6_set_ijp_mode
//g6_set_j_particle
//g6_set_j_particle_multisend
//g6_set_ti
//g6_set_xunit
//g6_setmccount
//g6_setup_njdata

#include <stdio.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "hermite-16x32.h"

#define NCHIP 4
#define LOCKFILE "/tmp/gdrlock.lock"

void g6_setup_njdata(int clusterid, int nbody)
{}
void g6_dummywait()
{}
void g6_setmccount(int dummy)
{}

    
void g6_set_ijp_mode(int clusterid,
		     int mode) /* mode=0: IP, 1: JP */
{
    // dummy
}
void g6_set_ijp_mode_(int * clusterid,
		      int * mode) /* mode=0: IP, 1: JP */
{
    g6_set_ijp_mode(*clusterid,* mode);
}




void g6_reset(int  clusterid)
{
    //dummy
}

void g6_reset_fofpga(int  clusterid)
{
    //dummy
}

	
void g6_change_cbjpmode_(unsigned int *	 clusterid,
			unsigned int *mode /*0: BCAST, 1: MCAST */)
{
    //dummy
}
void g6_change_cbjpmode(unsigned int 	 clusterid,
			unsigned int mode /*0: BCAST, 1: MCAST */)
{
    //dummy
}
	


void g6_change_clock(int clusterid, int mode)
     /* mode:0 for sleep, 1 for activate */
{
    //dummy
}

static int xunit = 51;

void g6_set_xunit_(int * newxunit)
{
  xunit = *newxunit;
}

void g6_set_xunit(int newxunit)
{
    g6_set_xunit_(&newxunit);
}

void g6_set_tunit_(int * newxunit)
{
}

void g6_set_tunit(int newxunit)
{
}

int g6_reinitialize(int clusterid)
{
  return 0;
}

static int opened=0;

int g6open(int clusterid)
{
  int fd;
  fprintf(stderr,"g6open called\n");
  if(opened == 0){
    fd = open(LOCKFILE, O_WRONLY|O_CREAT);
    flock(fd, LOCK_EX);
    SING_grape_init();
    opened = 1;
    close(fd);
  }
}


int g6_open_(int *clusterid)
{
  return g6open(*clusterid);
}

int g6_open(int clusterid)
{
  return g6open(clusterid);
}

int g6close(int clusterid)
{
  // might need something here as well
  int fd;

  if(opened ==1){
    fd = open(LOCKFILE, O_WRONLY|O_CREAT);
    flock(fd, LOCK_EX);
    SING_grape_close();
    opened = 0;
    close(fd);
  }
  return 0;
}


int g6_close_(int *clusterid)
{
    return g6close(*clusterid);
}

int g6_close(int clusterid)
{
    return g6close(clusterid);
}


static double ti_set;
void g6_set_ti_(int *clusterid, double *ti)
{
    //    fprintf(stderr,"g6_set_ti ti = %g\n", *ti);
    ti_set = *ti;
}

void g6_set_ti(int clusterid, double ti)
{
    g6_set_ti_(&clusterid, &ti);
}


int g6_initialize_jp_buffer(int clusterid, int size)
{
    // dummy
  return 0;
}
	
int g6_initialize_jp_buffer_(int* clusterid, int* size) 
{
    return g6_initialize_jp_buffer(*clusterid, *size);
}



int g6_flush_jp_buffer_and_multisend(int  clusterid, int nclusters)
{
    // dummy
    return 0;
}



int g6_set_j_particle_multisend_(int * clusterid,
				 int * nclusters,
				 int *address,
				 int *index,
				 double *tj, /* particle time */
				 double *dtj, /* particle time */
				 double *mass,
				 double a2by18[3], /* a2dot divided by 18 */
				 double a1by6[3], /* a1dot divided by 6 */
				 double aby2[3], /* a divided by 2 */
				 double v[3], /* velocity */
				 double x[3] /* position */)
{
    struct grape_j_particle_struct xj;
    xj.tj= *tj;
    xj.x0=x[0]; 
    xj.y0=x[1];
    xj.z0=x[2];
    xj.v0x= v[0];
    xj.v0y= v[1];
    xj.v0z= v[2];
    xj.ax2= aby2[0];
    xj.ay2= aby2[1];
    xj.az2= aby2[2];
    xj.jx6=a1by6[0];
    xj.jy6=a1by6[1];
    xj.jz6=a1by6[2];
    xj.sx18=a2by18[0];
    xj.sy18=a2by18[1];
    xj.sz18=a2by18[2];
    xj.mj= *mass;
    xj.eps2=0;
    xj.idxj=*index;
    SING_send_j_particle(&xj, *address);
    return 0;
}

int g6_set_j_particle_(int * clusterid,
		       int *address,
		       int *index,
		       double *tj, /* particle time */
		       double *dtj, /* particle time */
		       double *mass,
		       double a2by18[3], /* a2dot divided by 18 */
		       double a1by6[3], /* a1dot divided by 6 */
		       double aby2[3], /* a divided by 2 */
		       double v[3], /* velocity */
		       double x[3] /* position */)
{
    int one = 1;
    return g6_set_j_particle_multisend_(clusterid,&one,address,index,tj,dtj,mass,
			      a2by18, a1by6, aby2, v, x);
}


int g6_set_j_particle(int  clusterid,
		       int address,
		       int index,
		       double tj, /* particle time */
		       double dtj, /* particle time */
		       double mass,
		       double a2by18[3], /* a2dot divided by 18 */
		       double a1by6[3], /* a1dot divided by 6 */
		       double aby2[3], /* a divided by 2 */
		       double v[3], /* velocity */
		       double x[3] /* position */)
{
    return g6_set_j_particle_(&clusterid,&address,&index,&tj,&dtj,&mass,
			      a2by18, a1by6, aby2, v, x);
}

int g6_set_j_particle_mxonly(int clusterid, int address, int index, double mass, double x[3])
{
  double a2by18[3] = {0.0, 0.0, 0.0};
  double a1by6[3]  = {0.0, 0.0, 0.0};
  double aby2[3]   = {0.0, 0.0, 0.0};
  double v[3]      = {0.0, 0.0, 0.0};
  return g6_set_j_particle(clusterid, address, index, ti_set, 0.0, mass, a2by18, a1by6, aby2, v, x);
}

int g6_set_j_particle_multisend(int  clusterid,
				int  nclusters,
		       int address,
		       int index,
		       double tj, /* particle time */
		       double dtj, /* particle time */
		       double mass,
		       double a2by18[3], /* a2dot divided by 18 */
		       double a1by6[3], /* a1dot divided by 6 */
		       double aby2[3], /* a divided by 2 */
		       double v[3], /* velocity */
		       double x[3] /* position */)
{
    return g6_set_j_particle_multisend_(&clusterid,&nclusters,&address,&index,&tj,&dtj,&mass,
			      a2by18, a1by6, aby2, v, x);
}

    
int g6_flush_jp_buffer(int cluserid, int size)
{
  return 0;
}
 
int g6_npipes()
{
    return g6_npipes_();
}

int g6_npipes_()
{
    return SING_number_of_pe()*4;
}

int g6_print_chip_status_(int * clusterid)
{
    // dummy
    return 0;
}

int g6_print_chip_status(int clusterid)
{
    return  g6_print_chip_status_(&clusterid);
}

void insert_dummyparticle(int addr)
{
  double x[3] = {1e50, 1e50, 1e50};
  double v[3] = {0.0, 0.0, 0.0};
  double a2by18[3] = {0.0, 0.0, 0.0};
  double a1by6[3]  = {0.0, 0.0, 0.0};
  double aby2[3]   = {0.0, 0.0, 0.0};

  g6_set_j_particle(0, addr, -1, ti_set, 0.0, 0.0, a2by18, a1by6, aby2, v, x);
}
   

void g6calc_firsthalf0_(int * clusterid,
			int * nj,
			int * ni,
			int index[],
			double x[][3],
			double v[][3],
		      double fold[][3],
			double j6old[][3],
			double phiold[],
			double eps2[],
			double h2[],
			int * mode)
     /* mode: 0 --- standard mode, regard both eps2 and h2 as array
	      1 --  GRAPE-4 compatibility mode, eps2 is scalar
	      2 (bit loc 1) skip send_njp... future extension */
// pit uses mode 2....
{
    int niparallel = SING_number_of_pe()*4;
    int njreal = *nj/SING_number_of_bb()/NCHIP;
    int i, d;
    static struct grape_i_particle_struct xi[2048];
    
    d = 64 - ((*nj)%64);
    //printf("dummy : %d \n", d);
    if(d!=64){
      for(i=0;i<d;i++){
	insert_dummyparticle(i+(*nj));
      }
      njreal++;
    }
    for(i=0; i< *ni; i++){
	xi[i].ti= ti_set;
	xi[i].xi=x[i][0];
	xi[i].yi=x[i][1];
	xi[i].zi=x[i][2];
	xi[i].vxi=v[i][0];
	xi[i].vyi=v[i][1];
	xi[i].vzi=v[i][2];
	xi[i].idxi=index[i];
	xi[i].eps2i= (*mode & 1) ? eps2[0]:eps2[i];
    }
    for(i= *ni; i< niparallel; i++) xi[i]=xi[0];

    // h2 ...
    
    SING_send_i_particle(xi, niparallel);
    SING_grape_run(njreal);
}


void g6calc_firsthalf0(int clusterid,
		       int nj,
		       int ni,
		       int index[],
		       double xi[][3],
		       double vi[][3],
		       double fold[][3],
		       double j6old[][3],
		       double phiold[],
		       double eps2[],
		       double h2[],
		       int  mode)
{
     g6calc_firsthalf0_(&clusterid,&nj,&ni,index, xi, vi,fold,j6old,phiold,eps2,
			h2,&mode);
}

void g6calc_firsthalf_(int * clusterid,
		      int * nj,
		      int * ni,
		      int index[],
		      double xi[][3],
		      double vi[][3],
		      double fold[][3],
		      double j6old[][3],
		      double phiold[],
		      double *eps2,
		      double h2[])
{
    int mode = 1;
    g6calc_firsthalf0_(clusterid,nj,ni,index, xi, vi,fold,j6old,phiold,eps2,
			h2,&mode);
}
    


void g6calc_firsthalf(int  clusterid,
		      int  nj,
		      int  ni,
		      int index[],
		      double xi[][3],
		      double vi[][3],
		      double fold[][3],
		      double j6old[][3],
		      double phiold[],
		      double eps2,
		      double h2[])
{
    g6calc_firsthalf_(&clusterid, &nj,&ni,index, xi,vi,fold,j6old,phiold,
		     &eps2, h2);
}

static index_scalechanged = -2;

int g6_changed_index()
{
    return index_scalechanged;
}
int g6_changed_index_()
{
    return index_scalechanged;
}

int g6calc_lasthalf0a_(int * clusterid,
		     int * nj,
		     int * ni,
		     int index[],
		     double xi[][3],
		     double vi[][3],
		     double *eps2,
		     double h2[],
		     double acc[][3],
		     double jerk[][3],
		     double pot[],
		       int nnbindex[],
		     int *mode)
{
    int i,k;
    struct grape_result_struct result[2048];
    SING_get_result(result);
    for (i=0;i< *ni; i++){
	acc[i][0]=result[i].accx;
	acc[i][1]=result[i].accy;
	acc[i][2]=result[i].accz;
	jerk[i][0]=result[i].jerkx;
	jerk[i][1]=result[i].jerky;
	jerk[i][2]=result[i].jerkz;
	pot[i]    =result[i].pot;
	nnbindex[i]=result[i].nnb[1];
    }
    return 0;
}

int g6calc_lasthalf0(int clusterid,
		     int nj,
		     int ni,
		     int index[],
		     double xi[][3],
		     double vi[][3],
		     double *eps2,
		     double h2[],
		     double acc[][3],
		     double jerk[][3],
		     double pot[],
		     int mode)
{
    return   g6calc_lasthalf0_(&clusterid, &nj, &ni, index, xi, vi,eps2,h2,acc, jerk, pot,&mode);
}

int g6calc_lasthalf0a(int clusterid,
		      int nj,
		      int ni,
		      int index[],
		      double xi[][3],
		      double vi[][3],
		      double *eps2,
		      double h2[],
		      double acc[][3],
		      double jerk[][3],
		      double pot[],
		      int nnbindex[],
		      int mode)
{
    return   g6calc_lasthalf0a_(&clusterid, &nj, &ni, index, xi, vi,eps2,h2,acc, jerk, pot,
				nnbindex,&mode);
}


int g6calc_lasthalf2_(int * clusterid,
		     int * nj,
		     int * ni,
		     int index[],
		     double xi[][3],
		     double vi[][3],
		     double *eps2,
		     double h2[],
		     double acc[][3],
		     double jerk[][3],
		     double pot[],
		      int nnbindex[])
{
    int mode = 1;
    return   g6calc_lasthalf0a_(clusterid, nj, ni, index, xi, vi,eps2,h2,acc, jerk, pot,
				nnbindex,&mode);
}

int g6calc_lasthalf0_(int * clusterid,
		     int * nj,
		     int * ni,
		     int index[],
		     double xi[][3],
		     double vi[][3],
		     double *eps2,
		     double h2[],
		     double acc[][3],
		     double jerk[][3],
		     double pot[],
		     int * mode )
{
    int nnbindex[2048];
    return   g6calc_lasthalf0a_(clusterid, nj, ni, index, xi, vi,eps2,h2,acc, jerk, pot,
				nnbindex,mode);
}

int g6calc_lasthalf_(int * clusterid,
		     int * nj,
		     int * ni,
		     int index[],
		     double xi[][3],
		     double vi[][3],
		     double *eps2,
		     double h2[],
		     double acc[][3],
		     double jerk[][3],
		     double pot[])
{
    int mode = 1;
    return   g6calc_lasthalf0_(clusterid, nj, ni, index, xi, vi,eps2,h2,acc, jerk, pot,
				&mode);
}
    
int g6calc_lasthalf2(int  clusterid,
		     int  nj,
		     int  ni,
		     int index[],
		     double xi[][3],
		     double vi[][3],
		     double eps2,
		     double h2[],
		     double acc[][3],
		     double jerk[][3],
		     double pot[],
		    int nnbindex[])
{
    return     g6calc_lasthalf2_(&clusterid, &nj, &ni, index, xi, vi,
			       &eps2, h2, acc, jerk, pot,nnbindex);
}
int g6calc_lasthalf(int  clusterid,
		     int  nj,
		     int  ni,
		     int index[],
		     double xi[][3],
		     double vi[][3],
		     double eps2,
		     double h2[],
		     double acc[][3],
		     double jerk[][3],
		     double pot[])
{
    return     g6calc_lasthalf_(&clusterid, &nj, &ni, index, xi, vi,
			       &eps2, h2, acc, jerk, pot);
}
 


