#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sing_chip.h>

static SINGinfo sing;
int singchip_id = 0;

int SING_init(CTRLinfo *c) 
{
  // SING CP reset
  // CTRL_init(c);
  //CTRL_result_clear(c);
  // need to execute twice for unknown reason
#if SING_TEST
  CTRL_sing_chip_init(c);
  sleep(1);
  CTRL_sing_chip_init(c);
  sleep(1);
#endif
  return 0;
}



void SING_set_chipid(int id)
{
  singchip_id = id;
}

//only id 0 version
int SING_open()
{
    fprintf(stderr,"SING_open chipid = %d\n", singchip_id);
  sing.c = CTRL_open(singchip_id);
#ifdef SING_TEST
  CTRL_beginbcast(sing.c);
#endif
  if (sing.c == NULL) {
    fprintf(stderr, "SING_open: failed at CTRL_open\n");
    return -1;
  }
  if (SING_init(sing.c)) {
    fprintf(stderr, "SING_open: failed at SING_init\n");
    return -1;
  }
  init_ispinfo(&(sing.isp));
  init_eminfo(&(sing.em));
  init_lminfo(&(sing.lm));
  init_bminfo(&(sing.bm));
  //  fprintf(stderr," set reduction mode to 0\n");
  SING_set_reductionmode(&sing, 0);
  return 0;
}

SINGinfo *SING_openMC(int id) 
{
  SINGinfo *s;
  s = malloc(sizeof(SINGinfo));

  s->c = CTRL_open(id);
  if (s->c == NULL) {
    fprintf(stderr, "SING_open: failed at CTRL_open\n");
    return NULL;
  }
  if (SING_init(s->c)) {
    fprintf(stderr, "SING_open: failed at SING_init\n");
    return NULL;
  }
  init_ispinfo(&(s->isp));
  init_eminfo(&(s->em));
  init_lminfo(&(s->lm));
  init_bminfo(&(s->bm));
  //  fprintf(stderr," set reduction mode to 0\n");
  SING_set_reductionmode(s, 0);
  return s;
}

int SING_close()
{
  if (CTRL_close(sing.c)) {
    fprintf(stderr, "SING_close: failed at CTRL_close\n");
    return 0;
  }
  return 0;
}


int SING_closeMC(SINGinfo *s)
{
  if (CTRL_close(s->c)) {
    fprintf(stderr, "SING_close: failed at CTRL_close\n");
    return 0;
  }
  free(s);
  return 0;
}

int SING_number_of_pe() 
{
  return NPE;
}

int SING_number_of_bb() 
{
  return NBB;
}

int SING_number_of_active_bb() 
{
  return NBB;
}

int SING_EM_alloc(int length, const int *conversions, int size, int *err)
{
  return SING_EM_alloc_i(&(sing.em), sing.c, length, conversions, size, err);
}

int SING_EM_allocMC(SINGinfo *s, int length, const int *conversions, int size, int *err)
{
  return SING_EM_alloc_i(&(s->em), s->c, length, conversions, size, err);
}

int SING_EM_send(int areaid, int start, int ndata, void *hostp)
{
  return SING_EM_send_i(&(sing.em), sing.c, areaid, start, ndata, hostp);
}
int SING_EM_vsend(int areaid, int address[], int ndata, void *hostp)
{
  return SING_EM_vsend_i(&(sing.em), sing.c, areaid, address, ndata, hostp);
}

int SING_EM_sendMC(SINGinfo *s, int areaid, int start, int ndata, void *hostp)
{
  return SING_EM_send_i(&(s->em), s->c, areaid, start, ndata, hostp);
}
int SING_EM_vsendMC(SINGinfo *s, int areaid, int address[], int ndata, void *hostp)
{
  return SING_EM_vsend_i(&(s->em), s->c, areaid, address, ndata, hostp);
}


int SING_dmar_buffer_control(int nchips,
			     SINGinfo * s[nchips],
			     int flag)
{
    CTRLinfo *c[nchips];
    int i;
    int ichips;
    for(i=0;i<nchips;i++){
	c[i]=s[i]->c;
    }
    return  CTRL_write_multichip_usingdmar_buffered_control(nchips,c,flag);
}


void SING_DMAR_wait(int nchips, SINGinfo * s[nchips])
{
    int i;


    for(i=0;i<nchips;i++){
	int devid = s[i]->c->hib_id;
	hib_finish_dmarMC(devid);
    }
}
void SING_DMAR_wait_onechip(SINGinfo * s)
{
    int devid = s->c->hib_id;
    hib_finish_dmarMC(devid);
}

int SING_set_dmacheck_mode(int mode)
{
    return hib_set_dmacheck_mode(mode);
}
int SING_get_dmacheck_mode()
{
    return hib_get_dmacheck_mode();
}

void SING_initialize_dmawbufferMC(SINGinfo * s, int size)
{
    hib_initialize_dmawbufferMC(s->c->hib_id, size,
				s->c->h->dmaw_buf);	
}
    
int SING_remaining_words_dmawMC(SINGinfo * s)
{
    return hib_remaining_words_dmawMC(s->c->hib_id);
}
				    
void SING_register_dma_seqMC(SINGinfo *s, int seq_id, int wmode,
			     int size, UINT64 *buf)
{
    hib_register_dma_seqMC(s->c->hib_id,  seq_id, wmode, size, buf);
}


int SING_EM_free()
{
  return SING_EM_free_i(&(sing.em));
}

int SING_EM_freeMC(SINGinfo *s)
{
  return SING_EM_free_i(&(s->em));
}

int SING_EM_areas() 
{
  return SING_EM_areas_i(&(sing.em));
}

int SING_EM_areasMC(SINGinfo *s) 
{
  return SING_EM_areas_i(&(s->em));
}

void SING_EM_status()
{
  SING_EM_status_i(&(sing.em));
}

void SING_EM_statusMC(SINGinfo *s)
{
  SING_EM_status_i(&(s->em));
}

void SING_BM_status()
{
  SING_BM_status_i(&(sing.bm));
}

void SING_BM_statusMC(SINGinfo *s)
{
  SING_BM_status_i(&(s->bm));
}

int SING_LM_write_regconv(int length, int *conversions) 
{
  return SING_LM_write_regconv_i(&(sing.lm), sing.c, length, conversions);
}

int SING_LM_write_regconvMC(SINGinfo *s, int length, int *conversions) 
{
  return SING_LM_write_regconv_i(&(s->lm), s->c, length, conversions);
}

int SING_LM_write(void *hostp, int lmoffset, int peeach)
{
  return SING_LM_write_i(&(sing.lm), &(sing.isp), sing.c, hostp, lmoffset, peeach);
}

int SING_LM_writeMC(SINGinfo *s, void *hostp, int lmoffset, int peeach)
{
  return SING_LM_write_i(&(s->lm), &(s->isp), s->c, hostp, lmoffset, peeach);
}

int SING_iregister(int ninst, const char **insts)
{
  return SING_iregister_i(&(sing.isp), sing.c, ninst, insts);
}

int SING_iregisterMC(SINGinfo *s, int ninst, const char **insts)
{
    return SING_iregister_i(&(s->isp), s->c, ninst, (const char**)insts);
}

int SING_iflush()
{
  return SING_iflush_i(&(sing.isp));
}

int SING_iflushMC(SINGinfo *s)
{
  return SING_iflush_i(&(s->isp));
}

int SING_BM_alloc(int emid, int size, int *err) 
{
  return SING_BM_alloc_i(&(sing.bm), &(sing.em), emid, size, err);
}

int SING_BM_allocMC(SINGinfo *s, int emid, int size, int *err) 
{
  return SING_BM_alloc_i(&(s->bm), &(s->em), emid, size, err);
}

int SING_BM_free() 
{
  return SING_BM_free_i(&(sing.bm), &(sing.em));
}

int SING_BM_freeMC(SINGinfo *s) 
{
  return SING_BM_free_i(&(s->bm), &(s->em));
}

int SING_EM_BM_transfer(int areaid,
                        int emstart,
                        int bmstart,
                        int ndata,
                        int mode,
                        int bbid) 
{
  return SING_EM_BM_transfer_i(sing.c, &(sing.bm), &(sing.em), 
			       areaid, emstart, bmstart, ndata, mode, bbid);
}

int SING_EM_BM_transferMC(SINGinfo *s, int areaid, int emstart, int bmstart,
                        int ndata, int mode, int bbid) 
{
    //  return SING_EM_BM_transfer_i(sing.c, &(sing.bm), &(sing.em), 
    //			       areaid, emstart, bmstart, ndata, mode, bbid);
  return SING_EM_BM_transfer_i(s->c, &(s->bm), &(s->em), 
			       areaid, emstart, bmstart, ndata, mode, bbid);
}
int SING_EM_BM_transfer_setup(int idpid, int areaid, int emstart, int bmstart,
			      int ndata, int mode, int bbid) 
{
    return SING_EM_BM_transfer_setup_i(sing.c, &(sing.bm), &(sing.em), idpid,
				       areaid, emstart, bmstart, ndata, mode, bbid,0);
}

int SING_EM_BM_transfer_setupMC(SINGinfo *s, int idpid, int areaid, int emstart,
				int bmstart, int ndata, int mode, int bbid) 
{
    return SING_EM_BM_transfer_setup_i(s->c, &(s->bm), &(s->em), idpid, areaid, emstart,
				       bmstart, ndata, mode, bbid,0);
}
int SING_EM_BM_transfer_setupMC_andsend1(SINGinfo *s, int idpid, int areaid, int emstart,
				int bmstart, int ndata, int mode, int bbid) 
{
    return SING_EM_BM_transfer_setup_i(s->c, &(s->bm), &(s->em), idpid, areaid, emstart,
				       bmstart, ndata, mode, bbid,1);
}
 
int SING_execute(int init_id,
                int loop_id,
                int final_id,
                int loop_count,
                void *loop_out,
                void *final_out) 
{
  return SING_execute_i(&(sing.isp), sing.c, 
			init_id, loop_id, final_id, loop_count, loop_out, final_out);
}
int SING_execute2(int ncodes, int codes[],
                int loop_count,
                void *loop_out,
		  void *final_out, int sizes[]) 
{
  return SING_execute2_i(&(sing.isp), sing.c, 
			 ncodes, codes,  loop_count,
			 loop_out, final_out, sizes);
}

int SING_start(int code, int loop_count)
{
    return SING_start_code(&(sing.isp), 1, &(sing.c), code,  loop_count);
}

int SING_finish()
{
    return SING_finish_code(&(sing.isp), 1, &(sing.c));
}

int SING_startMC(SINGinfo *s, int code, int loop_count)
{
    return SING_start_code(&(s->isp), 1, &(s->c), code,  loop_count);
}

int SING_finishMC(SINGinfo *s)
{
    return SING_finish_code(&(s->isp), 1, &(s->c));
}


int SING_executeMC(SINGinfo *s,
		   int init_id,
                int loop_id,
                int final_id,
                int loop_count,
                void *loop_out,
                void *final_out) 
{
  return SING_execute_i(&(s->isp), s->c, 
			init_id, loop_id, final_id, loop_count, loop_out, final_out);
}

int SING_execute2MC(SINGinfo *s,
		    int ncodes,
		    int codes[],
		    int loop_count,
		    void *loop_out,
		    void *final_out,
		    int sizes[]) 
{
    //    fprintf(stderr, "call execute2_i\n");
  return SING_execute2_i(&(s->isp), s->c, 
			 ncodes, codes, loop_count, loop_out, final_out, sizes);
}

int SING_set_result_size(int dum, int size)
{
  return SING_set_result_size_i(&(sing.isp), dum, size);
}

int SING_set_result_sizeMC(SINGinfo *s, int dum, int size)
{
  return SING_set_result_size_i(&(s->isp), dum, size);
}

void SING_EM_fifoclr()
{
  SING_EM_fifoclr_i(sing.c);
}

void SING_EM_fifoclrMC(SINGinfo *s)
{
  SING_EM_fifoclr_i(s->c);
}

CTRLinfo* SING_get_ctrlinfo()
{
  return sing.c;
}

SINGinfo* SING_get_singinfo()
{
  return &sing;
}

void SING_coc_reg_write(SINGinfo *s, int address, UINT64 data)
{
    CTRL_coc_reg_write(s->c, address, data);
}
 
void SING_dump_syncstatus(int ndevs,
			  SINGinfo * s[ndevs],
			  char * c)
{
    int i;
    fprintf(stderr,"%s: ", c);
    for(i=0;i<ndevs;i++){
	int devid = s[i]->c->hib_id;
	hib_mem_writeMC(devid, 0x0, 0x10);
	fprintf(stderr, " [%d]= %02x", devid, hib_mem_readMC(devid, 0x60));
    }
    fprintf(stderr,"\n remaining DMAR words:");
    for(i=0;i<ndevs;i++){
	int devid = s[i]->c->hib_id;
	fprintf(stderr, " [%d]= %06x", devid, hib_mem_readMC(devid, 0x8));
    }
    fprintf(stderr,"\n");
    fprintf(stderr," remaining DMAW words:");
    for(i=0;i<ndevs;i++){
	int devid = s[i]->c->hib_id;
	fprintf(stderr, " [%d]= %06x", devid, hib_mem_readMC(devid, 0x18));
    }
    fprintf(stderr,"\n");
    fprintf(stderr," DMAW status:");
    for(i=0;i<ndevs;i++){
	int devid = s[i]->c->hib_id;
	fprintf(stderr, " [%d]= %08x", devid, hib_mem_readMC(devid, 0x1c));
    }
    fprintf(stderr,"\n");
    fprintf(stderr," coc regs:");
    for(i=0;i<ndevs;i++){
	int devid = s[i]->c->hib_id;
	fprintf(stderr, " [%d]= %08x", devid, hib_mem_readMC(devid, 0xc));
    }
    fprintf(stderr,"\n");
    fprintf(stderr," dma regs:");
    for(i=0;i<ndevs;i++){
	int devid = s[i]->c->hib_id;
	fprintf(stderr, " [%d]= %016lx", devid, hib_mem_read64MC(devid, 0x28));
    }
    fprintf(stderr,"\n");
}

static int syncsaved[4];
static int dmawsaved[4];
void SING_save_syncstatus(int ndevs,
			  SINGinfo * s[ndevs])
{
    int i;
    for(i=0;i<ndevs;i++){
	int devid = s[i]->c->hib_id;
	hib_mem_writeMC(devid, 0x0, 0x10);
	syncsaved[i]= hib_mem_readMC(devid, 0x60);
	dmawsaved[i]=hib_mem_readMC(devid, 0x18);
    }
}

void SING_dump_savedstatus(int ndevs,  char * c)
{
    int i;
    fprintf(stderr,"%s: ", c);
    for(i=0;i<ndevs;i++){
	fprintf(stderr, " [%d]= %02x", i, syncsaved[i]);
    }
    fprintf(stderr,"\n");
    fprintf(stderr," remaining DMAW words:");
    for(i=0;i<ndevs;i++){
	fprintf(stderr, " [%d]= %06x", i, dmawsaved[i]);
    }
    fprintf(stderr,"\n");
}

void SING_set_reductionmode(SINGinfo *s,  int f)
// f= 1 for 4-chip reduction mode
// otherwize for separate mode
{
    int i;
    
    if(f==1){
	CTRL_setodpcom(s->c, 0x0ffULL);
	CTRL_setodpcom(s->c, 0x200ULL);
    } else {
	CTRL_setodpcom(s->c, 0x0ffULL);
	CTRL_setodpcom(s->c, 0x000ULL);
    }
}

void SING_flush_piow_channelMC(SINGinfo * s)
{
    CTRL_flush_piow_channel(s->c);
}
void SING_flush_piow_channel()
{
    CTRL_flush_piow_channel((&sing)->c);
}


#ifdef SING_TEST

int hetconv[] = {0, 0, 0, 0, 0, 0,0,0,0,0,0,0,0, 0, 0, 0, 0, 0,0,0,0,0,0,0};
double data[] = {1, 2, 3, 4, -1, -2, -3, -4, 5, 6, 7, 8, 9, 10, 11, 12,13};

static void dump_sreg(int firstdev, int lastdev)
{
    int i,j;
    for(j=firstdev;j<lastdev;j++){
	printf("hib %d hibreg 0x28 = %8x %8x\n", j,
	       hib_mem_readMC(j, 0x2c), hib_mem_readMC(j, 0x28));
	for(i=0;i<16;i++){
	hib_mem_writeMC(j, 0x0, i);
	printf(" hib %d misc:0x%08lx: 0x%08x\n", j, i, hib_mem_readMC(j, 0x60));
    }    }

}

int main(int argc, char **argv) 
{
    int emid, bbid, err, ilid, i, j ;
    int firstdev=0;
    INSTLIST *instlist;
    double result[64];
    CTRLinfo *chip[4];
    
    
    if(argc == 2){
	firstdev = atoi(argv[1]);
	singchip_id = firstdev;
	fprintf(stderr,"first device = %d\n", firstdev);
    }
    
    
    SING_open();
    SING_EM_fifoclr();
    chip[0] = sing.c;
    for(i=1;i<4;i++){
	chip[i] = CTRL_open(i+firstdev);
    }
    
    CTRL_setodpcom(sing.c, 0x0ffULL);
    CTRL_setodpcom(sing.c, 0x200ULL);
    for(i=1;i<4;i++){
	CTRL_setodpcom(chip[i], 0x0ffULL);
	CTRL_setodpcom(chip[i], 0x200ULL);
    }
    
    usleep(20);
    emid = SING_EM_alloc(4, hetconv, 16, &err);
    bbid = SING_BM_alloc(emid, /* emid */
			 2, /* size */
			 &err);
    usleep(10);
    SING_EM_status();
    SING_BM_status();
    fprintf(stderr,"call SING_EM_send\n");
    SING_EM_send(emid, /* areaid */
		 0, /* start */
		 2, /* ndata */
		 data);
    usleep(30);
    fprintf(stderr,"call SING_EM_BM_transfer\n");
    SING_EM_BM_transfer(bbid, 
			0, /* emstart */
			0, /* bmstart */
			2, /* ndata */
			0, /* mode */
			-1 /* bbid */
			);
    fprintf(stderr,"end  SING_EM_BM_transfer\n");
    usleep(10);
    //SING_print_BM_i(&(sing.isp), sing.c, 0, 32, 0);
    
    ilid = SING_ialloc_i(&(sing.isp), 80);
    instlist = &(sing.isp.instlists[ilid]);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    
    SING_asm_add_RRN("bb", 0, 0, 1, "", "flt72to64", instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_RRN("bb", 0, 0, 1, "", "flt72to64", instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_RRN("bb", 0, 0, 1, "", "flt72to64", instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_RRN("bb", 0, 0, 1, "", "flt72to64", instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);

    fprintf(stderr,"call SING_register_instlist\n");
  
    SING_register_instlist(sing.c, instlist);
    //    dump_sreg(firstdev, firstdev+4);
    fprintf(stderr,"call SING_send_instlist\n");
    SING_send_instlist(sing.c, instlist,1,0);
    usleep(1000);
    SING_ipop_i(&(sing.isp));
  
    //    printf("\n\n");

    //    dump_sreg(firstdev, firstdev+4);
    int iret = 0;
    for(j=0;j<4;j++){
	iret = CTRL_read_as_many_as(chip[j], result, 4);
	printf("chip %d : ", j);
	for(i=0;i<4;i++){
	    printf("%e ", result[i]);
	}
	printf("  error=%d\n",iret);
	if (iret) j = 4;
    }
    //SING_print_BM_i(&(sing.isp), sing.c, 0, 32, 0);
  
  
    for(i=0;i<4;i++){
	CTRL_setodpcom(chip[i], 0x0ffULL);
	CTRL_setodpcom(chip[i], 0x000ULL);
    }
  
  
    CTRL_endbcast(sing.c);
    return iret;
}



#endif
