#include <stdio.h>
#include <string.h>
#include <typedef.h>
#include <ctrl.h>
#include <sing_isp.h>
#include <sing_conf.h>

#define ArraySize(x)    (sizeof(x) / sizeof(x[0]))

static const char *ialu_opcode[] = {
  "add", "sub", "inc", "dec",
  "not", "and", "or", "xor", "max", "min",
  "passa", "passb", "lsl", "lsr", "bsl", "bsr", "lnot"};

int SING_asm_RRN_process_aux(const char *aux, INST *inst) 
{
  while ('\0' != *aux) {
    switch (*aux) {
    case 'r':
      inst->rrn_round = 1;
      break;
    case 'o':
      inst->rrn_odpoe = 0;
      break;
    case 's':
      inst->rrn_sregen = 0;
      break;
    default:
      fprintf(stderr, "SING_asm_RRN_process_aux: unrecognizable letter '%c'\n", *aux);
      return -1;
    };
    aux++;
  }
  return 0;
}

int SING_asm_RRN_process_faddsub(const char *opcode, INST *inst) 
{
  /*
  inst->rrn_normala = 1;
  inst->rrn_normalb = 1;
  inst->rrn_normalo = 1;
  */

  opcode++; /* skip first 'f' */
  if (0 == strncmp("add", opcode, 3))
    ; /* do nothing */
  else if (0 == strncmp("sub", opcode, 3))
    inst->rrn_signb = 1;
  else
    return -1;

  opcode += 3; /* skip "add" or "sub" */
  while ('\0' != *opcode) {
    switch (*opcode) {
    case 'A':
      inst->rrn_normala = 0; break;
    case 'B':
      inst->rrn_normalb = 0; break;
    case 'S':
      inst->rrn_signb = 1; break;
    case 'O':
      inst->rrn_normalo = 0; break;
    default:
      return -1;
    }
    opcode++;
  }
  return 0;
}

int SING_asm_RRN_init(int adr, int wl, INST *inst) {
  /* wl and adr check */
  inst->idprrn_valid = 1;
  inst->idprrn_select = 1;
  inst->rrn_adr = adr;
  inst->rrn_n = 4;
  inst->rrn_reduc = 1;
  inst->rrn_wl = wl;
  inst->rrn_fsel = 0;
  inst->rrn_normala = inst->rrn_normalb = inst->rrn_normalo = 1;
  inst->rrn_signb = 0;
  inst->rrn_round = 0;
  inst->rrn_ialuop = 0;
  inst->rrn_unsigned = 0;
  inst->rrn_odpoe = 1;
  inst->rrn_sregen = 1;
  return 0;
}

/*
  SING_asm_RRN_process_opcode
  return value:
  0 normal
  1 opcode error
  2 bbadr error
  */
int SING_asm_RRN_process_opcode(const char *opcode, int bbadr, INST *inst) {
  int i;
  
  switch (opcode[0]) {
  case 'f':
    inst->rrn_fsel = 1;
    if (SING_asm_RRN_process_faddsub(opcode, inst)) return 1;
    break;
  case 'u':
    inst->rrn_unsigned = 1;
    /* continue to case 'i' */
  case 'i':
    inst->rrn_ialuop = -1; /* set dummy flag */
    for (i = 0; i < ArraySize(ialu_opcode); i++) {
      if (0 == strcmp(ialu_opcode[i], opcode + 1))
        inst->rrn_ialuop = i;
    }
    if (-1 == inst->rrn_ialuop) return 1;
    break;
  case 'b':
    if (0 != strcmp("bb", opcode)) return 1;
    if (bbadr < 0 || bbadr > SING_number_of_bb())
      return 2;
    inst->rrn_reduc = 0;
    inst->rrn_bbadr = bbadr;
    break;
  default:
    return 1;
  }
  return 0;
}

int SING_asm_RRN_process_conv(const char *conv, INST *inst) 
{
  const char *conv_list[] = {
    "flt72to64", "flt36to64", "fix72to64l", "fix72to64r", "fix72to64w"};
  int i;
  
  if ('\0' == conv[0]) {
    inst->rrn_conv = 0; /* when conv == "", corresponding field is set to 0 */
  } else {
    inst->rrn_conv = -1; /* set dummy flag */
    for (i = 0; i < ArraySize(conv_list); i++) {
      if (0 == strcmp(conv_list[i], conv))
        inst->rrn_conv = i;
    }
    if (-1 == inst->rrn_conv) {
      fprintf(stderr, "SING_asm_RRN_process_conv: undefined conversion\n"); return -1;
    }
  }
  return 0;
}

/*
  SING_asm_RRN
  string processing is rough.
  opcode etc. with unnecessary spaces aren't well recognized.
 */
int SING_asm_RRN(const char *opcode,
                 int bbadr, int adr, int wl, const char *aux, const char *conv,
                 INST *inst) {
  int err;

  SING_asm_RRN_init(adr, wl, inst);
  err = SING_asm_RRN_process_opcode(opcode, bbadr, inst);
  switch (err) {
  case 0:
    /* no problem */
    break;
  case 1:
    fprintf(stderr, "SING_asm_RRN: opcode \"%s\" is not defined\n", opcode);
    return -1;
    break;
  case 2:
    fprintf(stderr, "SING_asm_RRN: bb %d doesn't exist\n", bbadr);
    return -1;
    break;
  default:
    fprintf(stderr, "SING_asm_RRN: unknown opcode error\n");
    return -1;
  }
  if (SING_asm_RRN_process_aux(aux, inst)) {
    fprintf(stderr, "SING_asm_RRN: failed at SING_asm_RRN_process_aux\n"); 
    return -1;
  }
  if (SING_asm_RRN_process_conv(conv, inst)) {
    fprintf(stderr, "SING_asm_RRN: failed at SING_asm_RRN_process_conv\n"); 
    return -1;
  }
  return 0;
}

int SING_asm_add_RRN(const char *opcode,
                     int bbadr, int adr, int wl, const char *aux, const char *conv,
                     INSTLIST *instlist) {
  INST *inst;
  inst = SING_new_inst(instlist);
  SING_asm_RRN(opcode, bbadr, adr, wl, aux, conv, inst);
  /* error processing */
  return 0;
}

int SING_asm_process_var(const char* var, int* wl, char* type, int* vector) {
  *wl = 0;
  *type = '\0';
  *vector = 0;
  while ('\0' != *var) {
    switch (*var) {
    case 'l':
      *wl = 1;
      break;
    case 'v':
      *vector = 1;
      break;
    case 'b':
    case 'm':
    case 'r':
    case 't':
    case 'B': /* for $bb */
    case 'P': /* for $pe */
      *type = *var;
      break;
    default:
      fprintf(stderr, "SING_asm_process_var: unrecognizable letter '%c'\n", *var);
    };
    var++;
  }
  if ('\0' == *type) {
    fprintf(stderr, "SING_asm_process_var: var \"%s\" does not speficy variable type\n", var);
    return -1;
  }
  return 0;
}

int SING_asm_bm(int vlen,
		const char* srcsym, int srcaddr,
		const char* destsym, int destaddr,
		int peid,
		INST *inst) {
  char srctype, desttype;
  int srcwl, srcvector, destwl, destvector;

  inst->ll = vlen;
  if (SING_asm_process_var(srcsym, &srcwl, &srctype, &srcvector)) {
    /* error */
    return -1;
  }
  if (SING_asm_process_var(destsym, &destwl, &desttype, &destvector)) {
    /* error */
    return -1;
  }

  /*
  dbp("src: %d %c %d, dest: %d %c %d\n",
      srcwl, srcvar, srcvector,
      destwl, destvar, destvector);
  */

  if ('b' == srctype) {
    inst->bm_adr = srcaddr;
    inst->bm_wl = srcwl;
    /* don't sure if $t is OK as the destination */
    switch (desttype) {
    case 'm':
      inst->l_write = 1;
      inst->l_isel = 2; /* select BM */
      inst->l_adr = destaddr;
      inst->l_wl = destwl;
      if (destvector) inst->l_adri = 1;
      break;
    case 'r':
      inst->r_write = 1;
      inst->r_wwl = destwl;
      inst->r_wadr = destaddr;
      inst->r_isel = 2; /* select BM */
      if (destvector) inst->r_wadri = 1;
      break;
    default:
      /* error */
      return -1;
    }
  } else if ('b' == desttype) {
    if ('r' != srctype) {
      /* error */
      return -1;
    }
    inst->bm_write = 1;
    inst->bm_adr = destaddr;
    inst->bm_peadr = peid;
    inst->bm_wl = destwl;
    inst->r_radrb = srcaddr;
    inst->r_rwlb = srcwl;
    if (srcvector) inst->r_radrib = 1;
  } else {
    /* error */
    return -1;
  }
  return 0;
}

int SING_asm_add_bm(int vlen,
		    const char* srcsym, int srcaddr, 
		    const char* destsym, int destaddr,
		    int peid,
		    INSTLIST *instlist) {
  INST *inst;
  inst = SING_new_inst(instlist);
  SING_asm_bm(vlen, srcsym, srcaddr, destsym, destaddr, peid, inst);
  /* error processing */
  return 0;
}

int SING_asm_add_nop(INSTLIST* instlist) {
  SING_new_inst(instlist);
  return 0;
}

int SING_asm_op_process_faddialu_source(char src1type, char src2type, INST *inst) {
  /* '\0' is a dummy */
  const char isela_table[] = {'r', '\0', 't', 'm', 'P', 'B'};
  const char iselb_table[] = {'\0', 'r', 't', 'm'};
  int i;

  inst->fadd_isela = -1; /* set dummy flag */
  for (i = 0; i < ArraySize(isela_table); i++) {
    if (src1type == isela_table[i])
      inst->fadd_isela = i;
  }
  if (-1 == inst->fadd_isela) return -1;

  inst->fadd_iselb = -1; /* set dummy flag */
  for (i = 0; i < ArraySize(iselb_table); i++) {
    if (src2type == iselb_table[i])
      inst->fadd_iselb = i;
  }
  if (-1 == inst->fadd_iselb) return -1;

  return 0;
}

int SING_asm_op_process_ialu_opcode(const char* opcode, INST *inst) {
  int i;
  inst->ialu_ialuop = -1; /* set dummy flag */
  for (i = 0; i < ArraySize(ialu_opcode); i++) {
    if (0 == strcmp(ialu_opcode[i], opcode + 1))
      inst->ialu_ialuop = i;
  }
  if (-1 == inst->ialu_ialuop) return 1;
  return 0;  
}

int SING_asm_op_process_opcode(const char* opcode, char src1type, char src2type, int *dest_input, INST *inst) {
  enum {FMUL, FADDIALU};
  
  switch (opcode[0]) {
  case 'f':
    opcode++; /* skip first 'f' */
    if (0 == strncmp("mul", opcode, 3)) {
      *dest_input = FMUL;
      /* to be implemented */
    } else if (0 == strncmp("add", opcode, 3)) {
      *dest_input = FADDIALU;
      inst->fsel = 1;
      /* fadd option details are neglected for now */
      SING_asm_op_process_faddialu_source(src1type, src2type, inst);
    } else if (0 == strncmp("sub", opcode, 3)) {
      *dest_input = FADDIALU;
      inst->fsel = 1;
      inst->fadd_signb = 1;
      /* fsub option details are neglected for now */
      SING_asm_op_process_faddialu_source(src1type, src2type, inst);
    } else {
      /* error*/
      return -1;
    }
    break;
  case 'd':
    *dest_input = FMUL;
    if (0 == strncmp("mul0", opcode, 4)) {
      /* to be implemented */
    } else if (0 == strncmp("mul1", opcode, 4)) {
      /* to be implemented */
    }
  case 'u':
    inst->ialu_unsigned = 1;
    /* continue to case 'i' */
  case 'i':
    *dest_input = FADDIALU;
    SING_asm_op_process_faddialu_source(src1type, src2type, inst);
    SING_asm_op_process_ialu_opcode(opcode, inst);
    break;
  default:
    return -1;
  }
  return 0;
}

int SING_asm_op_process_source(int src1wl,
			       char src1type,
			       int src1vector,
			       int src1addr,
			       int src2wl,
			       char src2type,
			       int src2vector,
			       int src2addr,
			       INST *inst) {
  /* indirect access to LM is not supported for now */
  if ('m' == src1type && 'm' == src2type) {
    /* error */
    return -1;
  }
  if ('m' == src1type) {
    inst->l_adr = src1addr;
    if (src1vector) inst->l_adri = 1;
    inst->l_wl = src1wl;
  }
  if ('m' == src2type) {
    inst->l_adr = src2addr;
    if (src2vector) inst->l_adri = 1;
    inst->l_wl = src2wl;
  }
  if ('r' == src1type) {
    inst->r_radra = src1addr;
    if (src1vector) inst->r_radria = 1;
    inst->r_rwla = src1wl;
  }
  if ('r' == src2type) {
    inst->r_radrb = src2addr;
    if (src2vector) inst->r_radrib = 1;
    inst->r_rwlb = src2wl;
  }
  /* when source type is 't', 'P' or 'B', do nothing */
  return 0;
}

int SING_asm_op_process_destination(int destwl,
				    char desttype, 
				    int destvector, 
				    int destaddr,
				    int dest_input,
				    INST *inst) {
  switch (desttype) {
  case 'm':
    inst->l_isel = dest_input;
    inst->l_write = 1;
    inst->l_adr = destaddr;
    if (destvector) inst->l_adri = 1;
    inst->l_wl = destwl;
    break;
  case 'r':
    inst->r_write = 1;
    inst->r_isel = dest_input;
    inst->r_wadr = destaddr;
    if (destvector) inst->r_wadri = 1;
    inst->r_wwl = destwl;
    break;
  case 't':
    inst->t_write = 1;
    inst->t_isel = dest_input;
    break;
  default:
    /* error */
    return -1;
  }
  return 0;
}

int SING_asm_op(int vlen, const char* op,
		const char* src1sym, int src1addr,
		const char* src2sym, int src2addr,
		const char* destsym, int destaddr,
		INST *inst) {
  char src1type, src2type, desttype;
  int src1wl, src1vector, src2wl, src2vector, destwl, destvector;
  int dest_input;

  inst->ll = vlen;
  if (SING_asm_process_var(src1sym, &src1wl, &src1type, &src1vector)) {
    /* error */
    return -1;
  }
  if (SING_asm_process_var(src2sym, &src2wl, &src2type, &src2vector)) {
    /* error */
    return -1;
  }
  if (SING_asm_process_var(destsym, &destwl, &desttype, &destvector)) {
    /* error */
    return -1;
  }
  if (SING_asm_op_process_opcode(op, src1type, src2type, &dest_input, inst)) {
    /* error */
    return -1;
  }
  if (SING_asm_op_process_source(src1wl, src1type, src1vector, src1addr,
				 src2wl, src2type, src2vector, src2addr,
				 inst)) {
    /* error */
    return -1;
  }
  if (SING_asm_op_process_destination(destwl, desttype, destvector, destaddr, dest_input, inst)) {
    /* error */
    return -1;
  }
  
  return 0;
}

int SING_asm_add_op(int vlen, const char* op,
		    const char* src1sym, int src1addr,
		    const char* src2sym, int src2addr,
		    const char* destsym, int destaddr,
		    INSTLIST *instlist) {
  INST *inst;
  inst = SING_new_inst(instlist);
  SING_asm_op(vlen, op, src1sym, src1addr, src2sym, src2addr, destsym, destaddr, inst);
  /* error processing */
  return 0;
}

int SING_fprint_BM_low(SINGISP_INFO *info, CTRLinfo *c, FILE *stream, int saddr, int length, int bbid, int label_saddr) 
{
  /* error check */
  int i, j;
  int nblock = (length + 7) / 8;

  for (i = 0; i < nblock; i++) {
    int id = SING_ialloc_i(info, MAX_INST);
    INSTLIST* instlist = &((info->instlists)[id]);
    int addr =  saddr + 8 * i;
    double result[4];
    	
    SING_asm_add_RRN("bb", bbid, addr, 1, "", "flt72to64", instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_asm_add_nop(instlist);
    SING_register_instlist(c, instlist);
    SING_send_instlist(c, instlist, 1, 1);
    SING_ipop_i(info);
  
    CTRL_read_as_many_as(c, result, 4);
    fprintf(stream, " %03x:", label_saddr + 8 * i);
    for (j = 0; j < 4; j++) {
      /* concatenate upper 64 bit and lower 8 bit */
      //fprintf(stream, " %016llx%02x", result[j], result[j + 4] % (1 << 8));
      fprintf(stream, " %e ", result[j]);
    }
    fprintf(stream, "\n");
  }
  return 0;
}

int SING_fprint_BM_i(SINGISP_INFO *info, CTRLinfo *c, FILE *stream, int saddr, int length, int bbid) 
{
  return SING_fprint_BM_low(info, c, stream, saddr, length, bbid, saddr);
}

int SING_print_BM_i(SINGISP_INFO *info, CTRLinfo *c, int saddr, int length, int bbid) 
{
  return SING_fprint_BM_i(info, c, stdout, saddr, length, bbid);
}
