#include <stdio.h>
#include <ctrl.h>
#include <sing_embm.h>
#include <sing_conf.h>
#include <sing_chip.h>


void init_eminfo(SINGEM_INFO *info)
{
  info->em_addr_curr = 0;
  info->convaddr_curr = 0;
  info->em_narea = 0;
}

void init_bminfo(SINGBM_INFO *info)
{
  info->bm_narea = 0;
  info->bm_addr_curr = 0;
}


int SING_pack_conversions(int length, const int *conversions, UINT64 *table, int *convlength, int *nword) 
{
  int conv_prev = -1, conv, i, nshortword = 0;
  *convlength = 0;
  for (i = 0; i < length; i++) {
    conv = conversions[i];
    switch (conv) {
    case 2:
    case 7:
    case 8:
    case 9:
      if (conv_prev < 0) {
        conv_prev = conv;
	nshortword++;
      } else {
        table[*convlength] = (conv_prev << 4) + conv;
        (*convlength)++;
	nshortword++;;
        conv_prev = -1;
      }
      break;
    case 0:
    case 1:
    case 3:
    case 4:
    case 5:
    case 6:
      if (conv_prev < 0) {
        table[*convlength] = conv << 4;
        (*convlength)++;
	if (1 == conv || 6 == conv)
	  nshortword++;
	else
	  nshortword += 2;
      } else {
        fprintf(stderr, "SING_pack_conversions: type %d cannot appear here. %d element%s scanned by now. current table length is %d and a half\n", conv, i, (i <= 1) ? "" : "s", *convlength);
        return 2;
      }
      break;
    default:
      fprintf(stderr, "SING_pack_conversions: undefined conversion type\n");
      return 16;
    }
  }
  if (conv_prev >= 0) {
    fprintf(stderr, "SING_pack_conversion: lack of one more short word\n");
    return 2;
  }
  *nword = (nshortword + 1) / 2;
  //  fprintf(stderr," SING_pack_conversions: convlength=%d nwords=%d\n", *convlength, *nword);
  //  for(i=0;i< *convlength; i++){
  //      fprintf(stderr,"table[%d]=%4x\n",   i,      table[i] );
  //  }
  return 0;
}

int SING_EM_alloc_i(SINGEM_INFO *info, CTRLinfo *c, int length, const int *conversions, int size, int *err) 
{
  int id = info->em_narea, nword, convlength;
  int convaddr_next, em_addr_next;

  *err = SING_pack_conversions(length, conversions, info->convtable, &convlength, &nword);
  if (0 != *err) {
    fprintf(stderr, "SING_EM_alloc: failed in SING_pack_conversions\n");
    return -1;
  }

  em_addr_next = info->em_addr_curr + nword * size;
  if (em_addr_next >= MAX_EM_ADDR) {
    fprintf(stderr, "SING_EM_alloc: lack of EM address space\n");
    *err |= 1;
    return -1;
  }

  if (id >= MAX_EM_AREA) {
    fprintf(stderr, "SING_EM_alloc: no room for new EM area\n");
    *err |= 4;
    return -1;
  }

  convaddr_next = info->convaddr_curr + convlength;
  if (convaddr_next >= MAX_CONVADDR) {
    fprintf(stderr, "SING_EM_alloc: conversion table doesn't have enough space\n");
    *err |= 8;
    return -1;
  }

  (info->em_areas)[id].addr = info->em_addr_curr;
  (info->em_areas)[id].length = length;
  (info->em_areas)[id].size = size;
  (info->em_areas)[id].convaddr = info->convaddr_curr;
  (info->em_areas)[id].convlength = convlength;
  (info->em_areas)[id].nword = nword;
  CTRL_het_convtable(c, info->convaddr_curr, info->convtable, convlength);
  (info->em_areas)[id].bmid = -1;

  info->em_narea++;
  info->convaddr_curr = convaddr_next;
  info->em_addr_curr = em_addr_next;

  /*
  dbp("SING_EM_alloc: id = %d, addr = %x, length = %d, size = %d, convaddr = %x, convlength = %d, nword = %d\n",
      id,
      em_areas[id].addr,
      em_areas[id].length,
      em_areas[id].size,
      em_areas[id].convaddr,
      em_areas[id].convlength,
      em_areas[id].nword
      );
  */
  return id;
}

int SING_EM_free_i(SINGEM_INFO *info) 
{
  SING_EM_status_i(info);
  
  if (info->em_narea < 0) {
    fprintf(stderr, "SING_EM_free: the number of EM areas is %d (negative)\n", info->em_narea);
    return -1;
  } else if (0 == info->em_narea) {
    return 0; /* when EM area is not registered, do nothing and return 0 */
  } else {
    int emid_last = info->em_narea - 1;
    EM_AREA *e = &((info->em_areas)[emid_last]);
    if (e->bmid > 0) {
      fprintf(stderr, "SING_EM_free: free corressponding BM area %d beforehand in order to free EM area %d\n", e->bmid, emid_last);
      return -1;
    }
    info->em_narea--;
    info->em_addr_curr -= e->nword * e->size;
    info->convaddr_curr -= e->convlength;

    SING_EM_status_i(info);
    return info->em_narea;
  }
}

int SING_EM_areas_i(SINGEM_INFO *info) 
{
  return info->em_narea;
}

void SING_EM_status_i(SINGEM_INFO *info) 
{
  int i;
  printf("SING_EM_status: %d area%s allocated\n", info->em_narea, (1 >= info->em_narea) ? "" : "s");
  if (0 == info->em_narea) return;
  printf(" id addr convaddr convlength nword length size bmid\n");
  for (i = 0; i < info->em_narea; i++) {
    EM_AREA *e = &(info->em_areas)[i];
    printf(" %2d %4x %8x %10d %5d %6d %4d %4d\n",
	   i, e->addr, e->convaddr, e->convlength, e->nword, e->length, e->size, e->bmid);
  }
}

int SING_EM_send_i(SINGEM_INFO *info, CTRLinfo *c, int areaid, int start, int ndata, void *hostp) 
{
  EM_AREA *e;
  int i;

  if (areaid < 0 || areaid >= info->em_narea) {
    fprintf(stderr, "SING_EM_send: EM area %d is not defined. EM area id must be 0 - %d\n", areaid, info->em_narea - 1);
    return 1;
  }

  e = &(info->em_areas[areaid]);

  if (start < 0 || start >= e->size) {
    fprintf(stderr, "SING_EM_send: start position %d is out of EM area %d. it must be 0 - %d\n", start, areaid, e->size - 1);
    return 2;
  }

  if (ndata < 0 || start + ndata > e->size) {
    fprintf(stderr, "SING_EM_send: ndata %d is incorrect. end position %d is out of EM area %d or preceeds start position. it must be %d - %d\n", ndata, start + ndata, areaid, start, e->size - 1);
    return 4;
  }



  //  CTRL_het_convaddr(c, e->convaddr);
  //  CTRL_het_length(c, e->convlength);
  int nextloc=0;
  nextloc = CTRL_het_convaddr_shortmessage(c, e->convaddr,0,nextloc);
  int em_addr = e->addr + (start * e->nword);
  void *data = (UINT64 *)hostp;
  if (ndata > 1){
      CTRL_flush_piow_channel(c);
      CTRL_flush_piow_channel(c);
      nextloc=CTRL_het_length_shortmessage(c, e->convlength,1,nextloc);
      CTRL_flush_piow_channel(c);
      CTRL_flush_piow_channel(c);
      CTRL_het_write(c, em_addr, data, e->convlength * ndata);
      CTRL_flush_piow_channel(c);
      CTRL_flush_piow_channel(c);
  }else{
      CTRL_flush_piow_channel(c);
      CTRL_flush_piow_channel(c);
      nextloc=CTRL_het_length_shortmessage(c, e->convlength,0,nextloc);
      CTRL_flush_piow_channel(c);
      CTRL_flush_piow_channel(c);
      CTRL_het_write_shortmessage(c, em_addr, data, 1,nextloc,
				  e->convlength * ndata);
      
      CTRL_flush_piow_channel(c);
      CTRL_flush_piow_channel(c);
  }
  return 0;
}

int SING_EM_vsend_i(SINGEM_INFO *info, CTRLinfo *c, int areaid, int address[],
		    int ndata, void *hostp) 
{
  EM_AREA *e;
  int i;

  int len = 3;
  UINT64 dumdata[len];
  UINT64 dumheader;
  for(i=0;i<len;i++)dumdata[i]=0;
  dumheader = ((UINT64)len << 40) | (((UINT64)(7)) << 36) ; 

  if (areaid < 0 || areaid >= info->em_narea) {
    fprintf(stderr, "SING_EM_send: EM area %d is not defined. EM area id must be 0 - %d\n", areaid, info->em_narea - 1);
    return 1;
  }

  e = &(info->em_areas[areaid]);

  //
  // do not check address validity yet...
  //
  //
  // assumed buffer size = 384
  //

  //  CTRL_het_convaddr(c, e->convaddr);
  //  CTRL_het_length(c, e->convlength);
  int nextloc=0;
  //  nextloc = CTRL_het_convaddr_shortmessage(c, e->convaddr,0,nextloc);
  nextloc = CTRL_het_convaddr_shortmessage(c, e->convaddr,0,nextloc);
  for (i = 0; i<ndata; i++){
#if 0      
      SING_EM_send_i(info, c, areaid, address[i], 1,
		     (void*) (((UINT64*)hostp)+e->convlength*i));
#endif      
#if 1
      int em_addr = e->addr + (address[i] * e->nword);
      void *data = ((UINT64 *)hostp)+ i* e->convlength;
      int flushmode = 0;
      if (nextloc + e->nword*2 > 384 || i == ndata - 1)flushmode = 1;
      //      nextloc = CTRL_het_convaddr_shortmessage(c, e->convaddr,0,nextloc);
      nextloc=CTRL_het_length_shortmessage(c, e->convlength,0,nextloc);      
      nextloc = CTRL_het_write_shortmessage(c, em_addr, data, flushmode,nextloc,
					    e->convlength);
#endif      

  }
  return 0;
}

static int skip_areaid_setting = 0;

int SING_EM_set_skip_areaid(int skip)
{
    skip_areaid_setting = skip;
    //    fprintf(stderr,"skip_areaid=%d\n", skip);
    return     skip_areaid_setting;
}

void SING_EM_set_areaid_multichip(int nchips,
				 SINGinfo * s[nchips],
				 int areaid)
{
    SINGEM_INFO *info[nchips];
    CTRLinfo *c[nchips];
    EM_AREA *e[nchips];
    int i;
    int ichips;
    for(i=0;i<nchips;i++){
	info[i]=&(s[i]->em);
	c[i]=s[i]->c;
	e[i] = &(info[i]->em_areas[areaid]);
	
	CTRL_het_convaddr(c[i], e[i]->convaddr);
	CTRL_het_length(c[i], e[i]->convlength);
    }
}

int SING_EM_send_i_multichip(int nchips,
			     SINGinfo * s[nchips],
			     int areaid,
			     int start,
			     int ndata,
			     void *hostp,
			     int increment) 
{
    SINGEM_INFO *info[nchips];
    CTRLinfo *c[nchips];
    EM_AREA *e[nchips];
    static int old_areaid = -1;
    int i;
    int ichips;
    
    for(i=0;i<nchips;i++){
	info[i]=&(s[i]->em);
	c[i]=s[i]->c;
	e[i] = &(info[i]->em_areas[areaid]);
	if (skip_areaid_setting == 0){
	    if (i==0)fprintf(stderr,"SING_EM_send_i_multichip set new areaid %d\n",areaid);
	    hib_finish_dmarMC(s[i]->c->hib_id);
	    CTRL_het_convaddr(c[i], e[i]->convaddr);
	    CTRL_het_length(c[i], e[i]->convlength);
	}
    }
    
    //    fprintf(stderr,"calling CTRL_het_write_multichip\n");
    CTRL_het_write_multichip(nchips, c, e[0]->addr + (start * e[0]->nword),
			     hostp,  e[0]->convlength * ndata,increment);
    
    return 0;
}

/*
  SING_BM_alloc

  return value:
  non-negative  BB ID
  negative      error

  the value of *err:
  0  no problem
  1  not enough BM address space
  2  EM area specified by [emid] does not exist
  4  no room for another BM
 */

int SING_BM_alloc_i(SINGBM_INFO *bminfo, SINGEM_INFO *eminfo, int emid, int size, int *err) 
{
  int bmid = bminfo->bm_narea;
  int bm_addr_next;

  *err = 0;
  if (bmid >= MAX_BM_AREA) {
    fprintf(stderr, "SING_BM_alloc: no room for another BM area\n");
    *err |= 4;
    return -1;
  }

  if (emid < 0 || emid >= eminfo->em_narea) {
    fprintf(stderr, "SING_BM_alloc: EM AREA %d does not exist\n", emid);
    *err |= 2;
    return -1;
  }

  bm_addr_next = bminfo->bm_addr_curr + size * (eminfo->em_areas)[emid].nword;
  if (bm_addr_next >= MAX_BM_ADDR) {
    fprintf(stderr, "SING_BM_alloc: not enough BM address space\n");
    *err |= 1;
    return -1;
  }

  (eminfo->em_areas)[emid].bmid = bmid;
  (bminfo->bm_areas)[bmid].emid = emid;
  (bminfo->bm_areas)[bmid].addr = bminfo->bm_addr_curr;
  (bminfo->bm_areas)[bmid].size = size;
  //  dbp("SING_BM_alloc: bmid = %d, size = %d\n", bmid, size);

  bminfo->bm_narea++;
  bminfo->bm_addr_curr = bm_addr_next;
  return bmid;
}

int SING_BM_free_i(SINGBM_INFO *bminfo, SINGEM_INFO *eminfo) 
{
  if (bminfo->bm_narea < 0) {
    fprintf(stderr, "SING_BM_free: the number of BM areas is %d (negative)\n", bminfo->bm_narea);
    return -1;
  } else if (0 == bminfo->bm_narea) {
    return 0; /* when BM area is not registered, do nothing and return 0 */
  } else {
    bminfo->bm_narea--;
    BM_AREA *b = &((bminfo->bm_areas)[bminfo->bm_narea]);
    bminfo->bm_addr_curr -= (b->size)*(eminfo->em_areas)[b->emid].nword; 
    (eminfo->em_areas)[b->emid].bmid = -1;

    return bminfo->bm_narea;
  }
}

/*
  SING_EM_BM_transfer
  return value:
   0  normal
   1  emstart error
   2  bmstart error
   4  ndata error
   8  mode error
  16  areaid error

  re-think about ndata
  */
int SING_EM_BM_transfer_setup_i(CTRLinfo *c,
				SINGBM_INFO *bminfo,
				SINGEM_INFO *eminfo,
				int idpid,
				int areaid,
				int emstart,
				int bmstart,
				int ndata,
				int mode,
				int bbid,
				int start)
{
    int emid, nword, nbm, len, eit_register_mode;
    BM_AREA *b;
    EM_AREA *e;
    
    //  fprintf(stderr,"SING_EM_BM_transfer\n");
    
    if (areaid < 0 || areaid >= bminfo->bm_narea) {
	fprintf(stderr, "SING_EM_BM_trasfer: BM ID %d is unrecognizable\n", areaid);
	return 16;
    }
    b = &((bminfo->bm_areas)[areaid]);
    if (bmstart < 0 || bmstart >= b->size) {
	fprintf(stderr, "SING_EM_BM_transfer: bmstart %d is not the right index\n", bmstart);
	return 2;
    }
    
    emid = b->emid;
    e = &((eminfo->em_areas)[emid]);
    if (emstart < 0 || emstart >= e->size) {
	fprintf(stderr, "SING_EM_BM_transfer: emstart %d is not the right index\n", emstart);
	return 1;
    }
    nword = e->nword;
    if (ndata < 0 || bmstart + ndata > b->size || emstart + ndata > e->size) {
	fprintf(stderr, "SING_EM_BM_transfer: ndata %d is inappropriate\n", ndata);
	return 4;
    }
    
    //  dbp("SING_EM_BM_transfer: %d active bb\n", SING_number_of_active_bb());
    
    len = ndata * nword;
    /* error check; len must be 1 - 256 */
    if (0 == mode) {
	if (bbid >= 0) {
	    eit_register_mode = 0; nbm = bbid; 
	} else {
	    eit_register_mode = 1; nbm = 0; /* don't care about the value of nbm */
	}
    } else if (1 == mode) {
	eit_register_mode = 2; nbm = SING_number_of_active_bb();
    } else {
	fprintf(stderr, "SING_EM_BM_transfer: undefined mode %d\n", mode);
	return 8;
    }
    //  fprintf(stderr, "SING_EM_BM_transfer call CTRL_eit_register\n");
    
    CTRL_eit_register(c, idpid, /* idpid*/
		      e->addr + emstart * nword, /* emsadr */
		      0, /* emiadr0 */
		      nword, /* emiadr1 */
		      2*(b->addr + bmstart * nword), /* bmsadr */
		      nbm,
		      len, /* len */
		      eit_register_mode, /* mode */
		      start); /* start */
    
    //  fprintf(stderr, "SING_EM_BM_transfer end  CTRL_eit_register\n");
    return 0;
}

int SING_EM_BM_transfer_i(CTRLinfo *c,
			  SINGBM_INFO *bminfo,
			  SINGEM_INFO *eminfo,
			  int areaid,
			  int emstart,
			  int bmstart,
			  int ndata,
			  int mode,
			  int bbid)
{
    return SING_EM_BM_transfer_setup_i(c,bminfo,eminfo, 0, areaid, emstart, bmstart,
				     ndata, mode, bbid,1);
}

int SING_EM_BM_transfer_i_old(CTRLinfo *c,
			  SINGBM_INFO *bminfo,
			  SINGEM_INFO *eminfo,
			int areaid,
                        int emstart,
                        int bmstart,
                        int ndata,
                        int mode,
                        int bbid) {
  int emid, nword, nbm, len, eit_register_mode;
  BM_AREA *b;
  EM_AREA *e;

  //  fprintf(stderr,"SING_EM_BM_transfer\n");

  if (areaid < 0 || areaid >= bminfo->bm_narea) {
    fprintf(stderr, "SING_EM_BM_trasfer: BM ID %d is unrecognizable\n", areaid);
    return 16;
  }
  b = &((bminfo->bm_areas)[areaid]);
  if (bmstart < 0 || bmstart >= b->size) {
    fprintf(stderr, "SING_EM_BM_transfer: bmstart %d is not the right index\n", bmstart);
    return 2;
  }

  emid = b->emid;
  e = &((eminfo->em_areas)[emid]);
  if (emstart < 0 || emstart >= e->size) {
    fprintf(stderr, "SING_EM_BM_transfer: emstart %d is not the right index\n", emstart);
    return 1;
  }
  nword = e->nword;
  if (ndata < 0 || bmstart + ndata > b->size || emstart + ndata > e->size) {
    fprintf(stderr, "SING_EM_BM_transfer: ndata %d is inappropriate\n", ndata);
    return 4;
  }

  //  dbp("SING_EM_BM_transfer: %d active bb\n", SING_number_of_active_bb());
  
  len = ndata * nword;
  /* error check; len must be 1 - 256 */
  if (0 == mode) {
    if (bbid >= 0) {
      eit_register_mode = 0; nbm = bbid; 
    } else {
      eit_register_mode = 1; nbm = 0; /* don't care about the value of nbm */
    }
  } else if (1 == mode) {
    eit_register_mode = 2; nbm = SING_number_of_active_bb();
  } else {
    fprintf(stderr, "SING_EM_BM_transfer: undefined mode %d\n", mode);
    return 8;
  }
  //  fprintf(stderr, "SING_EM_BM_transfer call CTRL_eit_register\n");

  CTRL_eit_register(c, 0, /* idpid: for now, fixed to 0 */
		    e->addr + emstart * nword, /* emsadr */
                    0, /* emiadr0 */
                    nword, /* emiadr1 */
                    2*(b->addr + bmstart * nword), /* bmsadr */
                    nbm,
                    len, /* len */
                    eit_register_mode, /* mode */
                    1); /* start */
  
  //  fprintf(stderr, "SING_EM_BM_transfer end  CTRL_eit_register\n");
  return 0;
}

void SING_EM_fifoclr_i(CTRLinfo *c)
{
  UINT64 data = 0;
  CTRL_write(c, 2, 0xff, &data, 1); 
}

void SING_BM_status_i(SINGBM_INFO *info) 
{
  int i;
  printf("SING_BM_status: %d area%s allocated\n", info->bm_narea, (1 >= info->bm_narea) ? "" : "s");
  if (0 == info->bm_narea) return;
  printf(" id emid addr size\n");
  for (i = 0; i < info->bm_narea; i++) {
    BM_AREA *b = &(info->bm_areas)[i];
    printf(" %2d %4d %4x %4d\n", i, b->emid, b->addr, b->size);
  }
}

