#ifndef _SING_EMBM_H
#define _SING_EMBM_H

#include <sing_conf.h>

typedef struct {
  int addr;
  int convaddr;
  int convlength;
  int nword; /* the number of long words needed to contain one data structure */
  int length; /* data length of one structure */
  int size; /* the number of structures */
  int bmid; /* BM ID to which the EM is assigned. otherwise -1 */
} EM_AREA;

typedef struct {
  int emid;
  int addr;
  int size;
} BM_AREA;

typedef struct
{
  UINT64 convtable[MAX_CONVERSION];
  int em_addr_curr;
  int convaddr_curr;
  EM_AREA em_areas[MAX_EM_AREA];
  int em_narea;
} SINGEM_INFO;

typedef struct
{
  BM_AREA bm_areas[MAX_BM_AREA];
  int bm_narea;
  int bm_addr_curr;
}  SINGBM_INFO;


void init_eminfo(SINGEM_INFO *info);
void init_bminfo(SINGBM_INFO *info);

int SING_EM_alloc_i(SINGEM_INFO *info, CTRLinfo *c, int length, const int *conversions,
		    int size, int *err);
int SING_EM_free_i(SINGEM_INFO *info);
int SING_EM_areas_i(SINGEM_INFO *info);
void SING_EM_status_i(SINGEM_INFO *info);
void SING_BM_status_i(SINGBM_INFO *info);
int SING_EM_send_i(SINGEM_INFO *info, CTRLinfo *c, int areaid, int start,
		   int ndata, void *hostp);
int SING_EM_vsend_i(SINGEM_INFO *info, CTRLinfo *c, int areaid, int address[],
		    int ndata, void *hostp);

int SING_BM_alloc_i(SINGBM_INFO *bminfo, SINGEM_INFO *eminfo, int emid, int size, int *err);
int SING_BM_free_i(SINGBM_INFO *bminfo, SINGEM_INFO *eminfo);
int SING_EM_BM_transfer_i(CTRLinfo *c, SINGBM_INFO *bminfo, SINGEM_INFO *eminfo, int areaid, int emstart, int bmstart, int ndata, int mode, int bbid);

int SING_pack_conversions(int length, const int *conversions, UINT64 *table, int *convlength, int *nword);
void SING_EM_fifoclr_i(CTRLinfo *c);

int SING_EM_BM_transfer_setup_i(CTRLinfo *c,
				SINGBM_INFO *bminfo,
				SINGEM_INFO *eminfo,
				int idpid,
				int areaid,
				int emstart,
				int bmstart,
				int ndata,
				int mode,
				int bbid,
				int start);

#endif
