#include <stdio.h>
#include <string.h>
#include <ctrl.h>
#include <hibutil.h>
#include <sing_conf.h>
#include <sing_isp.h>

#define ArraySize(x)    (sizeof(x) / sizeof(x[0]))

#define MAX_STR 256

#ifdef DEBUG
void dbp(const char *fmt,  ...) {
  va_list argp;
  // if (dblevel > vpm_debug_level) return;
  // fprintf(stderr, "DBG%d: ",dblevel);
  va_start(argp, fmt);
  vfprintf(stderr, fmt, argp);
  va_end(argp);
}
#else
#define dbp(...) {}
#endif

static const int isp_nbit[] = {
  3, 3, 3, 1,       /* 0: ll, 1: m_imr, 2: m_omr, 3: m_ifsel, */
  1, 1, 2,          /* 4: t_write, 5: t_shortstop, 6: t_isel, */
  1, 2, 6, 1, 1,    /* 7: r_write, 8: r_isel, 9: r_wadr, 10: r_wadri, 11: r_wwl, */
  6, 1, 1, 6, 1, 1, /* 12: r_radra, 13: r_radria, 14: r_rwla, 15: r_radrb, 16: r_radrib, 17: r_rwlb, */
  1, 2, 9, 4, 1, 1, /* 18: l_write, 19: l_isel, 20: l_adr, 21: l_adri, 22: l_tregadr, 23: l_wl, */
  1, 1, 1, 1,       /* 24: fmul_shift50a, 25: fmul_round50a, 26: fmul_rounda, 27: fmul_normala, */
  1, 1, 1, 1,       /* 28: fmul_shift25b, 29: fmul_shift50b, 30: fmul_roundb, 31: fmul_normalb, */
  1, 1, 2, 2,       /* 32: fmul_round, 33: fmul_normalo, 34: fmul_isela, 35: fmul_iselb */
  1, 1, 1, 1, 1,    /* 36: fadd_normala, 37: fadd_normalb, 38: fadd_signb, 39: fadd_round, 40: fadd_normalo, */
  3, 2,             /* 41: fadd_isela, 42: fadd_iselb, */
  5, 1, 1,          /* 43: ialu_ialuop, 44: ialu_unsigned, 45: fsel, */
  1, 11, 5, 1,      /* 46: bm_write, 47: bm_adr, 48: bm_peadr, 49: bm_wl */
};

//For DRAM Load Instruction
static const int idpread_nbit[] = {1,5,1}; //|IDPRV(1)|IDPID{2-6)|IDPWAIT(1)|
static const int idprrn_nbit[] = {1, 1};
static const int idp_nbit[] = {5};
static const int rrn_nbit[] = {
  13, 8, 5, 1, 1, 1, 1, 1, 1, 1, 1, 5, 1, 1, 1, 3};


void init_ispinfo(SINGISP_INFO *info)
{
  info->ninstlist = 0;
  info->inst_addr_curr = 0;
}

int SING_print_inst_low_format(const int *nbit, int size, char *format, int maxlen) 
{
  int i, n, len = 0;
  for (i = 0; i < size; i++) {
    n = snprintf(format + len, maxlen - len, "%%%dx%s", 1 + (nbit[i] - 1) / 4, (size - 1 == i) ? "" : " ");
    if (n < 0) {
      fprintf(stderr, "SING_print_inst_low_format: format string too long\n");
      return -1;
    }
    len += n;
  }
  return len;
}

int SING_print_inst_low(const INST* inst, int without_header) 
{
  static char *isp_header[] = {
    "l m m m t t t r r  r r r  r r r  r r r l l   l l l l f f f f f f f f f f f f f f f f f f f  i i f b   b  b b",
    "l _ _ _ _ _ _ _ _  _ _ _  _ _ _  _ _ _ _ _   _ _ _ _ m m m m m m m m m m m m a a a a a a a  a a s m   m  m m",
    ": i o i w s i w i  w w w  r r r  r r r w i   a a t w u u u u u u u u u u u u d d d d d d d  l l e _   _  _ _",
    ": m m f r h s r s  a a w  a a w  a a w r s   d d r l l l l l l l l l l l l l d d d d d d d  u u l w   a  p w",
    ": r r s i o e i e  d d l  d d l  d d l i e   r r e : _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _  _ _ : r   d  e l",
    ": : : e t r l t l  r r :  r r a  r r b t l   : i g : s r r n s s r n r n i i n n s r n i i  i u : i   r  a :",
    ": : : l e t : e :  : i :  a i :  b i : e :   : : a : h o o o h h o o o o s s o o i o o s s  a n : t   :  d :",
    ": : : : : s : : :  : : :  : a :  : b : : :   : : d : i u u r i i u r u r e e r r g u r e e  l s : e   :  r :",
    ": : : : : t : : :  : : :  : : :  : : : : :   : : r : f n n m f f n m n m l l m m n n m l l  u i : :   :  : :",
    ": : : : : o : : :  : : :  : : :  : : : : :   : : : : t d d a t t d a d a a b a a b d a a b  o g : :   :  : :",
    ": : : : : p : : :  : : :  : : :  : : : : :   : : : : 5 5 a l 2 5 b l : l : : l l : : l : :  p n : :   :  : :",
    ": : : : : : : : :  : : :  : : :  : : : : :   : : : : 0 0 : a 5 0 : b : o : : a b : : o : :  : e : :   :  : :",
    ": : : : : : : : :  : : :  : : :  : : : : :   : : : : a a : : b b : : : : : : : : : : : : :  : d : :   :  : :",
  };

  static char *idprrn_header[] = {
    "i i i   r  r  r r r r r r r r r  r r r r r",
    "d d d   r  r  r r r r r r r r r  r r r r r",
    "p p p   n  n  n n n n n n n n n  n n n n n",
    "r r _   _  _  _ _ _ _ _ _ _ _ _  _ _ _ _ _",
    "r r i   a  n  b r w f n n s r n  i u o s c",
    "n n d   d  :  b e l s o o i o o  a n d r o",
    "_ _ :   r  :  a d : e r r g u r  l s p e n",
    "v s :   :  :  d u : l m m n n m  u i o g v",
    "a e :   :  :  r c : : a a b d a  o g e e :",
    "l l :   :  :  : : : : l l : : l  p n : n :",
    "i e :   :  :  : : : : a b : : o  : e : : :",
    "d c :   :  :  : : : : : : : : :  : d : : :",
    ": t :   :  :  : : : : : : : : :  : : : : :",
    ": : :   :  :  : : : : : : : : :  : : : : :",
  };

  int i;
  const int left_padding = 1, between_padding = 1;
  char format[MAX_STR];

  if (!without_header) {
    for (i = 0; i < ArraySize(isp_header); i++)
      printf("%*s%s%*s%s\n", left_padding, "", isp_header[i], between_padding, "", idprrn_header[i]);

    printf("\n");
  }

  printf("%*s", left_padding, "");

  SING_print_inst_low_format(isp_nbit, ArraySize(isp_nbit), format, MAX_STR);
  printf(format,
	 inst->ll, inst->m_imr, inst->m_omr, inst->m_ifsel,
	 inst->t_write, inst->t_shortstop, inst->t_isel,
	 inst->r_write, inst->r_isel, inst->r_wadr, inst->r_wadri, inst->r_wwl,
	 inst->r_radra, inst->r_radria, inst->r_rwla, inst->r_radrb, inst->r_radrib, inst->r_rwlb,
	 inst->l_write, inst->l_isel, inst->l_adr, inst->l_adri, inst->l_tregadr, inst->l_wl,
	 inst->fmul_shift50a, inst->fmul_round50a, inst->fmul_rounda, inst->fmul_normala,
	 inst->fmul_shift25b, inst->fmul_shift50b, inst->fmul_roundb, inst->fmul_normalb,
	 inst->fmul_round, inst->fmul_normalo, inst->fmul_isela, inst->fmul_iselb,
	 inst->fadd_normala, inst->fadd_normalb, inst->fadd_signb, inst->fadd_round, inst->fadd_normalo,
	 inst->fadd_isela, inst->fadd_iselb,
	 inst->ialu_ialuop, inst->ialu_unsigned, inst->fsel,
	 inst->bm_write, inst->bm_adr, inst->bm_peadr, inst->bm_wl);

  printf("%*s", between_padding, "");

  SING_print_inst_low_format(idprrn_nbit, ArraySize(idprrn_nbit), format, MAX_STR);
  printf(format,inst->idprrn_valid, inst->idprrn_select);

  SING_print_inst_low_format(idp_nbit, ArraySize(idp_nbit), format, MAX_STR);
  printf(format, inst->idp_id);

  SING_print_inst_low_format(rrn_nbit, ArraySize(rrn_nbit), format, MAX_STR);

  printf(format,
	 inst->rrn_adr, inst->rrn_n, inst->rrn_bbadr, inst->rrn_reduc, inst->rrn_wl, inst->rrn_fsel,
	 inst->rrn_normala, inst->rrn_normalb, inst->rrn_signb, inst->rrn_round, inst->rrn_normalo,
	 inst->rrn_ialuop, inst->rrn_unsigned, inst->rrn_odpoe, inst->rrn_sregen, inst->rrn_conv);

  printf("\n");
  return 0;
}

int SING_print_inst(const INST *inst)
{
  return SING_print_inst_low(inst, 1);
}

int SING_print_instlist(SINGISP_INFO *info, int ilid) 
{
  INSTLIST *instlist;
  /* error check */
  int i;

  instlist = &((info->instlists)[ilid]);
  fprintf(stderr, "instruction list %d:\n", ilid);
  for (i = 0; i < instlist->ninst; i++)
    SING_print_inst_low(&(instlist->insts[i]), i);
  return 0;
}

int SING_print_instlist_pack(SINGISP_INFO *info, int ilid) 
{
  INSTLIST *instlist;
  UINT64 *pack;
  int i;
  const int left_padding = 1;

  /* error check */
  instlist = &((info->instlists)[ilid]);
  pack = instlist->pack;
  fprintf(stderr, "SING_print_instlist_pack:\n");
  for (i = 0; i < instlist->ninst; i++) {
    fprintf(stderr, "%*s%016llx %016llx %016llx %016llx\n",
	    left_padding,
	    "",
	    pack[4 * i],
	    pack[4 * i + 1],
	    pack[4 * i + 2],
	    pack[4 * i + 3]);
  }
  return 0;
}


int SING_clear_inst(INST *inst) 
{
  memset(inst, 0, sizeof(INST));
  return 0;
}

int SING_scan_inst_str(const char *str, INST *inst) 
{
  int i = 0;
  char label[MAX_STR], format[2 * ArraySize(isp_nbit)+1];

  SING_clear_inst(inst);
  i = sscanf(str, "%s", label);
  if (1 != i || 0 != strncmp("ISP", label, 3)) {
    fprintf(stderr, "SING_scan_inst_str: ISP header field is not \"ISP\"\n");
    return -1;
  }

  for (i = 0; i < ArraySize(isp_nbit); i++) {
    format[2 * i] = '%'; format[2 * i + 1] = 'x';
  }
  format[2*ArraySize(isp_nbit)]='\0';

  str += 4;
  i = sscanf(str, format,
             &(inst->ll), &(inst->m_imr), &(inst->m_omr), &(inst->m_ifsel),
             &(inst->t_write), &(inst->t_shortstop), &(inst->t_isel),
             &(inst->r_write), &(inst->r_isel), &(inst->r_wadr), &(inst->r_wadri), &(inst->r_wwl),
             &(inst->r_radra), &(inst->r_radria), &(inst->r_rwla), &(inst->r_radrb), &(inst->r_radrib), &(inst->r_rwlb),
             &(inst->l_write), &(inst->l_isel), &(inst->l_adr), &(inst->l_adri), &(inst->l_tregadr), &(inst->l_wl),
             &(inst->fmul_shift50a), &(inst->fmul_round50a), &(inst->fmul_rounda), &(inst->fmul_normala),
             &(inst->fmul_shift25b), &(inst->fmul_shift50b), &(inst->fmul_roundb), &(inst->fmul_normalb),
             &(inst->fmul_round), &(inst->fmul_normalo), &(inst->fmul_isela), &(inst->fmul_iselb),
             &(inst->fadd_normala), &(inst->fadd_normalb), &(inst->fadd_signb), &(inst->fadd_round), &(inst->fadd_normalo),
             &(inst->fadd_isela), &(inst->fadd_iselb),
             &(inst->ialu_ialuop), &(inst->ialu_unsigned), &(inst->fsel),
             &(inst->bm_write), &(inst->bm_adr), &(inst->bm_peadr), &(inst->bm_wl));

  if (ArraySize(isp_nbit) != i) {
    fprintf(stderr, "SING_scan_inst_str: incomplete ISP fields(%d found, %d expected)\n", i, ArraySize(isp_nbit));
    return -1;
  }

  for (i = 0; i < ArraySize(isp_nbit); i++) { /* skip ISP field and move to IDP/RRN field */
    while (' ' != *str) {
      if ('\0' == str) {
        fprintf(stderr, "SING_scan_inst_str: unexpectedly reached the end of the instruction string\n");
        return -1;
      }
      str++;
    }
    while (' ' == *str) str++;
  }

  inst->idprv = 0;
  inst->idpwait = 0;
  inst->idprrn_valid = 0;
  if (1 != sscanf(str, "%s", label)) {
    //dbp("SING_scan_inst_str: ISP string is not accompanied by anything\n");
    return 0;
  }
  str += 4;
  if(0 == strncmp("rdp", label, 3)){
    inst->idprv = 1;
    inst->idpwait = 0;
    i = sscanf(str, "%x", &(inst->idprid));
    if(i!=1){
      fprintf(stderr, "SING_scan_inst_str : incomplete rdp fields \n");
      return -1;
    }
  } else if(0 == strncmp("wdp", label, 3)){
    inst->idprv = 0;
    inst->idpwait = 1;
  } else if (0 == strncmp("idp", label, 3)) {
    //    dbp("SING_scan_inst_str: IDP header found\n");
    inst->idprrn_valid = 1; inst->idprrn_select = 0;
    i = sscanf(str, "%x", &(inst->idp_id));
    inst->idprid = inst->idp_id;
    if (ArraySize(idp_nbit) != i) {
      fprintf(stderr, "SING_scan_inst_str: incomplete IDP fields((%d found, %d expected)\n", i, ArraySize(idp_nbit));
      return -1;
    }
  } else if (0 == strncmp("rrn", label, 3)) {
    //dbp("SING_scan_inst_str: RRN header found\n");
    inst->idprrn_valid = 1; inst->idprrn_select = 1;
    i = sscanf(str, "%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%s%x",
               &(inst->rrn_adr), &(inst->rrn_n), &(inst->rrn_bbadr), &(inst->rrn_reduc), &(inst->rrn_wl), &(inst->rrn_fsel),
               &(inst->rrn_normala), &(inst->rrn_normalb), &(inst->rrn_signb), &(inst->rrn_round), &(inst->rrn_normalo),
               &(inst->rrn_ialuop), &(inst->rrn_unsigned), &(inst->rrn_odpoe), &(inst->rrn_sregen), label, &(inst->rrn_conv));
    if (ArraySize(rrn_nbit) + 1 != i) { /* RRN field and "conv" */
      fprintf(stderr, "SING_scan_inst_str: incomplete RRN fields(%d found, %d expected)\n", i, ArraySize(rrn_nbit) + 1);
      return -1;
    }
    if (0 != strncmp("conv", label, 4)) {
      fprintf(stderr, "SING_scan_inst_str: \"conv\" is not in the right place\n");
    }
  };
  return 0;
}

UINT64 SING_bitmask(int start, int n) 
{
  return (((0x1ULL << n) - 1) << (start - n + 1));
}

int SING_bitwrite(int x, UINT64 *y, int start, int n) 
{
  int end = start - n;
  *y &= ~SING_bitmask(start, n);
  *y |= (UINT64)x << (end + 1);
  return end;
}

/*
  SING_pack_inst_low
  arguments:
  pack   UINT64 array into which fields is packed. should be cleared to 0 beforehand
  base   specifies where packing starts. for example, MSB is 63 and LSB is 0
  n      SING_pack_inst_low tries to pack the fields within pack[0 .. n - 1]. if it failed, it returns -1
  field  array of field going to be packed
  nbit   bit width of each field
  len    the number of fields
 */
int SING_pack_inst_low(UINT64 *pack, int *base, int n, const int *field, const int *nbit, int len) 
{
  int i, j, digit_left;

  j = 0;
  for (i = 0; i < len; i++) {
/*
    dbp(stderr, "*base = %2d, nbit[i] = %2d, ", *base, nbit[i]);
*/
    *base = SING_bitwrite(field[i], &pack[j], *base, nbit[i]);
    if (*base < 0) {
      j++;
      if (j >= n) {
        fprintf(stderr, "SING_pack_inst_low: unable to pack in %d UINT64%s\n", n, (1 == n) ? "" : "s");
        return -1;
      }
      digit_left = - (*base) - 1;
      if (digit_left > 0) SING_bitwrite(field[i] & SING_bitmask(digit_left - 1, digit_left), &pack[j], 63, digit_left);
      *base += 64;
    }
/*
    {
      int k;
      dbp("j = %d, ", j);
      for (k = 0; k < n; k++)
        dbp(stderr, "%016llx ", pack[k]);
      dbp(stderr, "\n");
    }
*/
  }
  return 0;
}

/*
 SING_pack_inst
  return value: zero when succeeded, non-zero when failed
 */
int SING_pack_inst(const INST *inst, UINT64 *pack) 
{
  int err = 0, base;

  int isp_field[] = {
    inst->ll, inst->m_imr, inst->m_omr, inst->m_ifsel,
    inst->t_write, inst->t_shortstop, inst->t_isel,
    inst->r_write, inst->r_isel, inst->r_wadr, inst->r_wadri, inst->r_wwl,
    inst->r_radra, inst->r_radria, inst->r_rwla, inst->r_radrb, inst->r_radrib, inst->r_rwlb,
    inst->l_write, inst->l_isel, inst->l_adr, inst->l_adri, inst->l_tregadr, inst->l_wl,
    inst->fmul_shift50a, inst->fmul_round50a, inst->fmul_rounda, inst->fmul_normala,
    inst->fmul_shift25b, inst->fmul_shift50b, inst->fmul_roundb, inst->fmul_normalb,
    inst->fmul_round, inst->fmul_normalo, inst->fmul_isela, inst->fmul_iselb,
    inst->fadd_normala, inst->fadd_normalb, inst->fadd_signb, inst->fadd_round, inst->fadd_normalo,
    inst->fadd_isela, inst->fadd_iselb,
    inst->ialu_ialuop, inst->ialu_unsigned, inst->fsel,
    inst->bm_write, inst->bm_adr, inst->bm_peadr, inst->bm_wl};

  int idpread_field[] = {
    inst->idprv, inst->idprid, inst->idpwait
  };
  int idprrn_field[] = {
    inst->idprrn_valid, inst->idprrn_select};

  int idp_field[] = {
    inst->idp_id};

  int rrn_field[] = {
    inst->rrn_adr, inst->rrn_n, inst->rrn_bbadr, inst->rrn_reduc, inst->rrn_wl, inst->rrn_fsel,
    inst->rrn_normala, inst->rrn_normalb, inst->rrn_signb, inst->rrn_round, inst->rrn_normalo,
    inst->rrn_ialuop, inst->rrn_unsigned, inst->rrn_odpoe, inst->rrn_sregen, inst->rrn_conv};

  pack[0] = pack[1] = pack[2] = pack[3] = 0;
  base = 63; err = SING_pack_inst_low(pack, &base, 2, isp_field, isp_nbit, ArraySize(isp_nbit));


  base = 63; err = SING_pack_inst_low(pack+2,&base, 2, idpread_field, idpread_nbit, ArraySize(idpread_nbit));
  
  err |= SING_pack_inst_low(pack + 2, &base, 2, idprrn_field, idprrn_nbit, ArraySize(idprrn_nbit));
  if (!inst->idprrn_valid) return err;

  if (0 == inst->idprrn_select)
    err |= SING_pack_inst_low(pack + 2, &base, 2, idp_field, idp_nbit, ArraySize(idp_nbit));
  else
    err |= SING_pack_inst_low(pack + 2, &base, 2, rrn_field, rrn_nbit, ArraySize(rrn_nbit));

  return err;
}

int SING_register_instlist_already_packed(CTRLinfo *c, const INSTLIST *instlist) 
{
  CTRL_coc_write(c, instlist->addr << 2, instlist->pack, 4 * instlist->ninst);
  return 0;
}

int SING_pack_instlist(INSTLIST* instlist) 
{
  int i;
  for (i = 0; i < instlist->ninst; i++) {
    if (SING_pack_inst(&(instlist->insts[i]), &(instlist->pack[4 * i]))) {
      fprintf(stderr, "SING_pack_instlist: failed in SING_pack_inst\n");
      return -1;
    }
  }
  return 0;
}

int SING_register_instlist(CTRLinfo *c, INSTLIST *instlist) 
{
  /* addr right? */
  SING_pack_instlist(instlist);
  SING_register_instlist_already_packed(c, instlist);
  return 0;
}

int SING_send_instlist(CTRLinfo *c, const INSTLIST *instlist, int niter, int nonblock) 
{
    //  fprintf(stderr, "addr = %x\n", instlist->addr);
  CTRL_coc_start(c, instlist->addr, instlist->ninst, niter, nonblock);
  return 0;
}

INST *SING_new_inst(INSTLIST *instlist) 
{
  INST *inst;

  if (instlist->ninst >= instlist->maxinst) {
    fprintf(stderr, "SING_new_inst: the instruction list is full\n");
    return (void *)0;
  }
  inst = &(instlist->insts[instlist->ninst]);
  instlist->ninst++;
  SING_clear_inst(inst);
  return inst;
}

int SING_delete_inst(INSTLIST *instlist) 
{
  if (instlist->ninst <= 0) {
    fprintf(stderr, "SING_delete_inst: instruction list is empty\n");
    return -1;
  }
  instlist->ninst--;
  return 0;
}

int SING_add_inst_str(const char* str, INSTLIST *instlist) 
{
  INST *inst;
  inst = SING_new_inst(instlist);
  if (!inst) {
    fprintf(stderr, "SING_add_inst_str: unable to add a new instruction\n");
    return -1;
  }
  if (SING_scan_inst_str(str, inst)) {
    fprintf(stderr, "SING_add_inst_str: failed in SING_scan_inst_str\n");
    return -1;
  }
  return 0;
}

int SING_add_inst(const INST *inst, INSTLIST *instlist) 
{
  INST *new_inst = SING_new_inst(instlist);

  if (!new_inst) {
    fprintf(stderr, "SING_add_inst: unable to add a new instruction\n");
    return -1;
  }
  *new_inst = *inst;
  /*
  if (SING_pack_inst(inst, instlist->pack + 4 * instlist->ninst)) {
    fprintf(stderr, "SING_add_inst: failed in SING_pack_inst\n");
    return -1;
  }
  */
  return 0;
}



int SING_iregister_i(SINGISP_INFO *info, CTRLinfo *c, int ninst, const char **insts) 
{
  int i, id;
  id = SING_ialloc_i(info, ninst);
  INSTLIST *instlist = &(info->instlists)[id];

  for (i = 0; i < ninst; i++) {
    if (0 != SING_add_inst_str(insts[i], instlist)) {
      fprintf(stderr, "SING_iregister: failed at SING_add_inst_str\n");
      return -1;
    }
  }
  SING_register_instlist(c, instlist);
  return id;
}

int SING_iflush_i(SINGISP_INFO *info) 
{
  info->ninstlist = 0;
  info->inst_addr_curr = 0;
  return 0;
}

/*
  SING_ialloc
  return value: instruction list ID when succeeded, negative value when failed

  error processing has to be re-implemented (?)
 */
int SING_ialloc_i(SINGISP_INFO *info, int maxinst) 
{
  int id = info->ninstlist;
  if (maxinst > MAX_INST) {
    fprintf(stderr, "SING_ialloc: maxinst is too large\n");
    return -1;
  }
  if (info->inst_addr_curr + maxinst >= MAX_INSTADDR) {
    dbp("inst_addr_curr = %d, maxinst = %d\n", info->inst_addr_curr, maxinst);
    fprintf(stderr, "SING_ialloc: short of instruction address space\n");
    return -1;
  }
  (info->instlists)[id].addr = info->inst_addr_curr;
  (info->instlists)[id].maxinst = maxinst;
  (info->instlists)[id].ninst = 0;

  if (info->ninstlist >= MAX_INSTLIST) {
    fprintf(stderr, "SING_ialloc: no room for another instruction list\n");
    return -1;
  }
  info->ninstlist++;
  info->inst_addr_curr += maxinst;
  return id;
}

int SING_ipop_i(SINGISP_INFO *info) 
{
  if (info->ninstlist <= 0) {
    fprintf(stderr, "SING_ipop: no instruction is registered\n");
    return -1;
  }
  info->ninstlist--;
  info->inst_addr_curr -= (info->instlists)[info->ninstlist].maxinst;
  return 0;
}

static inline UINT64 rdtsc2()
{
	UINT64 ret;
	asm __volatile__("rdtsc  \n\t" \
			 "shlq $32, %%rdx \n\t" \
			 "orq  %%rdx, %%rax \n\t" \
			 "movq %%rax, %0" : "=g"(ret) : :"rax", "rdx");
	return ret;
}

int SING_execute_i(SINGISP_INFO *info, 
		   CTRLinfo *c,
		   int init_id,
                int loop_id,
                int final_id,
                int loop_count,
                void *loop_out,
                void *final_out) 
{
  //CTRL_coc_reset(c);
  UINT64 *loop_res = (UINT64 *)loop_out;
  UINT64 start, init, loop, rrn;
  double elp_rrn;

  start = rdtsc2();
  if(init_id >= 0){
    SING_send_instlist(c, &(info->instlists)[init_id], 1, 1); 
    //SING_send_instlist(c, &(info->instlists)[init_id], 1, 0); 
  }
  init = rdtsc2();
  if(loop_id >= 0){
    if(loop_out){
      CTRL_read_as_many_as_async_start(c, loop_res, info->loop_result_size*loop_count);
    }
    SING_send_instlist(c, &(info->instlists)[loop_id], loop_count, 1);
    if (loop_out) {
      CTRL_read_as_many_as_async_end(c, loop_res, info->loop_result_size*loop_count);
      loop_res += (info->loop_result_size);
    } else {
#ifdef LOOPWAIT      
      while (CTRL_coc_busy(c)) {
      };
#endif     
    }
  }
  loop = rdtsc2();
  if(final_id >= 0){
    CTRL_read_as_many_as_async_start(c, final_out, info->result_size);
    SING_send_instlist(c, &(info->instlists)[final_id], 1, 1);
    CTRL_read_as_many_as_async_end(c, final_out, info->result_size);
    //CTRL_read_as_many_as(c, final_out, info->result_size);
  }
  rrn = rdtsc2();

#ifdef PERFMON
  elp_rrn = (double)(rrn-loop)/2.6e9;
  printf("init : %lld , loop : %lld, rrn : %lld \n", init-start, loop-init, rrn-loop);
  printf("rrn datarate : %e B /s \n", (info->result_size)*8.0/elp_rrn);
#endif
  return 0;
}

void nwait(int n)
{
    UINT64 first=rdtsc2();
    while (first + n > rdtsc2());
}
void SING_nwait(int n)
{
    nwait(n);
}

UINT64 SING_rdtsc()
{
    return rdtsc2();
}
int SING_start_code(SINGISP_INFO *info,
		    int nchips,
		    CTRLinfo *c[nchips],
		    int codeid,
		    int loop_count)
{
    int i;
    for(i=0;i<nchips; i++){
	SING_send_instlist(c[i], &(info->instlists)[codeid], loop_count, 1);
	
    }
    return 0;
}


int SING_finish_code(SINGISP_INFO *info,
		     int nchips,
		     CTRLinfo *c[nchips])
{
    int ic;
    for(ic=0;ic<nchips; ic++){
	while (CTRL_coc_busy(c[ic])) {};
    }
    return 0;
}


int SING_execute2_i_start(SINGISP_INFO *info,
			  int nchips,
			  CTRLinfo *c[nchips],
			  
			  int ncodes,
			  int codeids[],
			  int loop_count,
			  void *loop_out[nchips],
			  void *final_out[nchips],
			  int sizes[]) 
{
  //CTRL_coc_reset(c);
  UINT64 *loop_res = (UINT64 *)loop_out;
  UINT64 start, init, loop, rrn;
  double elp_rrn;
  int init_id = codeids[0];
  int loop_id = codeids[1];
  int final_id= codeids[2];
  int remaining_words[ncodes];
  int dma_location[ncodes];
  int total_size = 0;
  int i,nremain;
  for(i=0;i<ncodes-2;i++){
      total_size += sizes[i];
      dma_location[i]=total_size;
  }
  nremain = total_size;
  for(i=0;i<ncodes-2;i++){
      nremain -= sizes[i];
      remaining_words[i]=nremain;
  }
  int nccom = nchips;
  if (nchips == 4) {
      nccom = 1;
      CTRL_beginbcast(c[0]);
  }
  if(init_id >= 0){
      for(i=0;i<nccom; i++){
	  SING_send_instlist(c[i], &(info->instlists)[init_id], 1, 1);
      }
  }
  if(loop_id >= 0){
      for(i=0;i<nchips; i++){
	  if(loop_out[i]){
	      CTRL_read_as_many_as_async_start(c[i], loop_res,
					       info->loop_result_size*loop_count);
	  }
      }
      for(i=0;i<nccom; i++){
	  SING_send_instlist(c[i], &(info->instlists)[loop_id], loop_count, 1);
      }
  }
  if (nchips == 4)  CTRL_endbcast(c[0]);
  return 0;
}

int SING_execute2_i_end(SINGISP_INFO *info,
			int nchips,
			CTRLinfo *c[nchips],
			int ncodes,
			int codeids[],
			int loop_count,
			void *loop_out[nchips],
			void *final_out[nchips],
			int sizes[]) 
{
    //CTRL_coc_reset(c);
    UINT64 *loop_res = (UINT64 *)loop_out;
    UINT64 start, init, loop, rrn;
    double elp_rrn;
    int init_id = codeids[0];
    int loop_id = codeids[1];
    int final_id= codeids[2];
    int remaining_words[ncodes];
    int dma_location[ncodes];
    int total_size = 0;
    int i,nremain, ic;
    for(i=0;i<ncodes-2;i++){
	total_size += sizes[i];
	dma_location[i]=total_size;
    }
    nremain = total_size;
    for(i=0;i<ncodes-2;i++){
	nremain -= sizes[i];
	remaining_words[i]=nremain;
    }
    int nccom = nchips;
    int bcmode = 0;
    if (nchips == 4) {
	nccom = 1;
	bcmode = 1;
	CTRL_beginbcast(c[0]);
    }
    
    if(loop_id >= 0){
	for(ic=0;ic<nchips; ic++){
	    if (loop_out[ic]) {
		CTRL_read_as_many_as_async_end(c[ic], loop_res,
					       info->loop_result_size*loop_count);
		loop_res += (info->loop_result_size);
	    } 
	}
    }
    if(final_id >= 0){
	for(ic=0;ic<nchips; ic++){
	    //
	    // assumed that seqid = 0 is already registered
	    //
	    hib_start_dma_seqMC(c[ic]->hib_id, 0, 1);
	    
	}
	if (bcmode == 0){
	    for(i=2;i<ncodes;i++){
		for(ic=0;ic<nchips; ic++){
		    if (i>2){
			while (hib_remaining_words_dmawMC(c[ic]->hib_id)
			       > (total_size-dma_location[i-3])/8){
			}
		    }
		    SING_send_instlist(c[ic], &(info->instlists)[codeids[i]],
				       1, 1);
		}
	    }
	}else{
	    for(i=2;i<ncodes;i++){
		for(ic=nchips-1;ic<nchips; ic++){
		    if (i>2){
			while (hib_remaining_words_dmawMC(c[ic]->hib_id)
			       > (total_size-dma_location[i-3]+500)/8){
			    // 1000 here is unsafety factor
			    // start next DMA when remaining data is  500byts
			    SING_nwait(2000);
			}
		    }
		}
		SING_send_instlist(c[0], &(info->instlists)[codeids[i]],
				   1, 1);
	    }
	}
	for(ic=0;ic<nchips; ic++){
	    hib_finish_dmawMC(c[ic]->hib_id);
	    UINT64 *dst = (UINT64 *)final_out[ic];
	    UINT64 *src = c[ic]->h->dmaw_buf;
	    for (i=0; i< info->result_size; i++){
		dst[i] = src[i];
	    }
	}
    }else{
	for(ic=0;ic<nchips; ic++){
	    while (CTRL_coc_busy(c[ic])) {};
	}
    }
    CTRL_endbcast(c[0]);
	
    return 0;
}


int SING_execute2_i_end_old(SINGISP_INFO *info,
			int nchips,
			CTRLinfo *c[nchips],
			int ncodes,
			int codeids[],
			int loop_count,
			void *loop_out[nchips],
			void *final_out[nchips],
			int sizes[]) 
{
    //CTRL_coc_reset(c);
    UINT64 *loop_res = (UINT64 *)loop_out;
    UINT64 start, init, loop, rrn;
    double elp_rrn;
    int init_id = codeids[0];
    int loop_id = codeids[1];
    int final_id= codeids[2];
    int remaining_words[ncodes];
    int dma_location[ncodes];
    int total_size = 0;
    int i,nremain, ic;
    for(i=0;i<ncodes-2;i++){
	total_size += sizes[i];
	dma_location[i]=total_size;
    }
    nremain = total_size;
    for(i=0;i<ncodes-2;i++){
	nremain -= sizes[i];
	remaining_words[i]=nremain;
    }
    
    if(loop_id >= 0){
	for(ic=0;ic<nchips; ic++){
	    if (loop_out[ic]) {
		CTRL_read_as_many_as_async_end(c[ic], loop_res,
					       info->loop_result_size*loop_count);
		loop_res += (info->loop_result_size);
	    } 
	}
    }
    if(final_id >= 0){
	for(i=2;i<ncodes;i++){
	    for(ic=0;ic<nchips; ic++){
		if ((loop_out[ic]==NULL) && (i==2))
		    while (CTRL_coc_busy(c[ic])) {};
		if (i>2)hib_finish_dmawMC(c[ic]->hib_id);
		hib_start_dmawMC(c[ic]->hib_id, sizes[i-2]/8,
				 c[ic]->h->dmaw_buf+dma_location[i-3]/8);
		    
		SING_send_instlist(c[ic], &(info->instlists)[codeids[i]], 1, 1);
	    }
	}
	for(ic=0;ic<nchips; ic++){
	    hib_finish_dmawMC(c[ic]->hib_id);
	    UINT64 *dst = (UINT64 *)final_out[ic];
	    UINT64 *src = c[ic]->h->dmaw_buf;
	    for (i=0; i< info->result_size; i++){
		dst[i] = src[i];
	    }
	}
    }else{
	for(ic=0;ic<nchips; ic++){
	    while (CTRL_coc_busy(c[ic])) {};
	}
    }
	
    return 0;
}


int SING_execute2_i(SINGISP_INFO *info, 
		    CTRLinfo *c,
		    int ncodes,
		    int codeids[],
		    int loop_count,
		    void *loop_out,
		    void *final_out,
		    int sizes[]) 
{
    SING_execute2_i_start(info, 1, &c, ncodes, codeids, loop_count,
			  &loop_out, &final_out,sizes);
    //    fprintf(stderr,"call i_end\n");
    return SING_execute2_i_end(info, 1, &c, ncodes, codeids, loop_count,
			       &loop_out,&final_out,sizes);
}

    

int SING_execute_i_async(SINGISP_INFO *info, 
		   CTRLinfo *c,
		   int init_id,
                int loop_id,
                int final_id,
                int loop_count,
                void *loop_out,
                void *final_out) 
{
  //CTRL_coc_reset(c);
  UINT64 *loop_res = (UINT64 *)loop_out;
  UINT64 start, init, loop, rrn;
  double elp_rrn;

  start = rdtsc2();
  if(init_id >= 0){
    SING_send_instlist(c, &(info->instlists)[init_id], 1, 0); 
  }
  init = rdtsc2();
  if(loop_id >= 0){
    if(loop_out){
      CTRL_read_as_many_as_async_start(c, loop_res, info->loop_result_size*loop_count);
    }
    SING_send_instlist(c, &(info->instlists)[loop_id], loop_count, 1);
    if (loop_out) {
      CTRL_read_as_many_as_async_end(c, loop_res, info->loop_result_size*loop_count);
      loop_res += (info->loop_result_size);
    } else {
      /*
      while (CTRL_coc_busy(c)) {
      };
      */
    }
  }
  loop = rdtsc2();
  if(final_id >= 0){
    CTRL_read_as_many_as_async_start(c, final_out, info->result_size);
    SING_send_instlist(c, &(info->instlists)[final_id], 1, 1);
    /*
    CTRL_read_as_many_as_async_end(c, final_out, info->result_size);
    //CTRL_read_as_many_as(c, final_out, info->result_size);
    */
  }
  rrn = rdtsc2();

#ifdef PERFMON
  elp_rrn = (double)(rrn-loop)/2.6e9;
  printf("init : %lld , loop : %lld, rrn : %lld : rrnstart = %lld \n", init-start, loop-init, rrn-loop, rrn);
  //printf("rrn datarate : %e B /s \n", (info->result_size)*8.0/elp_rrn);
#endif
  
  return 0;
}

//unit of size is byte.
int SING_set_result_size_i(SINGISP_INFO *info, int dum, int size)
{
  info->loop_result_size = dum / 8;
  info->result_size = size / 8; 
  return 0;
}

