#ifndef _SING_ISP_H
#define _SING_ISP_H

#include <typedef.h>
#include <ctrl.h>
#include <sing_conf.h>

typedef struct {
  int ll, m_imr, m_omr, m_ifsel;
  int t_write, t_shortstop, t_isel;
  int r_write, r_isel, r_wadr, r_wadri, r_wwl;
  int r_radra, r_radria, r_rwla, r_radrb, r_radrib, r_rwlb;
  int l_write, l_isel, l_adr, l_adri, l_tregadr, l_wl;
  int fmul_shift50a, fmul_round50a, fmul_rounda, fmul_normala;
  int fmul_shift25b, fmul_shift50b, fmul_roundb, fmul_normalb;
  int fmul_round, fmul_normalo, fmul_isela, fmul_iselb;
  int fadd_normala, fadd_normalb, fadd_signb, fadd_round, fadd_normalo;
  int fadd_isela, fadd_iselb;
  int ialu_ialuop, ialu_unsigned, fsel;
  int bm_write, bm_adr, bm_peadr, bm_wl;

  int idprv, idprid, idpwait; //For DRAM cont.
  int idprrn_valid, idprrn_select;
  int idp_id;
  int rrn_adr, rrn_n, rrn_bbadr, rrn_reduc, rrn_wl, rrn_fsel;
  int rrn_normala, rrn_normalb, rrn_signb, rrn_round, rrn_normalo;
  int rrn_ialuop, rrn_unsigned, rrn_odpoe, rrn_sregen;
  int rrn_conv;
} INST;

typedef struct {
  int addr;
  int maxinst;
  int ninst;
  INST insts[MAX_INST];
  UINT64 pack[4 * MAX_INST];
} INSTLIST;  

typedef struct {
  INSTLIST instlists[MAX_INSTLIST];
  int ninstlist;
  int inst_addr_curr;
  int loop_result_size;
  int result_size;
} SINGISP_INFO;

void init_ispinfo(SINGISP_INFO *info);
int SING_iflush_i(SINGISP_INFO *info);
int SING_ialloc_i(SINGISP_INFO *info, int maxinst);
int SING_ipop_i(SINGISP_INFO *info);
int SING_iregister_i(SINGISP_INFO *info, CTRLinfo *c, int ninst, const char **insts);
int SING_send_instlist(CTRLinfo *c, const INSTLIST *instlist, int niter, int nonblock);
int SING_register_instlist(CTRLinfo *c, INSTLIST *instlist) ;
int SING_register_instlist_already_packed(CTRLinfo *c, const INSTLIST *instlist);
int SING_execute_i(SINGISP_INFO *info,
		   CTRLinfo *c,
		   int init_id,
		   int loop_id,
		   int final_id,
		   int loop_count,
		   void *loop_out,
		   void *final_out);

int SING_execute2_i(SINGISP_INFO *info,
		    CTRLinfo *c,
		    int ncodes, int codes[],
		    int loop_count, void *loop_out, void *final_out,
		    int sizes[]);
int SING_set_result_size_i(SINGISP_INFO *info, int dum, int size);

int SING_execute_i_async(SINGISP_INFO *info, CTRLinfo *c, int init_id, int loop_id,
			 int final_id, int loop_count, void *loop_out, void *final_out); 
INST *SING_new_inst(INSTLIST *instlist);

int SING_start_code(SINGISP_INFO *info,
		    int nchips,
		    CTRLinfo *c[nchips],
		    int codeid,
		    int loop_count);

int SING_finish_code(SINGISP_INFO *info,
		     int nchips,
		     CTRLinfo *c[nchips]);
void SING_nwait(int n);
UINT64 SING_rdtsc();

#endif
