#include <stdio.h>
#include <assert.h>
#include <ctrl.h>
#include <sing_isp.h>
#include <sing_conf.h>
#include <sing_embm.h>
#include <sing_lm.h>

void init_lminfo(SINGLM_INFO *info)
{
  info->lm_hostlen = 0;
  info->lm_length  = 0;
  info->lm_nword   = 0;
  info->firstsend  = 1;
}

void SING_LM_status(SINGLM_INFO *info) 
{
  printf("LM status:\n");
  printf(" length = %d, nword = %d\n", info->lm_length, info->lm_nword);
}

int SING_calc_convhostlen(int length, const int *conversions)
{
	int i, halfword;
	
	halfword = 0;
	for(i=0;i<length;i++){
		switch(conversions[i]){
			case 0:
			case 1:
			case 2:
			case 3:
			case 4:
			case 5:
			case 6:
				halfword += 2;
				break;
			
			case 7:
			case 8:
			case 9:
				halfword += 1;
				break;
		}
	}
	if(halfword%2 != 0){
		fprintf(stderr, "pack conversions : conversions is not allowed to odd 32bit word!\n");
		assert(0);
	}
	 
	return halfword/2;
}

int SING_LM_write_regconv_i(SINGLM_INFO *info, CTRLinfo *c, int length, int *conversions) 
{
  int err, convlength;
  
  info->lm_length = SING_calc_convhostlen(length, conversions);
  err = SING_pack_conversions(length, conversions, info->convtable, &convlength, &(info->lm_nword));
  
  CTRL_hit_convtable(c, 0, info->convtable, convlength);
  CTRL_hit_length(c, convlength);
  //  fprintf(stderr,"regconv %d %d %d\n", err, (int)info->lm_length, convlength);
  return 0;
}

static inline UINT64 rdtsc2()
{
	UINT64 ret;
	asm __volatile__("rdtsc  \n\t" \
			 "shlq $32, %%rdx \n\t" \
			 "orq  %%rdx, %%rax \n\t" \
			 "movq %%rax, %0" : "=g"(ret) : :"rax", "rdx");
	return ret;
}


int SING_LM_write_i(SINGLM_INFO *info, SINGISP_INFO *ispinfo, CTRLinfo *c, void *hostp, int lmoffset, int peeach) 
{
  int i, j, niter, ilid, n, lend;
  int bmoffset = 2 * TMP_BMADDR;
  UINT64 *p = (UINT64 *)hostp;
  INSTLIST *instlist;
  int bmsadr;
  UINT64 *pack, s, e, e_isp;
  

  ilid = SING_ialloc_i(ispinfo, MAX_INST);
  instlist = &(ispinfo->instlists)[ilid];
  if (0 == peeach) niter = 1; else niter = peeach;

    //  dbp("**** %d, %d, %d ****\n", peeach, lm_nword, peeach * lm_nword);

  /* send no more than 256 long word at one time */
  s = rdtsc2();
  CTRL_hit_reset_count(c);
  bmsadr = TMP_BMADDR;
  for (i = 0; i < niter; i++) {
#if 1
    CTRL_hit_header(c, info->lm_nword, /* len */
    				  2 * bmsadr, /* bmsadr */
    				  0, /*nbm */
    				  0, /* bbnm */
    				  0); /*seq*/
    CTRL_hit_data(c, p, info->lm_length);
#else
    CTRL_hit_header_data(c, info->lm_nword, 
			 2*bmsadr,
			 0 , 
			 0 ,
			 0 , 
			 p ,
			 info->lm_length);
#endif

    p += info->lm_length;
    bmsadr += info->lm_nword;	
  }
  e = rdtsc2();
#ifdef PERFMON
  printf("hit transfer .. %e B /s \n", (double)(info->lm_length*8*niter)/((e-s)/2.6e9));
#endif
  pack = instlist->pack;

  n = (info->lm_nword + 3) / 4;
  lend = 1 + ((info->lm_nword - 1) % 4);
  if (0 != peeach) {
    *pack++ = 0x8024000000000000;
    *pack++ = 0x0001470000000000;
    *pack++ = 0;
    *pack++ = 0;
    instlist->ninst++;
    *pack++ = 0x8024000000000000;
    *pack++ = 0x0001470000000000;
    *pack++ = 0;
    *pack++ = 0;
    instlist->ninst++;
    /* clear t register */
  }

  for (i = 0; i < niter; i++) {
    if (0 != peeach) {
      *pack++ = 0x8080000000000000;
      *pack++ = 0x0002470000000000;
      *pack++ = 0;
      *pack++ = 0;
      instlist->ninst++;
      *pack++ = 0x8024000000000000;
      *pack++ = 0x0001420000000000;
      *pack++ = 0;
      *pack++ = 0;
      instlist->ninst++;
      /* compare treg and PEid */
    }
    for (j = 0; j < n; j++) {
      UINT64 m_imr = (0 != peeach) ? 1 : 0;
      UINT64 ll = ((n - 1) == j) ? lend : 4;
      UINT64 l_adr =  lmoffset + j * 8;
      UINT64 bmadr = bmoffset + j * 8;
      *pack++ = 0x00000000006000a0 | (m_imr << 58) | (ll << 61) | (l_adr << 11);
      *pack++ = 0x0000000000100000 | (bmadr << 26); 
      *pack++ = 0;
      *pack++ = 0;
      instlist->ninst++;
      *pack++ = 0;
      *pack++ = 0;
      *pack++ = 0;
      *pack++ = 0;
      instlist->ninst++;
    }
    bmoffset += 2 * info->lm_nword;
  }

  SING_register_instlist_already_packed(c, instlist);
  
  SING_send_instlist(c, instlist, 1, 1);
  SING_ipop_i(ispinfo);

  /*
  while(CTRL_coc_busy(c)){
  	//printf("wait transfer ISP..\n");
  }
  */
  e_isp = rdtsc2();

#ifdef PERFMON
  printf("hit trans : %lld , isp_exec : %lld \n", e-s, e_isp-e);
#endif

  return 0;
}

int SING_LM_write_i_2(SINGLM_INFO *info, SINGISP_INFO *ispinfo, CTRLinfo *c, void *hostp, int lmoffset, int peeach) 
{
  int i, j, niter, ilid, n, lend;
  int bmoffset = 2 * TMP_BMADDR;
  UINT64 *p = (UINT64 *)hostp;
  INSTLIST *instlist;
  int bmsadr;
  UINT64 *pack, s, e, e_isp;
  

  if (0 == peeach) niter = 1; else niter = peeach;

    //  dbp("**** %d, %d, %d ****\n", peeach, lm_nword, peeach * lm_nword);

  /* send no more than 256 long word at one time */
  s = rdtsc2();
  //CTRL_hit_reset_count(c);
  bmsadr = TMP_BMADDR;
  CTRL_beginbuff(c);
#if 1
  for (i = 0; i < niter; i++) {
#if 0
    CTRL_hit_header(c, info->lm_nword, /* len */
    				  2 * bmsadr, /* bmsadr */
    				  0, /*nbm */
    				  0, /* bbnm */
    				  0); /*seq*/
    CTRL_hit_data(c, p, info->lm_length);
#else
    CTRL_hit_header_data(c, info->lm_nword*1, 
			 2*bmsadr,
			 0 , 
			 0 ,
			 0 , 
			 p ,
			 info->lm_length*1);
#endif

    p += info->lm_length*1;
    bmsadr += info->lm_nword*1;	
  }
#endif

  //printf("hit transfer .. %e B /s \n", (double)(info->lm_length*8*niter)/((e-s)/2.6e9));
#if 1
  if(info->firstsend==1){
    ilid = SING_ialloc_i(ispinfo, MAX_INST);
    instlist = &(ispinfo->instlists)[ilid];
    pack = instlist->pack;

    n = (info->lm_nword + 3) / 4;
    lend = 1 + ((info->lm_nword - 1) % 4);
    if (0 != peeach) {
      *pack++ = 0x8024000000000000;
      *pack++ = 0x0001470000000000;
      *pack++ = 0;
      *pack++ = 0;
      instlist->ninst++;
      *pack++ = 0x8024000000000000;
      *pack++ = 0x0001470000000000;
      *pack++ = 0;
      *pack++ = 0;
      instlist->ninst++;
      /* clear t register */
    }

    for (i = 0; i < NPE; i++) {
      if (0 != peeach) {
	*pack++ = 0x8080000000000000;
	*pack++ = 0x0002470000000000;
	*pack++ = 0;
	*pack++ = 0;
	instlist->ninst++;
	*pack++ = 0x8024000000000000;
	*pack++ = 0x0001420000000000;
	*pack++ = 0;
	*pack++ = 0;
	instlist->ninst++;
	/* compare treg and PEid */
      }
      for (j = 0; j < n; j++) {
	UINT64 m_imr = (0 != peeach) ? 1 : 0;
	UINT64 ll = ((n - 1) == j) ? lend : 4;
	UINT64 l_adr =  lmoffset + j * 8;
	UINT64 bmadr = bmoffset + j * 8;
	*pack++ = 0x00000000006000a0 | (m_imr << 58) | (ll << 61) | (l_adr << 11);
	*pack++ = 0x0000000000100000 | (bmadr << 26); 
	*pack++ = 0;
	*pack++ = 0;
	instlist->ninst++;
	*pack++ = 0;
	*pack++ = 0;
	*pack++ = 0;
	*pack++ = 0;
	instlist->ninst++;
      }
      bmoffset += 2 * info->lm_nword;
    }
    SING_register_instlist_already_packed(c, instlist);
    info->ipart_id = ilid;
    info->firstsend = 0;
  } else {
    instlist = &(ispinfo->instlists)[info->ipart_id];
  }
#ifdef POLLSENDI
  //SING_send_instlist(c, instlist, 1, 0);
#else
  SING_send_instlist(c, instlist, 1, 1);
#endif

  CTRL_flushbuff(c);
  e = rdtsc2();

  //SING_ipop_i(ispinfo);

  /*
  while(CTRL_coc_busy(c)){
  	//printf("wait transfer ISP..\n");
  }
  */
#endif

  e_isp = rdtsc2();
  //#define PERFMON
#ifdef PERFMON
  printf("hit trans : %lld , isp_exec : %lld , all : %lld \n", e-s, e_isp-e, e_isp-s);
#endif

  return 0;
}

int SING_LM_write_i_3(SINGLM_INFO *info, SINGISP_INFO *ispinfo, CTRLinfo *c, void *hostp, int lmoffset, int peeach, int nhead) 
{
  int i, j, niter, ilid, n, lend;
  int bmoffset = 2 * TMP_BMADDR;
  UINT64 *p = (UINT64 *)hostp;
  INSTLIST *instlist;
  int bmsadr;
  UINT64 *pack, s, e, e_isp;
  

  if (0 == peeach) niter = 1; else niter = peeach;

    //  dbp("**** %d, %d, %d ****\n", peeach, lm_nword, peeach * lm_nword);

  /* send no more than 256 long word at one time */
  s = rdtsc2();
  //CTRL_hit_reset_count(c);
  bmsadr = TMP_BMADDR;
  CTRL_beginbuff(c);
#if 1
  for (i = 0; i < niter; i+=nhead){
#if 0
    CTRL_hit_header(c, info->lm_nword, /* len */
    				  2 * bmsadr, /* bmsadr */
    				  0, /*nbm */
    				  0, /* bbnm */
    				  0); /*seq*/
    CTRL_hit_data(c, p, info->lm_length);
#else
    CTRL_hit_header_data(c, info->lm_nword*nhead, 
			 2*bmsadr,
			 0 , 
			 0 ,
			 0 , 
			 p ,
			 info->lm_length*nhead);
#endif

    p += info->lm_length*nhead;
    bmsadr += info->lm_nword*nhead;	
  }
#endif

  //printf("hit transfer .. %e B /s \n", (double)(info->lm_length*8*niter)/((e-s)/2.6e9));
#if 1
  if(info->firstsend==1){
    ilid = SING_ialloc_i(ispinfo, MAX_INST);
    instlist = &(ispinfo->instlists)[ilid];
    pack = instlist->pack;

    n = (info->lm_nword + 3) / 4;
    lend = 1 + ((info->lm_nword - 1) % 4);
    if (0 != peeach) {
      *pack++ = 0x8024000000000000;
      *pack++ = 0x0001470000000000;
      *pack++ = 0;
      *pack++ = 0;
      instlist->ninst++;
      *pack++ = 0x8024000000000000;
      *pack++ = 0x0001470000000000;
      *pack++ = 0;
      *pack++ = 0;
      instlist->ninst++;
      /* clear t register */
    }

    for (i = 0; i < NPE; i++) {
      if (0 != peeach) {
	*pack++ = 0x8080000000000000;
	*pack++ = 0x0002470000000000;
	*pack++ = 0;
	*pack++ = 0;
	instlist->ninst++;
	*pack++ = 0x8024000000000000;
	*pack++ = 0x0001420000000000;
	*pack++ = 0;
	*pack++ = 0;
	instlist->ninst++;
	/* compare treg and PEid */
      }
      for (j = 0; j < n; j++) {
	UINT64 m_imr = (0 != peeach) ? 1 : 0;
	UINT64 ll = ((n - 1) == j) ? lend : 4;
	UINT64 l_adr =  lmoffset + j * 8;
	UINT64 bmadr = bmoffset + j * 8;
	*pack++ = 0x00000000006000a0 | (m_imr << 58) | (ll << 61) | (l_adr << 11);
	*pack++ = 0x0000000000100000 | (bmadr << 26); 
	*pack++ = 0;
	*pack++ = 0;
	instlist->ninst++;
	*pack++ = 0;
	*pack++ = 0;
	*pack++ = 0;
	*pack++ = 0;
	instlist->ninst++;
      }
      bmoffset += 2 * info->lm_nword;
    }
    SING_register_instlist_already_packed(c, instlist);
    info->ipart_id = ilid;
    info->firstsend = 0;
  } else {
    instlist = &(ispinfo->instlists)[info->ipart_id];
  }
#ifdef POLLSENDI
  //SING_send_instlist(c, instlist, 1, 0);
#else
  SING_send_instlist(c, instlist, 1, 1);
#endif

  CTRL_flushbuff(c);
  e = rdtsc2();

  //SING_ipop_i(ispinfo);

  /*
  while(CTRL_coc_busy(c)){
  	//printf("wait transfer ISP..\n");
  }
  */
#endif

  e_isp = rdtsc2();
  //#define PERFMON
#ifdef PERFMON
  printf("hit trans : %lld , isp_exec : %lld , all : %lld \n", e-s, e_isp-e, e_isp-s);
#endif

  return 0;
}
