#!/usr/bin/env ruby
# checkalias.rb
#
# vsm alias checker
#
# Copyright Jun Makino All rights reserved.
# J. Makino, 2005/4/15
$LOAD_PATH.unshift(File.dirname($0))
$LOAD_PATH.push(ENV["ACSROOT"]+"/src/utilities") if ENV["ACSROOT"] != nil

require "clop.rb"
options_definition_string = <<-END

  Description: VSM alias checker
  Long description:
    Assembler alias checker program for VSM --- VPM assembly language
    (c) 2004- Jun Makino
  
    This program performs valious tests for the input file written in VSM, 
    and generates output machine code files. Maximum help is provided by 
    the command 
    "ruby #{$0} --help".


  Short name: 		-i
  Long name:		--input
  Value type:		string
  Variable name: 	infile
  Description:		Input VSM file name
  Long description:
    This option specifies the file name for input VSM file. If
    not specified, STDIN is used instead.



  Short name:		-l
  Long name:  		--list-alias
  Value type:  		bool
  Variable name:	aliaslisting
  Description:		List all defined aliases
  Long description:
    List all aliases if this option is specified, in the format
    name[regdef] first_line-final_line:first_address-last_address

  END

$clop = parse_command_line(options_definition_string, false)

class Alias
  attr_reader :name, :start, :end
  attr_accessor :first, :final
  def initialize(name, regstring)
    @name=name
    if regstring =~ /^\$(l*)(r)(\d+)(v*)$/
      @long =  $1 == "l"
      @start = $3.to_i
      @vect=  $4 == "v"
      @str = regstring[1,regstring.length]
      @end = @start+ ( @vect ? 4:1 )*  ( @long ? 1:1 ) - 1
    else
      raise "#{namestring} is not valid register "
    end
  end
  def to_s
    "#{@name}[#{@str}] #{@first}-#{@final}:#{@start}-#{@end}"
  end

end
input=  $clop.infile ?    open($clop.infile,"r") : STDIN
instring= input.gets(nil)
input.close
aliases = {}
regdata=[]
iline = 0
vlen = 4
instring.each{|s|
  a = s.split
  if a[0]=="alias"
    aliases[a[1]]=iline
    regdata[iline]=Alias.new(a[1],a[2])
  end
  iline += 1;
}

regdata.each{|x| 
  next unless x
  first=-1
  final=nil
  iline = 0
  vlen = 4
  avlen=[]
  instring.each{|s|
    iline += 1
    a = s.gsub(/;/," ; ").split
    vlen = a[1].to_i if a[0]=="vlen"
    avlen[iline]=vlen
    next if a[0]=="alias"  or a[0] =~ /(^\#)|^alias|^print/
      if a.index(x.name)
        first = iline if first == -1
        final = iline
      end
    if final
      x.first = first
      x.final = final
    end
  }
}
regdata.reject!{|x| x== nil or  x.final == nil}
regdata.each{|x| print x.to_s+"\n" if x} if $clop.aliaslisting
regdata.each{|x| 
  regdata.each{|y|
    next if x == y
    next if x.first > y.first
    next if x.first > y.final or   y.first > x.final 
    next if x.start > y.end or   y.start > x.end
    print "collision: ",x, " ", y, "\n"
  }
}
    

