#!/usr/bin/ruby
# vsm.rb
#
# vsm assembler
#
# Copyright Jun Makino All rights reserved.
# J. Makino, 2004/10/7-
$LOAD_PATH.unshift(File.dirname($0))
$LOAD_PATH.push(ENV["ACSROOT"]+"/src/utilities") if ENV["ACSROOT"] != nil

$default_idp_size=65536
$bememsize = 2048
$vsize = 4
require "clop.rb"

def printmessage(mtype, s)
  STDERR.print "#{mtype}: line #{$lineno} \"#{$current_line}\"\n"
  STDERR.print " "*8+s+"\n"
end
def printwarning(s)
  printmessage("Warning", s)
end
def printerror(s)
  printmessage("ERROR  ", s)
end

def remove_comment(s)
  s.gsub(/\#.*$/,"")
end

  def oneif(x)
    if x 
      1
    else
      0
    end
  end

class Instruction
  
  attr_reader :fsel, :faddflags, :ialuflags, :asmcode,
  :twused, :regwused, :lmemwused

  def initialize(s=nil)
    if s
      @asmcode = s if s.class ==String
    else
      @asmcode=""
    end
    @regaused =nil
    @regbused =nil
    @regwused =nil
    @regwisel =0
    @lmemrused =nil
    @lmemwused =nil
    @lmemwisel=0
    @bmrused =nil
    @bmwused =nil
    @trused =nil
    @twused =nil
    @twisel =0
    @fmulisel=[0,0]
    @fmulround=0
    @faddisel=[0,0]
    @faddround=0
    @fsel=0
    # set flags for arithmetic units and input muxes for them
    @faddflags=[1,1,0,1]
    @fmulflags=[0,0,1,1,0,0,1,1,1,1]
    @ialuflags=[0,0]
    @idpdata = nil
    @rdpdata = nil
    @wdpdata = nil
    @rrndata = nil
  end

  def muxcode(operand,faddaport)
    case operand.location 
    when /^t$/
      mux=2
      if @trused == nil
	@trused = operand
      else
	if operand != @trused
	  raise "Cannot use both of shortstop and normal value"
	end
      end
    when /^m$/
      if @lmemrused == nil
	mux=3
	@lmemrused = operand
      elsif operand == @lmemrused
	mux=3
      else
	raise "Cannot assign memory #{operand}, already used as #{@lmemrused}"
      end
    when /^r$/
      if operand==@regaused
	mux=0
      elsif operand==@regbused
	mux=1
      elsif @regaused == nil
	mux=0
	@regaused = operand
      elsif @regbused == nil
	mux=1
	@regbused = operand
      else
	raise "Cannot assign register #{operand}"
      end
    when /^pe$/
      mux=4
    when /^bb$/
      mux=5
    when /^fb$/
      mux=6
    else
      raise "Unknown location type for operand #{operand}"
    end
    if mux>3 and not faddaport
      raise "PB, BM specified for unsupported operand #{operand}"
    end
    mux
  end
  def setregb(operand)
    case operand.location 
    when /^r$/
      if operand==@regbused
	# operand reused
      elsif @regbused == nil
	@regbused = operand
      else
	raise "Cannot assign register #{operand}: line \"#{$current_line}\""
      end
    else
	raise "Non-register  #{operand} for bm instruction: line \"#{$current_line}\""
    end
  end
  
  def setstoragemux(operand,instype)
    round= operand.long ?  0:1
    if instype == "fmul"
      isel=0
      @fmulround |=round
    elsif instype == "fadd" or  instype == "ialu"
      isel=1
      @faddround |=round
    elsif instype == "bm" 
      if operand.location != "b"
	isel=2
      end
    else
      raise "Unknown instruction type #{instype}"
    end
    case operand.location 
    when /^t$/
      if @twused == nil
	@twisel=isel
	@twused = operand
      else
	raise "Output operand #{operand} already used in line: \"{#$current_line}\""
      end
    when /^m$/
      if @lmemwused == nil and @lmemrused == nil and 
	  @lmemwisel=isel
	@lmemwused = operand
      else
	raise "Output operand #{operand} already used in line: \"{#$current_line}\""
      end
    when /^r$/
      if @regwused == nil
	@regwisel=isel
	@regwused = operand
      else
	raise "Output operand #{operand} already used in line: \"{#$current_line}\""
      end
    when /^b$/
      #do nothing, modes are set in set_code
    else
      raise "Unknown location type for operand #{operand}"
    end
    nil
  end

  def is_using_jpdata(vsm)
    name = @bmrused
    return false unless name
    return false if name =~  /^\$/
    v = vsm.getvarwithname(name)
    while  v.addressstring =~ /\w+/
      v = vsm.getvarwithname(v.addressstring)
    end
    return false unless v.transferdata
    return false unless v.transferdata[0] == "elt" 
    true
  end
  
  def set_address_modes
    # t register --- write bit and shortstop bit
    @tregcode=[0,0]
    @tregcode[0] = 1 if @twused
    if @trused
      @tregcode[1] = 1 if @trused.shortstop
    end
    # grf
    @regacode=[0,0,0]
    @regbcode=[0,0,0]
    @regwcode=[0,0,0,0]
    @lmemcode=[0,0,0,0,0]
    @bmcode=[0,0,0,0]
    
    if @regaused
      @regacode=[@regaused.address, @regaused.vectorbit, @regaused.longbit]
    end
    if @regbused
      @regbcode=[@regbused.address, @regbused.vectorbit, @regbused.longbit]
    end
    if @regwused
      @regwcode=[1,@regwused.address, @regwused.vectorbit, @regwused.longbit]
    end
    if @lmemrused or  @lmemwused
      lmem = @lmemrused? @lmemrused:@lmemwused
      inc= lmem.vector ? 1:0
      inc *= lmem.stride
      write=@lmemrused ? 0:1
      @lmemcode=[write,lmem.address,inc, lmem.tregindirectbit, lmem.longbit]
    end
    if @bmrused or  @bmwused
      bm = @bmrused? @bmrused:@bmwused
      inc=0
      write=@bmrused ? 0:1
      @bmcode=[write,bm.address,@bmpeid,bm.longbit]
    end
  end      

  def set_immediate(source)
    bits = source.to_i(16).to_s(2)
    raise "source value #{source} exceed 36 bits for imm " if bits.length>36
    bits = "0"*(36-bits.length)+bits
    val=[]
    [6,1,1,6,1,1,1,1,1,1,
      1,1,1,1,1,1,2,2,1,1,
      1,1,1,1].each{|k|
      val.push bits[0,k].to_i(2)
      bits = bits[k,bits.length]
    }
    p val
    @regacode = val[0,3]
    @regbcode = val[3,3]
    @fmulflags= val[6,10]
    @fmulround= val[14]
    @fmulisel = val[16,2]
    @faddflags[0]=val[18]
    @faddflags[1]=val[19]
    @faddflags[2]=val[20]
    @faddround=val[21]
    @faddflags[3]=val[22]
    @faddisel[0]= val[23] << 2
  end

  def setvsm(vsm)
    @vsm=vsm
  end
  def setidp(opcode, operands, parameters=nil)
    @idpid = @vsm.emdata.idpid(operands[0])
    bcastmode = ""
    if operands[1]
      bcastmode = operands[1].to_i
      if bcastmode > 2 || bcastmode <0
        raise "parameter bcastmode = #{a[3]} not supported for  #{text} "
      end
    end
    bbid =   operands[2]? (operands[2].to_i):""
    @idpdata = "idp #{@idpid} #{bcastmode} #{bbid}"
    @vsm.emdata.setbcastmode(@idpid, bcastmode)
    false
  end

  def setrdp(opcode, operands, parameters=nil)
    @rdpid = @vsm.emdata.idpid(operands[0])
    @rdpdata = "rdp #{@rdpid}"
    false
  end

  def setwdp(opcode, operands, parameters=nil)
    @wdpdata = "wdp"
    false
  end

  def setrrn(opcode, operands, parameters=nil)
    convtype = @vsm.rrn_conversions(operands.pop)
    convtype = @vsm.rrn_conversions(operands.pop) if convtype == nil

    a = remove_comment(@vsm.rrnstring(operands)).split
    a.shift
    a.unshift("rrn")
    a.push("conv")
    a.push(convtype)
    @rrndata = a.join(" ")
    p @rrndata
    false
  end
  def set_code(opcode, operands, parameters=nil)
    @params=parameters
    imvalue = nil
    if opcode =~ /^idp$/
      return  setidp(opcode, operands, parameters)
    elsif opcode =~ /^rdp$/
      return      setrdp(opcode, operands, parameters)
    elsif opcode =~ /^wdp$/
      return      setwdp(opcode, operands, parameters)
    elsif opcode =~ /^rrn$/
      return      setrrn(opcode, operands, parameters)
    elsif opcode =~ /^fmul/
      @fmulflags=[0,0,0,1,0,0,0,1,1,1]
      fmulflags_default = @fmulflags.clone
      @fmulflags[2]=1 if operands[0].long 
      @fmulflags[6]=1 if operands[1].long 
      if  operands[2]
        @fmulflags[8]=0 if operands[2].long 
      else
        @fmulflags[8]=0
      end
      subtype= $'
#      print subtype + ": subtype \n"
      if subtype
      @fmulflags= fmulflags_default.clone
	(0..9).each{|i|
	  flag= "uvwAxyzBRO"[i]
	  if subtype.index(flag)
#            print "i=", i, "\n"
	    @fmulflags[i]=1-fmulflags_default[i]
	    @fmulflags[2]=0 if i==1
	  end
	}
      end
#      print @fmulflags,"\n"
      (0..1).each {|i| @fmulisel[i]=muxcode(operands[i],false)}
      instype= "fmul"
    elsif opcode =~ /dmul([01])/
      subtype = $1.to_i
      @fmulflags= [
	[0,1,0,1,0,0,0,1,0,0],
	[0,1,0,1,1,0,1,0,0,0]][subtype]
      (0..1).each {|i| @fmulisel[i]=muxcode(operands[i],false)}
      instype= "fmul"
    elsif opcode =~ /^(fadd|fsub)/
      @faddflags=[1,1,0,1]
      @faddflags[2]=1 if $1 == "fsub"
      subtype= $'
      if subtype
	@faddflags.each_index{|i|
	  flag= "ABSO"[i]
	  if subtype.index(flag)
	    @faddflags[i] = (@faddflags[i]== 1) ? 0:1
	  end
	}
      end
      @faddisel[0]=muxcode(operands[0],true)
      @faddisel[1]=muxcode(operands[1],false)
      instype= "fadd"
      @fsel=1
    elsif opcode =~/^([iu])/
      @ialuflags=[0,0]
      @ialuflags[1]=1 if $1 == "u"
      opcode = $'
      codes = %w{
         add sub inc   dec
         not and or    xor
         max min passa passb
         lsl lsr bsl   bsr
         lnot}
      opid = codes.index(opcode)
      if opid 
	@ialuflags[0]=opid
	instype= "ialu"
	@faddisel[0]=muxcode(operands[0],true)
	@faddisel[1]=muxcode(operands[1],false)
	@fsel=0
      elsif opcode == "mm"
	@ialuflags[0]=31;
	instype= "ialu"
	imvalue = operands[0]
	operands.unshift(imvalue)
	p opcode
	p operands
	p "imm read"
      else
	raise "unknown opcode #{opcode} for operands #{operands.join(" ")} line \"#{$current_line}\""
      end
    elsif opcode =~ /^bm/
      instype= "bm"
      if operands[0].location == "b"
	bmop=operands[0]
	peop=operands[1]
	@bmrused=bmop
	@bmpeid=0
	setstoragemux(peop,instype)
	raise "bm to bm "      if peop.location == "b"
      else
	peop=operands[0]
	bmop=operands[1]
	@bmwused=bmop
#	@faddisel[0]=muxcode(operands[0],true)
        setregb(operands[0])
	@bmpeid=operands[2].to_i
	raise "too few operands" if operands.size < 3
	raise "not bm  "      if bmop.location != "b"
	raise "bm to bm "      if peop.location == "b"
      end
    else
	raise "unknown opcode #{opcode} for operands #{operands.join(" ")} line \"#{$current_line}\""
    end
    # set flags for input muxes of storages
    if instype != "bm"
      operands[2..(operands.size-1)].each{|x| 
	setstoragemux(x,instype)
      }
    end
    #set address mode bits for storages
    set_address_modes
    #overwrite address etc if imvalue is set (immediate mode)
    set_immediate(imvalue) if imvalue
    true
  end

  def dump
    self.instance_variables.sort.each{|x| print x+" ";eval("p #{x}")}
  end

  def nop(parameters)
    @params=parameters
    @tregcode=[0,0]
    @twisel=0
    @regwisel=0
    @regwcode=[0,0,0,0]
    @regacode=[0,0,0]
    @regbcode=[0,0,0]
    @lmemwisel=0
    @lmemcode=[0,0,0,0,0]
    @fmulflags=[0,0,0,0,0,0,0,0,0,0]
    @fmulround=0
    @fmulisel=[0,0]
    @faddflags=[0,0,0,0,0]
    @faddround=0
    @faddisel=[0,0]
    @ialuflags=[0,0]
    @fsel=0
    @bmcode=[0,0,0,0]
  end
  def to_machinecode
    @mcode=[]
    @mcode.push([@params.vlen, "ll"])
    @mcode.push([@params.mi, "m_imr"])
    @mcode.push([@params.mo, "m_omr"])
    @mcode.push([@params.mf, "m_ifsel"])
    @mcode.push([@tregcode[0] ,"t_write"])
    @mcode.push([@tregcode[1] ,"t_shortstop"])
    @mcode.push([@twisel ,"t_isel"])
    @mcode.push([@regwcode[0] ,"r_write"])
    @mcode.push([@regwisel ,"r_isel"])
    @mcode.push([@regwcode[1] ,"r_wadr"])
    @mcode.push([@regwcode[2] ,"r_wadri"])
    @mcode.push([@regwcode[3], "r_wwl"])
    @mcode.push([@regacode[0] ,"r_radra"])
    @mcode.push([@regacode[1] ,"r_radria"])
    @mcode.push([@regacode[2] ,"r_rwla"])
    @mcode.push([@regbcode[0] ,"r_radrb"])
    @mcode.push([@regbcode[1] ,"r_radrib"])
    @mcode.push([@regbcode[2] ,"r_rwlb"])
    @mcode.push([@lmemcode[0] ,"l_write"])
    @mcode.push([@lmemwisel ,"l_isel"])
    @mcode.push([@lmemcode[1] ,"l_adr"])
    @mcode.push([@lmemcode[2] ,"l_adri"])
    @mcode.push([@lmemcode[3] ,"l_tregadr"])
    @mcode.push([@lmemcode[4] ,"l_wl"])
    @mcode.push([@fmulflags[0] ,"fmul_shift50a"])
    @mcode.push([@fmulflags[1] ,"fmul_round50a"])
    @mcode.push([@fmulflags[2] ,"fmul_rounda"])
    @mcode.push([@fmulflags[3] ,"fmul_normala"])
    @mcode.push([@fmulflags[4] ,"fmul_shift25b"])
    @mcode.push([@fmulflags[5] ,"fmul_shift50b"])
    @mcode.push([@fmulflags[6] ,"fmul_roundb"])
    @mcode.push([@fmulflags[7] ,"fmul_normalb"])
    @mcode.push([@fmulround ,"fmul_round"])
    @mcode.push([@fmulflags[9] ,"fmul_normalo"])
    @mcode.push([@fmulisel[0] ,"fmul_isela"])
    @mcode.push([@fmulisel[1] ,"fmul_iselb"])
    @mcode.push([@faddflags[0] ,"fadd_normala"])
    @mcode.push([@faddflags[1] ,"fadd_normalb"])
    @mcode.push([@faddflags[2] ,"fadd_signb"])
    @mcode.push([@faddround ,   "fadd_round"])
    @mcode.push([@faddflags[3] ,"fadd_normalo"]) 
    @mcode.push([@faddisel[0] ,"fadd_isela"])
    @mcode.push([@faddisel[1]  ,"fadd_iselb"])
    @mcode.push([@ialuflags[0],"ialu_ialuop"])
    @mcode.push([@ialuflags[1],"ialu_unsigned"])
    @mcode.push([@fsel ,"fsel"])
    @mcode.push([@bmcode[0],"bm_write"])
    @mcode.push([@bmcode[1],"bm_adr"])
    @mcode.push([@bmcode[2],"bm_peadr"])
    @mcode.push([@bmcode[3],"bm_wl"])
    return @mcode
  end

  def emit
    str = ""
    longfields={
      "r_radra" =>2,
      "r_radrb" =>2,
      "r_wadr"  =>2,
      "l_adr"   =>3,
      "ialu_ialuop"   =>2,
      "bm_adr"  =>3,
      "bm_peadr"=>2}
    str +=  "ISP"
    to_machinecode
    @mcode.each{|x|
      len=1
      len=longfields[x[1]] if longfields[x[1]]
      data=x[0].to_s(16)
      while data.length <= len
	data = " "+data
      end
      str += data
    }
    str +=  " "+@idpdata if @idpdata
    str +=  " "+@rdpdata if @rdpdata
    str +=  " "+@wdpdata if @wdpdata
    str +=  " "+@rrndata if @rrndata
    str += " # #{@asmcode}\n"
    print str
    str
  end

  def check_code(previnst, prepreinst)
# perform various tests...
    if previnst.twused # previous inst wrote treg
      if @trused and ( not @trused.shortstop )
        printwarning("$t read used though previous line writes $t")
      end
    else
      if @trused and  @trused.shortstop
        printerror("$ti read used though previous line does not write $t")
      end
    end
    if previnst.regwused 
      if @regaused  == previnst.regwused or @regbused  == previnst.regwused
        printwarning("grf read from address written in previous line")
      end
    end
    if previnst.lmemwused and @lmemrused
      printerror("lmem read performed after lmem write without idle ")
    end
    if prepreinst
      if prepreinst.lmemwused and @lmemrused
        printerror("lmem read performed after lmem write without two idles ")
      end
    end
  end
end


class AddressExpression
  attr_reader :location, :long, :vector, :address, :shortstop, :pe, :bm,
    :tregindirect, :lmemstride
  def initialize(s)
    @location=nil
    @long=nil
    @vector=nil
    @address=nil
    @shortstop=nil
    @pe=nil
    @bm=nil
    @tregindirect=nil
    @lmemstride = 1
    self.parse(s)
  end
  def parse(s)
    sold = s
    i=0
    raise "#{s} is not a valid address expression " unless s[0..0] == "$"
    if s[1..2] == "pe"
      @location= "pe"
      @pe=true
      @long = true
      return self
    elsif s[1..2] == "bb"
      @location= "bb"
      @bm=true
      @long = true
      return self
    elsif s[1..2] == "fb"
      @location= "fb"
      @fb=true
      @long = true
      return self
    elsif s[1..2] == "lt"
      @location= "m"
      @long = true
      @tregindirect=true
      @address=0
      return self
    elsif s[1..2] == "st"
      @location= "m"
      @tregindirect=true
      @address=0
      return self
    elsif s[1..2] == "lr"
      @location= "r"
      @long = true
      i=3
    elsif s[1..2] == "lm"
      @location= "m"
      @long = true
      i=3
    elsif s[1..2] == "lb"
      @location= "b"
      @long = true
      i=3
    elsif s[1..1] == "r"
      @location= "r"
      i=2
    elsif s[1..1] == "m"
      @location= "m"
      i=2
    elsif s[1..1] == "b"
      @location= "b"
      i=2
    elsif s[1..1]== "t"
      @location= "t"
      @long=true
      @vector=true
      if s.length==3
	if s[2..2] == "i"
	  @shortstop=true
	else
	  raise "#{s} is not a valid address expression "
	end
      elsif s.length!=2
	raise "#{s} is not a valid address expression "
      else
	@shortstop=false
      end
      return self
    end
    s = s[i..(s.size-1)]
    vloc = s.index("v")
    if vloc
      @vector=true
      if vloc == s.size-1
	@lmemstride=1
      else
	@lmemstride =s[(vloc+1)..(s.size)].to_i
      end
      addval=s[0..(vloc-1)].to_i
    else
      @vector=false
      addval=s.to_i
    end
    @address=addval
  end
end



class Var
  attr_reader :location, :long, :vector, :address, :shortstop, :name,
    :len, :constant_value, :lmemstride, :transferdata, :addressstring

  def initialize
    @location=nil
    @long=nil
    @vector=nil
    @address=nil
    @shortstop=nil
    @lmemstride=1
    @transferdata=nil
  end

  def emit
    unless @constant_value
      print "DUM "
      print @location
      print @long ? "l" : "s"
      printf("%5x: ",@address)
      print @name
      print "[4]" if @vector
      print "[#{@len}]" if @len>1
      print "\n"
    end
  end
  def emit_consts
    if @constant_value
      print "DUM "
      print @location
      print @long ? "l" : "s"
      printf("%5x: ",@address)
      print @name
      print " = ",@constant_value.join("\""),"\""
      print "\n"
    end
  end

  def tregindirect
    nil
  end
  def indexed(i)
    v = Var.new
    v.location = @location
    v.long  = @long 
    v.vector  = @vector 
    v.address  = @address 
    v.shortstop  = @shortstop 
    v.name = @name
    v.constant_value = @constant_value
    v.len = @len
    
    inc = 1
    inc = 2 if @long 
    v.name += "[#{i}]"
    v.len=1
    if @len
      inc *= @len  
    else
      v.vector=false
    end
    v.address += inc*i
    v
  end

  def convert_constant(constant)
    @location="m"
    @vector=false
    if constant[0] =~ /l/
      @long=true
    else
      @long=false
    end
    @name=constant[2]
    @address=nil
    @constant_value=constant[0..1]
    self
  end

  def parse(s)
    a=remove_comment(s).split
    first=a.shift
    if first== "var"
      @location="m"
    elsif  first== "bvar"
      @location="b"
    else
      raise "variable definition, unknown type #{first}"
    end
    token=a.shift
    if token=="vector"
      @vector=true
      token=a.shift
    else
      @vector=false
    end
    if token== "long"
      @long=true
      token=a.shift
    elsif token== "short"
      @long=false
      token=a.shift
    end
    @name=token
    @len=1
    if @name =~ /(\w+)(\[)(.*)(\])/
      @name=$1
      len= $3
      if len != /^\d+$/ 
	raise "illegal variable name in \"#{s}\""
      end
      @len=len.to_i
    end
    token=a.shift
    @address=nil
    while token
      if token !~ /^(elt|hlt|rrn).*$/
	@addressstring=token
	token=a.shift
      else
	@transferdata = [token]+a
	token=nil
      end
    end
    @constant_value=nil
  end

  def assign_address(start_value)
    return start_value if @address 
    if @addressstring
      case @addressstring
      when /\d+/
	@address=@addressstring.to_i
      when /\w+/
	@address=@addressstring.address
	if @address == nil
	  raise "Unknown variable name #{@addressstring} in address field"
	end
      end
      return start_value
    end
    index= (@location == "m") ? 0:1 
    address=start_value[index]
    if @long and (address &1) == 1
      address += 1
    end
    @address=address
    length = (@long? 2:1) *(@vector? 4:1) 
    start_value[index]=address+length
    start_value
  end

  def find_variable_with_name(name)
    index = nil
    if name =~ /(\w+)(\[)(.*)(\])/
      name = $1
      index = $3.to_i
    end
    @variables.each{|x|
      if x.name==name
	return x.indexed(index) if index
	return x
      end
    }
    raise "variable name #{name} not defined"
  end
end

class String


  def storageobject
    if self =~ /^\$/
      return AddressExpression.new(self)
    else
      return $vsm.getvarwithname(self)
    end
  end      

  
  def address
    storageobject.address
  end      
  def stride
    storageobject.lmemstride
  end      
  def location
    x = storageobject
    x.location
  end      
  def len
    storageobject.len
  end
  def long
    storageobject.long 
  end
  def longbit
    long ? 1:0
  end
  def vector
    storageobject.vector
  end       
  def vectorbit
    vector ? 1:0
  end
  def shortstop
    storageobject.shortstop
  end      
  def shortstopbit
    shortstop ? 1:0
  end
  def tregindirect
    storageobject.tregindirect
  end      
  def tregindirectbit
    tregindirect ? 1:0
  end

end

class EMdata

  def initialize
    @ems=[]
    @bcastmode=[]
  end

  def check(data, text)
    raise "too few parameters in  #{text} "    if data[2]==nil
    @ems.each{|x|
      raise "evar name collision for #{data[0]} " if data[0]==x[0]
      raise "evar field collision for #{data[1]} " if data[1]==x[1]
    }
  end
  def complete(a,text)
    a[3] = 0 if a[3]==nil
    a[4] = -1 if a[4]==nil
    a[3] = a[3].to_i
    a[4] = a[4].to_i
    if a[3] < 0 ||  a[3] > 2
      raise "parameter bcastmode = #{a[3]} not supported for  #{text} "
    end
    a
  end

  def setbcastmode(id, mode)
    @bcastmode[id]=mode
  end
  def register(parms, text)
    parms.shift
    newdata=complete(parms,text)
    check(newdata,text)
    @ems.push(newdata)
  end

  def idpid(name)
    if name =~ /^(\d)+$/
      return name.to_i
    else
      @ems.each{|x|
        if name == x[0]
          return x[1].to_i
        end
      }
      raise "Unknown idp parameter #{name}"
    end
  end

  def idpsize(index)
    @ems.each{|x|
      if index.to_i == x[1].to_i
        return x[2].to_i
      end
    }
    # default values, for index=0,  default. Else NPE
    (index == 0) ?  $default_idp_size : $clop.npe
  end
  def idpparm(index,id)
    @ems.each{|x|
      if index.to_i == x[1].to_i
        return x[2+id].to_i
      end
    }
    raise "idp index #{index} not defined "
  end

  def idpbcastmode(index)
    if @bcastmode[index] != nil
      return @bcastmode[index].to_i 
    end
    idpparm(index, 1)
  end
  def idpbbid(index)
    idpparm(index, 2)
  end
end


class Vsm

  attr_accessor :previous_instruction
  attr_reader :emdata

  @@em_bm_conversions = {
    "flt64to72"  =>  0,
    "flt64to36"  =>  1,
    "flt32to36"  =>  2,
    "fix64to72l" =>  3,
    "fix64to72rs"=>  4,
    "fix64to72ru"=>  5,
    "fix64to36"  =>  6,
    "fix32to36l" =>  7,
    "fix32to36rs"=>  8,
    "fix32to36ru"=>  9}

  @@em_bm_type_to_conversion = {
    "f"  =>["double", "", "flt64to36"],
    "fl" =>["double", "", "flt64to72"],
    "i"  =>["INT64",   "",  "fix64to36"],
    "il" =>["INT64",   "", "fix64to72rs"],
    "h"  =>["UINT64",  "", "fix64to36"],
    "hl"  =>["UINT64", "[2]", ["fix64to36","fix64to36",9]]}

  @@em_bm_hosttype = {
    "flt64to72"  =>  "double",
    "flt64to36"  =>  "double",
    "flt32to36"  =>  "float",
    "fix64to72l" =>  "UINT64",
    "fix64to72rs"=>  "INT64",
    "fix64to72ru"=>  "UINT64",
    "fix64to36"  =>  "UINT64",
    "fix32to36l" =>  "UINT32",
    "fix32to36rs"=>  "INT32",
    "fix32to36ru"=>  "UINT32"}

  @@rrn_conversions = {
    "flt72to64"   =>  0,
    "flt36to64"   =>  1,
    "fix72to64l"  =>  2,
    "fix72to64r"  =>  3,
    "fix72to64w"  =>  4}

  @@rrn_hosttype = {
    "flt72to64"   =>  ["double", ""],
    "flt36to64"   =>  ["double", ""],
    "fix72to64l"  =>  ["UINT64",  ""],
    "fix72to64r"  =>  ["UINT64",  ""],
    "fix72to64w"  =>  ["UINT64", "[2]"]}

  @@alias_table = Hash.new

  @previous_instruction = nil
  @pre_previous_instruction = nil

  def rrn_conversions(s)
    @@rrn_conversions[s]
  end

  def initialize
    @number_of_constants=0
    @constants = []
    @variables = []
    @params=Parameters.new
    @params.mi=0
    @params.mo=0
    @params.mf=0
    @params.vlen=1
    @codestate = "initialization"
    @initialization_code = []
    @loop_initialization_code = []
    @loop_body_code = []
    @loop_finalization_code = []
    @loop_extra_finalization_codes = []
    @loop_extra_initialization_codes = []
    @initialization_rrn = []
    @loop_initialization_rrn = []
    @loop_body_rrn = []
    @loop_finalization_rrn = []
    @loop_extra_finalization_rrns = []
    @extra_id = -1
    @extra_initialization_id = -1
    @prefix= "SING"
  end
  
  def find_constant(s)
    if s =~ /([idfh]l*)\"(.+)\"/
      return [$1,$2]
    end
    nil
  end
  
  def generate_name
    s = "_const_"+@number_of_constants.to_s
    @number_of_constants+= 1
    s
  end
  def register_constant(type, value)
    name=generate_name
    @constants.push([type, value, name])
  end
  
  def find_and_set_prefix(a)
    a.each{|s|
      x = s.split
      @prefix = x[1] if x[0] == "prefix"
    }
  end
  def find_and_register_constant(a)
    a.map{|x| remove_comment(x)}.each{|s|
      consts=find_constant(s)
      if consts
	isregistered=false
	@constants.each{|x|
	  isregistered=true if x[0]==consts[0] and x[1]==consts[1]
	}
	
	register_constant(*consts) unless isregistered
      end
    }
    a
  end

  def replace_constant(a)
    b=[]
    @constants.each{|x|
      conststring= x[0]+"\""+x[1]+"\""
      pattern= Regexp.new( conststring)
      a.each_index{|i|
	a[i]=a[i].gsub(pattern,x[2])
      }
    }
    a
  end

  def add_eltx(a)
    b=[]
    ($clop.nhltloop).times{|iloop|
      a.each{|x|
        b.push(x) if iloop == 0
        if x =~ /var (.*) hlt /
          y = x.split
          i = y.index("hlt")
          y[i]= "elt1"
          y[i-1] = y[i-1]+"_b"+(iloop.to_s)
          y[0] = "bvar"
          b.push(y.join(" "))
        end
      }
    }
    b
  end
        

  def find_and_register_variables(a)
    a.map{|x| remove_comment(x)}.each{|s|
      token=s.split.shift
      if token =~ /^b*var$/
	x=Var.new
	x.parse(s)
	@variables.push(x)
      end
    }
  end
  def add_constants_to_variables
    @constants.each{|x|
      @variables.push(Var.new.convert_constant(x))
    }
  end

  def reorder_variables
    hltvars=[]
    eltvars=[]
    oldvars=[]

    @variables.each{|x|
      if x.transferdata
        type, conv = x.transferdata
        if type == "hlt"
          hltvars.push(x)
        elsif type =~ /elt/
          i = type[3,3].to_i
          eltvars[i] = [] if eltvars[i]== nil
          eltvars[i].push(x)
        else
          oldvars.push(x)
        end
      else
        oldvars.push(x)
      end
    }
    @variables = hltvars
    eltvars.each{|x|      @variables += x if x }
    @variables += oldvars

  end
    
  def assign_addresses_to_variables
    reorder_variables
    address=[0,0] #local and broadcast
    @variables.each{|x|
      address = x.assign_address(address)
    }
  end


  def emit_single_idp_data_packet(x)
    suffix = x.long ? "L" : "S"
    v=x.constant_value[1]
    case x.constant_value[0]
    when /^f/
      print "FLT#{suffix} ", sprintf("%25.17g",eval(v)), " #", v,"\n"
    when /^i/
      print "INT#{suffix} ", eval(v), "\n"
    when /^h/
      print "HEX#{suffix} ", v, "\n"
    end
  end

  def create_grape_setup_for_constants

    ndata=0;
    structstring= "struct #{@prefix}_grape_consts{\n"
    assignstring= ""
    conversionstring="int conversions[]={"
    firstaddress=nil;
    @variables.each{|x|
      if x.constant_value
	firstaddress = x.address unless firstaddress;
	a = @@em_bm_type_to_conversion[x.constant_value[0]]
	structstring += "  "+a[0]+ " "+x.name + a[1]+ ";\n";
	conv=a[2];
	valstr=x.constant_value[1]
	lhs = "    constants.#{x.name}"
	if conv.class==Array
	  conversionstring+= @@em_bm_conversions[conv[0]].to_s+","+
	    @@em_bm_conversions[conv[1]].to_s+","
	  len=valstr.length
	  if len > 9
	    msb= valstr[0..(valstr.length-10)];
	    valstr=valstr[(valstr.length-10)..(valstr.length-1)];
	    assignstring += "#{lhs}[0] = 0x#{msb}LL;\n"
	  else
	    assignstring += "#{lhs}[0] = 0;\n"
	  end
	  assignstring += "#{lhs}[1] = 0x#{valstr}LL;\n"
          ndata += 2
	else
	  conversionstring+= @@em_bm_conversions[conv].to_s+","
          ndata+= 1
	  if x.constant_value[0] == "h"
	    assignstring += "#{lhs} = 0x#{valstr}LL;\n"
	  elsif x.constant_value[0][0] == "i"
	    assignstring += "#{lhs} = #{eval(valstr)}LL;\n"
	  else
	    assignstring += "#{lhs} = #{eval(valstr)};\n"
	  end
	end
      end
    }
    structstring += "};\n"
    structstring += "struct #{@prefix}_grape_consts constants;\n"
    conversionstring = conversionstring.chop + "};\n"
    if $grapeout 
      if ndata > 0
	$grapeout.print <<-END
#{structstring}
int #{@prefix}_send_vsm_constants()
{
    int ndata = #{ndata};
    int firstaddress= #{firstaddress};
    #{conversionstring}
#{assignstring}
    SING_LM_write_regconv(ndata,conversions);
    SING_LM_write(&constants, firstaddress, 0);
}
int #{@prefix}_send_vsm_constantsMC(SINGinfo * s)
{
    int ndata = #{ndata};
    int firstaddress= #{firstaddress};
    #{conversionstring}
#{assignstring}
    SING_LM_write_regconvMC(s, ndata,conversions);
    SING_LM_writeMC(s, &constants, firstaddress, 0);
}
        END
      else
	$grapeout.print "int #{@prefix}_send_vsm_constants(){}\n"
      end
    end
  end

  def create_grape_setup_for_j_particles
    if $grapeout
      $grapehout.print "\#define grape_j_particle_struct  #{@prefix}_elt_struct0\n"
      $grapeout.print <<-END
int #{@prefix}_send_j_particle(struct #{@prefix}_elt_struct0 *jp, int index)
{
      return #{@prefix}_send_elt_data0(jp, index);
}
int #{@prefix}_send_j_particleMC(int chipid, struct #{@prefix}_elt_struct0 *jp, int index)
{
      return #{@prefix}_send_elt_data0MC(chipid, jp, index);
}
      END
    end
  end

  def max_transfer_code(name)
    maxnum=0
    regexp = Regexp.new(name + "(.*)")
    @variables.each{|x|
      if x.transferdata
	type, conv = x.transferdata
	if type =~ regexp
	  numtext=$1
	  num =  numtext.length > 0 ? numtext.to_i : 0
	  maxnum = num if num > maxnum
	end
      end
    }
    maxnum
  end

  def process_evar(a)
    @emdata = EMdata.new
    a.each{|s|
      x = s.split
      @emdata.register(x,s) if x[0] == "evar" 
    }
  end
  

  def create_grape_setup_for_i_particles(index)
    ndata=0;
    suffix=  index.to_s
    structname ="struct #{@prefix}_hlt_struct#{suffix}"
    structstring= structname + "{\n"
    vstructname = "struct #{@prefix}_hlt_vector_struct#{suffix}"
    vstructstring= vstructname + "{\n"
    assignstring= ""
    conversionstring="int conversions[]={"
    firstaddress=nil
    @variables.each{|x|
      if x.transferdata
	type, conv = x.transferdata
	if type =~ /^hlt(.*)/
	  numtext=$1
	  num =  numtext.length > 0 ? numtext.to_i : 0
	  if num == index
	    if  @@em_bm_hosttype[conv] == nil
	      raise "nonexistent converion #{conv} "+
		"specified for variable #{x.name}" 
	    end
	    firstaddress = x.address unless firstaddress;
	    raise "hlt variable #{x.name} must be of vector type" unless x.vector
	    vstr= "[4]"
	    vstructstring += "  "+@@em_bm_hosttype[conv]+ " "+x.name  +
	      vstr+ ";\n";
	    structstring += "  "+@@em_bm_hosttype[conv]+ " "+x.name  + ";\n";
	    assignstring += "   vp->#{x.name}[k]= p->#{x.name};\n"
	    conversionstring+= (@@em_bm_conversions[conv].to_s+",")*4
	    ndata+= 1
	  end
	end
      end
    }
    structstring += "};\n"
    vstructstring += "};\n"
    conversionstring = conversionstring.chop + "};\n"
    if $grapeout
      sendipheader=
	"int #{@prefix}_send_i_particle(#{structname} *ip,\n"+
	"                         int n)"
      sendipheaderMC=
      "int #{@prefix}_send_i_particleMC(int chipid, #{structname} *ip,\n"+
      "                         int n)"
      sendipheaderMCP=
      "int #{@prefix}_send_i_particleMCP(int nchips, int chipid[],\n"+
      "                          #{structname} *ip,\n"+
      "                          int n)"

      $grapehout.print structstring, vstructstring, sendipheader,";\n"
      $grapehout.print sendipheaderMC,";\n", sendipheaderMCP,";\n" 
      $grapehout.print "\#define grape_i_particle_struct #{structname.split[1]}\n"
      $grapehout.print "\#define grape_i_particle_vector_struct #{vstructname.split[1]}\n"
      $grapehout.print "int #{@prefix}_npipes();\n"
      $grapeout.print <<-END


static #{@prefix}_copy_ip_to_vip(#{structname} *p,
                                 #{vstructname} *vp,
			         int k)
{
#{assignstring}
}

#{sendipheader}
{
    #{vstructname} ipv[NPEMAX];
    int i;
    for(i=0;i<n;i++){
        int iv = i/4;
	int k  = i%4;
        #{@prefix}_copy_ip_to_vip(ip+i, ipv+iv, k);
    }
    i = (n+3)/4;
    SING_EM_fifoclr();
    #{@prefix}_send_elt_data1_m((struct #{@prefix}_elt_struct1 *)ipv, 0, i);
    SING_EM_fifoclr();
    SING_LM_write(ipv, #{firstaddress}, i);
}

#{sendipheaderMC}
{
    #{vstructname} ipv[NPEMAX];
    int i;
    for(i=0;i<n;i++){
        int iv = i/4;
      int k  = i%4;
        #{@prefix}_copy_ip_to_vip(ip+i, ipv+iv, k);
    }
    i = (n+3)/4;
    SING_EM_fifoclrMC(chip[chipid]);
    SING_send_elt_data1_mMC(chipid,(struct SING_elt_struct1 *)ipv, 0, i);
    SING_flush_piow_channelMC(chip[chipid]);
    SING_EM_BM_transfer_setupMC_andsend1(chip[chipid], emid1a[chipid], bmid1a[chipid],
				0,0,1,0,-1);
    SING_startMC(chip[chipid], ids[#{allcodes.length-1}], 1);
    SING_finishMC(chip[chipid]);
    SING_startMC(chip[chipid], ids[#{allcodes.length-2}], i);
    SING_flush_piow_channelMC(chip[chipid]);
    SING_flush_piow_channelMC(chip[chipid]);
    SING_nwait(10000);
    SING_finishMC(chip[chipid]);
}

#{sendipheaderMCP}
// currently support J-parallel only
//
{
    #{vstructname} ipv[NPEMAX];
    int i;
    int nend = n+#{4* $clop.nhltloop} -1;
    if (nend > SING_npipes()){
	nend = SING_npipes();
    }
    for(i=0;i<nend;i++){
        int iv = i/4;
      int k  = i%4;
        #{@prefix}_copy_ip_to_vip(ip+i, ipv+iv, k);
    }
    i = (n+ #{4* $clop.nhltloop} -1)/#{4* $clop.nhltloop};
    int ic;
     SING_nwait(10000-i*200);
     // need to wait in case DRAM is written recently?

    if (nchips == 4) CTRL_beginbcast(chip[chipid[0]]->c);


    SING_EM_fifoclrMC(chip[chipid[0]]);

    //    UINT64 t0 = SING_rdtsc();


    #{@prefix}_send_elt_data1_mMC(chipid[0], (struct #{@prefix}_elt_struct1 *)ipv,
			    0, i);
    //    UINT64 t1 = SING_rdtsc();
    SING_EM_BM_transfer_setupMC_andsend1(chip[chipid[0]], emid1a[chipid[0]],
				bmid1a[chipid[0]], 0,0,1,0,-1);
    SING_nwait(20000);
    //BEWARE: you might need to increase the wait parameter above
    // to make the library works on some machines
    //    UINT64 t2 = SING_rdtsc();

    if (nchips == 4)     CTRL_coc_setflags(3,-1,-1);
    SING_startMC(chip[chipid[0]], ids[#{allcodes.length-1}], 1);
    //    UINT64 t2a = SING_rdtsc();
    for(ic=0;ic<nchips; ic++)SING_finishMC(chip[chipid[ic]]);
//    SING_finishMC(chip[chipid[0]]);
//    SING_nwait(10000);

    //    UINT64 t3 = SING_rdtsc();

    if (nchips == 4)     CTRL_coc_setflags(3,-1,-1);

    SING_startMC(chip[chipid[0]], ids[#{allcodes.length-2}], i);

//    SING_finishMC(chip[chipid[0]]);
//    SING_nwait(1000);
    for(ic=0;ic<nchips; ic++)SING_finishMC(chip[chipid[ic]]);
    
    if (nchips == 4) CTRL_endbcast(chip[chipid[0]]->c);

    //    UINT64 t4 = SING_rdtsc();
#if 0
    fprintf(stderr, "i= %d MCP cycles = %e %e %e %e %e\\n", 
           i,
           (double)(t1-t0),
           (double)(t2-t1),
           (double)(t2a-t2),
           (double)(t3-t2a),
           (double)(t4-t3));
#endif
}

void #{@prefix}_register_i_particle_conversion()
{
    #{conversionstring}
    SING_LM_write_regconv(#{ndata*4},conversions);
}
void #{@prefix}_register_i_particle_conversionMC(SINGinfo * s)
{
    #{conversionstring}
    SING_LM_write_regconvMC(s, #{ndata*4},conversions);
}

int #{@prefix}_npipes()
{
    return #{$clop.npe * 4};
}


      END
    end
  end

  def create_setup_for_hlt(index)
    ndata=0;
    suffix= index == 0 ? "" : index.to_s
    structstring= "struct #{@prefix}_hlt_struct#{suffix}{\n"
    conversionstring="int conversions[]={"
    firstaddress=nil
    @variables.each{|x|
      if x.transferdata
	type, conv = x.transferdata
	if type =~ /^hlt(.*)/
	  numtext=$1
	  num =  numtext.length > 0 ? numtext.to_i : 0
	  if num == index
	    if  @@em_bm_hosttype[conv] == nil
	      raise "nonexistent converion #{conv} "+
		"specified for variable #{x.name}" 
	    end
	    firstaddress = x.address unless firstaddress;
	    vstr=  x.vector ? "[4]" :  ""
	    structstring += 
	      "  "+@@em_bm_hosttype[conv]+ " "+x.name + vstr + ";\n";

	    (1..(  x.vector ? 4 : 1 )).each{
	      conversionstring+= (@@em_bm_conversions[conv].to_s+",")
	      ndata+= 1;
	    }
	  end
	end
      end
    }
    structstring += "};\n"
    conversionstring = conversionstring.chop + "};\n"
    if $grapeout
      sendipheader=
	"int #{@prefix}_send_hlt_data#{suffix}(struct #{@prefix}_hlt_struct#{suffix} *ip,"+
        "\n                         int peindex)"
      sendipheaderMC=
	"int #{@prefix}_send_hlt_data#{suffix}MC(int chipid, struct #{@prefix}_hlt_struct#{suffix} *ip,"+
        "\n                         int peindex)"
      $grapehout.print structstring, sendipheader,";\n"

      $grapeout.print <<-END



#{sendipheader}
{
    #{conversionstring}
    SING_LM_write_regconv(#{ndata},conversions);
    SING_LM_write(ip, #{firstaddress}, peindex);
}
      END
    end
  end

  def create_setup_for_elt(index)
    ndata=0;
#    p "setup for elt #{index}"
#    suffix= index == 0 ? "" : index.to_s
    suffix= index.to_s
    structstring= "struct #{@prefix}_elt_struct#{suffix}{\n"
    conversionstring="int conversions[]={"
    firstaddress=nil
    @variables.each{|x|
      if x.transferdata
	type, conv = x.transferdata
	if type =~ /^elt(.*)/
	  numtext=$1
	  num =  numtext.length > 0 ? numtext.to_i : 0
	  if num == index
	    if  @@em_bm_hosttype[conv] == nil
	      raise "nonexistent converion #{conv} "+
		"specified for variable #{x.name}" 
	    end
	    firstaddress = x.address unless firstaddress;
	    vstr=  x.vector ? "[4]" :  ""
	    structstring += 
	      "  "+@@em_bm_hosttype[conv]+ " "+x.name + vstr + ";\n";

	    (1..(  x.vector ? 4 : 1 )).each{
	      conversionstring+= (@@em_bm_conversions[conv].to_s+",")
	      ndata+= 1;
	    }
	  end
	end
      end
    }
    structstring += "};\n"
    conversionstring = conversionstring.chop + "};\n"
    if $grapeout
      sendipheader=
	"int #{@prefix}_send_elt_data#{suffix}(struct #{@prefix}_elt_struct#{suffix} *ip,"+
        " int index_in_EM)"
      sendipheaderMC=
	"int #{@prefix}_send_elt_data#{suffix}MC(int chipid, struct #{@prefix}_elt_struct#{suffix} *ip,"+
        " int index_in_EM)"
      moveipheader= "int #{@prefix}_move_elt_data#{suffix}(int index)"
      moveipheaderMC= "int #{@prefix}_move_elt_data#{suffix}MC(int chipid, int index)"
      setipheader= "int #{@prefix}_setup_elt_data#{suffix}()"
      setipheaderMC= "int #{@prefix}_setup_elt_data#{suffix}MC(int chipid)"
      idpsizefuncheader= "int #{@prefix}_elt_size#{suffix}()"
      $grapehout.print "\n",structstring, sendipheader,";\n"
      $grapehout.print setipheaderMC,";\n"
      $grapehout.print sendipheaderMC,";\n"
      $grapehout.print moveipheader,";\n"
      $grapehout.print moveipheaderMC,";\n"
      $grapehout.print idpsizefuncheader,";\n"
      sendipheader_m=
	"int #{@prefix}_send_elt_data#{suffix}_m(struct #{@prefix}_elt_struct#{suffix} *ip,"+
        " int index_in_EM, int ndata)"
      sendipheader_mMC=
	"int #{@prefix}_send_elt_data#{suffix}_mMC(int chipid, struct #{@prefix}_elt_struct#{suffix} *ip,"+
        " int index_in_EM, int ndata)"
      $grapehout.print sendipheader_m,";\n\n"
      sendipheader_vMC=
	"int #{@prefix}_send_elt_data#{suffix}_vMC(int chipid, struct #{@prefix}_elt_struct#{suffix} *ip,"+
        " int index[], int ndata)"
      $grapehout.print sendipheader_mMC,";\n\n"
      $grapehout.print sendipheader_vMC,";\n\n"

      $grapeout.print <<-END



static int emid#{suffix};
static int bmid#{suffix};
static int emid#{suffix}a[NCHIPS];
static int bmid#{suffix}a[NCHIPS];
#{setipheader}
{
    int err;
    #{conversionstring}
    int bcastmode = #{@emdata.idpbcastmode(index)};
    emid#{suffix}=SING_EM_alloc(#{ndata}, conversions, #{@emdata.idpsize(index)}, &err);
    if (emid#{suffix} < 0){
        fprintf(stderr,"register jp, EMalloc err = %d\\n\",err);
        exit(-1);
    }
    bmid#{suffix}=SING_BM_alloc(emid#{suffix},1,&err);
    if (bmid#{suffix} < 0){
    fprintf(stderr,"register jp, BMalloc err = %d\\n\",err);
          exit(-1);
    }
    SING_EM_BM_transfer_setup(emid#{suffix},bmid#{suffix},0,0,1,bcastmode,-1);

}

#{setipheaderMC}
{
    int err;
    #{conversionstring}
    int bcastmode = #{@emdata.idpbcastmode(index)};
    emid#{suffix}=SING_EM_allocMC(chip[chipid],#{ndata}, conversions, #{@emdata.idpsize(index)}, &err);
    if (emid#{suffix} < 0){
        fprintf(stderr,"register jp, EMalloc err = %d\\n\",err);
        exit(-1);
    }
    bmid#{suffix}=SING_BM_allocMC(chip[chipid],emid#{suffix},1,&err);
    if (bmid#{suffix} < 0){
    fprintf(stderr,"register jp, BMalloc err = %d\\n\",err);
          exit(-1);
    }
    emid#{suffix}a[chipid]=emid#{suffix};
    bmid#{suffix}a[chipid]=bmid#{suffix};
    SING_EM_BM_transfer_setupMC(chip[chipid],emid#{suffix},bmid#{suffix},0,0,1,bcastmode,-1);
}

#{idpsizefuncheader}
{
    return #{@emdata.idpsize(index)};
}

#{sendipheader}
{
    int ret;
    ret = SING_EM_send(emid#{suffix},index_in_EM,1,ip);
    if (ret){
        fprintf(stderr,"SING_EM_send returns: %d\\n\", ret);
         exit(-1);
    }
}

#{sendipheaderMC}
{
    int ret;
    ret = SING_EM_vsendMC(chip[chipid], emid#{suffix}a[chipid],&index_in_EM,1,ip);
    if (ret){
        fprintf(stderr,"SING_EM_send returns: %d\\n\", ret);
         exit(-1);
    }
}

#{sendipheader_m}
{
    int ret;

    ret = SING_EM_send(emid#{suffix},index_in_EM,ndata,ip);
    
    if (ret){
        fprintf(stderr,"SING_EM_send returns: %d\\n\", ret);
         exit(1);
    }

    return 0;
}

#{sendipheader_mMC}
{
    int ret;

    ret = SING_EM_sendMC(chip[chipid], emid#{suffix}a[chipid],index_in_EM,ndata,ip);
    
    if (ret){
        fprintf(stderr,"SING_EM_send returns: %d\\n\", ret);
         exit(1);
    }

    return 0;
}

#{sendipheader_vMC}
{
    int ret;

    ret = SING_EM_vsendMC(chip[chipid], emid#{suffix}a[chipid],index,ndata,ip);
    
    if (ret){
        fprintf(stderr,"SING_EM_send returns: %d\\n\", ret);
         exit(1);
    }

    return 0;
}

#{moveipheader}
{
    SING_EM_BM_transfer(bmid#{suffix},index,0,1,1,-1);
}
#{moveipheaderMC}
{
    SING_EM_BM_transferMC(chip[chipid],  bmid#{suffix}a[chipid],index,0,1,1,-1);
}
      END
    end
  end

  def process_hlt
    maxindex=max_transfer_code("hlt")
    create_grape_setup_for_i_particles(0)
    (1..maxindex).each{|i| create_setup_for_hlt(i)}
  end
  def process_elt
    $grapeout.print "static int ids[#{allcodes.length}];\n" if $grapeout
    maxindex=max_transfer_code("elt")
#    p "maxindex for elt = #{maxindex}"
    (maxindex+1).times{|i| create_setup_for_elt(i)}
  end

  def rrnstring(tokenarray)
    s = tokenarray.join(" ")
    x= parse_rrn_line_to_array(tokenarray, s)
    rrn_string(x,s)
  end

  def variables_with_type(s)
    list=[]
    @variables.each{|x|
      if x.transferdata
	type, conv = x.transferdata
	if type == s
          list.push(x)
        end
      end
    }
    list
  end

#   def variables_with_rrn
#     list=[]
#     @variables.each{|x|
#       if x.transferdata
# 	type, conv = x.transferdata
# 	if type == "rrn"
# 	  if  @@rrn_hosttype[conv] == nil
# 	    raise "nonexistent converion #{conv} "+
# 	      "specified for variable #{x.name}" 
# 	  end
#           list.push(x)
#         end
#       end
#     }
#     list
#   end
  def variables_with_rrn
    variables_with_type("rrn")
  end
  def variables_with_hlt
    variables_with_type("hlt")
  end


  def generate_finalization_from_rrnvariables
    s = ["loop finalization"]
    s.push("nop")
    s.push("nop")
    npe = $clop.npe
    i = 0
    variables_with_rrn.each{|x|
      incsize = $vsize * ( x.vector ? 2:1 )
      startaddress = $bememsize - incsize*npe
      t = x.long ? "l" : ""
      
      s.push "upassa #{x.name} #{x.name} $#{t}r0v"
      s.push("nop")
      if $clop.koike_runtime
        rrn_location = 3 
        nop_count=npe+1 
#
# Wait here is necessary to make SING generate output,
# but not enough to make FPGAs pass date to PCIe, becase
# PCIe (with x4) has half the bandwidth of the SING chip.
#
      else
        rrn_location = 1
        nop_count=npe-1
      end
      npe.times{|ipe|
        inst = "bm $#{t}r0v $lb#{startaddress + incsize*ipe}v  #{ipe}"
        if ipe == rrn_location
          inst += " ; rrn #{x.transferdata[2]} $lb#{startaddress}v  #{npe*$vsize} #{x.transferdata[1]} #{x.name}"
        end
        s.push(inst)
      }
      s.push("nop #{nop_count}")
      i += 1;
      if (i % 4 == 0 ) && (i < variables_with_rrn.length)
        s.push "loop extra finalization #{i/4}-1"
      end
    }
#    s.push("nop 3")
    s
  end


  def generate_extra_initialization_from_hltvariables
    s = ["loop extra initialization"]
    ($clop.nhltloop).times{|iloop|
      s.push("mi 0")
      s.push("moi 1")
      s.push("uxor $peid $lr0v $t ")
      s.push("moi 0")
      s.push("uinc  $lr0v $lr0v $lr0v")
      s.push("mi 1")
      i = 0
      variables_with_hlt.each{|x|
        s.push "bm #{x.name}_b#{iloop} #{x.name}"
      }
    }
    s.push("mi 0 ")
    s.push("idp 1 1")
    s.push("rdp 1")
    s.push("nop 10") 
    s.push("wdp")
    s.push("nop 10")
    s.push("loop extra initialization")
    s.push("nop 10")
    s.push("uxor  $lr0v $lr0v $lr0v ; rdp 1")
    s.push("nop 10")
#    s.push("idp 1 1")
#    s.push("rdp 1")
#    s.push("nop 2") 
    s.push("wdp")
    s.push("nop 10") 
    s
  end



  def create_grape_setup_for_result
      ndata=0;
    structstring= "struct grape_result_struct{\n"
    vstructstring= "struct grape_result_vector_struct{\n"
    assignstring= ""
    rrninstructions=[]
    conversionstring="int conversions[]={"
    firstaddress=nil

    @variables.each{|x|
      if x.transferdata
	type, conv = x.transferdata
	if type == "rrn"
	  if  @@rrn_hosttype[conv] == nil
	    raise "nonexistent converion #{conv} "+
	      "specified for variable #{x.name}" 
	  end
	  rrndata=x.transferdata
	  rrndata[1]=rrndata[2]
	  rrndata[2]=x.name
	  rrndata.shift
	  rrninstructions.push(rrnstring(rrndata))
	  firstaddress = x.address unless firstaddress;
	  raise "rrn variable #{x.name} must be of vector type" unless x.vector
	  vstr= "[(4*NPE)]"
	  vstructstring += "  "+@@rrn_hosttype[conv][0]+ " "+x.name  +
	    vstr+@@rrn_hosttype[conv][1]+ ";\n";
	  structstring += "  "+@@rrn_hosttype[conv][0]+ " "+x.name +
	    @@rrn_hosttype[conv][1] + ";\n";
	  if @@rrn_hosttype[conv][1] == ""
	    assignstring += "   p->#{x.name}=vp->#{x.name}[k];\n"
	  elsif  @@rrn_hosttype[conv][1] == "[2]"
	    2.times{|i|
	      assignstring += 
		"   p->#{x.name}[#{i}]=vp->#{x.name}[k][#{i}];\n"
	    }
	  else
	    raise "unimplemented data type for variable #{x.name}"
	  end
	  conversionstring+= (@@rrn_conversions[conv].to_s+",")
	  ndata+= 1
	end
      end
    }
    structstring += "};\n"
    vstructstring += "};\n"
    conversionstring = conversionstring.chop + "};\n"
    rrnstring = rrninstructions.collect{|y| "\""+y+"\","}.join("\n").chop
    if $grapeout
      grheader="int #{@prefix}_get_result(struct grape_result_struct *rp)"
      grheaderMC="int #{@prefix}_get_resultMC(int chipid, struct grape_result_struct *rp)"
      $grapehout.print structstring,vstructstring,grheader,";\n"
      $grapeout.print <<-END

static char* rrnstring[]={
#{rrnstring}
};
static #{@prefix}_copy_vr_to_r(struct grape_result_vector_struct *vp,
                         struct grape_result_struct *p,
                         int k)
{
#{assignstring}
}

static int n_rrn;

#{grheader}
{
    struct grape_result_vector_struct rvp;
    int i;
    int n = (NPE*4);
    SING_LM_read((UINT32*)&rvp);

    for(i=0;i<n;i++){
        #{@prefix}_copy_vr_to_r(&rvp, rp+i,i);
    }
}
#{grheaderMC}
{
    struct grape_result_vector_struct rvp;
    int i;
    int n = (NPE*4);
    SING_LM_readMC(chip[chipid], (UINT32*)&rvp);

    for(i=0;i<n;i++){
        #{@prefix}_copy_vr_to_r(&rvp, rp+i,i);
    }
}

void #{@prefix}_register_result_conversion()
{
    #{conversionstring}
    SING_LM_read_regconv(#{ndata},conversions,rrnstring,
                          #{firstaddress},1,0);
}

      END
    end
  end

  def sort_constants_by_word_length
    shorts=[]
    longs=[]
    @constants.each{|x|
      tmpvar = Var.new.convert_constant(x)
      if tmpvar.constant_value
	if tmpvar.long
          longs.push(x)
        else
          shorts.push(x)
        end
      end
    }
    @constants = longs + shorts
  end  
  
  def fix_nwords_of_constants
    sort_constants_by_word_length
    #  first count length
    len=0
    @constants.each{|x|
      tmpvar = Var.new.convert_constant(x)
      if tmpvar.constant_value
	if tmpvar.long
	  len = ((len+1)/2)*2+2
	else
	  len += 1
	end
      end
    }
    print "constant length = ", len, "\n";
    if len%2 != 0
      register_constant("f", "0.0");
      len += 1;
      print "constant length = ", len, "\n";
    end
  end
  def emit_idp_packets_for_constants
    #  first count length
    len=0
    @variables.each{|x|
      if x.constant_value
	if x.long
	  len = ((len+1)/2)*2+2
	else
	  len += 1
	end
      end
    }
    if len > 0 
      print "IDP ", ((len+1)/2), " 0 0 0 0 # IDP constant packets\n"
      lastconst=nil
      @variables.each{|x|       
        if x.constant_value
          emit_single_idp_data_packet(x)
          lastconst=x
        end
      }
      emit_single_idp_data_packet(lastconst)if len%2 != 0
    end
  end

  def emit_isp_for_constants
    @params.vlen=1
    offset=-1
    @variables.each{|x|
      if x.constant_value
	offset = x.address if offset == -1
	prefix = x.long ? "l" : ""
	asmcode = "bm $#{prefix}b#{x.address-offset}  #{x.name}"
	inst=Instruction.new(asmcode)
	a = asmcode.split
	opcode=a.shift
	inst.set_code(opcode,a,@params)
	inst.emit
      end
    }
  end
  
  def getvarwithname(s)
    @variables.each{|x|
      return x if x.name == s
    }
    raise "Unknown variable #{s} in line \"#{$current_line}\""
  end    

  def print_variables
    @variables.each{|x|  x.emit }
    @variables.each{|x|  x.emit_consts  }
  end
    
  
  def parse_idp_line(a,s)
#  IDP header line format
#  IDP varname [bbn [bbnmask]]
#  need to convert to
#  IDP len addr bbn bbnmask
    if a[0] =~ /^\d+/
      length= a[0].to_i
      header = a.join(" ")
    else
      bbn = 0
      bbnmask = 0
      seq = ""
      varname=a[0]
      address = a[0].address
      length= a[0].len
      length *= 4 if a[0].vector
      if a.size > 1
        if a[a.size-1] == "SEQ" 
          seq = "1"
          a.pop
        end
      end
      bbn = a[1].to_i if a.size > 1
      bbnmask = a[2].to_i if a.size > 2
      header = [length, address, bbn, bbnmask,seq].join(" ")
    end
    print "IDP ", header, " #", s, "\n"
  end

  def parse_rrn_line_to_array(a,s)
#  RRN line format
#  RRN fadd/ialuop/number memloc length/round/odpoe/sregen
#
#  last field written as [n][r][o][s]
    varname=a[1]
    params=a[2]
    length= varname.vector ? 4:1
    if params =~ /^(\d+)/
      length= $1.to_i
    end
    round  = (params =~ /r/) ? 1:0
    odpoe  = (params =~ /o/) ? 0:1
    sregen = (params =~ /s/) ? 0:1
    destcode = "$lr0v"
    destcode = "$r0v" if round
    sourcecode = "$r0v"
    sourcecode = "$lr0v" if varname.long
    if a[0] =~ /^(\d+)$/
      nbb= a[0].to_i
      instcode="iadd"
      reduc=0
    else
      nbb= 0
      instcode= a[0]
      reduc=1
    end
    x=Instruction.new(s)
    x.set_code(instcode,[sourcecode,sourcecode,destcode])
    a=[]
    a.push([varname.address,"adr"])
    a.push([length,"n"])
    a.push([nbb,"bbadr"])
    a.push([reduc, "reduc"])
    a.push([oneif(varname.long), "wl"])
    a.push([x.fsel, "fsel"])
    a.push([x.faddflags[0], "normala"])
    a.push([x.faddflags[1], "normalb"])
    a.push([x.faddflags[2], "signb"])
    a.push([round, "round"])
    a.push([x.faddflags[3], "normalo"])
    a.push([x.ialuflags[0], "ialuop"])
    a.push([x.ialuflags[1], "unsigned"])
    a.push([odpoe, "odpoe"])
    a.push([sregen, "sregen"])
    a
  end

  def parse_rrn_line(a,s)
    a=parse_rrn_line_to_array(a,s)
    emit_rrn(a,s)
  end

  def rrn_string(a,s)
    longfields={
      "adr" =>4,
      "n" =>4,
      "ialuop"   =>2}
    rrns= "RRN"

    a.each{|x|
      len=1
      len=longfields[x[1]] if longfields[x[1]]
      data=x[0].to_s(16)
      while data.length <= len
	data = " "+data
      end
      rrns+=  data
    }
    rrns+ " # #{s}"
  end
  def emit_rrn(a,s)
    print "IDP 1 1000  0 0\n"
    print rrn_string(a,s)+"\n"

    a.each{|x|
      len=1
      len=longfields[x[1]] if longfields[x[1]]
      data=x[0].to_s(16)
      while data.length <= len
	data = " "+data
      end
      print data
    }
    print " # #{s}\n"
  end


  def parse_debugprint_line(a,s)
#  debugprint vpm-sim line format
#   DBGP bbid peid LM/GRF/TREG/BM format address count varname
#  debugprint vsm line format
#  print [bbid] [peid] fmt varname
    str =  "DBGP #{a[1]} #{a[2]}"
    varname=replace_alias([a[4]])[0]
    count = varname.vector ? 4:1
    case varname.location
    when "b"
      str += " BM "
    when "m"
      str+= " LM "
    when "t"
      str+= " TREG "
    when "r"
      str+= " GRF "
    else
      raise "Unparsable debug line #{s}"
    end
    format = a[3]
    format += "l" if varname.long && format.length==1
    adr = varname.location == "t"? 0:varname.address;
    str+= format+ " "+adr.to_s+ " "+count.to_s+ " "+varname+ "\n"
    print str
    str
  end

  def emit_rrn(a,s)
    longfields={
      "adr" =>4,
      "n" =>4,
      "ialuop"   =>2}
    print "IDP 1 1000  0 0\n"
    print "RRN"

    a.each{|x|
      len=1
      len=longfields[x[1]] if longfields[x[1]]
      data=x[0].to_s(16)
      while data.length <= len
	data = " "+data
      end
      print data
    }
    print " # #{s}\n"
  end



  def parse_line(s)
    $current_line = s
    a=remove_comment(s).split
    return nil if a.size==0
    case a[0]
    when /^initialization$/
      @codestate = "initialization"
    when /^loop$/
      if a[1] == "initialization"
	@codestate = "loop initialization"
      elsif  a[1] == "body"
	@codestate = "loop body"
      elsif  a[1] == "finalization"
	@codestate = "loop finalization"
      elsif  a[2] == "finalization" && a[1]=="extra"
	@codestate = "loop extra finalization"
        @extra_id += 1
        @loop_extra_finalization_codes[@extra_id]=[]
        @loop_extra_finalization_rrns[@extra_id]=[]
      elsif  a[2] == "initialization" && a[1]=="extra"
	@codestate = "loop extra initialization"
        @extra_initialization_id += 1
        @loop_extra_initialization_codes[@extra_initialization_id]=[]
      else
	raise "Unknown instruction: #{s}"
      end
    when /^mi$/
      @params.mi= a[1].to_i
    when /^mo([if])$/
      @params.mo= a[1].to_i
      if $1 == "f" 
	@params.mf= 1
      else
	@params.mf= 0
      end
    when /^vlen$/
      @params.vlen= a[1].to_i
    when /^(bm|rdp|wdp|rrn|fmul(\w*)|fadd(\w*)|dmul\d|fsub|[iu]\w+)$/
      inst=Instruction.new(s)
      inst.setvsm(self)
      instset=false
      s.split(";").each{|x|
	a=remove_comment(x).split
        register_rrn(a)
	opcode=a.shift
        if opcode
          instset = true if inst.set_code(opcode,replace_alias(a),@params)
        end
      }
      inst.nop(@params) unless instset
      register_instruction(inst.emit)
      if inst.is_using_jpdata(self) 
	@eltlocation = @loop_body_code.size 
      end
      if @previous_instruction
        inst.check_code(@previous_instruction, @pre_previous_instruction) 
      end
      @pre_previous_instruction=@previous_instruction
      @previous_instruction=inst
    when /^(nop)$/
      count=1
      count = a[1].to_i if a[1]
      count.times{
        inst=Instruction.new(s)
        inst.nop(@params)
        register_instruction(inst.emit)
        @pre_previous_instruction=@previous_instruction
        @previous_instruction=inst
      }
    when /^(bvar|var|evar)$/
      # variables are registered. Do nothing
    when /^(IDP)$/
      a.shift
      parse_idp_line(a,s)
    when /^(RRN)$/
      a.shift
      parse_rrn_line(a,s)
    when /^(((HEX|FLT|INT)((S|L)*)(\s*)))$/
      print s+"\n"
    when /^ISP_/
      print s+"\n"
    when /^(DBG|DBGL|ECHO)$/
      print s+"\n"
      register_instruction(s+"\n")
    when /^(print)$/
      register_instruction(parse_debugprint_line(a,s))
    when /^(alias)$/
      register_alias(a[1],a[2])
    when /^(prefix)$/
      STDERR.print a, "\n" if $DEBUG
      @prefix = a[1]
    else
      raise "Unparsable line #{s}"
    end
  end

  def register_alias(name, entity)
    @@alias_table[name]=entity
  end
  def replace_alias(a)
    a.collect{|x|   @@alias_table[x] ? @@alias_table[x]:x  }
  end
  def register_instruction(s)
    case @codestate
    when  "initialization"
      @initialization_code.push(s.chomp)
    when  "loop initialization"
      @loop_initialization_code.push(s.chomp)
    when  "loop body"
      @loop_body_code.push(s.chomp)
    when  "loop finalization"
      @loop_finalization_code.push(s.chomp)
    when  "loop extra finalization"
      @loop_extra_finalization_codes[@extra_id].push(s.chomp)
    when  "loop extra initialization"
      @loop_extra_initialization_codes[@extra_initialization_id].push(s.chomp)
    else
      raise "Unknown codestate #{@codestate}"
    end
  end
  def register_rrn(a)
    return if a[0] != "rrn"
    s = a.join(" ")
    case @codestate
    when  "initialization"
      @initialization_rrn.push(s.chomp)
    when  "loop initialization"
      @loop_initialization_rrn.push(s.chomp)
    when  "loop body"
      @loop_body_rrn.push(s.chomp)
    when  "loop finalization"
      @loop_finalization_rrn.push(s.chomp)
    when  "loop extra finalization"
      @loop_extra_finalization_rrns[@extra_id].push(s.chomp)
    else
      raise "Unknown codestate #{@codestate}"
    end
  end

  def decode_regrrns(s)
    vresult=""
    result=""
    assign=""
    index=0
    len=1
    sizes=[]
    size = 0;
    maxindex = s.size
    s.each{|x|
      a = x.split
      name = a.pop
      type = name
      convtype = @@rrn_conversions[type]
      if convtype == nil
        type = a.pop
        convtype = @@rrn_conversions[type]
      else
        name = "x#{index}"
      end
      a.shift
      len = parse_rrn_line_to_array(a, "")[1][0]
      vresult += "  #{@@rrn_hosttype[type][0]} #{name}[#{len}]#{@@rrn_hosttype[type][1]};\n";
      xsize = 0
      xsize = 8 if @@rrn_hosttype[type][0] == "double"
      xsize = 8 if @@rrn_hosttype[type][0] == "UINT64"
      xsize = 4 if @@rrn_hosttype[type][0] == "float"
      if xsize == 0
        raise "decode_regrrn: data type #{@@rrn_hosttype[type][0]} not implemented yet"
      end
      xsize *= 2 if @@rrn_hosttype[type][1] == "[2]"
      xsize *= len
      size += xsize
        print "size= #{size}, index=#{index}\n"
        if ((index % 4) == 3) || (index == (maxindex -1))
          sizes.push(size)
          size=0
        end
      result += "  #{@@rrn_hosttype[type][0]} #{name}#{@@rrn_hosttype[type][1]};\n";
      if @@rrn_hosttype[type][1] == ""
        assign += "   p->#{name}=vp->#{name}[k];\n"
      elsif  @@rrn_hosttype[type][1] == "[2]"
        2.times{|i|
          assign += "   p->#{name}[#{i}]=vp->#{name}[k][#{i}];\n"
        }
      else
        raise "unimplemented data type for name #{name}"
      end
      index += 1
    }
    [vresult,result,assign,len,sizes]
  end

  def writeout_regrrn_functions(vstr, str, assign,length, prefix)
    return unless $grapeout
    return if vstr == ""
    $grapehout.print <<-END
struct #{@prefix}_#{prefix}_struct{
#{str}
};
struct #{@prefix}_#{prefix}_vectorstruct{
#{vstr}
};
END
    $grapeout.print <<-END
static #{@prefix}_copy_vr_to_r(struct #{@prefix}_#{prefix}_vectorstruct *vp,
                              struct #{@prefix}_#{prefix}_struct *p,
                         int k)
{
#{assign}
}
END
  return  if prefix != "result"
  $grapehout.print "int #{@prefix}_get_result(struct #{@prefix}_result_struct *rp);\n"
  $grapehout.print "int #{@prefix}_get_resultMC(int chipid, struct #{@prefix}_result_struct *rp);\n"
  $grapehout.print "\#define  grape_result_struct #{@prefix}_result_struct \n"
  $grapeout.print <<-END
static struct #{@prefix}_result_vectorstruct #{@prefix}_rvp;
static struct #{@prefix}_result_vectorstruct #{@prefix}_rvpa[NCHIPS];

int #{@prefix}_get_result(struct #{@prefix}_result_struct *rp)
{
    int i;
    for(i=0;i<#{length};i++){
        #{@prefix}_copy_vr_to_r(&#{@prefix}_rvp, rp+i,i);
    }
}
int #{@prefix}_get_resultMC(int chipid, struct #{@prefix}_result_struct *rp)
{
    int i;
    for(i=0;i<#{length};i++){
        #{@prefix}_copy_vr_to_r(#{@prefix}_rvpa+chipid, rp+i,i);
    }
}
END
  end

  def generate_functions_for_regrrns
    @loop_result_rrn = @loop_finalization_rrn + @loop_extra_finalization_rrns.flatten
    [["body", "loopout"],["result", "result"]].each{|x|
      rrnstring = eval("@loop_#{x[0]}_rrn")
      prefix = x[1]
      vstr, str, assign, len,@rrn_sizes =decode_regrrns(rrnstring)
      writeout_regrrn_functions(vstr, str, assign,len, prefix)
      
    }
  end
  def testprint
    print "Enter testprint: initilization\n"
    print @initialization_code.join("\n"),"\n"
    print "testprint: loop initilization\n"
    print @loop_initialization_code.join("\n"),"\n"
    print "testprint: loop body\n"
    print @loop_body_code.join("\n"),"\n"
    print @eltlocation, "\n"
  end

  def edit_loop_initialization_for_idp(s)
    if $grapeout and (not $type2output)
      s.each_index{|i|
        if s[i] =~ /loop initialization/
          return s.insert(i+1,"idp 0") 
        end
      }
    end
    s
  end

  def insert_idp(code, location,id)
    return if location < 0 
    s = code[location]
    code[location]=s.sub(/#/," idp #{id} #") 
  end
  def insert_rdp_and_wdp(code, rdp_location,wdp_location,id)
    return if rdp_location < 0 
    s = code[rdp_location]
    code[rdp_location]=s.sub(/#/," rdp #{id} #") 
    wdp_location = rdp_location+ 1    if wdp_location <= rdp_location
    s = code[wdp_location]
    code[wdp_location]=s.sub(/#/," wdp #") 
  end

  def edit_loop_body_for_idp
    if $grapeout and (not $type2output)
      insert_idp(@loop_body_code, @eltlocation,0)
      if $clop.koike_runtime
        insert_rdp_and_wdp(@loop_body_code, @eltlocation+1,
                           (@loop_body_code.length)-4 ,0)
      end
    end
  end

  def edit_initialization_for_koike_runtime(code)
    STDERR.print "enter edit init for koike\n"
    code.map!{|x| x.sub(/ idp /," rdp ")}

    loc_to_insert_wdp = (code.length) - 3
    loc_to_insert_wdp = 1 if loc_to_insert_wdp <= 0
    code[loc_to_insert_wdp]=code[loc_to_insert_wdp].sub(/#/," wdp #") 
  end
  def edit_loop_initialization_for_koike_runtime
    if  $clop.koike_runtime
      edit_initialization_for_koike_runtime(@loop_initialization_code)
    end
  end

  def allcodes
      [@initialization_code,@loop_initialization_code,
        @loop_body_code,@loop_finalization_code]+@loop_extra_finalization_codes+@loop_extra_initialization_codes
  end


  def create_grape_code_routines
    if $grapeout
      i=0;
      codes=[nil,nil,nil]
      eltlocs=[-1,-1,@eltlocation,-1]
      lines=[0,0,0]      
      allcodes.each{|x|
	if x != nil && x.size>0
	  codes[i]= x.collect{|y| "\""+y+"\","}.join("\n").chop
	  lines[i]=x.size
	end
        i+=1;
      }
      codes.each_index{|i|
	if codes[i]
	  $grapeout.print "static char* codes_#{i}[]={\n"
	  $grapeout.print codes[i],"\n"
	  $grapeout.print "};\n"
	end
      }
      initheader="void #{@prefix}_grape_init()"
      initheaderMC="void #{@prefix}_grape_initMC(int chipid)"
      if $clop.koike_runtime 
           initheader="void #{@prefix}_grape_init(int chipid)"
      end
      closeheader="void #{@prefix}_grape_close()"
      closeheaderMC="void #{@prefix}_grape_closeMC(int chipid)"
      runheader="int #{@prefix}_grape_run(int n)"
      runheader2="int #{@prefix}_type2_run(int n, void * loopout)"
      runheaderMC="int #{@prefix}_grape_runMC(int chipid, int n)"
      runheaderMCP="int #{@prefix}_grape_runMCP(int nchips, int chipid[], int n)"
      runheader2MC="int #{@prefix}_type2_runMC(int chipid, int n, void * loopout)"

      waitdmarheader = "int #{@prefix}_wait_DMAR(int nchips, int chipid[])"
      eltsetupcode=""
      (max_transfer_code("elt")+1).times{|i|
        eltsetupcode += "    #{@prefix}_setup_elt_data#{i}();\n"
      }
      eltsetupcodeMC=""
      (max_transfer_code("elt")+1).times{|i|
        eltsetupcodeMC += "    #{@prefix}_setup_elt_data#{i}MC(chipid);\n"
      }
      
      $grapehout.print initheader,";\n",runheader,";\n",runheader2,";\n"
      $grapehout.print initheaderMC,";\n",runheaderMC,";\n",runheader2MC,";\n"
      $grapehout.print waitdmarheader,";\n"
      $grapehout.print runheaderMCP,";\n"
      $grapehout.print closeheader,";\n", closeheaderMC,";\n";
      if (not $type2output)
        set_result_size_args = "0,sizeof(#{@prefix}_rvp)"
      else
        set_result_size_args = "sizeof(struct #{@prefix}_loopout_vectorstruct),0"
      end

      $grapeout.print <<-END
static #{@prefix}_register_grape_codes(int id[])
{
    int k;
    for (k=0;k< #{codes.size};k++) id[k]=-1;
    SING_iflush();
END
      codes.each_index{|i|
	if codes[i]
	  $grapeout.print "    id[#{i}]=SING_iregister(#{lines[i]},(const char**)codes_#{i});\n"
	end
      };

      $grapeout.print "}\n"

        $grapeout.print <<-END
static #{@prefix}_register_grape_codesMC(SINGinfo *s, int id[])
{
    int k;
    for (k=0;k< #{codes.size};k++) id[k]=-1;
    SING_iflush();
END
      codes.each_index{|i|
      if codes[i]
        $grapeout.print "    id[#{i}]=SING_iregisterMC(s,#{lines[i]},(const char**)codes_#{i});\n"
      end
      };
    $grapeout.print "}\n"      

      $grapeout.print <<-END  

int * #{@prefix}_codes_pointer()
{
    return ids;
}

int DMAW_size = (#{@rrn_sizes.join("+")})/8;

#{initheader}
{
    #{$clop.koike_runtime ? "SING_set_chipid(chipid);" : ""}
    SING_open();
    #{@prefix}_send_vsm_constants();
#{eltsetupcode}
    #{@prefix}_register_i_particle_conversion();
    #{@prefix}_register_grape_codes(ids);
    SING_set_result_size(#{set_result_size_args});
    SING_set_dmacheck_mode(1);
}
#{initheaderMC}
{
    SINGinfo * s = chip[chipid]=SING_openMC(chipid);
    #{@prefix}_send_vsm_constantsMC(s);
#{eltsetupcodeMC}
    #{@prefix}_register_i_particle_conversionMC(s);
    #{@prefix}_register_grape_codesMC(s,ids);
    SING_set_result_sizeMC(s, #{set_result_size_args});
    SING_set_dmacheck_mode(1);
    SING_register_dma_seqMC(s, 0, 1, DMAW_size, s->c->h->dmaw_buf);
    CTRL_coc_setflags(0,-1,-1);
}

void #{@prefix}_use_default_singstruct(int chipid)
{
    chip[chipid] = SING_get_singinfo();
}

END
        if (not $type2output)
          $grapeout.print <<-END

#{runheader}
{
    int rrn_sizes[]={#{@rrn_sizes.join(",")}};
    #{$clop.koike_runtime ? "SING_EM_fifoclr();\n    SING_move_elt_data0(0);" : ""}
    return SING_execute2(#{@rrn_sizes.length+2}, ids+1, n,
                        NULL, (void*) &#{@prefix}_rvp,
                        rrn_sizes);
}
#{runheaderMC}
{
    int rrn_sizes[]={#{@rrn_sizes.join(",")}};
    #{$clop.koike_runtime ? "SING_EM_fifoclrMC(chip[chipid]);\n    SING_move_elt_data0MC(chipid,0);" : ""}
    SING_initialize_dmawbufferMC(chip[chipid],sizeof(SING_rvp)/8);

    return SING_execute2MC(chip[chipid], #{@rrn_sizes.length+2}, ids+1, n,
                           NULL, (void*) (#{@prefix}_rvpa+chipid),
                           rrn_sizes);
}

#{runheaderMCP}
{
    int rrn_sizes[]={#{@rrn_sizes.join(",")}};
    int i;
    static int firstcall = 1;
    if (nchips == 4){
    //    if (0){
	CTRL_beginbcast(chip[chipid[0]]->c);
	SING_EM_fifoclrMC(chip[chipid[0]]);
	SING_nwait(1000);
	if (firstcall){
	    firstcall=0;
	    SING_nwait(100000);
	}

	#{@prefix}_move_elt_data0MC(chipid[0],0);
	CTRL_endbcast(chip[chipid[0]]->c);
    }else{
	for(i=0;i<nchips; i++)SING_EM_fifoclrMC(chip[chipid[i]]);
	for(i=0;i<nchips; i++)#{@prefix}_move_elt_data0MC(chipid[i],0);
    }
    for(i=0;i<nchips; i++){
	SING_initialize_dmawbufferMC(chip[chipid[i]],sizeof(SING_rvp)/8);
    }
    SINGinfo *s[NCHIPS];
    SINGISP_INFO * info[NCHIPS];
    CTRLinfo *c[NCHIPS];
    void * final_out[NCHIPS];    
    void * loop_out[NCHIPS];
    
    for(i=0;i<nchips; i++){
	s[i] = chip[chipid[i]];
	info[i] = &(s[i]->isp);
	c[i] = s[i]->c;
	final_out[i] = (void*) (#{@prefix}_rvpa+chipid[i]);
	loop_out[i] = NULL;
    }
    if (nchips == 4)     CTRL_coc_setflags(3,-1,-1);
    SING_execute2_i_start(info[0], nchips, c, #{@rrn_sizes.length+2}, ids+1, n,
			  loop_out, final_out,  rrn_sizes);
    if (nchips == 4)     CTRL_coc_setflags(1,-1,-1);
    SING_execute2_i_end(info[0], nchips, c, #{@rrn_sizes.length+2}, ids+1, n,
			loop_out,  final_out,  rrn_sizes);
    return 0;
}

#{waitdmarheader}
{
    SINGinfo *s[NCHIPS];
    int i;
    for(i=0;i<nchips; i++){
	s[i] = chip[chipid[i]];
    }
    
    SING_DMAR_wait(nchips, s);
    return 0;
}

END
        else
          $grapeout.print <<-END
#{runheader2}
{
    return SING_execute(ids[1],ids[2],ids[3], n,
                        loopout, NULL);
}
#{runheader2MC}
{
    return SING_executeMC(chip[chipid], ids[1],ids[2],ids[3], n,
                        loopout, NULL);
}
END
        end
        if $clop.koike_runtime
          $grapeout.print <<-END
#{closeheader}
{
    SING_close();
}
#{closeheaderMC}
{
    SING_closeMC(chip[chipid]);
}
END
        end     
      end
    end
  end
  
machinecode_header =<<END
DUM
DUM l m m m t t t r r  r r r  r r r  r r r l l   l l l l f f f f f f f f f f f f f f f f f f f  i i f b   b  b b
DUM l _ _ _ _ _ _ _ _  _ _ _  _ _ _  _ _ _ _ _   _ _ _ _ m m m m m m m m m m m m a a a a a a a  a a s m   m  m m
DUM : i o i w s i w i  w w w  r r r  r r r w i   a a t w u u u u u u u u u u u u d d d d d d d  l l e _   _  _ _
DUM : m m f r h s r s  a a w  a a w  a a w r s   d d r l l l l l l l l l l l l l d d d d d d d  u u l w   a  p w
DUM : r r s i o e i e  d d l  d d l  d d l i e   r r e : _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _  _ _ : r   d  e l
DUM : : : e t r l t l  r r :  r r a  r r b t l   : i g : s r r n s s r n r n i i n n s r n i i  i u : i   r  a :
DUM : : : l e t : e :  : i :  a i :  b i : e :   : : a : h 0 o o h h o o o o s s o o i o o s s  a n : t   :  d :
DUM : : : : : s : : :  : : :  : a :  : b : : :   : : d : i u u r i i u r u r e e r r g u r e e  l s : e   :  r :
DUM : : : : : t : : :  : : :  : : :  : : : : :   : : r : f n n m f f n m n m l l m m n n m l l  u i : :   :  : :
DUM : : : : : o : : :  : : :  : : :  : : : : :   : : : : t d d a t t d a d a a b a a b d a a b  o g : :   :  : :
DUM : : : : : p : : :  : : :  : : :  : : : : :   : : : : 5 5 a l 2 5 b l : l : : l l : : l : :  p n : :   :  : :
DUM : : : : : : : : :  : : :  : : :  : : : : :   : : : : 0 0 : a 5 0 : b : o : : a b : : o : :  : e : :   :  : :
DUM : : : : : : : : :  : : :  : : :  : : : : :   : : : : a a : : b b : : : : : : : : : : : : :  : d : :   :  : :
END

class Parameters
  attr_accessor :mi, :mo, :mf, :vlen
end



options_definition_string = <<-END

  Description: VSM assembler
  Long description:
    Assembler program for VSM --- VPM assembly language
    (c) 2004- Jun Makino
  
    This program assembles the input file written in VSM, and generates
    output machine code files. Maximum help is provided by the command
    "ruby vsm.rb --help".


  Short name: 		-g
  Long name:		--grape_output
  Value type:		string
  Variable name: 	grape_out_file
  Description:		Output file name for grape type output
  Long description:
    This option specifies the file name for GRAPE-type output mode. If
    not specified, GRAPE-type output is not generated.



  Short name: 		-t
  Long name:		--type2_output
  Value type:		bool
  Variable name: 	type2_output
  Description:		Output is type 2
  Long description:
    This option specifies that the created code is type 2, which means
    default generation of idp etc are suppressed.


  Short name: 		-k
  Long name:		--koike_runtime_support
  Value type:		bool
  Default value:        false
  Variable name: 	koike_runtime
  Description:		Output is for runtime support by Koike
  Long description:
    This option specifies that the created code is not for simulator
    but for the runtime code by Koike.    


  Short name: 		-i
  Long name:		--input
  Value type:		string
  Variable name: 	infile
  Description:		Input VSM file name
  Long description:
    This option specifies the file name for input VSM file. If
    not specified, STDIN is used instead.


  Short name: 		-p
  Long name:		--number_of_pes
  Value type:		int
  Default value:        2
  Variable name: 	npe
  Description:		Number of PEs
  Long description:
    This option specifies the number of PEs per BB, which is used in automatic
    generation of the finalization code. 



  Short name: 		-u
  Long name:		--unroll_hlt_transfer_loop
  Value type:		int
  Default value:        1
  Variable name: 	nhltloop
  Description:		unroll factor for EM-BM transfer loop for hlt variables
  Long description:
    This option specifies how many times the EM-BM transfer loop is unrolled.
    Max possible value is the same as the number of PEs, and the number of PEs
    must be an integer multiple of this number.  However, currently no error
    check is done. 

    Large value might give some performance improvement, for i-particle (hlt)
    transfer, at the cost of more use of the code memory of controll processor.



  Short name: 		-e
  Long name:		--erb_text
  Value type:		string
  Default value:        
  Variable name: 	erb_text
  Description:		Text evaluated by erb
  Long description:
    This option specifies the ruby script which is evaluated in erb at the top 
    of the source file.

  END

$clop = parse_command_line(options_definition_string, false)

input=  $clop.infile ?    open($clop.infile,"r") : STDIN
$type2output= $clop.type2_output
if $clop.grape_out_file
  fname = $clop.grape_out_file
  dirname=File.dirname(fname)
  basename=File.basename(fname, ".c")
  $cname=dirname+"/"+basename+".c"
  $hname=dirname+"/"+basename+".h"
  $grapeout=open($cname,"w+")
  $grapehout=open($hname,"w+")
else
  $grapeout=nil
  $type2output=nil
end

if $grapeout
  $grapeout.print <<-END
\#include <stdio.h>
\#include <stdlib.h>
\#include <stdarg.h>
\#include \"#{$hname}\"
\#ifndef NJP
\#    define NJP 65536
\#endif
END
  if $clop.koike_runtime
    $grapeout.print <<-END
\#ifndef NCHIPS
\#    define NCHIPS 8
\#endif
\#include <sing_chip.h>
SINGinfo * chip[NCHIPS];
END
    $grapehout.print <<-END
\#include "typedef.h"
\#include "sing.h"
\#include "sing_chip.h"
\#include "singMC.h"
\#include "ctrl.h"
END
  else
    $grapehout.print "\#include \"vpm-local.h\"\n"
    $grapehout.print "\#include \"vpm-sim-defs.h\"\n"
    $grapehout.print "\#include \"vpm-board-sim-defs.h\"\n"
  end
end

require "erb"
s = input.gets(nil)
s = "<% #{$clop.erb_text}%>" + s
erb = ERB.new(s, $SAFE)
s= erb.result.split("\n")

$vsm = Vsm.new
$vsm.find_and_set_prefix(s)
s=$vsm.find_and_register_constant(s)
$vsm.fix_nwords_of_constants
s=$vsm.replace_constant(s)
#print s.join("\n")
s=$vsm.add_eltx(s)
$vsm.find_and_register_variables(s)
$vsm.add_constants_to_variables
$vsm.assign_addresses_to_variables
$vsm.print_variables
$vsm.emit_idp_packets_for_constants
$vsm.create_grape_setup_for_constants
$vsm.process_evar(s)
$vsm.create_grape_setup_for_j_particles
$vsm.emit_isp_for_constants
print machinecode_header
s += $vsm.generate_finalization_from_rrnvariables if  $vsm.variables_with_rrn != []
s=$vsm.edit_loop_initialization_for_idp(s)
s += $vsm.generate_extra_initialization_from_hltvariables
$lineno=1
s.each{|x| 
  p x
  $vsm.parse_line(x)
  $lineno += 1
}
$vsm.process_elt
$vsm.process_hlt
$vsm.generate_functions_for_regrrns
$vsm.edit_loop_initialization_for_koike_runtime
$vsm.edit_loop_body_for_idp
$vsm.create_grape_code_routines

